Package: e2tree
Title: Explainable Ensemble Trees
Version: 1.2.0
Authors@R: c(
    person(given = "Massimo",
           family = "Aria",
           role = c("aut", "cre", "cph"),
           email = "aria@unina.it",
           comment = c(ORCID = "0000-0002-8517-9411")),
           person(given = "Agostino",
           family = "Gnasso",
           role = c("aut", "cph"),
           email = "agostino.gnasso@unina.it",
           comment = c(ORCID = "0000-0002-8046-3923"))
           )
Description: The Explainable Ensemble Trees 'e2tree' approach has been proposed by Aria et al. (2024) <doi:10.1007/s00180-022-01312-6>. It aims to explain and interpret decision tree ensemble models using a single tree-like structure. 'e2tree' is a new way of explaining an ensemble tree trained through 'randomForest' or 'xgboost' packages.
License: MIT + file LICENSE
URL: https://github.com/massimoaria/e2tree
BugReports: https://github.com/massimoaria/e2tree/issues
Encoding: UTF-8
Imports: ape, dplyr, parallel, future.apply, ggplot2, Matrix,
        partitions, purrr, rpart.plot, tidyr, Rcpp
LazyData: true
LinkingTo: Rcpp
Suggests: doParallel, foreach, htmlwidgets, jsonlite, knitr, partykit,
        gbm, lightgbm, randomForest, ranger, xgboost, rmarkdown,
        RSpectra, testthat (>= 3.0.0), visNetwork
VignetteBuilder: knitr
Config/testthat/edition: 3
Depends: R (>= 3.5)
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-05-15 12:20:50 UTC; massimoaria
Author: Massimo Aria [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-8517-9411>),
  Agostino Gnasso [aut, cph] (ORCID:
    <https://orcid.org/0000-0002-8046-3923>)
Maintainer: Massimo Aria <aria@unina.it>
Repository: CRAN
Date/Publication: 2026-05-15 13:00:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-05-15 13:58:18 UTC; unix
Archs: e2tree.so.dSYM

Package: e2tree
Title: Explainable Ensemble Trees
Version: 1.2.0
Authors@R: c(
    person(given = "Massimo",
           family = "Aria",
           role = c("aut", "cre", "cph"),
           email = "aria@unina.it",
           comment = c(ORCID = "0000-0002-8517-9411")),
           person(given = "Agostino",
           family = "Gnasso",
           role = c("aut", "cph"),
           email = "agostino.gnasso@unina.it",
           comment = c(ORCID = "0000-0002-8046-3923"))
           )
Description: The Explainable Ensemble Trees 'e2tree' approach has been proposed by Aria et al. (2024) <doi:10.1007/s00180-022-01312-6>. It aims to explain and interpret decision tree ensemble models using a single tree-like structure. 'e2tree' is a new way of explaining an ensemble tree trained through 'randomForest' or 'xgboost' packages.
License: MIT + file LICENSE
URL: https://github.com/massimoaria/e2tree
BugReports: https://github.com/massimoaria/e2tree/issues
Encoding: UTF-8
Imports: ape, dplyr, parallel, future.apply, ggplot2, Matrix,
        partitions, purrr, rpart.plot, tidyr, Rcpp
LazyData: true
LinkingTo: Rcpp
Suggests: doParallel, foreach, htmlwidgets, jsonlite, knitr, partykit,
        gbm, lightgbm, randomForest, ranger, xgboost, rmarkdown,
        RSpectra, testthat (>= 3.0.0), visNetwork
VignetteBuilder: knitr
Config/testthat/edition: 3
Depends: R (>= 3.5)
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-05-15 12:20:50 UTC; massimoaria
Author: Massimo Aria [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-8517-9411>),
  Agostino Gnasso [aut, cph] (ORCID:
    <https://orcid.org/0000-0002-8046-3923>)
Maintainer: Massimo Aria <aria@unina.it>
Repository: CRAN
Date/Publication: 2026-05-15 13:00:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-15 14:18:28 UTC; unix
Archs: e2tree.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(as.rpart,e2tree)
S3method(e2splits,e2tree)
S3method(extract_terminal_nodes,catboost.CatBoost)
S3method(extract_terminal_nodes,catboost.Model)
S3method(extract_terminal_nodes,default)
S3method(extract_terminal_nodes,gbm)
S3method(extract_terminal_nodes,lgb.Booster)
S3method(extract_terminal_nodes,randomForest)
S3method(extract_terminal_nodes,ranger)
S3method(extract_terminal_nodes,xgb.Booster)
S3method(fitted,e2tree)
S3method(get_ensemble_predictions,catboost.CatBoost)
S3method(get_ensemble_predictions,catboost.Model)
S3method(get_ensemble_predictions,default)
S3method(get_ensemble_predictions,gbm)
S3method(get_ensemble_predictions,lgb.Booster)
S3method(get_ensemble_predictions,randomForest)
S3method(get_ensemble_predictions,ranger)
S3method(get_ensemble_predictions,xgb.Booster)
S3method(get_ensemble_type,catboost.CatBoost)
S3method(get_ensemble_type,catboost.Model)
S3method(get_ensemble_type,default)
S3method(get_ensemble_type,gbm)
S3method(get_ensemble_type,lgb.Booster)
S3method(get_ensemble_type,randomForest)
S3method(get_ensemble_type,ranger)
S3method(get_ensemble_type,xgb.Booster)
S3method(measures,eValidation)
S3method(nodes,e2tree)
S3method(plot,e2tree)
S3method(plot,eValidation)
S3method(plot,loi)
S3method(plot,loi_perm)
S3method(predict,e2tree)
S3method(print,e2tree)
S3method(print,eValidation)
S3method(print,loi)
S3method(print,loi_perm)
S3method(proximity,eValidation)
S3method(residuals,e2tree)
S3method(summary,e2tree)
S3method(summary,eValidation)
S3method(summary,loi)
S3method(summary,loi_perm)
export(as.rpart)
export(createDisMatrix)
export(e2splits)
export(e2tree)
export(ePredTree)
export(eValidation)
export(extract_terminal_nodes)
export(get_ensemble_predictions)
export(get_ensemble_type)
export(loi)
export(loi_perm)
export(measures)
export(nodes)
export(plot_e2tree)
export(plot_e2tree_click)
export(plot_e2tree_vis)
export(print_e2tree_summary)
export(proximity)
export(roc)
export(rpart2Tree)
export(save_e2tree_html)
export(vimp)
import(future.apply)
import(ggplot2)
import(parallel)
import(partitions)
import(purrr)
import(rpart.plot)
import(stats)
importFrom(Matrix,"%&%")
importFrom(Matrix,Matrix)
importFrom(Matrix,all.equal)
importFrom(Matrix,as.array)
importFrom(Matrix,as.matrix)
importFrom(Matrix,rowSums)
importFrom(Matrix,sparseMatrix)
importFrom(Rcpp,evalCpp)
importFrom(Rcpp,sourceCpp)
importFrom(dplyr,"%>%")
importFrom(dplyr,across)
importFrom(dplyr,arrange)
importFrom(dplyr,as_tibble)
importFrom(dplyr,count)
importFrom(dplyr,desc)
importFrom(dplyr,distinct)
importFrom(dplyr,do)
importFrom(dplyr,first)
importFrom(dplyr,group_by)
importFrom(dplyr,group_by_at)
importFrom(dplyr,if_else)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,mutate_if)
importFrom(dplyr,n)
importFrom(dplyr,pick)
importFrom(dplyr,relocate)
importFrom(dplyr,rename)
importFrom(dplyr,rename_with)
importFrom(dplyr,row_number)
importFrom(dplyr,select)
importFrom(dplyr,slice_head)
importFrom(dplyr,slice_tail)
importFrom(dplyr,starts_with)
importFrom(dplyr,summarize)
importFrom(dplyr,tibble)
importFrom(dplyr,top_n)
importFrom(grDevices,colorRampPalette)
importFrom(grDevices,dev.cur)
importFrom(grDevices,dev.new)
importFrom(grDevices,devAskNewPage)
importFrom(grDevices,rainbow)
importFrom(graphics,abline)
importFrom(graphics,axis)
importFrom(graphics,barplot)
importFrom(graphics,box)
importFrom(graphics,hist)
importFrom(graphics,image)
importFrom(graphics,layout)
importFrom(graphics,legend)
importFrom(graphics,lines)
importFrom(graphics,locator)
importFrom(graphics,mtext)
importFrom(graphics,par)
importFrom(graphics,points)
importFrom(graphics,text)
importFrom(tidyr,complete)
importFrom(tidyr,drop_na)
importFrom(tidyr,pivot_longer)
importFrom(tidyr,pivot_wider)
importFrom(tidyr,replace_na)
importFrom(utils,globalVariables)
importFrom(utils,head)
importFrom(utils,setTxtProgressBar)
importFrom(utils,tail)
importFrom(utils,txtProgressBar)
useDynLib(e2tree)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# Explainable Ensemble Trees (e2tree)

<!-- badges: start -->

[![R-CMD-check](https://github.com/massimoaria/e2tree/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/massimoaria/e2tree/actions/workflows/R-CMD-check.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/e2tree)](https://CRAN.R-project.org/package=e2tree)
[![](http://cranlogs.r-pkg.org/badges/grand-total/e2tree)](https://cran.r-project.org/package=e2tree)

<!-- badges: end -->

<p align="center">

<img src="man/figures/e2tree_logo.png" width="400"  />
</p>

The **Explainable Ensemble Trees** (**e2tree**) key idea consists of the
definition of an algorithm to represent every ensemble approach based on
decision trees model using a single tree-like structure. The goal is to
explain the results from the ensemble algorithm while preserving its
level of accuracy, which always outperforms those provided by a decision
tree. The proposed method is based on identifying the relationship
tree-like structure explaining the classification or regression paths
summarizing the whole ensemble process. There are two main advantages of
e2tree: - building an explainable tree that ensures the predictive
performance of an RF model - allowing the decision-maker to manage with
an intuitive structure (such as a tree-like structure).

In this example, we focus on Random Forest but, again, the algorithm can
be generalized to every ensemble approach based on decision trees.

## Setup

You can install the **developer version** of e2tree from
[GitHub](https://github.com) with:

``` r
install.packages("remotes")
remotes::install_github("massimoaria/e2tree")
```

You can install the **released version** of e2tree from
[CRAN](https://CRAN.R-project.org) with:

``` r
if (!require("e2tree", quietly=TRUE)) install.packages("e2tree")
```

``` r
require(e2tree)
require(randomForest)
require(ranger)
require(dplyr)
require(ggplot2)
if (!(require(rsample, quietly=TRUE))){install.packages("rsample"); require(rsample, quietly=TRUE)}
#> 
#> The downloaded binary packages are in
#>  /var/folders/lh/7mrcjmgx18l44jd6m848qrjc0000gn/T//RtmpEumgLu/downloaded_packages
options(dplyr.summarise.inform = FALSE)
```

## S3 Classes and Methods

The **e2tree** package uses a proper S3 class system. The main classes
and their associated methods are:

| Class | Methods |
|----|----|
| `e2tree` | `print`, `summary`, `plot`, `predict`, `fitted`, `residuals`, `as.rpart`, `nodes`, `e2splits` |
| `eValidation` | `print`, `summary`, `plot`, `measures`, `proximity` |
| `loi` | `print`, `summary`, `plot` |
| `loi_perm` | `print`, `summary`, `plot` |

E2Tree objects can also be converted to other formats for
interoperability:

- `as.rpart()` converts to `rpart` format for use with `rpart.plot`
- `as.party()` converts to `partykit`’s `constparty` format (if partykit
  is installed)

## Example 1: IRIS dataset (Classification)

Starting from the IRIS dataset, we train an ensemble tree using the
randomForest package and then use e2tree to obtain an explainable tree
synthesis of the ensemble classifier.

``` r
# Set random seed to make results reproducible:
set.seed(0)

# Initialize the split
iris_split <- iris %>% initial_split(prop = 0.6)
iris_split
#> <Training/Testing/Total>
#> <90/60/150>
# Assign the data to the correct sets
training <- iris_split %>% training()
validation <- iris_split %>% testing()
response_training <- training[,5]
response_validation <- validation[,5]
```

Train a Random Forest model with 1000 weak learners

``` r
# Perform training with "ranger" or "randomForest" package:
## RF with "ranger" package
ensemble <- ranger(Species ~ ., data = training, num.trees = 1000, importance = 'impurity')

## RF with "randomForest" package
#ensemble = randomForest(Species ~ ., data = training, importance = TRUE, proximity = TRUE)
```

Create the dissimilarity matrix between observations:

``` r
D = createDisMatrix(ensemble, data = training, label = "Species", parallel = list(active = FALSE, no_cores = NULL))
```

Build an explainable tree for RF:

``` r
setting=list(impTotal=0.1, maxDec=0.01, n=2, level=5)
tree <- e2tree(Species ~ ., data = training, D, ensemble, setting)
```

### S3 methods for e2tree objects

The `e2tree` class supports standard S3 methods for inspecting the
fitted model:

**Print** — compact model overview:

``` r
print(tree)
#> 
#>   Explainable Ensemble Tree (E2Tree)
#>   -----------------------------------
#>   Task:            Classification
#>   Response:        Species
#>   Predictors:      4 (Sepal.Length, Sepal.Width, Petal.Length, Petal.Width)
#>   Observations:    90
#>   Nodes:           11 (total), 6 (terminal)
#>   Max depth:       5
#>   Split variables: Petal.Length, Petal.Width
#>   Classes:         setosa, versicolor, virginica
```

**Summary** — full model details including terminal nodes and decision
rules:

``` r
summary(tree)
#> 
#> ====================================================================== 
#>                     E2TREE MODEL SUMMARY
#> ====================================================================== 
#> 
#> MODEL INFORMATION
#> ---------------------------------------- 
#>   Task:              Classification
#>   Response:          Species
#>   Observations:      90
#>   Total Nodes:       11
#>   Terminal Nodes:    6
#>   Max Depth:         5
#>   Split Variables:   Petal.Length, Petal.Width
#>   Classes:           setosa, versicolor, virginica
#> 
#> 
#> TERMINAL NODES
#> ---------------------------------------------------------------------- 
#>   Node      Prediction            n  Purity      Wt
#> ------------------------------------------------------- 
#>   2         setosa               33  100.0%      --
#>   12        versicolor           22  100.0%      --
#>   26        versicolor            2  100.0%      --
#>   54        versicolor            2   50.0%      --
#>   55        virginica             2  100.0%      --
#>   7         virginica            29  100.0%      --
#> 
#> 
#> DECISION RULES
#> ---------------------------------------------------------------------- 
#> 
#> Rule 1 (Node 2, n=33):
#>     IF   Petal.Length <=1.9
#>   THEN: setosa
#> 
#> Rule 2 (Node 12, n=22):
#>     IF   Petal.Length >1.9
#>     AND  Petal.Width <=1.7
#>     AND  Petal.Length <=4.7
#>   THEN: versicolor
#> 
#> Rule 3 (Node 26, n=2):
#>     IF   Petal.Length >1.9
#>     AND  Petal.Width <=1.7
#>     AND  Petal.Length >4.7
#>     AND  Petal.Length <=5
#>   THEN: versicolor
#> 
#> Rule 4 (Node 54, n=2):
#>     IF   Petal.Length >1.9
#>     AND  Petal.Width <=1.7
#>     AND  Petal.Length >4.7
#>     AND  Petal.Length >5
#>     AND  Petal.Length <=5.1
#>   THEN: versicolor
#> 
#> Rule 5 (Node 55, n=2):
#>     IF   Petal.Length >1.9
#>     AND  Petal.Width <=1.7
#>     AND  Petal.Length >4.7
#>     AND  Petal.Length >5
#>     AND  Petal.Length >5.1
#>   THEN: virginica
#> 
#> Rule 6 (Node 7, n=29):
#>     IF   Petal.Length >1.9
#>     AND  Petal.Width >1.7
#>   THEN: virginica
#> 
#> ======================================================================
```

**Plot** — tree visualization via `rpart.plot`:

``` r
plot(tree, ensemble)
```

<img src="man/figures/README-unnamed-chunk-11-1.png" alt="" width="100%" />

### Accessor functions

Accessor functions provide a clean interface to extract components
without exposing the internal structure:

``` r
# Extract terminal nodes
nodes(tree, terminal = TRUE)
#>    node  n       pred prob  impTotal impChildren decImp decImpSur variable
#> 2     2 33     setosa  1.0 0.0222880          NA     NA        NA     <NA>
#> 12   12 22 versicolor  1.0 0.2026426          NA     NA        NA     <NA>
#> 26   26  2 versicolor  1.0 0.5207338          NA     NA        NA     <NA>
#> 54   54  2 versicolor  0.5 0.5794982          NA     NA        NA     <NA>
#> 55   55  2  virginica  1.0 0.5404635          NA     NA        NA     <NA>
#> 7     7 29  virginica  1.0 0.1514348          NA     NA        NA     <NA>
#>    split splitLabel variableSur splitLabelSur parent children terminal
#> 2     NA       <NA>        <NA>          <NA>      1       NA     TRUE
#> 12    NA       <NA>        <NA>          <NA>      6       NA     TRUE
#> 26    NA       <NA>        <NA>          <NA>     13       NA     TRUE
#> 54    NA       <NA>        <NA>          <NA>     27       NA     TRUE
#> 55    NA       <NA>        <NA>          <NA>     27       NA     TRUE
#> 7     NA       <NA>        <NA>          <NA>      3       NA     TRUE
#>                                                                                                                                obs
#> 2  4, 5, 8, 11, 14, 17, 21, 23, 26, 27, 29, 30, 32, 35, 37, 39, 42, 43, 44, 46, 47, 48, 49, 57, 60, 62, 64, 72, 73, 80, 81, 84, 85
#> 12                                             2, 6, 7, 10, 13, 20, 24, 33, 51, 54, 55, 56, 61, 65, 68, 75, 76, 77, 83, 86, 87, 90
#> 26                                                                                                                          12, 50
#> 54                                                                                                                          22, 79
#> 55                                                                                                                          69, 89
#> 7                  1, 3, 9, 15, 16, 18, 19, 25, 28, 31, 34, 36, 38, 40, 41, 45, 52, 53, 58, 59, 63, 66, 67, 70, 71, 74, 78, 82, 88
#>                                                                                                       path
#> 2                                                                                       Petal.Length <=1.9
#> 12                                            !Petal.Length <=1.9 & Petal.Width <=1.7 & Petal.Length <=4.7
#> 26                        !Petal.Length <=1.9 & Petal.Width <=1.7 & !Petal.Length <=4.7 & Petal.Length <=5
#> 54  !Petal.Length <=1.9 & Petal.Width <=1.7 & !Petal.Length <=4.7 & !Petal.Length <=5 & Petal.Length <=5.1
#> 55 !Petal.Length <=1.9 & Petal.Width <=1.7 & !Petal.Length <=4.7 & !Petal.Length <=5 & !Petal.Length <=5.1
#> 7                                                                 !Petal.Length <=1.9 & !Petal.Width <=1.7
#>    ncat pred_val    yval2.V1    yval2.V2    yval2.V3    yval2.V4    yval2.V5
#> 2    NA        1  1.00000000 33.00000000  0.00000000  0.00000000  1.00000000
#> 12   NA        2  2.00000000  0.00000000 22.00000000  0.00000000  0.00000000
#> 26   NA        2  2.00000000  0.00000000  2.00000000  0.00000000  0.00000000
#> 54   NA        2  2.00000000  0.00000000  1.00000000  1.00000000  0.00000000
#> 55   NA        3  3.00000000  0.00000000  0.00000000  2.00000000  0.00000000
#> 7    NA        3  3.00000000  0.00000000  0.00000000 29.00000000  0.00000000
#>       yval2.V6    yval2.V7 yval2.nodeprob
#> 2   0.00000000  0.00000000     0.36666667
#> 12  1.00000000  0.00000000     0.24444444
#> 26  1.00000000  0.00000000     0.02222222
#> 54  0.50000000  0.50000000     0.02222222
#> 55  0.00000000  1.00000000     0.02222222
#> 7   0.00000000  1.00000000     0.32222222

# Extract split information
str(e2splits(tree), max.level = 1)
#> List of 2
#>  $ splits: num [1:5, 1:5] 90 57 28 6 4 -1 -1 -1 -1 -1 ...
#>   ..- attr(*, "dimnames")=List of 2
#>  $ csplit: NULL
```

### Coercion to other formats

E2Tree objects can be converted to standard tree formats for use with
other packages:

``` r
# Convert to rpart format
rpart_obj <- as.rpart(tree, ensemble)

# Convert to partykit format (if installed)
if (requireNamespace("partykit", quietly = TRUE)) {
  party_obj <- partykit::as.party(tree)
  plot(party_obj)
}
```

<img src="man/figures/README-unnamed-chunk-13-1.png" alt="" width="100%" />

### Prediction

Use the standard `predict()` method for prediction on new data:

``` r
# Predict on validation set
pred <- predict(tree, newdata = validation, target = "virginica")
head(pred)
#>      fit accuracy score
#> 1 setosa        1     0
#> 2 setosa        1     0
#> 3 setosa        1     0
#> 4 setosa        1     0
#> 5 setosa        1     0
#> 6 setosa        1     0
```

Comparison of predictions (training sample) of RF and e2tree

``` r
# Training predictions
pred_train <- predict(tree, newdata = training, target = "virginica")

# "ranger" package
table(pred_train$fit, ensemble$predictions)
#>             
#>              setosa versicolor virginica
#>   setosa         33          0         0
#>   versicolor      0         23         3
#>   virginica       0          1        30

# "randomForest" package
#table(pred_train$fit, ensemble$predicted)
```

Comparison of predictions (training sample) of RF and correct response

``` r
# "ranger" package
table(ensemble$predictions, response_training)
#>             response_training
#>              setosa versicolor virginica
#>   setosa         33          0         0
#>   versicolor      0         22         2
#>   virginica       0          3        30

## "randomForest" package
#table(ensemble$predicted, response_training)
```

Comparison of predictions (training sample) of e2tree and correct
response

``` r
table(pred_train$fit, response_training)
#>             response_training
#>              setosa versicolor virginica
#>   setosa         33          0         0
#>   versicolor      0         25         1
#>   virginica       0          0        31
```

Fitted values for the training data:

``` r
head(fitted(tree))
#> [1] "virginica"  "versicolor" "virginica"  "setosa"     "setosa"    
#> [6] "versicolor"
```

### Variable importance

Variable importance is automatically detected as classification or
regression:

``` r
V <- vimp(tree, training)
V$vimp
#> # A tibble: 2 × 9
#>   Variable     MeanImpurityDecrease MeanAccuracyDecrease `ImpDec_ setosa`
#>   <chr>                       <dbl>                <dbl>            <dbl>
#> 1 Petal.Length                0.365             2.22e- 2            0.317
#> 2 Petal.Width                 0.214             1.41e-16           NA    
#> # ℹ 5 more variables: `ImpDec_ versicolor` <dbl>, `ImpDec_ virginica` <dbl>,
#> #   `AccDec_ setosa` <dbl>, `AccDec_ versicolor` <dbl>,
#> #   `AccDec_ virginica` <dbl>
V$g_imp
```

<img src="man/figures/README-unnamed-chunk-19-1.png" alt="" width="100%" />

``` r
V$g_acc
```

<img src="man/figures/README-unnamed-chunk-19-2.png" alt="" width="100%" />

### Prediction on validation sample

``` r
ensemble.pred <- predict(ensemble, validation[,-5])

pred_val <- predict(tree, newdata = validation, target = "virginica")
```

Comparison of predictions (validation sample) of RF and e2tree

``` r
## "ranger" package
table(pred_val$fit, ensemble.pred$predictions)
#>             
#>              setosa versicolor virginica
#>   setosa         17          0         0
#>   versicolor      0         26         0
#>   virginica       0          0        17

## "randomForest" package
#table(pred_val$fit, ensemble.pred$predicted)
```

Comparison of predictions (validation sample) of e2tree and correct
response

``` r
table(pred_val$fit, response_validation)
#>             response_validation
#>              setosa versicolor virginica
#>   setosa         17          0         0
#>   versicolor      0         24         2
#>   virginica       0          1        16
roc_res <- roc(response_validation, pred_val$score, target="virginica")
```

<img src="man/figures/README-unnamed-chunk-22-1.png" alt="" width="100%" />

``` r
roc_res$auc
#> [1] 0.9325397
```

## Validation of the E2Tree Structure

A critical question when using E2Tree is: *how well does the single tree
capture the structure of the original ensemble?*

Assessing the fidelity of this reconstruction requires measuring
**agreement** between the ensemble and E2Tree proximity matrices — a
fundamentally different question from measuring their **association**.
The distinction parallels the classical one between *correlation* and
*concordance* in method comparison studies (Bland & Altman, 1986; Lin,
1989): two proximity matrices can be perfectly correlated yet
systematically disagree in their actual values. The Mantel test, being
scale-invariant, would declare perfect association in such a case. But
for E2Tree validation, we need to know whether the *actual proximity
values* are faithfully reproduced.

The `eValidation()` function supports two approaches via the `test`
argument:

- `test = "mantel"`: The classical Mantel test for *association*
- `test = "measures"`: A family of divergence/similarity measures for
  *agreement*
- `test = "both"` (default): Both approaches

### Divergence and similarity measures

| Measure | Type | Range | What it measures |
|----|----|----|----|
| **nLoI** | divergence | \[0, 1\] | Normalized Loss of Interpretability — weighted divergence with diagnostic decomposition |
| **Hellinger** | divergence | \[0, 1\] | Hellinger distance — robust to sparse matrices |
| **wRMSE** | divergence | \[0, 1\] | Weighted RMSE — emphasizes high-proximity regions |
| **RV** | similarity | \[0, 1\] | RV coefficient — global structural similarity (scale-invariant) |
| **SSIM** | similarity | \[-1, 1\] | Structural Similarity Index — captures local block patterns |

All measures are tested simultaneously using a **unified row/column
permutation test**.

### Running the validation

``` r
val <- eValidation(training, tree, D, test = "both", graph = FALSE, n_perm = 999, seed = 42)
```

**Print** — compact results with Mantel test and all measures:

``` r
print(val)
#> 
#> ##############################################################################
#>    E2Tree Validation
#> ##############################################################################
#> 
#>   Matrix dimension:    90 x 90
#>   Pairs:               4005
#> 
#>   Mantel test:         z = 1046.09, p = 0.0010
#> 
#> ------------------------------------------------------------------------------
#>   Measure          Type        Observed    Null mean    Z-stat     p-value
#> ------------------------------------------------------------------------------
#>   nLoI           [div]     0.0480     0.4126     -56.63   0.0010 ** 
#>   Hellinger      [div]     0.2154     0.6250     -81.19   0.0010 ** 
#>   wRMSE          [div]     0.1700     0.8406    -142.36   0.0010 ** 
#>   RV             [sim]     0.9674     0.3294     +57.04   0.0010 ** 
#>   SSIM           [sim]     0.6420     0.0088    +116.47   0.0010 ** 
#> ------------------------------------------------------------------------------
#>   Permutations: 999 (row/column), conf.level: 95%
#> 
#>   LoI Decomposition (per-pair avg):  mean_in = 0.007366,  mean_out = 0.064683
#> 
#>   [div] = divergence (lower=better), [sim] = similarity (higher=better)
#>   Signif. codes: 0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1
#> 
#> ##############################################################################
```

**Summary** — includes the LoI diagnostic decomposition:

``` r
summary(val)
#> 
#> ##############################################################################
#>    E2Tree Validation
#> ##############################################################################
#> 
#>   Matrix dimension:    90 x 90
#>   Pairs:               4005
#> 
#>   Mantel test:         z = 1046.09, p = 0.0010
#> 
#> ------------------------------------------------------------------------------
#>   Measure          Type        Observed    Null mean    Z-stat     p-value
#> ------------------------------------------------------------------------------
#>   nLoI           [div]     0.0480     0.4126     -56.63   0.0010 ** 
#>   Hellinger      [div]     0.2154     0.6250     -81.19   0.0010 ** 
#>   wRMSE          [div]     0.1700     0.8406    -142.36   0.0010 ** 
#>   RV             [sim]     0.9674     0.3294     +57.04   0.0010 ** 
#>   SSIM           [sim]     0.6420     0.0088    +116.47   0.0010 ** 
#> ------------------------------------------------------------------------------
#>   Permutations: 999 (row/column), conf.level: 95%
#> 
#>   LoI Decomposition (per-pair avg):  mean_in = 0.007366,  mean_out = 0.064683
#> 
#>   [div] = divergence (lower=better), [sim] = similarity (higher=better)
#>   Signif. codes: 0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1
#> 
#> ##############################################################################
#> 
#> 
#> 
#> ##############################################################################
#>    Loss of Interpretability (LoI) --Decomposition
#> ##############################################################################
#> 
#>   nLoI (normalized):   0.0480
#>   LoI (raw):           192.1100
#>   n = 90, pairs = 4005 (within: 1168, between: 2837)
#> 
#> ------------------------------------------------------------------------------
#>   Component              Total       Pairs     Mean/pair
#> ------------------------------------------------------------------------------
#>   Within-node (LoI_in)    8.6039       1168     0.007366
#>   Between-node (LoI_out)183.5062       2837     0.064683
#> ------------------------------------------------------------------------------
#> 
#>   Per-pair interpretation (comparable across components):
#> 
#>     mean_in  = 0.007366  (avg calibration error within nodes)
#>     mean_out = 0.064683  (avg ensemble proximity lost by separation)
#> 
#>   Diagnostic: LOW mean_out. The partition correctly separates pairs
#>   that have low ensemble proximity --the tree structure is well-placed.
#> 
#>   Diagnostic: LOW mean_in. Within-node proximity values closely
#>   match the ensemble --excellent calibration.
#> 
#> ##############################################################################
```

**Plot** — heatmaps, null distribution, and LoI decomposition:

``` r
plot(val)
```

<img src="man/figures/README-unnamed-chunk-26-1.png" alt="" width="100%" />

### Extracting results with accessors

Use accessor functions instead of direct `$` access:

``` r
# Validation measures table
measures(val)
#>      method       type   observed   null_mean     null_sd     z_stat p_value
#> 1      nLoI divergence 0.04796755 0.412605505 0.006438700  -56.63223   0.001
#> 2 Hellinger divergence 0.21542068 0.625021454 0.005044737  -81.19368   0.001
#> 3     wRMSE divergence 0.16995523 0.840584726 0.004710777 -142.36068   0.001
#> 4        RV similarity 0.96743746 0.329429934 0.011185735   57.03760   0.001
#> 5      SSIM similarity 0.64198676 0.008751186 0.005436671  116.47487   0.001
#>     ci_lower  ci_upper
#> 1 0.04013923 0.0626473
#> 2 0.20911191 0.2271914
#> 3 0.16400739 0.1810256
#> 4 0.94219513 0.9816048
#> 5 0.62833640 0.6503210

# Proximity matrices
prox <- proximity(val, type = "both")
str(prox, max.level = 1)
#> List of 2
#>  $ ensemble: num [1:90, 1:90] 1 0.947 0.951 0.917 0.92 ...
#>   ..- attr(*, "dimnames")=List of 2
#>  $ e2tree  : num [1:90, 1:90] 1 1 1 1 1 1 1 1 1 1 ...
#>   ..- attr(*, "dimnames")=List of 2
```

### The nLoI Decomposition

The nLoI is unique among the measures because it decomposes into two
interpretable components:

- **LoI_in** (within-node): measures how well the E2Tree reproduces the
  ensemble’s proximity values for pairs it groups *together*.

- **LoI_out** (between-node): measures the ensemble proximity lost for
  pairs that E2Tree *separates* into different nodes.

Since the number of within-node and between-node pairs can differ
dramatically, the `loi()` function reports **per-pair averages**
(`mean_in` and `mean_out`) that enable meaningful comparison:

``` r
O <- proximity(val, type = "ensemble")
O_hat <- proximity(val, type = "e2tree")

result <- loi(O, O_hat)
summary(result)
#> 
#> ##############################################################################
#>    Loss of Interpretability (LoI) --Decomposition
#> ##############################################################################
#> 
#>   nLoI (normalized):   0.0480
#>   LoI (raw):           192.1100
#>   n = 90, pairs = 4005 (within: 1168, between: 2837)
#> 
#> ------------------------------------------------------------------------------
#>   Component              Total       Pairs     Mean/pair
#> ------------------------------------------------------------------------------
#>   Within-node (LoI_in)    8.6039       1168     0.007366
#>   Between-node (LoI_out)183.5062       2837     0.064683
#> ------------------------------------------------------------------------------
#> 
#>   Per-pair interpretation (comparable across components):
#> 
#>     mean_in  = 0.007366  (avg calibration error within nodes)
#>     mean_out = 0.064683  (avg ensemble proximity lost by separation)
#> 
#>   Diagnostic: LOW mean_out. The partition correctly separates pairs
#>   that have low ensemble proximity --the tree structure is well-placed.
#> 
#>   Diagnostic: LOW mean_in. Within-node proximity values closely
#>   match the ensemble --excellent calibration.
#> 
#> ##############################################################################
```

The per-pair averages provide actionable diagnostics:

- **mean_out \> 0.3**: the tree is splitting apart pairs with
  substantial ensemble proximity — consider more terminal nodes
- **mean_out \< 0.1**: the partition correctly separates low-proximity
  pairs — tree structure is well-placed
- **mean_in \> 0.1**: within-node calibration error is high — check
  proximity estimation
- **mean_in \< 0.01**: excellent within-node match between E2Tree and
  ensemble

### Standalone LoI permutation test

For a quick significance assessment:

``` r
perm <- loi_perm(O, O_hat, n_perm = 999, seed = 42)
print(perm)
#> 
#> ==============================================================================
#>    Permutation Test for Loss of Interpretability (LoI)
#> ==============================================================================
#> 
#>   Observed nLoI:       0.0480
#>   Null mean:           0.4128
#>   Null SD:             0.0062
#>   Z-statistic:         -58.5019
#> 
#> ------------------------------------------------------------------------------
#>   Hypothesis Test (H1: nLoI < expected by chance)
#> ------------------------------------------------------------------------------
#>   p-value:             0.0010 **
#>   95% CI:             [0.0399, 0.0646]
#>   Permutations:        999 (row/column)
#> 
#> ------------------------------------------------------------------------------
#>   Decomposition (per-pair averages)
#> ------------------------------------------------------------------------------
#>   mean_in  (within):   0.007366  (n_pairs = 1168)
#>   mean_out (between):  0.064683  (n_pairs = 2837)
#> 
#> ==============================================================================
#>   Signif. codes: 0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1
```

``` r
plot(perm)
```

<img src="man/figures/README-unnamed-chunk-30-1.png" alt="" width="100%" />

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE   # examples require optional packages; set eval=TRUE locally
)

## ----xgb-clf------------------------------------------------------------------
# library(e2tree)
# if (!require("xgboost")) install.packages("xgboost",
#                                                repos="https://cran.r-project.org")
# library(xgboost)
# 
# data(iris)
# set.seed(42)
# n  <- floor(0.75 * nrow(iris))
# tr <- iris[sample(nrow(iris), n), ]
# va <- iris[setdiff(seq_len(nrow(iris)), as.integer(rownames(tr))), ]
# 
# # XGBoost requires a numeric matrix and 0-indexed integer labels
# X_tr <- as.matrix(tr[, 1:4])
# y_tr <- as.integer(tr$Species) - 1L
# dm_tr <- xgb.DMatrix(data = X_tr, label = y_tr)
# 
# ensemble <- xgb.train(
#   params  = list(objective  = "multi:softmax",
#                  num_class  = 3,
#                  max_depth  = 4,
#                  eta        = 0.3),
#   data    = dm_tr,
#   nrounds = 100,
#   verbose = 0
# )
# 
# # Attach the response back to the data.frame so the formula in e2tree()
# # can find it; createDisMatrix() will use it to annotate the dissimilarity
# # matrix (in classification, `label` is optional but recommended).
# tr_xgb         <- tr[, 1:4]
# tr_xgb$Species <- tr$Species
# 
# D <- createDisMatrix(ensemble, data = tr_xgb, label = "Species",
#                      parallel = list(active = FALSE, no_cores = 1))
# 
# setting  <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
# tree_xgb <- e2tree(Species ~ ., data = tr_xgb, D = D,
#                    ensemble = ensemble, setting = setting)
# print(tree_xgb)

## ----xgb-reg------------------------------------------------------------------
# library(xgboost)
# 
# data(mtcars)
# set.seed(42)
# n  <- floor(0.75 * nrow(mtcars))
# tr <- mtcars[sample(nrow(mtcars), n), ]
# 
# X_tr  <- as.matrix(tr[, -1])
# y_tr  <- tr$mpg
# dm_tr <- xgb.DMatrix(data = X_tr, label = y_tr)
# 
# ensemble <- xgb.train(
#   params  = list(objective = "reg:squarederror", max_depth = 4, eta = 0.3),
#   data    = dm_tr,
#   nrounds = 100,
#   verbose = 0
# )
# 
# # `data = tr` carries the response column too; the XGBoost adapter
# # automatically trims the matrix to the features used at training time.
# D    <- createDisMatrix(ensemble, data = tr, label = "mpg",
#                         parallel = list(active = FALSE, no_cores = 1))
# tree <- e2tree(mpg ~ ., data = tr, D = D, ensemble = ensemble,
#                setting = list(impTotal = 0.1, maxDec = 1e-6, n = 2, level = 5))
# print(tree)

## ----gbm-clf------------------------------------------------------------------
# if (!require("gbm")) install.packages("gbm",
#                                                repos="https://cran.r-project.org")
# library(gbm)
# 
# data(iris)
# set.seed(42)
# df <- iris
# df$is_setosa     <- as.integer(df$Species == "setosa")
# df$is_setosa_fct <- factor(df$is_setosa, levels = c(0L, 1L))
# n  <- floor(0.75 * nrow(df))
# tr <- df[sample(nrow(df), n), ]
# 
# ensemble <- gbm(is_setosa ~ Sepal.Length + Sepal.Width + Petal.Length + Petal.Width,
#                 data              = tr,
#                 distribution      = "bernoulli",
#                 n.trees           = 200,
#                 interaction.depth = 4,
#                 verbose           = FALSE)
# 
# D    <- createDisMatrix(ensemble,
#                         data     = tr[, c("Sepal.Length","Sepal.Width",
#                                           "Petal.Length","Petal.Width",
#                                           "is_setosa_fct")],
#                         label    = "is_setosa_fct",
#                         parallel = list(active = FALSE, no_cores = 1))
# tree <- e2tree(is_setosa_fct ~ Sepal.Length + Sepal.Width +
#                                Petal.Length + Petal.Width,
#                data = tr, D = D, ensemble = ensemble,
#                setting = list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5))
# print(tree)

## ----gbm-reg------------------------------------------------------------------
# library(gbm)
# 
# data(mtcars)
# set.seed(42)
# n  <- floor(0.75 * nrow(mtcars))
# tr <- mtcars[sample(nrow(mtcars), n), ]
# 
# ensemble <- gbm(mpg ~ ., data = tr,
#                 distribution      = "gaussian",
#                 n.trees           = 200,
#                 interaction.depth = 4,
#                 n.minobsinnode    = 2,
#                 bag.fraction      = 0.8,
#                 verbose           = FALSE)
# 
# D    <- createDisMatrix(ensemble, data = tr, label = "mpg",
#                         parallel = list(active = FALSE, no_cores = 1))
# tree <- e2tree(mpg ~ ., data = tr, D = D, ensemble = ensemble,
#                setting = list(impTotal = 0.1, maxDec = 1e-6, n = 2, level = 5))
# print(tree)

## ----lgb-clf------------------------------------------------------------------
# if (!require("lightgbm")) install.packages("lightgbm",
#                                                repos="https://cran.r-project.org")
# library(lightgbm)
# 
# data(iris)
# set.seed(42)
# n  <- floor(0.75 * nrow(iris))
# tr <- iris[sample(nrow(iris), n), ]
# 
# X_tr <- as.matrix(tr[, 1:4])
# y_tr <- as.integer(tr$Species) - 1L
# ds   <- lgb.Dataset(X_tr, label = y_tr)
# 
# ensemble <- lgb.train(
#   params  = list(objective  = "multiclass",
#                  num_class  = 3,
#                  num_leaves = 15,
#                  verbose    = -1),
#   data    = ds,
#   nrounds = 100
# )
# 
# tr_lgb         <- tr[, 1:4]
# tr_lgb$Species <- tr$Species
# 
# D <- createDisMatrix(ensemble, data = tr_lgb, label = "Species",
#                      parallel = list(active = FALSE, no_cores = 1))
# 
# tree <- e2tree(Species ~ ., data = tr_lgb, D = D, ensemble = ensemble,
#                setting = list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5))
# print(tree)

## ----lgb-reg------------------------------------------------------------------
# library(lightgbm)
# 
# data(mtcars)
# set.seed(42)
# n  <- floor(0.75 * nrow(mtcars))
# tr <- mtcars[sample(nrow(mtcars), n), ]
# 
# X_tr <- as.matrix(tr[, -1])
# y_tr <- tr$mpg
# ds   <- lgb.Dataset(X_tr, label = y_tr)
# 
# ensemble <- lgb.train(
#   params  = list(objective        = "regression",
#                  num_leaves       = 8,
#                  min_data_in_leaf = 2,
#                  learning_rate    = 0.1,
#                  verbose          = -1),
#   data    = ds,
#   nrounds = 200
# )
# 
# # Pass the response column to createDisMatrix() via `label`. The
# # LightGBM adapter selects the columns it needs through the booster's
# # stored feature names, so any extra columns in `data` are ignored.
# D    <- createDisMatrix(ensemble, data = tr, label = "mpg",
#                         parallel = list(active = FALSE, no_cores = 1))
# tree <- e2tree(mpg ~ ., data = tr, D = D, ensemble = ensemble,
#                setting = list(impTotal = 0.1, maxDec = 1e-6, n = 2, level = 5))
# print(tree)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)

## ----eval = FALSE-------------------------------------------------------------
# # Install from CRAN (when available):
# install.packages("e2tree")
# 
# # Or install the development version from GitHub:
# # devtools::install_github("massimoaria/e2tree")

## ----classification-setup-----------------------------------------------------
library(e2tree)

data(iris)
set.seed(42)
smp_size <- floor(0.75 * nrow(iris))
train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
training <- iris[train_ind, ]
validation <- iris[-train_ind, ]

## ----classification-ensemble--------------------------------------------------
if (!require("randomForest")) install.packages("randomForest", 
                                               repos="https://cran.r-project.org")

library(randomForest)
  
ensemble <- randomForest(Species ~ ., data = training,
                         importance = TRUE, proximity = TRUE)
ensemble

## ----classification-dismatrix-------------------------------------------------
D <- createDisMatrix(ensemble, data = training, label = "Species",
                     parallel = list(active = FALSE, no_cores = 1))
dim(D)

## ----classification-tree------------------------------------------------------
setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
tree <- e2tree(Species ~ ., training, D, ensemble, setting)

## ----classification-print-----------------------------------------------------
print(tree)

## ----classification-summary---------------------------------------------------
summary(tree)

## ----classification-nodes-----------------------------------------------------
# Terminal nodes only
nodes(tree, terminal = TRUE)

## ----classification-plot, fig.width = 8, fig.height = 6-----------------------
plot(tree, ensemble = ensemble, main = "E2Tree - Iris Classification")

## ----classification-rpart, fig.width = 8, fig.height = 6----------------------
# Convert to rpart
rpart_obj <- as.rpart(tree, ensemble)

# Convert to partykit (if installed)
if (requireNamespace("partykit", quietly = TRUE)) {
  party_obj <- partykit::as.party(tree)
  plot(party_obj)
}

## ----classification-predict---------------------------------------------------
pred <- predict(tree, newdata = validation)
head(pred)

## ----classification-fitted----------------------------------------------------
head(fitted(tree))

## ----classification-vimp------------------------------------------------------
vi <- vimp(tree, data = training)
vi$vimp
vi$g_imp

## ----regression-setup---------------------------------------------------------
data(mtcars)
set.seed(123)
smp_size <- floor(0.75 * nrow(mtcars))
train_ind <- sample(seq_len(nrow(mtcars)), size = smp_size)
training_reg <- mtcars[train_ind, ]
validation_reg <- mtcars[-train_ind, ]

## ----regression-ensemble------------------------------------------------------
ensemble_reg <- randomForest(mpg ~ ., data = training_reg, ntree = 500,
                             importance = TRUE, proximity = TRUE)

## ----regression-dismatrix-----------------------------------------------------
D_reg <- createDisMatrix(ensemble_reg, data = training_reg, label = "mpg",
                         parallel = list(active = FALSE, no_cores = 1))

## ----regression-tree----------------------------------------------------------
setting_reg <- list(impTotal = 0.1, maxDec = 1e-6, n = 2, level = 5)
tree_reg <- e2tree(mpg ~ ., training_reg, D_reg, ensemble_reg, setting_reg)
print(tree_reg)

## ----regression-plot, fig.width = 8, fig.height = 6---------------------------
plot(tree_reg, ensemble = ensemble_reg, main = "E2Tree - MPG Regression")

## ----regression-predict-------------------------------------------------------
pred_reg <- predict(tree_reg, newdata = validation_reg)
head(pred_reg)

## ----regression-residuals-----------------------------------------------------
res <- residuals(tree_reg)
summary(res)

## ----val-mantel---------------------------------------------------------------
val_mantel <- eValidation(training, tree, D, test = "mantel", graph = FALSE)
print(val_mantel)

## ----val-measures-------------------------------------------------------------
val_meas <- eValidation(training, tree, D, test = "measures",
                        graph = FALSE, n_perm = 499)
print(val_meas)

## ----val-measures-table-------------------------------------------------------
measures(val_meas)

## ----val-plot, fig.width = 8, fig.height = 7----------------------------------
val_full <- eValidation(training, tree, D, test = "both",
                        graph = FALSE, n_perm = 499)
plot(val_full)

## ----loi-classification-------------------------------------------------------
prox <- proximity(val_full)
result <- loi(prox$ensemble, prox$e2tree)
summary(result)

## ----loi-plot-----------------------------------------------------------------
plot(result)

## ----loi-perm-----------------------------------------------------------------
perm <- loi_perm(prox$ensemble, prox$e2tree, n_perm = 499, seed = 42)
print(perm)
plot(perm)

## ----val-regression-----------------------------------------------------------
val_reg <- eValidation(training_reg, tree_reg, D_reg, test = "measures",
                       graph = FALSE, n_perm = 499)
print(val_reg)

## ----session-info-------------------------------------------------------------
sessionInfo()

utils::globalVariables(".")
#' @import stats
#' @import parallel
#' @import ggplot2
#' @import future.apply
#' @import purrr
#' @import partitions
#' @import rpart.plot
#' @importFrom Rcpp sourceCpp evalCpp
#' @importFrom dplyr %>%
#' @importFrom dplyr if_else
#' @importFrom dplyr distinct
#' @importFrom dplyr pick
#' @importFrom dplyr row_number
#' @importFrom dplyr across
#' @importFrom dplyr tibble
#' @importFrom dplyr as_tibble
#' @importFrom dplyr first
#' @importFrom dplyr arrange
#' @importFrom dplyr do
#' @importFrom dplyr n
#' @importFrom dplyr count
#' @importFrom dplyr desc
#' @importFrom dplyr group_by
#' @importFrom dplyr mutate
#' @importFrom dplyr mutate_if
#' @importFrom dplyr rename
#' @importFrom dplyr rename_with
#' @importFrom dplyr summarize
#' @importFrom dplyr select
#' @importFrom dplyr left_join
#' @importFrom dplyr top_n
#' @importFrom dplyr relocate
#' @importFrom dplyr slice_head
#' @importFrom dplyr slice_tail
#' @importFrom dplyr starts_with
#' @importFrom dplyr group_by_at
#' @importFrom tidyr complete
#' @importFrom tidyr replace_na
#' @importFrom tidyr drop_na
#' @importFrom tidyr pivot_longer
#' @importFrom tidyr pivot_wider
#' @importFrom Matrix %&%
#' @importFrom Matrix all.equal
#' @importFrom Matrix as.array
#' @importFrom Matrix as.matrix
#' @importFrom Matrix Matrix
#' @importFrom Matrix rowSums
#' @importFrom Matrix sparseMatrix
#' @importFrom grDevices colorRampPalette dev.cur dev.new devAskNewPage rainbow
#' @importFrom graphics abline axis box hist image legend lines locator mtext par points
#' @importFrom utils globalVariables head setTxtProgressBar tail txtProgressBar
#' @useDynLib e2tree
#'
.onLoad <- function(libname, pkgname) {
  # Conditionally register as.party method for partykit
  register_partykit_method <- function(...) {
    if (requireNamespace("partykit", quietly = TRUE)) {
      registerS3method(
        "as.party",
        "e2tree",
        as.party.e2tree,
        envir = asNamespace("partykit")
      )
    }
  }
  setHook(packageEvent("partykit", "onLoad"), register_partykit_method)
  # Also register now if partykit is already loaded
  if (isNamespaceLoaded("partykit")) {
    registerS3method(
      "as.party",
      "e2tree",
      as.party.e2tree,
      envir = asNamespace("partykit")
    )
  }
}

.onAttach <- function(...) {
  packageStartupMessage(
    "Explainable Ensemble Trees (E2Tree) \n\n",
    "If you use e2tree in research, please cite: \n\n",
    "- Aria, M., Gnasso, A., Iorio, C., & Pandolfo, G. (2024). Explainable ensemble trees.\nComputational Statistics, 39(1), 3-19. DOI: 10.1007/s00180-022-01312-6\n\n",
    "- Aria, M., Gnasso, A., Iorio, C., & Fokkema, M. (2026). Extending Explainable Ensemble Trees to Regression Contexts. Applied Stochastic Models in Business and Industry, 42(1), e70064. DOI: 10.1002/asmb.70064\n\n"
  )
}

utils::globalVariables(c("node", "Y", "p", "variable", "decImp", "splitLabel", "ID", "index", "Wt", "prob")) # to avoid CRAN check errors for tidyverse programming

#' Explainable Ensemble Tree
#'
#' It creates an explainable tree for Random Forest. Explainable Ensemble Trees (E2Tree) aimed to generate a “new tree” that can explain and represent the relational structure between the response variable and the predictors. This lead to providing a tree structure similar to those obtained for a decision tree exploiting the advantages of a dendrogram-like output.
#'
#' @param formula is a formula describing the model to be fitted, with a response but no interaction terms.
#' @param data a data frame containing the variables in the model. It is a data frame in which to interpret the variables named in the formula.
#' @param D is the dissimilarity matrix. This is a dissimilarity matrix measuring the discordance between two observations concerning a given classifier of a random forest model. The dissimilarity matrix is obtained with the \link{createDisMatrix} function.
#' @param ensemble is an ensemble tree object (for the moment ensemble works only with random forest objects)
#' @param setting is a list containing the set of stopping rules for the tree building procedure.
#' \tabular{lll}{
#' \code{impTotal}\tab   \tab The threshold for the impurity in the node\cr
#' \code{maxDec}\tab   \tab The threshold for the maximum impurity decrease of the node\cr
#' \code{n}\tab   \tab The minimum number of the observations in the node \cr
#' \code{level}\tab   \tab The maximum depth of the tree (levels) \cr}
#' Default is \code{setting=list(impTotal=0.1, maxDec=0.01, n=2, level=5)}.
#'
#' @return A e2tree object, which is a list with the following components:
#' \tabular{lll}{
#' \code{tree}\tab   \tab  A data frame representing the main structure of the tree aimed at explaining and graphically representing the relationships and interactions between the variables used to perform an ensemble method. \cr
#' \code{call}\tab   \tab The matched call\cr
#' \code{terms}\tab   \tab A list of terms and attributes \cr
#' \code{control}\tab   \tab A list containing the set of stopping rules for the tree building procedure  \cr
#' \code{varimp}\tab   \tab A list containing a table and a plot for the variable importance. Variable importance refers to a quantitative measure that assesses the contribution of individual variables within a predictive model towards accurate predictions. It quantifies the influence or impact that each variable has on the model's overall performance. Variable importance provides insights into the relative significance of different variables in explaining the observed outcomes and aids in understanding the underlying relationships and dynamics within the model \cr}
#'
#' @examples
#' \donttest{
#' ## Classification:
#' data(iris)
#'
#' # Create training and validation set:
#' smp_size <- floor(0.75 * nrow(iris))
#' train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
#' training <- iris[train_ind, ]
#' validation <- iris[-train_ind, ]
#' response_training <- training[,5]
#' response_validation <- validation[,5]
#'
#' # Perform training:
#' ## "randomForest" package
#' ensemble <- randomForest::randomForest(Species ~ ., data=training,
#' importance=TRUE, proximity=TRUE)
#'
#' ## "ranger" package
#' if (requireNamespace("ranger", quietly = TRUE)) {
#'   ensemble <- ranger::ranger(Species ~ ., data = iris,
#'     num.trees = 1000, importance = 'impurity')
#' }
#'
#' D <- createDisMatrix(ensemble, data=training, label = "Species",
#'                               parallel = list(active=FALSE, no_cores = 1))
#'
#' setting=list(impTotal=0.1, maxDec=0.01, n=2, level=5)
#' tree <- e2tree(Species ~ ., training, D, ensemble, setting)
#'
#'
#'
#' ## Regression
#' data("mtcars")
#'
#' # Create training and validation set:
#' smp_size <- floor(0.75 * nrow(mtcars))
#' train_ind <- sample(seq_len(nrow(mtcars)), size = smp_size)
#' training <- mtcars[train_ind, ]
#' validation <- mtcars[-train_ind, ]
#' response_training <- training[,1]
#' response_validation <- validation[,1]
#'
#' # Perform training
#' ## "randomForest" package
#' ensemble = randomForest::randomForest(mpg ~ ., data=training, ntree=1000,
#' importance=TRUE, proximity=TRUE)
#'
#' ## "ranger" package
#' if (requireNamespace("ranger", quietly = TRUE)) {
#'   ensemble <- ranger::ranger(formula = mpg ~ ., data = training,
#'     num.trees = 1000, importance = "permutation")
#' }
#'
#' D = createDisMatrix(ensemble, data=training, label = "mpg",
#'                                parallel = list(active=FALSE, no_cores = 1))
#'
#' setting=list(impTotal=0.1, maxDec=(1*10^-6), n=2, level=5)
#' tree <- e2tree(mpg ~ ., training, D, ensemble, setting)
#'
#' }
#'
#' @export


e2tree <- function(formula, data, D, ensemble, setting=list(impTotal=0.1, maxDec=0.01, n=2, level=5)){
  
  # === Input Validation ===
  
  # Validate formula
  if (!inherits(formula, "formula")) {
    stop("Error: 'formula' must be a valid formula object.")
  }
  
  # Validate data
  if (!is.data.frame(data) || nrow(data) == 0) {
    stop("Error: 'data' must be a non-empty data frame.")
  }
  
  # Validate D (dissimilarity matrix)
  if (!is.matrix(D) || nrow(D) != ncol(D)) {
    stop("Error: 'D' must be a square dissimilarity matrix.")
  }
  
  # Validate ensemble: keep the legacy message for the common NULL / generic-list
  # cases so existing user code and tests that match those exact strings keep
  # working; for any other unsupported class we delegate to the adapter
  # dispatch (which lists the supported backends).
  if (is.null(ensemble) || is.na(ensemble_backend(ensemble))) {
    stop("Error: 'ensemble' must be a trained 'randomForest' or 'ranger' model.",
         call. = FALSE)
  }
  type <- tryCatch(
    get_ensemble_type(ensemble),
    error = function(e) stop(conditionMessage(e), call. = FALSE)
  )
  if (!tolower(type) %in% c("classification", "regression")) {
    stop("Error: 'type' in ensemble object must be 'classification' or 'regression'.",
         call. = FALSE)
  }

  # Verify (D, ensemble) consistency: createDisMatrix tags D with the backend
  # used to build it.  A mismatch usually means the user has reused an old D
  # with a different model, which yields meaningless splits.
  d_backend <- attr(D, "ensemble_backend")
  ens_backend <- ensemble_backend(ensemble)
  if (!is.null(d_backend) && !is.na(d_backend) && !is.na(ens_backend) &&
      d_backend != ens_backend) {
    warning(sprintf(
      paste0("e2tree(): D was built from a '%s' ensemble but a '%s' was ",
             "passed. Re-run createDisMatrix() with the same ensemble for ",
             "consistent results."),
      d_backend, ens_backend
    ), call. = FALSE)
  }


  # Validate setting
  if (!is.list(setting) || !all(c("impTotal", "maxDec", "n", "level") %in% names(setting))) {
    stop("Error: 'setting' must be a list with keys: 'impTotal', 'maxDec', 'n', and 'level'.")
  }
  
  # Ensure all setting values are numeric and positive
  if (!all(sapply(setting, is.numeric)) || any(unlist(setting) <= 0)) {
    stop("Error: All values in 'setting' must be positive numeric values.")
  }
  
  # === Proceed with the function ===
  
  row.names(data) = NULL
  Call <- match.call()
  mf <- match.call(expand.dots = FALSE)
  m <- match(c("formula", "data"), names(mf), 0L)
  mf <- mf[c(1L, m)]
  mf$drop.unused.levels <- TRUE
  mf[[1L]] <- quote(stats::model.frame)
  mf <- eval(mf, parent.frame())
  Terms <- attributes(mf)$terms
  
  response <- mf[,1]
  X <- ordered2factor(mf[,-1]) # Convert ordered factors to regular factors (to change!!!)
  
  # create type object
  type <- get_ensemble_type(ensemble)

  setting$tMax <- 2L^(setting$level + 1L) - 1L
  
  ## identify qualitative variable and the number of categories:
  # Determine classes of all variables in X
  #var_classes <- unlist(lapply(X,class))
  var_classes <- get_classes(X)
  # Identify indices of factors and character variables
  ind <- which(var_classes %in% c("factor","character"))
  # Calculate the number of unique categories for each factor and character variable
  ncat <- NULL
  if (length(ind)>0){
    for (i in 1:length(ind)){
      ncat[i] <- length(unique(X[,ind[i]]))
    }
    names(ncat) <- names(X)[ind]
  }
  #ncat <- (sapply(X[,ind], function(x) length(unique(x))))
  #ncat <- (apply(X[,ind],2, function(x) length(unique(x))))
  # Update var_classes with the number of categories for character and factor variables
  var_classes[names(ncat)] = ncat
  # Set other variable types to -1
  var_classes[!names(var_classes) %in% names(ncat)] = -1
  # Convert var_classes values to numeric
  var_classes <- as.numeric(var_classes)
  # Ensure the names of var_classes match those of X
  names(var_classes) <- names(X)
  
  ## Generate the split matrix S from the original predictor matrix X
  max_cat <- if (!is.null(setting$max_cat)) setting$max_cat else 10
  res <- split(X, max_cat = max_cat)
  S <- res$S
  l <- res$lab
  rm(res)
  
  # Initilize node vector
  nodes <- rep(1,nrow(S))
  
  # list of no-terminal nodes
  nterm <- 1
  m <- unique(response)
  
  if (type == "classification") {
    m_label <- paste("V", seq(1,length(m)*2), sep="")
    labels <- c("node","n", "pred", "prob", "impTotal", "impChildren","decImp","decImpSur","variable" ,"split", "splitLabel", "variableSur", "splitLabelSur","parent","children","terminal", "obs", "path", "ncat",m_label)
    info <- data.frame(matrix(NA,setting$tMax,length(labels)))
    names(info) <- labels
    indv <- (ncol(info)-length(m_label)+1):ncol(info)
  } else {
    labels <- c("node","n", "pred", "prob", "impTotal", "impChildren","decImp","decImpSur","variable" ,"split", "splitLabel", "variableSur", "splitLabelSur","parent","children","terminal", "obs", "path", "ncat")
    info <- data.frame(matrix(NA,setting$tMax,length(labels)))
    names(info) <- labels
  }
  
  N <- NULL

  # Pre-compute training predictions once (used for regression MSE per node)
  ensemble_preds <- if (tolower(type) == "regression") {
    get_ensemble_predictions(ensemble, data, tolower(type))
  } else {
    NULL
  }

  ### variance in the root node
  vart1 = ifelse(type=="regression", variance(response), NA)

  while(length(nterm)>0){
    t <- tail(nterm,1)
    #print(t)
    N <- c(N,t)
    index <- which(nodes == t)
    
    ### check Stopping Rules----
    results <- eStoppingRules(D,index, t, setting, response, ensemble, vart1)
    
    ### Compute the response in the node
    switch(type,
           classification={
             res <- moda(response[index])
           },
           regression={
             res <- c(mean(response[index]), sum((response[index] - ensemble_preds[index])^2)) # mean and MSE
           })
    ###
    
    info$pred[t] <- res[1]   # Mean for regression
    info$prob[t] <- as.numeric(res[2])  # MSE for regression
    info$node[t] <- t
    info$parent[t] <- floor(t/2)
    info$n[t] <- length(index)
    info$impTotal[t] <- results$impTotal
    
    ##### statistics
    if (type == "classification") {
      yval <- data.frame(Y=response[index]) %>%
        mutate(Y=factor(Y, levels=levels(m))) %>%
        group_by(Y) %>%
        summarize(n=n(),
                  p=n/length(response[index])) %>%
        complete(Y, fill = list(n = 0, p = 0)) %>%
        select(n,p) %>%
        as.matrix() %>%
        c()
      
      info[t,indv] <- yval
    }
    
    ###
    
    if (isTRUE(results$sRule)){
      #list(imp=NA,decImp=NA, split=NA, splitLabel=NA, terminal=TRUE, impTotal=results$impTotal, obs=index, path=NA)
      info$terminal[t] <- TRUE
      #info$impTotal[t] <- results$impTotal
      info$obs[t] <- list(index)
      info$path[t] <- paths(info,t)
      
      #### add prediction
      #info[[t]]$pred <- prediction(Y[index])
      ####
      
    } else{
      imp<- eImpurity(D,index,S)
      s <- which.min(imp)
      v <- gsub(" <=.*| %in%.*","",names(s))
      s2 <- which.min(imp[!(l %in% v)])
      
      # max decrease of impurity
      decImp <- results$impTotal-imp[s]
      decImpSur <- results$impTotal-imp[s2]
      
      #check for max impurity stopping rule
      if (decImp<=setting$maxDec){
        #info[[t]] <- list(imp=NA,decImp=NA, split=NA, splitLabel=NA, terminal=TRUE, parent= floor(t/2), children=NA, impTotal=results$impTotal, obs=index, path=NA)
        #info$parent[t] <- floor(t/2)
        info$terminal[t] <- TRUE
        #info$impTotal[t] <- results$impTotal
        info$obs[t] <- list(index)
        info$path[t] <- paths(info,t)
      } else if (suppressWarnings(Wtest(Y=response[index], X=S[index,s], p.value=0.05, type = type))){
        # Stopping Rule with Mann-Whitney for regression case
        # if it is regression, check that the hypothesis that the two distributions in tL and tR are equal is rejected
        
        info$impChildren[t] <- as.numeric(imp[s])
        info$decImp[t] <- as.numeric(decImp)
        info$decImpSur[t] <- as.numeric(decImpSur)
        info$split[t] <- as.numeric(s)
        info$splitLabel[t] <- names(s)
        info$splitLabelSur[t] <- names(s2)
        info$terminal[t] <- FALSE
        info$parent[t] <- floor(t/2)
        info$children[t] <- list(c(t*2,t*2+1))
        info$impTotal[t] <- results$impTotal
        info$obs[t] <- list(index)
        info$path[t] <- paths(info,t)
        info$variable[t] <- gsub(" <=.*| %in%.*","",info$splitLabel[t])
        info$ncat[t] <- var_classes[info$variable[t]]
        
        nodes[index] <- (nodes[index]*2+1)-S[index,s]
        nterm <- c(nterm, sort(unique(nodes[index]), decreasing=T))
      }else{
        info$terminal[t] <- TRUE
        #info$impTotal[t] <- results$impTotal
        info$obs[t] <- list(index)
        info$path[t] <- paths(info,t)
      }
    }
    
    #print(info$path[t])
    nterm <- setdiff(nterm,t)
    
  }
  info <- info %>% drop_na(node)
  info$pred_val <- as.numeric(factor(info$pred))
  info$variableSur <- gsub(" <=.*| %in%.*","",info$splitLabelSur)
  
  if (type == "classification") {
    yval2 <- as.matrix(info[,indv])
    ypred <- info$pred_val
    nodeprob <- info$n/first(info$n)
    info <- info[,-indv]
    
    
    #info$yval2 <- cbind(ypred, yval2, nodeprob)
    yval2 <- cbind(ypred,yval2)
    #names(yval2) <- paste("V",seq(ncol(yval2)),sep="")
    attr(yval2,"dimnames")[[2]] <- paste("V",seq(ncol(yval2)),sep="")
    info$yval2 <- cbind(yval2, nodeprob)
  }
  #ylevels <- as.character(unique(response))
  ylevels <- levels(mf[[1]]) #### I need this to preserve the orginal attributes
  row.names(info) <- info$node
  info <- info[as.character(N),]
  
  ## normalized variance in nodes for regression
  if (type == "regression"){
    info$Wt <- NULL
    maxvar <- diff(range(response))^2L / 9L
    size <- length(response)
    ## Vectorized Wt computation — no per-node filtering
    node_rows <- match(info$node, info$node)  # identity map since row.names already set
    wt_vals <- vapply(seq_len(nrow(info)), function(r) {
      indice <- unlist(info$obs[[r]])
      v <- 1 - (variance(response[indice]) / (maxvar * length(indice) / size))
      min(max(v, 0), 1)
    }, numeric(1))
    info$Wt <- wt_vals
    info <- info %>% relocate(Wt, .after=prob)
  }
  
  
  object <- csplit_str(info,X,ncat, call=Call, terms=Terms, control=setting, ylevels=ylevels)

  object$N <- N

  ## Store fitted values and response for predict/fitted/residuals methods
  n_obs <- length(response)
  fv <- if (type == "classification") rep(NA_character_, n_obs) else rep(NA_real_, n_obs)
  terminal_nodes <- info[info$terminal == TRUE, , drop = FALSE]
  for (i in seq_len(nrow(terminal_nodes))) {
    obs_idx <- unlist(terminal_nodes$obs[[i]])
    if (length(obs_idx) > 0) fv[obs_idx] <- terminal_nodes$pred[i]
  }
  object$fitted.values <- fv
  object$y <- response

  ## Store training data for predict method
  object$data <- data

  class(object) <- c("e2tree", "list")

  # Propagate backend tag from D so coercion methods can detect mismatches.
  attr(object, "ensemble_backend") <- attr(D, "ensemble_backend")

  return(object)
}


paths <- function(info,t){
  path <- NULL
  while(t[1]>1){
    tp <- floor(t[1]/2)
    symb <- ifelse((t[1] %% 2)==0,"","!")
    t <- c(tp,t)
    path <- c(paste(symb,info$splitLabel[tp],sep=""),path)
    #<- paste(path," AND ", symb,info[[tp]]$splitLabel, sep="",collapse="")
  }
  path <- paste(path,sep="",collapse=" & ")
  return(path)
}


### creation objects for rpart.plot ----
csplit_str <- function(info,X,ncat, call, terms, control, ylevels){
  
  var_lev <- var_ord <- attribute <- list()
  for (i in names(ncat)){
    var_lev[[i]] <- seq(ncat[i])
    names(var_lev[[i]]) <- unique(X[,i])
    attribute[[i]] <- as.character(unique(X[,i]))
  }
  
  # ordVar <- get_classes(X)
  # ordVar <- names(ordVar[ordVar=="ordered"])
  # for (i in names(ordVar)){
  #   var_ord[[i]] <- seq(length(levels(X[[i]])))
  #   names(var_ord[[i]]) <- unique(levels(X[[i]]))
  #   attribute[[i]] <- as.character(levels(X[[i]]))
  # }
  
  #qualitative splits
  splits <- info %>%
    dplyr::filter(!is.na(variable)) %>%
    dplyr::select(n,ncat,variable,decImp, splitLabel)
  
  #splits <- info[!is.na(info$variable), c("n","ncat","variable","decImp", "splitLabel")]
  # index for numerical predictors
  if(nrow(info)>1){
    varnumerics <- strsplit(splits$splitLabel[splits$ncat==-1],"<=")
    splits$index[splits$ncat==-1] <- unlist(lapply(varnumerics, function(x) x[2]))
  }
  
  #splits$index <- suppressWarnings(as.numeric(as.numeric(gsub("[^[:digit:]., ]", "",splits$splitLabel))))
  splits$index[splits$ncat!=-1] <- seq(sum(splits$ncat!=-1))
  splits$index <- as.numeric(splits$index)
  
  
  
  catsplits <- splits %>%
    dplyr::filter(ncat!=-1) %>%
    dplyr::mutate(ID = row_number()) %>%
    dplyr::group_by(ID) %>%
    dplyr::mutate(modal = gsub(paste0(variable," %in% "),"",splitLabel))
  
  splits <- splits %>%
    #data.frame() %>%
    dplyr::select(n,ncat,decImp,index) %>%
    dplyr::rename(improve = decImp,
                  count = n) %>%
    dplyr::mutate(adj=0) %>%
    as.matrix()
  
  attr(splits,"dimnames")[[1]]  <- info$variable[!is.na(info$variable)]
  
  ### creation of csplit
  if (nrow(catsplits)>0){
    csplit <- matrix(2,nrow(catsplits),max(catsplits$ncat))
    
    for (i in 1:nrow(csplit)){
      modal <- eval(parse(text=catsplits$modal[i]))
      vec <- var_lev[[catsplits$variable[i]]]
      ind <- vec[modal]
      csplit[i,ind] <- 1
      ind <- vec[setdiff(names(vec),modal)]
      csplit[i,ind] <- 3
    }
    csplit <- as.matrix(csplit)
  } else {
    csplit = NULL
  }
  
  object <- list(tree=info, csplit=csplit,splits=splits, call=call, terms=terms, control=control)
  
  attr(object,"xlevels") <- attribute
  attr(object,"ylevels") <- ylevels
  
  return(object)
}





## Variance — delegates to shared utility
variance <- e2_variance





Wtest = function(Y, X, type, p.value=0.05){
  switch (type,
          "classification" = {resp=TRUE},
          "regression" = {resp <- wilcox.test(Y ~ X)$p.value <= 0.05}
  )
  return(resp)
}

# get_classes <- function(df) {
#   dfClasses <- data.frame(
#     column = names(df),
#     class = sapply(df, function(x) paste(class(x), collapse = ", "))
#   ) %>%
#     mutate(type = class,
#            type = ifelse(sapply(df, is.factor), "factor",
#                          ifelse(sapply(df, is.character), "character", type))) %>%
#     select(column, type)
#   classes <- dfClasses$type
#   names(classes) <- dfClasses$column
#   return(classes)
# }

get_classes <- function(df) {
  classes <- vapply(df, function(x) {
    cl <- class(x)
    if ("ordered" %in% cl) return("ordered")
    if (is.character(x)) return("character")
    paste(cl, collapse = ", ")
  }, character(1))
  return(classes)
}

ordered2factor <- function(X){
  varClass <- get_classes(X)
  varClass <- names(varClass[varClass=="ordered"])
  for (i in varClass){
    class(X[[i]]) <- "factor"
  }
  return(X)
}

library(testthat)
library(e2tree)
library(randomForest)

test_that("e2tree works correctly for classification", {
  set.seed(42)
  
  # Prepare data
  data(iris)
  train_idx <- sample(seq_len(nrow(iris)), size = 0.75 * nrow(iris))
  training <- iris[train_idx, ]
  
  # Train Random Forest
  ensemble <- randomForest(Species ~ ., data = training, importance = TRUE, proximity = TRUE)
  
  # Create dissimilarity matrix
  D <- createDisMatrix(ensemble, data = training, label = "Species", parallel = list(active=FALSE, no_cores = 1))
  
  # Define settings
  setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
  
  # Run e2tree
  tree <- e2tree(Species ~ ., training, D, ensemble, setting)
  
  # Tests
  expect_type(tree, "list")  # Must return a list
  expect_true("tree" %in% names(tree))  # Must contain a tree object
  expect_true(is.data.frame(tree$tree))  # The tree should be a data frame
})

test_that("e2tree works correctly for regression", {
  set.seed(42)
  
  # Prepare data
  data(mtcars)
  train_idx <- sample(seq_len(nrow(mtcars)), size = 0.75 * nrow(mtcars))
  training <- mtcars[train_idx, ]
  
  # Train Random Forest
  ensemble <- randomForest(mpg ~ ., data = training, importance = TRUE, proximity = TRUE)
  
  # Create dissimilarity matrix
  D <- createDisMatrix(ensemble, data = training, label = "mpg", parallel = list(active=FALSE, no_cores = 1))
  
  # Define settings
  setting=list(impTotal=0.1, maxDec=(1*10^-6), n=2, level=5)
  
  # Run e2tree
  tree <- e2tree(mpg ~ ., training, D, ensemble, setting)
  
  # Tests
  expect_type(tree, "list")  # Must return a list
  expect_true("tree" %in% names(tree))  # Must contain a tree object
  expect_true(is.data.frame(tree$tree))  # The tree should be a data frame
})

test_that("e2tree handles incorrect input types", {
  set.seed(42)
  
  data(iris)
  train_idx <- sample(seq_len(nrow(iris)), size = 0.75 * nrow(iris))
  training <- iris[train_idx, ]
  
  ensemble <- randomForest(Species ~ ., data = training, importance = TRUE, proximity = TRUE)
  D <- createDisMatrix(ensemble, data = training, label = "Species", parallel = list(active=FALSE, no_cores = 1))
  setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)

  # Test incorrect inputs
  expect_error(e2tree(NULL, training, D, ensemble, setting), 
               "Error: 'formula' must be a valid formula object.")
  
  expect_error(e2tree(Species ~ ., NULL, D, ensemble, setting), 
               "Error: 'data' must be a non-empty data frame.")
  
  expect_error(e2tree(Species ~ ., training, NULL, ensemble, setting), 
               "Error: 'D' must be a square dissimilarity matrix.")
  
  expect_error(e2tree(Species ~ ., training, D, NULL, setting), 
               "Error: 'ensemble' must be a trained 'randomForest' or 'ranger' model.")
  
  expect_error(e2tree(Species ~ ., training, D, ensemble, NULL), 
               "Error: 'setting' must be a list with keys: 'impTotal', 'maxDec', 'n', and 'level'.")
  
  ensemble$type <- "unknown_type"  # Modify to invalid type
  expect_error(e2tree(Species ~ ., training, D, ensemble, setting), 
               "Error: 'type' in ensemble object must be 'classification' or 'regression'.")
})

test_that("e2tree handles incorrect settings", {
  set.seed(42)
  
  data(iris)
  train_idx <- sample(seq_len(nrow(iris)), size = 0.75 * nrow(iris))
  training <- iris[train_idx, ]
  
  ensemble <- randomForest(Species ~ ., data = training, importance = TRUE, proximity = TRUE)
  D <- createDisMatrix(ensemble, data = training, label = "Species", parallel = list(active=FALSE, no_cores = 1))

  # Incorrect settings
  bad_setting1 <- list(impTotal = -0.1, maxDec = 0.01, n = 2, level = 5)
  bad_setting2 <- list(impTotal = "high", maxDec = 0.01, n = 2, level = 5)
  
  expect_error(e2tree(Species ~ ., training, D, ensemble, bad_setting1), 
               "Error: All values in 'setting' must be positive numeric values.")
  
  expect_error(e2tree(Species ~ ., training, D, ensemble, bad_setting2), 
               "Error: All values in 'setting' must be positive numeric values.")
})

library(testthat)
library(e2tree)
library(randomForest)

test_that("vimp works correctly with valid inputs (classification case)", {
  set.seed(42)
  
  # Prepare data
  data(iris)
  train_idx <- sample(seq_len(nrow(iris)), size = 0.75 * nrow(iris))
  training <- iris[train_idx, ]
  
  # Train Random Forest
  ensemble <- randomForest(Species ~ ., data = training, importance = TRUE, proximity = TRUE)
  
  # Create dissimilarity matrix
  D <- createDisMatrix(ensemble, data = training, label = "Species", parallel = list(active=FALSE, no_cores = 1))
  
  # Define settings
  setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
  
  # Generate e2tree model
  fit <- e2tree(Species ~ ., training, D, ensemble, setting)
  
  # Run vimp function
  vimp_result <- vimp(fit, training, type = "classification")
  
  # Tests
  expect_type(vimp_result, "list")  # Should return a list
  expect_true("vimp" %in% names(vimp_result))
  expect_true(is.data.frame(vimp_result$vimp))  # vimp should be a data frame
})

test_that("vimp handles incorrect input types (classification case)", {
  set.seed(42)
  
  data(iris)
  train_idx <- sample(seq_len(nrow(iris)), size = 0.75 * nrow(iris))
  training <- iris[train_idx, ]
  
  ensemble <- randomForest(Species ~ ., data = training, importance = TRUE, proximity = TRUE)
  D <- createDisMatrix(ensemble, data = training, label = "Species", parallel = list(active=FALSE, no_cores = 1))
  setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
  fit <- e2tree(Species ~ ., training, D, ensemble, setting)

  # Test incorrect inputs
  expect_error(vimp(NULL, training, "classification"), 
               "Error: 'fit' must be an 'e2tree' object.")
  
  expect_error(vimp(fit, NULL, "classification"), 
               "Error: 'data' must be a non-empty data frame.")
  
  expect_error(vimp(fit, training, "unknown_type"))
})

test_that("vimp handles missing response variable in data (classification case)", {
  set.seed(42)
  
  data(iris)
  train_idx <- sample(seq_len(nrow(iris)), size = 0.75 * nrow(iris))
  training <- iris[train_idx, ]
  
  ensemble <- randomForest(Species ~ ., data = training, importance = TRUE, proximity = TRUE)
  D <- createDisMatrix(ensemble, data = training, label = "Species", parallel = list(active=FALSE, no_cores = 1))
  setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
  fit <- e2tree(Species ~ ., training, D, ensemble, setting)

  # Remove response variable from training data
  training_no_species <- training[, -which(names(training) == "Species")]

  expect_error(vimp(fit, training_no_species, "classification"), 
               "Error: The response variable from 'fit' is not found in 'data'.")
})













test_that("vimp works correctly with valid inputs (regression case)", {
  set.seed(42)
  
  # Prepare data
  data(mtcars)
  train_idx <- sample(seq_len(nrow(mtcars)), size = 0.75 * nrow(mtcars))
  training <- mtcars[train_idx, ]
  
  # Train Random Forest
  ensemble <- randomForest(mpg ~ ., data = training, importance = TRUE, proximity = TRUE)
  
  # Create dissimilarity matrix
  D <- createDisMatrix(ensemble, data = training, label = "mpg", parallel = list(active=FALSE, no_cores = 1))
  
  # Define settings
  setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
  
  # Generate e2tree model
  fit <- e2tree(mpg ~ ., training, D, ensemble, setting)
  
  # Run vimp function
  vimp_result <- vimp(fit, training, type = "regression")
  
  # Tests
  expect_type(vimp_result, "list")  # Should return a list
  expect_true("vimp" %in% names(vimp_result))
  expect_true(is.data.frame(vimp_result$vimp))  # vimp should be a data frame
})

test_that("vimp handles incorrect input types (regression case)", {
  set.seed(42)
  
  data(mtcars)
  train_idx <- sample(seq_len(nrow(mtcars)), size = 0.75 * nrow(mtcars))
  training <- mtcars[train_idx, ]
  
  ensemble <- randomForest(mpg ~ ., data = training, importance = TRUE, proximity = TRUE)
  D <- createDisMatrix(ensemble, data = training, label = "mpg", parallel = list(active=FALSE, no_cores = 1))
  setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
  fit <- e2tree(mpg ~ ., training, D, ensemble, setting)
  
  # Test incorrect inputs
  expect_error(vimp(NULL, training, "classification"), 
               "Error: 'fit' must be an 'e2tree' object.")
  
  expect_error(vimp(fit, NULL, "classification"), 
               "Error: 'data' must be a non-empty data frame.")
  
  expect_error(vimp(fit, training, "unknown_type"))
})

test_that("vimp handles missing response variable in data (regression case)", {
  set.seed(42)
  
  data(mtcars)
  train_idx <- sample(seq_len(nrow(mtcars)), size = 0.75 * nrow(mtcars))
  training <- mtcars[train_idx, ]
  
  ensemble <- randomForest(mpg ~ ., data = training, importance = TRUE, proximity = TRUE)
  D <- createDisMatrix(ensemble, data = training, label = "mpg", parallel = list(active=FALSE, no_cores = 1))
  setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
  fit <- e2tree(mpg ~ ., training, D, ensemble, setting)
  
  # Remove response variable from training data
  training_no_mpg <- training[, -which(names(training) == "mpg")]
  
  expect_error(vimp(fit, training_no_mpg, "classification"), 
               "Error: The response variable from 'fit' is not found in 'data'.")
})

library(testthat)
library(e2tree)
library(randomForest)

test_that("eValidation works correctly with valid inputs (classification case)", {
  set.seed(42)

  # Prepare data
  data(iris)
  train_idx <- sample(seq_len(nrow(iris)), size = 0.75 * nrow(iris))
  training <- iris[train_idx, ]

  # Train Random Forest
  ensemble <- randomForest(Species ~ ., data = training, importance = TRUE, proximity = TRUE)

  # Create dissimilarity matrix
  D <- createDisMatrix(ensemble, data = training, label = "Species", parallel = list(active=FALSE, no_cores = 1))

  # Define settings
  setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)

  # Generate e2tree model
  fit <- e2tree(Species ~ ., training, D, ensemble, setting)

  # Run eValidation
  comparison <- eValidation(training, fit, D, graph=FALSE)

  # Tests
  expect_type(comparison, "list")  # Should return a list
  expect_true("mantel_test" %in% names(comparison))
})

test_that("eValidation handles incorrect input types (classification case)", {
  set.seed(42)

  data(iris)
  train_idx <- sample(seq_len(nrow(iris)), size = 0.75 * nrow(iris))
  training <- iris[train_idx, ]

  ensemble <- randomForest(Species ~ ., data = training, importance = TRUE, proximity = TRUE)
  D <- createDisMatrix(ensemble, data = training, label = "Species", parallel = list(active=FALSE, no_cores = 1))
  setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
  fit <- e2tree(Species ~ ., training, D, ensemble, setting)

  # Test incorrect inputs
  expect_error(eValidation(NULL, fit, D, graph=FALSE),
               "Error: 'data' must be a non-empty data frame.")

  expect_error(eValidation(training, NULL, D, graph=FALSE),
               "Error: 'fit' must be an 'e2tree' object.")

  expect_error(eValidation(training, fit, NULL, graph=FALSE),
               "Error: 'D' must be a square dissimilarity matrix.")

  expect_error(eValidation(training, fit, matrix(1, 5, 5), graph=FALSE),
               "Error: The number of rows in 'data' must match the dimensions of 'D'.")
})


test_that("eValidation works correctly with valid inputs (regression case)", {
  set.seed(42)

  # Prepare data
  data(mtcars)
  train_idx <- sample(seq_len(nrow(mtcars)), size = 0.75 * nrow(mtcars))
  training <- mtcars[train_idx, ]

  # Train Random Forest
  ensemble <- randomForest(mpg ~ ., data = training, importance = TRUE, proximity = TRUE)

  # Create dissimilarity matrix
  D <- createDisMatrix(ensemble, data = training, label = "mpg", parallel = list(active=FALSE, no_cores = 1))

  # Define settings
  setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)

  # Generate e2tree model
  fit <- e2tree(mpg ~ ., training, D, ensemble, setting)

  # Run eValidation
  comparison <- eValidation(training, fit, D, graph=FALSE)

  # Tests
  expect_type(comparison, "list")  # Should return a list
  expect_true("mantel_test" %in% names(comparison))
})

test_that("eValidation handles incorrect input types (regression case)", {
  set.seed(42)

  data(mtcars)
  train_idx <- sample(seq_len(nrow(mtcars)), size = 0.75 * nrow(mtcars))
  training <- mtcars[train_idx, ]

  ensemble <- randomForest(mpg ~ ., data = training, importance = TRUE, proximity = TRUE)
  D <- createDisMatrix(ensemble, data = training, label = "mpg", parallel = list(active=FALSE, no_cores = 1))
  setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
  fit <- e2tree(mpg ~ ., training, D, ensemble, setting)

  # Test incorrect inputs
  expect_error(eValidation(NULL, fit, D, graph=FALSE),
               "Error: 'data' must be a non-empty data frame.")

  expect_error(eValidation(training, NULL, D, graph=FALSE),
               "Error: 'fit' must be an 'e2tree' object.")

  expect_error(eValidation(training, fit, NULL, graph=FALSE),
               "Error: 'D' must be a square dissimilarity matrix.")

  expect_error(eValidation(training, fit, matrix(1, 5, 5), graph=FALSE),
               "Error: The number of rows in 'data' must match the dimensions of 'D'.")
})

library(testthat)
library(e2tree)
library(randomForest)

test_that("rpart2Tree works correctly with valid inputs (classification case)", {
  set.seed(42)
  
  # Prepare data
  data(iris)
  train_idx <- sample(seq_len(nrow(iris)), size = 0.75 * nrow(iris))
  training <- iris[train_idx, ]
  
  # Train Random Forest
  ensemble <- randomForest(Species ~ ., data = training, importance = TRUE, proximity = TRUE)
  
  # Create dissimilarity matrix
  D <- createDisMatrix(ensemble, data = training, label = "Species", parallel = list(active=FALSE, no_cores = 1))
  
  # Define settings
  setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
  
  # Generate e2tree model
  fit <- e2tree(Species ~ ., training, D, ensemble, setting)
  
  # Run rpart2Tree
  rpart_obj <- rpart2Tree(fit, ensemble)
  
  # Tests
  expect_type(rpart_obj, "list")  # Should return a list
  expect_true("frame" %in% names(rpart_obj))
  expect_true("where" %in% names(rpart_obj))
  expect_true("variable.importance" %in% names(rpart_obj))
})

test_that("rpart2Tree handles incorrect input types (classification case)", {
  set.seed(42)
  
  data(iris)
  train_idx <- sample(seq_len(nrow(iris)), size = 0.75 * nrow(iris))
  training <- iris[train_idx, ]
  
  ensemble <- randomForest(Species ~ ., data = training, importance = TRUE, proximity = TRUE)
  D <- createDisMatrix(ensemble, data = training, label = "Species", parallel = list(active=FALSE, no_cores = 1))
  setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
  fit <- e2tree(Species ~ ., training, D, ensemble, setting)

  # Test incorrect inputs
  expect_error(rpart2Tree(NULL, ensemble), 
               "Error: 'fit' must be an 'e2tree' object.")
  
  expect_error(rpart2Tree(fit, NULL), 
               "Error: 'ensemble' must be a trained 'randomForest' or 'ranger' model.")
  
  expect_error(rpart2Tree(fit, list()), 
               "Error: 'ensemble' must be a trained 'randomForest' or 'ranger' model.")
})

test_that("rpart2Tree handles invalid ensemble type (classification case)", {
  set.seed(42)
  
  data(iris)
  train_idx <- sample(seq_len(nrow(iris)), size = 0.75 * nrow(iris))
  training <- iris[train_idx, ]
  
  ensemble <- randomForest(Species ~ ., data = training, importance = TRUE, proximity = TRUE)
  D <- createDisMatrix(ensemble, data = training, label = "Species", parallel = list(active=FALSE, no_cores = 1))
  setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
  fit <- e2tree(Species ~ ., training, D, ensemble, setting)
  ensemble$type <- "unknown_type"  # Modify to invalid type
  expect_error(rpart2Tree(fit, ensemble), "Error: 'type' in ensemble object must be either 'classification' or 'regression'.")


})





test_that("rpart2Tree works correctly with valid inputs (regression case)", {
  set.seed(42)
  
  # Prepare data
  data(mtcars)
  train_idx <- sample(seq_len(nrow(mtcars)), size = 0.75 * nrow(mtcars))
  training <- mtcars[train_idx, ]
  
  # Train Random Forest
  ensemble <- randomForest(mpg ~ ., data = training, importance = TRUE, proximity = TRUE)
  
  # Create dissimilarity matrix
  D <- createDisMatrix(ensemble, data = training, label = "mpg", parallel = list(active=FALSE, no_cores = 1))
  
  # Define settings
  setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
  
  # Generate e2tree model
  fit <- e2tree(mpg ~ ., training, D, ensemble, setting)
  
  # Run rpart2Tree
  rpart_obj <- rpart2Tree(fit, ensemble)
  
  # Tests
  expect_type(rpart_obj, "list")  # Should return a list
  expect_true("frame" %in% names(rpart_obj))
  expect_true("where" %in% names(rpart_obj))
  expect_true("variable.importance" %in% names(rpart_obj))
})

test_that("rpart2Tree handles incorrect input types (regression case)", {
  set.seed(42)
  
  data(mtcars)
  train_idx <- sample(seq_len(nrow(mtcars)), size = 0.75 * nrow(mtcars))
  training <- mtcars[train_idx, ]
  
  ensemble <- randomForest(mpg ~ ., data = training, importance = TRUE, proximity = TRUE)
  D <- createDisMatrix(ensemble, data = training, label = "mpg", parallel = list(active=FALSE, no_cores = 1))
  setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
  fit <- e2tree(mpg ~ ., training, D, ensemble, setting)
  
  # Test incorrect inputs
  expect_error(rpart2Tree(NULL, ensemble), 
               "Error: 'fit' must be an 'e2tree' object.")
  
  expect_error(rpart2Tree(fit, NULL), 
               "Error: 'ensemble' must be a trained 'randomForest' or 'ranger' model.")
  
  expect_error(rpart2Tree(fit, list()), 
               "Error: 'ensemble' must be a trained 'randomForest' or 'ranger' model.")
})

test_that("rpart2Tree handles invalid ensemble type (regression case)", {
  set.seed(42)
  
  data(mtcars)
  train_idx <- sample(seq_len(nrow(mtcars)), size = 0.75 * nrow(mtcars))
  training <- mtcars[train_idx, ]
  
  ensemble <- randomForest(mpg ~ ., data = training, importance = TRUE, proximity = TRUE)
  D <- createDisMatrix(ensemble, data = training, label = "mpg", parallel = list(active=FALSE, no_cores = 1))
  setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
  fit <- e2tree(mpg ~ ., training, D, ensemble, setting)
  ensemble$type <- "unknown_type"  # Modify to invalid type
  expect_error(rpart2Tree(fit, ensemble), "Error: 'type' in ensemble object must be either 'classification' or 'regression'.")
  
  
})

library(testthat)
library(e2tree)
library(randomForest)

test_that("ePredTree works correctly for classification", {
  set.seed(123)
  
  data(iris)
  train_idx <- sample(seq_len(nrow(iris)), size = 0.75 * nrow(iris))
  training <- iris[train_idx, ]
  validation <- iris[-train_idx, ]
  
  ensemble <- randomForest(Species ~ ., data=training, proximity=TRUE)
  D <- createDisMatrix(ensemble, data=training, label="Species", parallel = list(active=FALSE, no_cores = 1))
  setting <- list(impTotal=0.1, maxDec=0.01, n=2, level=5)
  tree <- e2tree(Species ~ ., training, D, ensemble, setting)
  
  pred <- ePredTree(tree, validation, target="setosa")
  
  expect_s3_class(pred, "data.frame")
  expect_true(all(c("fit", "accuracy", "score") %in% names(pred)))
  expect_equal(nrow(pred), nrow(validation))
  expect_type(pred$fit, "character")
})

test_that("ePredTree works correctly for regression", {
  set.seed(123)
  
  data(mtcars)
  train_idx <- sample(seq_len(nrow(mtcars)), size = 0.75 * nrow(mtcars))
  training <- mtcars[train_idx, ]
  validation <- mtcars[-train_idx, ]
  
  ensemble <- randomForest(mpg ~ ., data=training, proximity=TRUE)
  D <- createDisMatrix(ensemble, data=training, label="mpg", parallel = list(active=FALSE, no_cores = 1))
  setting <- list(impTotal=0.1, maxDec=1e-6, n=2, level=5)
  tree <- e2tree(mpg ~ ., training, D, ensemble, setting)
  
  pred <- ePredTree(tree, validation)
  
  expect_s3_class(pred, "data.frame")
  expect_true(all(c("fit", "accuracy", "score") %in% names(pred)))
  expect_equal(nrow(pred), nrow(validation))
  expect_type(pred$fit, "double") 
  expect_true(all(is.na(pred$accuracy)))
  expect_true(all(is.na(pred$score)))
})

#' ROC Curve
#'
#' Computes and plots the Receiver Operating Characteristic (ROC) curve for a binary classification model, 
#' along with the Area Under the Curve (AUC). The ROC curve is a graphical representation of a classifier’s 
#' performance across all classification thresholds.
#'
#' @param response is the response variable vector
#' @param scores is the probability vector of the prediction
#' @param target is the target response class
#'
#' @return an object.
#'
#'
#' @examples
#' \donttest{
#' ## Classification:
#' data(iris)
#'
#' # Create training and validation set:
#' smp_size <- floor(0.75 * nrow(iris))
#' train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
#' training <- iris[train_ind, ]
#' validation <- iris[-train_ind, ]
#' response_training <- training[,5]
#' response_validation <- validation[,5]
#'
#' # Perform training:
#' ensemble <- randomForest::randomForest(Species ~ ., data=training, 
#' importance=TRUE, proximity=TRUE)
#' 
#' D <- createDisMatrix(ensemble, data=training, label = "Species", 
#'                             parallel = list(active=FALSE, no_cores = 1))
#' 
#' setting=list(impTotal=0.1, maxDec=0.01, n=2, level=5)
#' tree <- e2tree(Species ~ ., training, D, ensemble, setting)
#'
#' pr <- ePredTree(tree, validation, target="setosa")
#' 
#' roc(response_training, scores = pr$score, target = "setosa")
#'  
#' }
#'
#' @export
#'
roc <- function(response, scores, target="1"){
  labels <- as.numeric(response==target)
  AUC <- auc_probability(labels, scores)
  labels <- labels[order(scores, decreasing=TRUE)]
  R <- data.frame(TPR=cumsum(labels)/sum(labels), FPR=cumsum(!labels)/sum(!labels), labels)

  p <- R %>% ggplot(aes(x=.data$FPR,y=.data$TPR)) +
    geom_rect(aes(xmin=0, xmax=1, ymin=0, ymax=1), alpha=0, color="grey")+
    geom_line(linewidth=0.5, color="firebrick")+
    geom_abline(intercept=0,slope=1, linetype="dashed")+
    annotate("text",x = 0.75, y = 0.25, label = paste("AUC = ",round(AUC,3)))+
    scale_x_continuous(breaks = c(0,0.25,0.5,0.75,1),labels = c("1.00","0.75","0.50","0.25","0.00"))+
    labs(title="Roc Curve", x = "Specificity", y = "Sensitivity") +
    theme_minimal()
  plot(p)

results <- list(roc_data=R, auc=AUC)
return(results)
}


auc_probability <- function(labels, scores, N=1e7){
  # Exact AUC via Wilcoxon-Mann-Whitney statistic — O(n log n), deterministic
  labels <- as.logical(labels)
  n1 <- sum(labels)
  n0 <- sum(!labels)
  if (n1 == 0 || n0 == 0) return(NA_real_)
  r <- rank(scores, ties.method = "average")
  (sum(r[labels]) - n1 * (n1 + 1) / 2) / (n1 * n0)
}

library(testthat)
library(ggplot2)  

test_that("roc() returns expected output structure and values", {
  set.seed(123)
  
  # data
  response <- rep(c("yes", "no"), each = 50)
  scores <- c(runif(50, 0.6, 1), runif(50, 0, 0.4))  
  target <- "yes"
  
  # function
  result <- roc(response, scores, target)
  
  # structure
  expect_type(result, "list")
  expect_named(result, c("roc_data", "auc"))
  
  # check on dataframe
  expect_s3_class(result$roc_data, "data.frame")
  expect_true(all(c("TPR", "FPR", "labels") %in% names(result$roc_data)))
  expect_equal(nrow(result$roc_data), length(response))
  
  # AUC 
  expect_gt(result$auc, 0.9)
})

test_that("roc() handles minimal input correctly", {
  response <- c("yes", "no")
  scores <- c(0.9, 0.1)
  target <- "yes"
  
  result <- roc(response, scores, target)
  
  expect_type(result, "list")
  expect_true(result$auc == 1 || result$auc == 0.5)  
})

# ============================================================================
# Utility functions shared across the e2tree package
# ============================================================================

#' Population Variance
#' @keywords internal
e2_variance <- function(x) {
  sum((x - mean(x))^2) / length(x)
}

#' Check Availability of Suggested Packages
#' @keywords internal
check_package <- function(pkg) {
  if (!requireNamespace(pkg, quietly = TRUE)) {
    stop(
      sprintf("Package '%s' is required but not installed. Please install it with: install.packages('%s')", pkg, pkg),
      call. = FALSE
    )
  }
}

#' Identify the canonical class of a supported ensemble model
#'
#' Returns one of \code{"randomForest"}, \code{"ranger"}, \code{"xgb.Booster"},
#' \code{"lgb.Booster"}, \code{"gbm"}, \code{"catboost.CatBoost"} or
#' \code{"catboost.Model"} (the same class used by the S3 adapter dispatch),
#' or \code{NA_character_} when no supported class is matched.
#'
#' @keywords internal
ensemble_backend <- function(ensemble) {
  supported <- c("randomForest", "ranger", "xgb.Booster",
                 "lgb.Booster", "gbm",
                 "catboost.CatBoost", "catboost.Model")
  hit <- intersect(supported, class(ensemble))
  if (length(hit) == 0L) return(NA_character_)
  hit[1L]
}

#' Validate the output of \code{extract_terminal_nodes()}
#'
#' Boosting backends store their tree structures in opaque containers; a tiny
#' API change can silently produce a malformed leaf matrix (e.g. all zeros),
#' yielding a degenerate dissimilarity matrix without raising any error.
#' This function asserts the shape and type contract so problems surface
#' immediately at extraction time rather than much later, after the C++
#' co-occurrence call has already produced nonsense.
#'
#' Contract: \code{nodes} must be a \code{data.frame} with \code{nrow(data)}
#' rows and at least one column; every column must be coercible to integer;
#' at least one column must contain more than one distinct value.
#'
#' @keywords internal
validate_terminal_nodes <- function(nodes, data, backend = NA_character_) {
  if (!is.data.frame(nodes)) {
    stop(sprintf(
      "extract_terminal_nodes(%s) must return a data.frame; got '%s'.",
      backend, paste(class(nodes), collapse = "/")
    ), call. = FALSE)
  }
  if (nrow(nodes) != nrow(data)) {
    stop(sprintf(
      "extract_terminal_nodes(%s) returned %d rows but data has %d.",
      backend, nrow(nodes), nrow(data)
    ), call. = FALSE)
  }
  if (ncol(nodes) < 1L) {
    stop(sprintf(
      "extract_terminal_nodes(%s) returned 0 columns; expected one column per tree.",
      backend
    ), call. = FALSE)
  }
  numeric_cols <- vapply(nodes, function(col) is.numeric(col) || is.integer(col),
                         logical(1L))
  if (!all(numeric_cols)) {
    bad <- names(nodes)[!numeric_cols]
    stop(sprintf(
      "extract_terminal_nodes(%s): non-numeric leaf columns detected (%s).",
      backend, paste(bad, collapse = ", ")
    ), call. = FALSE)
  }
  # Detect the GBM-style silent failure: every leaf column collapses to a
  # single value, which turns the dissimilarity matrix into all-zeros.
  uniq <- vapply(nodes, function(col) length(unique(col)), integer(1L))
  if (all(uniq <= 1L)) {
    stop(sprintf(
      paste0("extract_terminal_nodes(%s) produced degenerate output: every ",
             "tree maps all observations to the same leaf. This usually ",
             "signals an internal API mismatch with the upstream package."),
      backend
    ), call. = FALSE)
  }
  invisible(TRUE)
}

# ============================================================================
# Adapter Layer for Ensemble Models
#
# Three S3 generics provide a unified interface over diverse ensemble backends.
# To add support for a new model class, implement the three methods below for
# that class — no changes to the core algorithm are required.
#
#   1. get_ensemble_type(ensemble)           → "classification" | "regression"
#   2. extract_terminal_nodes(ensemble, data) → data.frame (n_obs × n_trees)
#   3. get_ensemble_predictions(ensemble, data, type) → numeric vector (n_obs)
#
# All three generics share the same `default` failure mode and the same list
# of supported classes; `.unsupported_backend()` centralises that message so
# the supported-classes list is maintained in exactly one place.
# ============================================================================

.SUPPORTED_BACKENDS <- c(
  "randomForest", "ranger", "xgb.Booster",
  "lgb.Booster", "gbm", "catboost.CatBoost", "catboost.Model"
)

.unsupported_backend <- function(generic, ensemble) {
  stop(sprintf(
    "%s(): unsupported ensemble class '%s'.\nSupported classes: %s.",
    generic,
    paste(class(ensemble), collapse = ", "),
    paste(.SUPPORTED_BACKENDS, collapse = ", ")
  ), call. = FALSE)
}


# ============================================================================
# 1.  get_ensemble_type()
# ============================================================================

#' Determine Task Type from a Trained Ensemble Model
#'
#' Returns \code{"classification"} or \code{"regression"} depending on the
#' objective used to train the ensemble.
#'
#' @param ensemble A trained ensemble model. Supported classes:
#'   \code{randomForest}, \code{ranger}, \code{xgb.Booster},
#'   \code{lgb.Booster}, \code{gbm}, \code{catboost.CatBoost}.
#' @return Character scalar: \code{"classification"} or \code{"regression"}.
#' @export
get_ensemble_type <- function(ensemble) {
  UseMethod("get_ensemble_type")
}

#' @export
#' @method get_ensemble_type randomForest
get_ensemble_type.randomForest <- function(ensemble) {
  ensemble$type
}

#' @export
#' @method get_ensemble_type ranger
get_ensemble_type.ranger <- function(ensemble) {
  tolower(ensemble$treetype)
}

#' @export
#' @method get_ensemble_type xgb.Booster
get_ensemble_type.xgb.Booster <- function(ensemble) {
  check_package("xgboost")
  # $params may be empty in xgboost >= 2.x; fall back to xgb.config()
  obj <- ensemble$params$objective
  if (is.null(obj) || identical(obj, "")) {
    obj <- tryCatch({
      cfg <- xgboost::xgb.config(ensemble)
      if (is.list(cfg)) {
        # xgboost >= 2.x: config is an R list
        cfg[["learner"]][["learner_train_param"]][["objective"]]
      } else {
        # xgboost < 2.x: config is a JSON character string
        m <- regexpr('"objective"\\s*:\\s*"([^"]+)"', cfg, perl = TRUE)
        if (m == -1L) NULL else sub('.*"objective"\\s*:\\s*"([^"]+)".*', "\\1",
                                    regmatches(cfg, m))
      }
    }, error = function(e) NULL)
  }
  if (is.null(obj) || identical(obj, "")) {
    obj <- tryCatch(xgboost::xgb.attr(ensemble, "objective"), error = function(e) NULL)
  }
  if (is.null(obj) || identical(obj, "")) {
    stop("Cannot determine objective from xgb.Booster.", call. = FALSE)
  }
  cls_prefixes <- c("binary:", "multi:")
  if (any(sapply(cls_prefixes, function(p) startsWith(obj, p)))) {
    return("classification")
  }
  "regression"
}

#' @export
#' @method get_ensemble_type lgb.Booster
get_ensemble_type.lgb.Booster <- function(ensemble) {
  check_package("lightgbm")
  obj <- ensemble$params$objective
  if (is.null(obj)) stop("Cannot determine objective from lgb.Booster.", call. = FALSE)
  cls_objs <- c("binary", "multiclass", "softmax", "multiclassova",
                "ovr", "multiclass_ova", "cross_entropy")
  if (any(sapply(cls_objs, function(x) startsWith(tolower(obj), x)))) {
    return("classification")
  }
  "regression"
}

#' @export
#' @method get_ensemble_type gbm
get_ensemble_type.gbm <- function(ensemble) {
  check_package("gbm")
  dist <- ensemble$distribution$name
  reg_dists <- c("gaussian", "laplace", "tdist", "quantile",
                 "huberized", "tweedie", "poisson", "gamma", "coxph")
  if (dist %in% reg_dists) "regression" else "classification"
}

#' @export
#' @method get_ensemble_type catboost.CatBoost
get_ensemble_type.catboost.CatBoost <- function(ensemble) {
  .catboost_get_ensemble_type(ensemble)
}

#' @export
#' @method get_ensemble_type catboost.Model
get_ensemble_type.catboost.Model <- function(ensemble) {
  .catboost_get_ensemble_type(ensemble)
}

# Resolve catboost exports without `catboost::` so R CMD check does not
# require it as a declared dependency (catboost is not on CRAN).
.catboost_fn <- function(name) {
  check_package("catboost")
  getExportedValue("catboost", name)
}

# Internal worker shared by both CatBoost class names.  Older releases of the
# catboost R package returned objects of class \code{catboost.CatBoost}; from
# release 1.2.x the class is \code{catboost.Model}.  Both names are supported.
.catboost_get_ensemble_type <- function(ensemble) {
  check_package("catboost")
  params <- tryCatch(
    .catboost_fn("catboost.get_model_params")(ensemble),
    error = function(e) NULL
  )
  loss <- .catboost_extract_loss(params)
  if (is.null(loss)) {
    # Last-resort fallback: inspect the user-side params attribute that
    # `catboost.train()` typically attaches to the returned model.
    loss <- .catboost_extract_loss(attr(ensemble, "params"))
  }
  if (is.null(loss) || !is.character(loss) || length(loss) == 0L) {
    stop("Cannot determine 'loss_function' from CatBoost model.", call. = FALSE)
  }
  cls_losses <- c("Logloss", "CrossEntropy", "MultiClass", "MultiClassOneVsAll")
  if (any(startsWith(loss[1L], cls_losses))) {
    return("classification")
  }
  "regression"
}

# Pull the loss function name out of a CatBoost params list.  Across catboost
# R-package versions the entry can be:
#   * a plain character scalar      ("MultiClass")
#   * a named list with "type"      (list(type = "MultiClass", ...))
#   * a list whose first element is the name (legacy odd serialisations)
#   * absent entirely (NULL params)
# This helper canonicalises all four shapes to a length-one character or NULL.
.catboost_extract_loss <- function(params) {
  if (is.null(params)) return(NULL)
  loss <- params[["loss_function"]]
  if (is.null(loss) && !is.null(params[["loss_function_params"]])) {
    loss <- params[["loss_function_params"]]
  }
  if (is.null(loss)) return(NULL)
  if (is.list(loss)) {
    nm <- loss[["type"]]
    if (is.null(nm)) nm <- loss[["name"]]
    if (is.null(nm) && length(loss) > 0L) nm <- loss[[1L]]
    loss <- nm
  }
  if (is.null(loss)) return(NULL)
  loss <- as.character(loss)
  if (length(loss) == 0L || all(is.na(loss)) || all(!nzchar(loss))) return(NULL)
  loss[1L]
}

#' @export
#' @method get_ensemble_type default
get_ensemble_type.default <- function(ensemble) {
  .unsupported_backend("get_ensemble_type", ensemble)
}


# ============================================================================
# 2.  extract_terminal_nodes()
# ============================================================================

#' Extract Terminal Node Assignments from an Ensemble Model
#'
#' Returns a \code{data.frame} with \code{n_obs} rows and \code{n_trees}
#' columns where each cell is the terminal-node index assigned to that
#' observation by that tree.
#'
#' @param ensemble A trained ensemble model.
#' @param data A \code{data.frame} of observations (may include the response
#'   column; it is ignored internally).
#' @return A \code{data.frame} with \code{n_obs} rows and \code{n_trees} columns
#'   of integer terminal-node identifiers.
#' @export
extract_terminal_nodes <- function(ensemble, data) {
  UseMethod("extract_terminal_nodes")
}

#' @export
#' @method extract_terminal_nodes randomForest
extract_terminal_nodes.randomForest <- function(ensemble, data) {
  as.data.frame(
    attr(predict(ensemble, newdata = data, nodes = TRUE), "nodes")
  )
}

#' @export
#' @method extract_terminal_nodes ranger
extract_terminal_nodes.ranger <- function(ensemble, data) {
  if (is.null(ensemble$forest)) {
    stop("ranger model must be trained with write.forest = TRUE.", call. = FALSE)
  }
  as.data.frame(
    predict(ensemble, data, type = "terminalNodes", num.threads = 1L)$predictions
  )
}

#' @export
#' @method extract_terminal_nodes xgb.Booster
extract_terminal_nodes.xgb.Booster <- function(ensemble, data) {
  check_package("xgboost")
  dm <- .xgb_dmatrix(ensemble, data)
  as.data.frame(predict(.xgb_strip_wrapper(ensemble), dm, predleaf = TRUE))
}

#' @export
#' @method extract_terminal_nodes lgb.Booster
extract_terminal_nodes.lgb.Booster <- function(ensemble, data) {
  check_package("lightgbm")
  feat <- .lgb_feature_names(ensemble)
  X    <- as.matrix(data[, feat, drop = FALSE])
  as.data.frame(ensemble$predict(X, predleaf = TRUE))
}

#' @export
#' @method extract_terminal_nodes gbm
extract_terminal_nodes.gbm <- function(ensemble, data) {
  check_package("gbm")
  n_trees <- ensemble$n.trees
  n_obs   <- nrow(data)
  X       <- data[, ensemble$var.names, drop = FALSE]

  # Convert factor columns to factors with the same levels as in training
  for (v in ensemble$var.names) {
    if (is.character(X[[v]])) X[[v]] <- as.factor(X[[v]])
  }

  leaf_mat <- matrix(0L, nrow = n_obs, ncol = n_trees)
  for (t in seq_len(n_trees)) {
    leaf_mat[, t] <- .gbm_leaf_indices(ensemble$trees[[t]], X, ensemble$c.splits)
  }
  as.data.frame(leaf_mat)
}

#' @export
#' @method extract_terminal_nodes catboost.CatBoost
extract_terminal_nodes.catboost.CatBoost <- function(ensemble, data) {
  .catboost_extract_terminal_nodes(ensemble, data)
}

#' @export
#' @method extract_terminal_nodes catboost.Model
extract_terminal_nodes.catboost.Model <- function(ensemble, data) {
  .catboost_extract_terminal_nodes(ensemble, data)
}

.catboost_extract_terminal_nodes <- function(ensemble, data) {
  check_package("catboost")

  # Build a catboost pool (without label)
  pool <- .catboost_pool_from_data(ensemble, data)

  n_trees <- .catboost_n_trees(ensemble)
  n_obs   <- nrow(data)
  leaf_mat <- matrix(0L, nrow = n_obs, ncol = n_trees)

  # Predict raw score from each individual tree and discretise into leaf IDs.
  # CatBoost R API does not expose leaf indices directly; raw scores per tree
  # are unique per leaf and serve as a faithful proxy.  For multi-class
  # objectives the per-tree prediction is a matrix of shape
  # (n_obs x n_classes); we collapse it to a single key per observation by
  # concatenating the rounded raw scores, so two observations sharing the
  # same leaf still receive the same integer identifier.
  for (t in seq_len(n_trees)) {
    raw <- .catboost_fn("catboost.predict")(
      ensemble, pool,
      prediction_type = "RawFormulaVal",
      ntree_start     = t - 1L,
      ntree_end       = t
    )
    if (is.matrix(raw) || (!is.null(dim(raw)) && length(dim(raw)) == 2L)) {
      # Multi-class: catboost.predict() returns an (n_obs x n_classes) matrix.
      raw_mat <- as.matrix(raw)
      if (nrow(raw_mat) != n_obs && ncol(raw_mat) == n_obs) {
        raw_mat <- t(raw_mat)
      }
      raw_mat <- round(raw_mat, digits = 10L)
      key <- apply(raw_mat, 1L, paste, collapse = "|")
    } else {
      raw_vec <- as.numeric(raw)
      if (length(raw_vec) == n_obs) {
        key <- as.character(round(raw_vec, digits = 10L))
      } else {
        stop(sprintf(
          "extract_terminal_nodes(catboost): unexpected raw-score length (%d) for tree %d (expected %d or an n_obs-row matrix).",
          length(raw_vec), t, n_obs), call. = FALSE)
      }
    }
    leaf_mat[, t] <- as.integer(factor(key))
  }
  as.data.frame(leaf_mat)
}

#' @export
#' @method extract_terminal_nodes default
extract_terminal_nodes.default <- function(ensemble, data) {
  .unsupported_backend("extract_terminal_nodes", ensemble)
}


# ============================================================================
# 3.  get_ensemble_predictions()
# ============================================================================

#' Get Training Predictions from an Ensemble Model
#'
#' Returns a numeric vector of length \code{n_obs} with the ensemble's
#' prediction for every training observation. For models that store
#' out-of-bag (OOB) predictions (\code{randomForest}, \code{ranger}) the
#' stored OOB vector is returned; for other models in-sample predictions
#' are computed from the training data.
#'
#' @param ensemble A trained ensemble model.
#' @param data The training \code{data.frame} that was used to fit the model.
#' @param type Character: \code{"classification"} or \code{"regression"}.
#' @return Numeric vector of length \code{nrow(data)}.
#' @export
get_ensemble_predictions <- function(ensemble, data, type) {
  UseMethod("get_ensemble_predictions")
}

#' @export
#' @method get_ensemble_predictions randomForest
get_ensemble_predictions.randomForest <- function(ensemble, data, type) {
  as.numeric(ensemble$predicted)
}

#' @export
#' @method get_ensemble_predictions ranger
get_ensemble_predictions.ranger <- function(ensemble, data, type) {
  as.numeric(ensemble$predictions)
}

#' @export
#' @method get_ensemble_predictions xgb.Booster
get_ensemble_predictions.xgb.Booster <- function(ensemble, data, type) {
  check_package("xgboost")
  as.numeric(predict(.xgb_strip_wrapper(ensemble), .xgb_dmatrix(ensemble, data)))
}

#' @export
#' @method get_ensemble_predictions lgb.Booster
get_ensemble_predictions.lgb.Booster <- function(ensemble, data, type) {
  check_package("lightgbm")
  feat <- .lgb_feature_names(ensemble)
  X <- as.matrix(data[, feat, drop = FALSE])
  as.numeric(ensemble$predict(X))
}

#' @export
#' @method get_ensemble_predictions gbm
get_ensemble_predictions.gbm <- function(ensemble, data, type) {
  check_package("gbm")
  as.numeric(
    gbm::predict.gbm(ensemble, newdata = data,
                     n.trees = ensemble$n.trees, type = "response")
  )
}

#' @export
#' @method get_ensemble_predictions catboost.CatBoost
get_ensemble_predictions.catboost.CatBoost <- function(ensemble, data, type) {
  .catboost_get_ensemble_predictions(ensemble, data, type)
}

#' @export
#' @method get_ensemble_predictions catboost.Model
get_ensemble_predictions.catboost.Model <- function(ensemble, data, type) {
  .catboost_get_ensemble_predictions(ensemble, data, type)
}

.catboost_get_ensemble_predictions <- function(ensemble, data, type) {
  check_package("catboost")
  pool      <- .catboost_pool_from_data(ensemble, data)
  pred_type <- if (type == "classification") "Class" else "RawFormulaVal"
  as.numeric(.catboost_fn("catboost.predict")(ensemble, pool, prediction_type = pred_type))
}

#' @export
#' @method get_ensemble_predictions default
get_ensemble_predictions.default <- function(ensemble, data, type) {
  .unsupported_backend("get_ensemble_predictions", ensemble)
}


# ============================================================================
# Internal helpers
# ============================================================================

# Convenience wrapper: build a double-precision predictor matrix and wrap it
# in an xgb.DMatrix in one call. Used by both extract_terminal_nodes() and
# get_ensemble_predictions() for the xgb.Booster backend.
.xgb_dmatrix <- function(ensemble, data) {
  X <- .xgb_predictor_matrix(ensemble, data)
  xgboost::xgb.DMatrix(data = X)
}

# In xgboost >= 3.x, models built via xgboost::xgboost() carry the additional
# class "xgboost" with its own predict method that rejects xgb.DMatrix inputs.
# To force dispatch to predict.xgb.Booster (which we need for `predleaf`), we
# strip the wrapper class for the duration of the prediction call.
.xgb_strip_wrapper <- function(ensemble) {
  cls <- class(ensemble)
  if ("xgboost" %in% cls && "xgb.Booster" %in% cls) {
    class(ensemble) <- setdiff(cls, "xgboost")
  }
  ensemble
}


# Build a double-precision predictor matrix for xgb.DMatrix.
# In xgboost >= 3.x, `ensemble$feature_names` is NULL; we fall back to all
# numeric columns, trimming to `num_feature` if data contains extra columns
# (e.g., the response variable appended by the user).
.xgb_predictor_matrix <- function(ensemble, data) {
  feat <- ensemble$feature_names
  if (length(feat) > 0L) {
    X <- as.matrix(data[, feat, drop = FALSE])
  } else {
    # xgboost >= 3.x: feature names not stored — use num_feature from config
    num_feat <- tryCatch({
      cfg <- xgboost::xgb.config(ensemble)
      as.integer(cfg[["learner"]][["learner_model_param"]][["num_feature"]])
    }, error = function(e) -1L)

    num_cols <- vapply(as.data.frame(data), is.numeric, logical(1L))
    X <- as.matrix(as.data.frame(data)[, num_cols, drop = FALSE])

    if (num_feat > 0L && ncol(X) > num_feat) {
      # Extra numeric columns present (e.g., response appended at the end)
      X <- X[, seq_len(num_feat), drop = FALSE]
    }
  }
  storage.mode(X) <- "double"
  X
}


# Extract feature names from an lgb.Booster model.
# In lightgbm < 4.x they are in private$used_feature_names; in 4.x they must
# be parsed from the dump_model() JSON string.
.lgb_feature_names <- function(ensemble) {
  # Old API (lightgbm < 4.x)
  feat <- tryCatch(
    ensemble$.__enclos_env__$private$used_feature_names,
    error = function(e) NULL
  )
  if (!is.null(feat) && length(feat) > 0L) return(feat)

  # New API (lightgbm >= 4.x): parse feature_names from the model JSON dump
  dmp <- tryCatch(ensemble$dump_model(), error = function(e) NULL)
  if (is.null(dmp)) {
    stop("Cannot determine feature names from lgb.Booster model.", call. = FALSE)
  }
  # Locate "feature_names":[...] and extract the array content
  tag_pos <- regexpr('"feature_names"', dmp, fixed = TRUE)
  if (tag_pos == -1L) {
    stop("Cannot parse feature_names from lgb.Booster dump.", call. = FALSE)
  }
  rest        <- substring(dmp, tag_pos)
  open_pos    <- regexpr("[", rest, fixed = TRUE)
  close_pos   <- regexpr("]", rest, fixed = TRUE)
  array_inner <- substring(rest, open_pos + 1L, close_pos - 1L)
  # array_inner is like: "cyl","disp","hp",...
  vals <- gsub('"', "", unlist(strsplit(array_inner, ",")), fixed = TRUE)
  trimws(vals)
}


# Traverse a single gbm tree and return the (1-indexed) leaf node for each obs.
# tree     : data.frame from ensemble$trees[[t]]
# X        : data.frame of predictor columns (already subset to var.names)
# c_splits : ensemble$c.splits (list of categorical split directions)
.gbm_leaf_indices <- function(tree, X, c_splits) {
  n_obs <- nrow(X)
  # Root node is row 1 (gbm uses 0-based indices internally).
  # ensemble$trees[[t]] is a plain list (not a data.frame): access by position.
  # [[1]]=SplitVar, [[2]]=SplitCodePred, [[3]]=LeftNode,
  # [[4]]=RightNode, [[5]]=MissingNode
  split_var  <- tree[[1L]]
  split_code <- tree[[2L]]
  left_node  <- tree[[3L]]
  right_node <- tree[[4L]]
  miss_node  <- tree[[5L]]

  current_nodes <- rep(0L, n_obs)

  repeat {
    sv <- split_var[current_nodes + 1L]
    # All observations have reached a terminal node
    if (all(sv == -1L)) break

    next_nodes <- current_nodes

    # Vectorised traversal: process each distinct non-terminal node
    non_terminal <- unique(current_nodes[sv != -1L])

    for (node_idx in non_terminal) {
      in_node  <- which(current_nodes == node_idx)
      sv_val   <- split_var[node_idx + 1L]       # variable index (0-based)
      sc_val   <- split_code[node_idx + 1L]       # split value / cat index
      left_n   <- left_node[node_idx + 1L]
      right_n  <- right_node[node_idx + 1L]
      miss_n   <- miss_node[node_idx + 1L]

      var_name <- names(X)[sv_val + 1L]  # convert to 1-based
      x_vals   <- X[[var_name]][in_node]

      if (is.numeric(x_vals)) {
        is_miss   <- is.na(x_vals)
        goes_left <- !is_miss & (x_vals < sc_val)
      } else {
        # Categorical variable: sc_val indexes into c_splits
        cat_split <- c_splits[[as.integer(sc_val) + 1L]]
        lvl_idx   <- match(as.character(x_vals), levels(X[[var_name]]))
        is_miss   <- is.na(lvl_idx)
        goes_left <- !is_miss & (cat_split[lvl_idx] == -1L)
      }

      next_nodes[in_node[is_miss]]              <- miss_n
      next_nodes[in_node[!is_miss & goes_left]] <- left_n
      next_nodes[in_node[!is_miss & !goes_left]] <- right_n
    }

    current_nodes <- next_nodes
  }

  current_nodes + 1L  # return 1-indexed leaf IDs
}


# Build a catboost.Pool from a data.frame, stripping the response column if
# present.  The label column name is stored as an attribute on the model by
# e2tree/createDisMatrix callers; if absent we strip nothing.
.catboost_pool_from_data <- function(ensemble, data) {
  label_col <- attr(ensemble, "e2tree_label")
  if (!is.null(label_col) && label_col %in% colnames(data)) {
    data <- data[, setdiff(colnames(data), label_col), drop = FALSE]
  }
  .catboost_fn("catboost.load_pool")(data = data)
}


# Return the number of trees in a CatBoost model.
.catboost_n_trees <- function(ensemble) {
  params <- tryCatch(
    .catboost_fn("catboost.get_model_params")(ensemble),
    error = function(e) NULL
  )
  n <- .catboost_extract_n_trees(params)
  if (is.null(n)) n <- .catboost_extract_n_trees(attr(ensemble, "params"))
  if (is.null(n)) {
    # Some catboost versions expose the count via dedicated helpers.
    n <- tryCatch(
      as.integer(.catboost_fn("catboost.get_model_params")(ensemble)$boosting_options$iterations),
      error = function(e) NULL
    )
  }
  if (is.null(n) || !is.finite(n) || n <= 0L) {
    stop("Cannot determine number of trees from catboost model.", call. = FALSE)
  }
  as.integer(n)
}

# Walk a CatBoost params list looking for an `iterations` field.  Different
# catboost releases place it at the top level, under `boosting_options`, or
# under `tree_learner_options`; this helper handles all three.
.catboost_extract_n_trees <- function(params) {
  if (is.null(params) || !is.list(params)) return(NULL)
  if (!is.null(params$iterations)) return(as.integer(params$iterations[1L]))
  if (is.list(params$boosting_options) && !is.null(params$boosting_options$iterations)) {
    return(as.integer(params$boosting_options$iterations[1L]))
  }
  if (is.list(params$tree_learner_options) && !is.null(params$tree_learner_options$iterations)) {
    return(as.integer(params$tree_learner_options$iterations[1L]))
  }
  NULL
}

library(testthat)
library(e2tree)

# ============================================================================
# Multi-backend integration tests for createDisMatrix() and the adapter layer.
#
# These tests assert the *invariants* that any supported backend must satisfy:
#   - D is a square dense matrix of the right size,
#   - D has zeros on the diagonal,
#   - D is approximately symmetric,
#   - D is non-degenerate (i.e. it is not the all-zero matrix — this would
#     indicate the silent leaf-extraction bug fixed for gbm),
#   - D carries the ensemble_backend attribute set by createDisMatrix(),
#   - validate_terminal_nodes() rejects malformed adapter output.
#
# Each backend is skipped automatically when the corresponding R package is
# not available, so the file is safe to run in any CI environment.
# ============================================================================

assert_dis_invariants <- function(D, n, backend) {
  expect_true(is.matrix(D))
  expect_equal(dim(D), c(n, n))
  expect_equal(as.numeric(diag(D)), rep(0, n),
               info = sprintf("backend=%s: diagonal must be zero", backend))
  expect_lt(max(abs(D - t(D))), 1e-8,
            label = sprintf("backend=%s: D must be symmetric", backend))
  # Non-degenerate: at least one off-diagonal entry must be non-zero. An
  # all-zero D is the smoking gun of an extract_terminal_nodes() failure.
  expect_gt(sum(D), 0,
            label = sprintf("backend=%s: D must be non-degenerate", backend))
  expect_equal(attr(D, "ensemble_backend"), backend)
}

prep_iris <- function() {
  set.seed(42)
  data(iris)
  idx <- sample(seq_len(nrow(iris)), size = floor(0.6 * nrow(iris)))
  iris[idx, ]
}

prep_mtcars <- function() {
  set.seed(42)
  data(mtcars)
  idx <- sample(seq_len(nrow(mtcars)), size = floor(0.7 * nrow(mtcars)))
  mtcars[idx, ]
}

# ----------------------------------------------------------------------------
# randomForest (regression) — baseline sanity (covered elsewhere for clf)
# ----------------------------------------------------------------------------
test_that("randomForest regression: D invariants hold", {
  skip_if_not_installed("randomForest")
  tr <- prep_mtcars()
  ens <- randomForest::randomForest(mpg ~ ., data = tr, ntree = 50)
  D <- createDisMatrix(ens, data = tr, label = "mpg",
                       parallel = list(active = FALSE, no_cores = 1))
  assert_dis_invariants(D, nrow(tr), "randomForest")
})

# ----------------------------------------------------------------------------
# ranger
# ----------------------------------------------------------------------------
test_that("ranger classification: D invariants hold", {
  skip_if_not_installed("ranger")
  tr <- prep_iris()
  ens <- ranger::ranger(Species ~ ., data = tr, num.trees = 100,
                        write.forest = TRUE)
  D <- createDisMatrix(ens, data = tr, label = "Species",
                       parallel = list(active = FALSE, no_cores = 1))
  assert_dis_invariants(D, nrow(tr), "ranger")
})

test_that("ranger regression: D invariants hold", {
  skip_if_not_installed("ranger")
  tr <- prep_mtcars()
  ens <- ranger::ranger(mpg ~ ., data = tr, num.trees = 100,
                        write.forest = TRUE)
  D <- createDisMatrix(ens, data = tr, label = "mpg",
                       parallel = list(active = FALSE, no_cores = 1))
  assert_dis_invariants(D, nrow(tr), "ranger")
})

# ----------------------------------------------------------------------------
# xgboost
# ----------------------------------------------------------------------------
test_that("xgb.Booster regression: D invariants hold", {
  skip_if_not_installed("xgboost")
  tr <- prep_mtcars()
  X <- as.matrix(tr[, setdiff(names(tr), "mpg")])
  y <- tr$mpg
  # xgb.train returns a plain xgb.Booster (no `xgboost` wrapper class) and is
  # API-stable across xgboost 1.x/2.x/3.x; use it for clean test output.
  dtrain <- xgboost::xgb.DMatrix(data = X, label = y)
  ens <- xgboost::xgb.train(
    params = list(objective = "reg:squarederror"),
    data = dtrain, nrounds = 30, verbose = 0
  )
  D <- createDisMatrix(ens, data = tr, label = "mpg",
                       parallel = list(active = FALSE, no_cores = 1))
  assert_dis_invariants(D, nrow(tr), "xgb.Booster")
})

test_that("xgb.Booster (high-level wrapper) is also handled", {
  skip_if_not_installed("xgboost")
  tr <- prep_mtcars()
  X <- as.matrix(tr[, setdiff(names(tr), "mpg")])
  y <- tr$mpg
  # The high-level xgboost::xgboost() returns class c("xgboost","xgb.Booster")
  # in xgboost >= 3.x. The adapter must dispatch to the booster method even
  # when the wrapper class is present.
  suppressWarnings(
    ens <- xgboost::xgboost(data = X, label = y, nrounds = 20,
                            objective = "reg:squarederror", verbose = 0)
  )
  D <- createDisMatrix(ens, data = tr, label = "mpg",
                       parallel = list(active = FALSE, no_cores = 1))
  assert_dis_invariants(D, nrow(tr), "xgb.Booster")
})

# ----------------------------------------------------------------------------
# gbm — historically the failing backend; explicit guard
# ----------------------------------------------------------------------------
test_that("gbm regression: D is not the all-zero degenerate matrix", {
  skip_if_not_installed("gbm")
  tr <- prep_mtcars()
  # Loosen gbm defaults so the small sample is acceptable.
  ens <- gbm::gbm(mpg ~ ., data = tr, distribution = "gaussian",
                  n.trees = 50, interaction.depth = 3, verbose = FALSE,
                  bag.fraction = 0.9, n.minobsinnode = 2)
  D <- createDisMatrix(ens, data = tr, label = "mpg",
                       parallel = list(active = FALSE, no_cores = 1))
  assert_dis_invariants(D, nrow(tr), "gbm")
})

# ----------------------------------------------------------------------------
# lightgbm
# ----------------------------------------------------------------------------
test_that("lgb.Booster regression: D invariants hold", {
  skip_if_not_installed("lightgbm")
  tr <- prep_mtcars()
  X <- as.matrix(tr[, setdiff(names(tr), "mpg")])
  y <- tr$mpg
  dtrain <- lightgbm::lgb.Dataset(data = X, label = y)
  # Default min_data_in_leaf=20 collapses tiny samples to a single leaf,
  # which would (correctly) be flagged as degenerate by the validator.
  ens <- lightgbm::lgb.train(
    params = list(objective = "regression", verbose = -1,
                  min_data_in_leaf = 2L, min_data_in_bin = 1L),
    data = dtrain, nrounds = 30
  )
  D <- createDisMatrix(ens, data = tr, label = "mpg",
                       parallel = list(active = FALSE, no_cores = 1))
  assert_dis_invariants(D, nrow(tr), "lgb.Booster")
})

# ----------------------------------------------------------------------------
# Adapter contract (no backend dependency required)
# ----------------------------------------------------------------------------
test_that("validate_terminal_nodes rejects degenerate output", {
  bad <- data.frame(Tree1 = rep(0L, 10), Tree2 = rep(0L, 10))
  expect_error(
    e2tree:::validate_terminal_nodes(bad, data.frame(x = 1:10), backend = "fake"),
    "degenerate"
  )
})

test_that("validate_terminal_nodes rejects wrong row count", {
  bad <- data.frame(Tree1 = 1:5)
  expect_error(
    e2tree:::validate_terminal_nodes(bad, data.frame(x = 1:10), backend = "fake"),
    "rows"
  )
})

test_that("validate_terminal_nodes rejects non-numeric columns", {
  bad <- data.frame(Tree1 = letters[1:10], stringsAsFactors = FALSE)
  expect_error(
    e2tree:::validate_terminal_nodes(bad, data.frame(x = 1:10), backend = "fake"),
    "non-numeric"
  )
})

test_that("default adapter methods raise an informative error", {
  bogus <- structure(list(), class = "not_a_real_ensemble")
  expect_error(get_ensemble_type(bogus), "Supported classes")
  expect_error(extract_terminal_nodes(bogus, data.frame(x = 1:3)), "Supported classes")
  expect_error(get_ensemble_predictions(bogus, data.frame(x = 1:3), "regression"),
               "Supported classes")
})

# ----------------------------------------------------------------------------
# End-to-end pipeline: D → e2tree → as.rpart for boosting backends.
# Catches regressions in the full surrogate-tree workflow when the underlying
# ensemble is not random-forest-like.
# ----------------------------------------------------------------------------
test_that("xgboost full pipeline (D -> e2tree -> as.rpart) works", {
  skip_if_not_installed("xgboost")
  tr <- prep_mtcars()
  X <- as.matrix(tr[, setdiff(names(tr), "mpg")])
  y <- tr$mpg
  dtrain <- xgboost::xgb.DMatrix(data = X, label = y)
  ens <- xgboost::xgb.train(
    params = list(objective = "reg:squarederror"),
    data = dtrain, nrounds = 20, verbose = 0
  )
  D <- createDisMatrix(ens, data = tr, label = "mpg",
                       parallel = list(active = FALSE, no_cores = 1))
  setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
  fit <- e2tree(mpg ~ ., tr, D, ens, setting)
  expect_s3_class(fit, "e2tree")
  expect_equal(attr(fit, "ensemble_backend"), "xgb.Booster")
  rp <- as.rpart(fit, ens)
  expect_true(all(c("frame", "where") %in% names(rp)))
})

test_that("xgboost binary classification: D invariants hold", {
  skip_if_not_installed("xgboost")
  set.seed(7)
  data(iris)
  df <- iris[iris$Species != "virginica", ]
  df$Species <- factor(df$Species)
  idx <- sample(seq_len(nrow(df)), floor(0.6 * nrow(df)))
  tr <- df[idx, ]
  X <- as.matrix(tr[, 1:4])
  y <- as.integer(tr$Species) - 1L
  dtrain <- xgboost::xgb.DMatrix(data = X, label = y)
  ens <- xgboost::xgb.train(
    params = list(objective = "binary:logistic"),
    data = dtrain, nrounds = 20, verbose = 0
  )
  D <- createDisMatrix(ens, data = tr, label = "Species",
                       parallel = list(active = FALSE, no_cores = 1))
  assert_dis_invariants(D, nrow(tr), "xgb.Booster")
})

test_that("createDisMatrix rejects unsupported ensembles before adapter call", {
  expect_error(
    createDisMatrix(structure(list(), class = "fake_model"),
                    data = mtcars, label = "mpg"),
    "unsupported ensemble class"
  )
})

test_that("createDisMatrix rejects non-data.frame data", {
  skip_if_not_installed("randomForest")
  ens <- randomForest::randomForest(mpg ~ ., data = mtcars, ntree = 10)
  expect_error(
    createDisMatrix(ens, data = as.matrix(mtcars), label = "mpg"),
    "must be a valid data frame"
  )
})

test_that("createDisMatrix rejects label not in data", {
  skip_if_not_installed("randomForest")
  ens <- randomForest::randomForest(mpg ~ ., data = mtcars, ntree = 10)
  expect_error(
    createDisMatrix(ens, data = mtcars, label = "not_a_column"),
    "valid column name"
  )
})

test_that("regression createDisMatrix requires non-NULL label", {
  skip_if_not_installed("randomForest")
  ens <- randomForest::randomForest(mpg ~ ., data = mtcars, ntree = 10)
  expect_error(
    createDisMatrix(ens, data = mtcars, label = NULL),
    "label.*required"
  )
})

test_that("D from chunked path is bit-identical to non-chunked path", {
  skip_if_not_installed("randomForest")
  set.seed(42)
  ens <- randomForest::randomForest(mpg ~ ., data = mtcars, ntree = 30)
  D_full <- createDisMatrix(ens, data = mtcars, label = "mpg",
                            parallel = list(active = FALSE, no_cores = 1))
  D_chunk <- createDisMatrix(ens, data = mtcars, label = "mpg",
                             parallel = list(active = FALSE, no_cores = 1),
                             chunk_size = 8)
  expect_equal(unname(D_full), unname(D_chunk))
})

test_that("ensemble_backend() returns NA for unknown classes", {
  expect_true(is.na(e2tree:::ensemble_backend(list())))
  expect_true(is.na(e2tree:::ensemble_backend(NULL)))
  expect_equal(e2tree:::ensemble_backend(
    structure(list(), class = "randomForest")), "randomForest")
})

test_that("C++ entry points reject non-numeric tree columns", {
  bad <- data.frame(OBS = as.character(1:4),
                    Tree1 = c("a","b","a","b"),
                    resp = c(1,2,3,4),
                    stringsAsFactors = FALSE)
  expect_error(
    e2tree:::compute_all_cooccurrences_cpp("regression", bad, 1L, 1L, 1.0),
    "must be integer or double"
  )
})

test_that("e2tree warns when D and ensemble come from different backends", {
  skip_if_not_installed("randomForest")
  skip_if_not_installed("ranger")
  tr <- prep_iris()
  rf <- randomForest::randomForest(Species ~ ., data = tr, ntree = 50)
  rg <- ranger::ranger(Species ~ ., data = tr, num.trees = 50,
                       write.forest = TRUE)
  D_rf <- createDisMatrix(rf, data = tr, label = "Species",
                          parallel = list(active = FALSE, no_cores = 1))
  setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
  expect_warning(
    e2tree(Species ~ ., tr, D_rf, rg, setting),
    "different backends|backend|consistent"
  )
})

library(testthat)
library(e2tree)

# ===========================================================================
# Helpers
# ===========================================================================

.make_rf_clf <- function() {
  skip_if_not_installed("randomForest")
  library(randomForest)
  set.seed(1)
  randomForest(Species ~ ., data = iris, ntree = 50)
}

.make_rf_reg <- function() {
  skip_if_not_installed("randomForest")
  library(randomForest)
  set.seed(1)
  randomForest(mpg ~ ., data = mtcars, ntree = 50)
}

.make_ranger_clf <- function() {
  skip_if_not_installed("ranger")
  library(ranger)
  set.seed(1)
  ranger(Species ~ ., data = iris, num.trees = 50, write.forest = TRUE)
}

.make_ranger_reg <- function() {
  skip_if_not_installed("ranger")
  library(ranger)
  set.seed(1)
  ranger(mpg ~ ., data = mtcars, num.trees = 50, write.forest = TRUE)
}

.make_xgb_clf <- function() {
  skip_if_not_installed("xgboost")
  library(xgboost)
  set.seed(1)
  X  <- as.matrix(iris[, 1:4])
  y  <- as.integer(iris$Species) - 1L
  dm <- xgb.DMatrix(data = X, label = y)
  xgb.train(
    params = list(objective = "multi:softmax", num_class = 3,
                  max_depth = 3, eta = 0.3),
    data = dm, nrounds = 20, verbose = 0
  )
}

.make_xgb_reg <- function() {
  skip_if_not_installed("xgboost")
  library(xgboost)
  set.seed(1)
  X  <- as.matrix(mtcars[, -1])
  y  <- mtcars$mpg
  dm <- xgb.DMatrix(data = X, label = y)
  xgb.train(
    params = list(objective = "reg:squarederror", max_depth = 3, eta = 0.3),
    data = dm, nrounds = 20, verbose = 0
  )
}

.make_gbm_clf <- function() {
  skip_if_not_installed("gbm")
  library(gbm)
  set.seed(1)
  df <- iris
  df$Species <- as.integer(df$Species == "setosa")
  gbm(Species ~ ., data = df, distribution = "bernoulli",
      n.trees = 50, interaction.depth = 3, verbose = FALSE)
}

.make_gbm_reg <- function() {
  skip_if_not_installed("gbm")
  library(gbm)
  set.seed(1)
  # mtcars has only 32 rows; the gbm defaults (bag.fraction=0.5,
  # n.minobsinnode=10) violate `nTrain*bag.fraction > 2*n.minobsinnode+1`,
  # so we relax both for the test fixture.
  gbm(mpg ~ ., data = mtcars, distribution = "gaussian",
      n.trees = 50, interaction.depth = 3, verbose = FALSE,
      bag.fraction = 0.9, n.minobsinnode = 2)
}

.make_lgb_clf <- function() {
  skip_if_not_installed("lightgbm")
  library(lightgbm)
  set.seed(1)
  X  <- as.matrix(iris[, 1:4])
  y  <- as.integer(iris$Species) - 1L
  ds <- lgb.Dataset(X, label = y)
  lgb.train(
    params = list(objective = "multiclass", num_class = 3,
                  num_leaves = 8, verbose = -1),
    data = ds, nrounds = 20
  )
}

.make_lgb_reg <- function() {
  skip_if_not_installed("lightgbm")
  library(lightgbm)
  set.seed(1)
  X  <- as.matrix(mtcars[, -1])
  y  <- mtcars$mpg
  ds <- lgb.Dataset(X, label = y)
  lgb.train(
    params = list(objective = "regression", num_leaves = 8, verbose = -1),
    data = ds, nrounds = 20
  )
}


# ===========================================================================
# Tests: get_ensemble_type()
# ===========================================================================

test_that("get_ensemble_type — randomForest", {
  m <- .make_rf_clf()
  expect_equal(get_ensemble_type(m), "classification")

  m2 <- .make_rf_reg()
  expect_equal(get_ensemble_type(m2), "regression")
})

test_that("get_ensemble_type — ranger", {
  m <- .make_ranger_clf()
  expect_equal(get_ensemble_type(m), "classification")

  m2 <- .make_ranger_reg()
  expect_equal(get_ensemble_type(m2), "regression")
})

test_that("get_ensemble_type — xgboost", {
  m <- .make_xgb_clf()
  expect_equal(get_ensemble_type(m), "classification")

  m2 <- .make_xgb_reg()
  expect_equal(get_ensemble_type(m2), "regression")
})

test_that("get_ensemble_type — gbm", {
  m <- .make_gbm_clf()
  expect_equal(get_ensemble_type(m), "classification")

  m2 <- .make_gbm_reg()
  expect_equal(get_ensemble_type(m2), "regression")
})

test_that("get_ensemble_type — lightgbm", {
  m <- .make_lgb_clf()
  expect_equal(get_ensemble_type(m), "classification")

  m2 <- .make_lgb_reg()
  expect_equal(get_ensemble_type(m2), "regression")
})

test_that("get_ensemble_type — unsupported class throws error", {
  expect_error(get_ensemble_type(list()), "unsupported ensemble class")
})


# ===========================================================================
# Tests: extract_terminal_nodes()
# ===========================================================================

test_that("extract_terminal_nodes — randomForest returns correct dimensions", {
  m   <- .make_rf_clf()
  res <- extract_terminal_nodes(m, iris)
  expect_true(is.data.frame(res))
  expect_equal(nrow(res), nrow(iris))
  expect_equal(ncol(res), m$ntree)
  expect_true(all(sapply(res, is.numeric)))
})

test_that("extract_terminal_nodes — ranger returns correct dimensions", {
  m   <- .make_ranger_clf()
  res <- extract_terminal_nodes(m, iris)
  expect_true(is.data.frame(res))
  expect_equal(nrow(res), nrow(iris))
  expect_equal(ncol(res), m$num.trees)
})

test_that("extract_terminal_nodes — xgboost returns correct dimensions", {
  m   <- .make_xgb_clf()
  res <- extract_terminal_nodes(m, iris[, 1:4])
  expect_true(is.data.frame(res))
  expect_equal(nrow(res), nrow(iris))
  # number of columns = n_rounds * n_class for multi:softmax with predleaf
  expect_gt(ncol(res), 0)
})

test_that("extract_terminal_nodes — gbm returns correct dimensions", {
  m   <- .make_gbm_reg()
  res <- extract_terminal_nodes(m, mtcars)
  expect_true(is.data.frame(res))
  expect_equal(nrow(res), nrow(mtcars))
  expect_equal(ncol(res), m$n.trees)
  expect_true(all(unlist(res) > 0))  # 1-indexed leaf IDs
})

test_that("extract_terminal_nodes — lightgbm returns correct dimensions", {
  m   <- .make_lgb_reg()
  res <- extract_terminal_nodes(m, as.data.frame(as.matrix(mtcars[, -1])))
  expect_true(is.data.frame(res))
  expect_equal(nrow(res), nrow(mtcars))
  expect_gt(ncol(res), 0)
})


# ===========================================================================
# Tests: get_ensemble_predictions()
# ===========================================================================

test_that("get_ensemble_predictions — randomForest returns numeric vector", {
  m   <- .make_rf_reg()
  preds <- get_ensemble_predictions(m, mtcars, "regression")
  expect_true(is.numeric(preds))
  expect_equal(length(preds), nrow(mtcars))
})

test_that("get_ensemble_predictions — ranger returns numeric vector", {
  m     <- .make_ranger_reg()
  preds <- get_ensemble_predictions(m, mtcars, "regression")
  expect_true(is.numeric(preds))
  expect_equal(length(preds), nrow(mtcars))
})

test_that("get_ensemble_predictions — xgboost returns numeric vector", {
  m     <- .make_xgb_reg()
  preds <- get_ensemble_predictions(m, mtcars[, -1], "regression")
  expect_true(is.numeric(preds))
  expect_equal(length(preds), nrow(mtcars))
})

test_that("get_ensemble_predictions — gbm returns numeric vector", {
  m     <- .make_gbm_reg()
  preds <- get_ensemble_predictions(m, mtcars, "regression")
  expect_true(is.numeric(preds))
  expect_equal(length(preds), nrow(mtcars))
})

test_that("get_ensemble_predictions — lightgbm returns numeric vector", {
  m     <- .make_lgb_reg()
  preds <- get_ensemble_predictions(m, as.data.frame(as.matrix(mtcars[, -1])), "regression")
  expect_true(is.numeric(preds))
  expect_equal(length(preds), nrow(mtcars))
})


# ===========================================================================
# Tests: createDisMatrix() backward-compatibility (RF and ranger unchanged)
# ===========================================================================

test_that("createDisMatrix — randomForest classification still works", {
  skip_if_not_installed("randomForest")
  library(randomForest)
  set.seed(42)
  n  <- floor(0.75 * nrow(iris))
  tr <- iris[sample(nrow(iris), n), ]
  m  <- randomForest(Species ~ ., data = tr, ntree = 50)
  D  <- createDisMatrix(m, data = tr, label = "Species",
                        parallel = list(active = FALSE, no_cores = 1))
  expect_true(is.matrix(D))
  expect_equal(dim(D), c(nrow(tr), nrow(tr)))
  expect_true(all(diag(D) == 0))
})

test_that("createDisMatrix — randomForest regression still works", {
  skip_if_not_installed("randomForest")
  library(randomForest)
  set.seed(42)
  n  <- floor(0.75 * nrow(mtcars))
  tr <- mtcars[sample(nrow(mtcars), n), ]
  m  <- randomForest(mpg ~ ., data = tr, ntree = 50)
  D  <- createDisMatrix(m, data = tr, label = "mpg",
                        parallel = list(active = FALSE, no_cores = 1))
  expect_true(is.matrix(D))
  expect_equal(dim(D), c(nrow(tr), nrow(tr)))
})

#' Predict Responses Using an Explainable Ensemble Tree
#'
#' Predicts classification and regression tree responses.
#'
#' \strong{Deprecated:} Use \code{\link{predict.e2tree}} instead.
#'
#' @param fit An e2tree object.
#' @param data A data frame with new observations.
#' @param target Target class for classification scoring.
#'
#' @return A data frame with predictions.
#'
#' @examples
#' \donttest{
#' ## Classification:
#' data(iris)
#'
#' # Create training and validation set:
#' smp_size <- floor(0.75 * nrow(iris))
#' train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
#' training <- iris[train_ind, ]
#' validation <- iris[-train_ind, ]
#' response_training <- training[,5]
#' response_validation <- validation[,5]
#'
#' # Perform training:
#' ensemble <- randomForest::randomForest(Species ~ ., data=training,
#' importance=TRUE, proximity=TRUE)
#'
#' D <- createDisMatrix(ensemble, data=training, label = "Species",
#'                              parallel = list(active=FALSE, no_cores = 1))
#'
#' setting=list(impTotal=0.1, maxDec=0.01, n=2, level=5)
#' tree <- e2tree(Species ~ ., training, D, ensemble, setting)
#'
#' ## Preferred method:
#' predict(tree, newdata = validation, target = "1")
#'
#' ## Legacy function (deprecated):
#' ePredTree(tree, validation, target = "1")
#'
#'
#' ## Regression
#' data("mtcars")
#'
#' # Create training and validation set:
#' smp_size <- floor(0.75 * nrow(mtcars))
#' train_ind <- sample(seq_len(nrow(mtcars)), size = smp_size)
#' training <- mtcars[train_ind, ]
#' validation <- mtcars[-train_ind, ]
#' response_training <- training[,1]
#' response_validation <- validation[,1]
#'
#' # Perform training
#' ensemble = randomForest::randomForest(mpg ~ ., data=training, ntree=1000,
#' importance=TRUE, proximity=TRUE)
#'
#' D = createDisMatrix(ensemble, data=training, label = "mpg",
#'                               parallel = list(active=FALSE, no_cores = 1))
#'
#' setting=list(impTotal=0.1, maxDec=(1*10^-6), n=2, level=5)
#' tree <- e2tree(mpg ~ ., training, D, ensemble, setting)
#'
#' ## Preferred method:
#' predict(tree, newdata = validation)
#'
#' ## Legacy function (deprecated):
#' ePredTree(tree, validation)
#'
#' }
#'
#' @export
#'
ePredTree <- function(fit, data, target="1") {
  tree <- fit$tree

  row.names(data) <- NULL
  n <- nrow(data)

  # Determine if the model is classification or regression
  is_classification <- !is.numeric(tree$pred)

  # Initialize the prediction dataframe (use correct type for fit column)
  if (is_classification) {
    pred <- data.frame(fit = rep(NA_character_, n), accuracy = NA_real_, score = NA_real_, row.names = 1:n)
  } else {
    pred <- data.frame(fit = rep(NA_real_, n), accuracy = NA_real_, score = NA_real_, row.names = 1:n)
  }

  # ---- Tree-traversal prediction (replaces eval/parse) ----
  # Start all observations at root node (node = 1)
  node_assignment <- rep(1L, n)

  # Internal (non-terminal) nodes, indexed by node number for fast lookup
  internal <- tree[tree$terminal == FALSE, , drop = FALSE]
  terminal <- tree[tree$terminal == TRUE, , drop = FALSE]

  # --- Pre-parse all split rules ONCE (avoid regex per prediction) ---
  split_cache <- parse_all_splits(internal$splitLabel)
  names(split_cache) <- internal$node

  # Traverse tree: assign each observation to a terminal node
  # Use named lookup for O(1) access to internal node rows
  internal_lookup <- setNames(seq_len(nrow(internal)), internal$node)

  max_iter <- nrow(tree) * 2  # safety limit
  iter <- 0
  while (iter < max_iter) {
    iter <- iter + 1
    in_internal <- node_assignment %in% internal$node
    if (!any(in_internal)) break

    active_nodes <- unique(node_assignment[in_internal])

    for (nd in active_nodes) {
      obs_idx <- which(node_assignment == nd)
      rule <- split_cache[[as.character(nd)]]
      if (is.null(rule)) next

      goes_left <- apply_split_rule(data, obs_idx, rule)

      node_assignment[obs_idx[goes_left]]  <- nd * 2L
      node_assignment[obs_idx[!goes_left]] <- nd * 2L + 1L
    }
  }

  # Now assign predictions from terminal nodes
  for (i in seq_len(nrow(terminal))) {
    nd <- terminal$node[i]
    obs_idx <- which(node_assignment == nd)
    if (length(obs_idx) == 0) next

    pred$fit[obs_idx] <- terminal$pred[i]

    if (is_classification) {
      prob_val <- as.numeric(terminal$prob[i])
      pred$accuracy[obs_idx] <- prob_val
      score_val <- prob_val
      if (!is.na(terminal$pred[i]) && terminal$pred[i] != target) {
        score_val <- 1 - prob_val
      }
      pred$score[obs_idx] <- score_val
    }
  }

  return(pred)
}


# Pre-parse all split labels into structured rules (called once)
parse_all_splits <- function(split_labels) {
  lapply(split_labels, function(lbl) {
    if (grepl(" <= ?", lbl)) {
      parts <- regmatches(lbl, regexec("^(.+?) <=(.+)$", lbl))[[1]]
      list(type = "numeric",
           var  = trimws(parts[2]),
           thr  = as.numeric(trimws(parts[3])))
    } else if (grepl(" %in% ", lbl)) {
      parts <- regmatches(lbl, regexec("^(.+?) %in% (.+)$", lbl))[[1]]
      cats_str <- parts[3]
      ## Parse c('a','b') once; safe because split labels are generated internally
      cats <- eval(parse(text = cats_str))
      list(type = "categorical",
           var  = trimws(parts[2]),
           cats = cats)
    } else {
      list(type = "unknown")
    }
  })
}

# Apply a pre-parsed split rule to data (no regex, no eval/parse)
apply_split_rule <- function(data, obs_idx, rule) {
  if (rule$type == "numeric") {
    return(data[[rule$var]][obs_idx] <= rule$thr)
  } else if (rule$type == "categorical") {
    return(as.character(data[[rule$var]][obs_idx]) %in% rule$cats)
  } else {
    warning("Unknown split format in cached rule")
    return(rep(TRUE, length(obs_idx)))
  }
}

#!/usr/bin/env Rscript
# =============================================================================
# e2tree — Comprehensive Simulation Test Suite
# =============================================================================
# Covers:
#   1.  Adapter correctness  (get_ensemble_type, extract_terminal_nodes,
#                              get_ensemble_predictions)
#   2.  Backward compatibility — RF/ranger leaf extraction matches native APIs;
#       D matrix is identical across two runs with the same model
#   3.  Dissimilarity matrix mathematical properties (symmetry, [0,1], zero
#       diagonal, non-trivial) for all available backends
#   4.  Core logic consistency — weighted co-occurrence formula is respected;
#       intra-class pairs have lower mean D than inter-class pairs
#   5.  Full e2tree pipeline — tree builds, predict() works, accuracy/RMSE
#       better than a trivial baseline
#   6.  Cross-backend coherence — D matrices from different backends on the
#       same data should be positively correlated (Spearman > 0.5)
#   7.  Determinism — same model + same data → identical D every call
#
# Usage:
#   Rscript tests/simulation_tests.R
# =============================================================================

suppressPackageStartupMessages(library(e2tree))

# ─── Test runner ─────────────────────────────────────────────────────────────

.results  <- list()
.pass_cnt <- 0L
.fail_cnt <- 0L

PASS <- function(name) {
  cat(sprintf("  \033[32m[PASS]\033[0m  %s\n", name))
  .pass_cnt <<- .pass_cnt + 1L
  .results[[length(.results) + 1L]] <<- list(name = name, status = "PASS")
}

FAIL <- function(name, reason = "") {
  cat(sprintf("  \033[31m[FAIL]\033[0m  %s", name))
  if (nzchar(reason)) cat(sprintf("  →  %s", reason))
  cat("\n")
  .fail_cnt <<- .fail_cnt + 1L
  .results[[length(.results) + 1L]] <<- list(name = name, status = "FAIL",
                                              reason = reason)
}

SKIP <- function(name, reason = "") {
  cat(sprintf("  \033[33m[SKIP]\033[0m  %s  →  %s\n", name, reason))
  .results[[length(.results) + 1L]] <<- list(name = name, status = "SKIP",
                                              reason = reason)
}

check <- function(name, expr) {
  ok <- tryCatch(isTRUE(expr), error = function(e) FALSE)
  if (ok) PASS(name) else FAIL(name, deparse(substitute(expr)))
}

section <- function(title) {
  cat(sprintf("\n\033[1;34m══════  %s  ══════\033[0m\n", title))
}

# ─── Shared data & settings ──────────────────────────────────────────────────

set.seed(42)

# Train / validation split — iris (classification)
n_iris     <- nrow(iris)
idx_iris   <- sample(n_iris, floor(0.75 * n_iris))
tr_iris    <- iris[idx_iris, ]
te_iris    <- iris[-idx_iris, ]

# Train / validation split — iris regression (Sepal.Length ~ rest)
tr_iris_reg <- tr_iris
te_iris_reg <- te_iris

# Train / validation split — mtcars (regression)
n_cars    <- nrow(mtcars)
idx_cars  <- sample(n_cars, floor(0.75 * n_cars))
tr_cars   <- mtcars[idx_cars, ]
te_cars   <- mtcars[-idx_cars, ]

SETTING_CLF <- list(impTotal = 0.1, maxDec = 0.01,  n = 5, level = 5)
SETTING_REG <- list(impTotal = 0.1, maxDec = 1e-6,  n = 3, level = 5)
PAR_OFF     <- list(active = FALSE, no_cores = 1)


# =============================================================================
# SECTION 1 — randomForest
# =============================================================================
section("1 · randomForest")

if (requireNamespace("randomForest", quietly = TRUE)) {
  suppressPackageStartupMessages(library(randomForest))
  set.seed(42)
  rf_clf <- randomForest(Species ~ ., data = tr_iris,  ntree = 100)
  rf_reg <- randomForest(mpg    ~ ., data = tr_cars, ntree = 100)

  # ── 1a. Adapter: get_ensemble_type ──
  check("RF clf  → 'classification'",
        get_ensemble_type(rf_clf) == "classification")
  check("RF reg  → 'regression'",
        get_ensemble_type(rf_reg) == "regression")

  # ── 1b. Adapter: extract_terminal_nodes ──
  nodes_rf <- extract_terminal_nodes(rf_clf, tr_iris)
  check("RF extract nodes — data.frame",       is.data.frame(nodes_rf))
  check("RF extract nodes — nrow = n_train",   nrow(nodes_rf) == nrow(tr_iris))
  check("RF extract nodes — ncol = ntree",     ncol(nodes_rf) == rf_clf$ntree)
  check("RF extract nodes — no NA",            !anyNA(nodes_rf))
  check("RF extract nodes — positive integers",
        all(unlist(nodes_rf) > 0) && all(unlist(nodes_rf) == floor(unlist(nodes_rf))))

  # ── 1c. Backward compatibility: adapter matches native RF predict(nodes=TRUE) ──
  native_nodes <- attr(predict(rf_clf, tr_iris, nodes = TRUE), "nodes")
  check("RF nodes identical to native predict(nodes=TRUE)",
        max(abs(as.matrix(nodes_rf) - unname(native_nodes))) == 0)

  # ── 1d. Adapter: get_ensemble_predictions ──
  preds_rf <- get_ensemble_predictions(rf_reg, tr_cars, "regression")
  check("RF preds — numeric vector",       is.numeric(preds_rf))
  check("RF preds — length = n_train",     length(preds_rf) == nrow(tr_cars))
  check("RF preds — no NA / Inf",          all(is.finite(preds_rf)))

  # ── 1e. createDisMatrix properties (classification) ──
  D_rf_clf <- createDisMatrix(rf_clf, tr_iris, "Species", parallel = PAR_OFF)
  check("RF D clf — square matrix",        is.matrix(D_rf_clf) && nrow(D_rf_clf) == ncol(D_rf_clf))
  check("RF D clf — dim = n_train × n",    all(dim(D_rf_clf) == nrow(tr_iris)))
  check("RF D clf — diagonal = 0",         all(diag(D_rf_clf) == 0))
  check("RF D clf — symmetric",            isSymmetric(D_rf_clf))
  check("RF D clf — values in [0,1]",      all(D_rf_clf >= 0) && all(D_rf_clf <= 1))
  check("RF D clf — non-trivial",          sum(D_rf_clf) > 0)

  # ── 1f. createDisMatrix properties (regression) ──
  D_rf_reg <- createDisMatrix(rf_reg, tr_cars, "mpg", parallel = PAR_OFF)
  check("RF D reg — diagonal = 0",         all(diag(D_rf_reg) == 0))
  check("RF D reg — symmetric",            isSymmetric(D_rf_reg))
  check("RF D reg — values in [0,1]",      all(D_rf_reg >= 0) && all(D_rf_reg <= 1))

  # ── 1g. Determinism — same model, same data → identical D ──
  D_rf_clf2 <- createDisMatrix(rf_clf, tr_iris, "Species", parallel = PAR_OFF)
  check("RF D clf — deterministic (run 1 == run 2)",
        identical(D_rf_clf, D_rf_clf2))

  # ── 1h. Core logic: intra-class D < inter-class D ──
  species <- tr_iris$Species
  idx_same  <- which(outer(species, species, `==`) & upper.tri(D_rf_clf))
  idx_diff  <- which(!outer(species, species, `==`) & upper.tri(D_rf_clf))
  check("RF D clf — intra-class mean D < inter-class mean D",
        mean(D_rf_clf[idx_same]) < mean(D_rf_clf[idx_diff]))

  # ── 1i. Full pipeline (classification) ──
  tree_rf_clf <- e2tree(Species ~ ., tr_iris, D_rf_clf, rf_clf, SETTING_CLF)
  check("RF e2tree clf — builds",          inherits(tree_rf_clf, "e2tree"))
  check("RF e2tree clf — at least 1 node", sum(!is.na(tree_rf_clf$tree$node)) >= 1)
  pred_rf_clf <- predict(tree_rf_clf, te_iris)
  check("RF e2tree clf — predict returns data.frame", is.data.frame(pred_rf_clf))
  acc_rf <- mean(pred_rf_clf$fit == te_iris$Species)
  check("RF e2tree clf — accuracy > 60%",  acc_rf > 0.60)

  # ── 1j. Full pipeline (regression) ──
  tree_rf_reg <- e2tree(mpg ~ ., tr_cars, D_rf_reg, rf_reg, SETTING_REG)
  pred_rf_reg <- predict(tree_rf_reg, te_cars)
  rmse_rf <- sqrt(mean((as.numeric(pred_rf_reg$fit) - te_cars$mpg)^2))
  baseline_rmse <- sd(tr_cars$mpg)
  check("RF e2tree reg — RMSE < sd(y_train)",  rmse_rf < baseline_rmse)

} else {
  SKIP("randomForest", "package not installed")
}


# =============================================================================
# SECTION 2 — ranger
# =============================================================================
section("2 · ranger")

if (requireNamespace("ranger", quietly = TRUE)) {
  suppressPackageStartupMessages(library(ranger))
  set.seed(42)
  rng_clf <- ranger(Species ~ ., data = tr_iris, num.trees = 100,
                    write.forest = TRUE)
  rng_reg <- ranger(mpg    ~ ., data = tr_cars, num.trees = 100,
                    write.forest = TRUE)

  # ── 2a. Adapter types ──
  check("ranger clf → 'classification'",   get_ensemble_type(rng_clf) == "classification")
  check("ranger reg → 'regression'",       get_ensemble_type(rng_reg) == "regression")

  # ── 2b. extract_terminal_nodes ──
  nodes_rng <- extract_terminal_nodes(rng_clf, tr_iris)
  check("ranger nodes — nrow = n_train",   nrow(nodes_rng) == nrow(tr_iris))
  check("ranger nodes — ncol = num.trees", ncol(nodes_rng) == rng_clf$num.trees)

  # Backward compatibility: matches native ranger predict(type="terminalNodes")
  native_rng <- predict(rng_clf, tr_iris, type = "terminalNodes")$predictions
  check("ranger nodes identical to native terminalNodes",
        max(abs(as.matrix(nodes_rng) - unname(native_rng))) == 0)

  # ── 2c. Dissimilarity matrix properties ──
  D_rng_clf <- createDisMatrix(rng_clf, tr_iris, "Species", parallel = PAR_OFF)
  check("ranger D clf — diagonal = 0",     all(diag(D_rng_clf) == 0))
  check("ranger D clf — symmetric",        isSymmetric(D_rng_clf))
  check("ranger D clf — values in [0,1]",  all(D_rng_clf >= 0) && all(D_rng_clf <= 1))

  D_rng_reg <- createDisMatrix(rng_reg, tr_cars, "mpg", parallel = PAR_OFF)
  check("ranger D reg — diagonal = 0",     all(diag(D_rng_reg) == 0))
  check("ranger D reg — symmetric",        isSymmetric(D_rng_reg))
  check("ranger D reg — values in [0,1]",  all(D_rng_reg >= 0) && all(D_rng_reg <= 1))

  # ── 2d. Determinism ──
  D_rng_clf2 <- createDisMatrix(rng_clf, tr_iris, "Species", parallel = PAR_OFF)
  check("ranger D clf — deterministic",    identical(D_rng_clf, D_rng_clf2))

  # ── 2e. Cross-backend coherence: RF vs ranger D should be correlated ──
  if (exists("D_rf_clf")) {
    r_spearman <- cor(as.vector(D_rf_clf[upper.tri(D_rf_clf)]),
                      as.vector(D_rng_clf[upper.tri(D_rng_clf)]),
                      method = "spearman")
    check("RF vs ranger D clf — Spearman > 0.50", r_spearman > 0.50)
  }

  # ── 2f. Full pipeline ──
  tree_rng_clf <- e2tree(Species ~ ., tr_iris, D_rng_clf, rng_clf, SETTING_CLF)
  acc_rng <- mean(predict(tree_rng_clf, te_iris)$fit == te_iris$Species)
  check("ranger e2tree clf — accuracy > 60%", acc_rng > 0.60)

  tree_rng_reg <- e2tree(mpg ~ ., tr_cars, D_rng_reg, rng_reg, SETTING_REG)
  rmse_rng <- sqrt(mean((as.numeric(predict(tree_rng_reg, te_cars)$fit) - te_cars$mpg)^2))
  check("ranger e2tree reg — RMSE < sd(y_train)",  rmse_rng < sd(tr_cars$mpg))

} else {
  SKIP("ranger", "package not installed")
}


# =============================================================================
# SECTION 3 — XGBoost
# =============================================================================
section("3 · XGBoost")

if (requireNamespace("xgboost", quietly = TRUE)) {
  suppressPackageStartupMessages(library(xgboost))

  # ── Build models ──
  set.seed(42)
  X_clf  <- as.matrix(tr_iris[, 1:4])
  y_clf  <- as.integer(tr_iris$Species) - 1L
  xgb_clf <- xgb.train(
    params = list(objective = "multi:softmax", num_class = 3,
                  max_depth = 4, eta = 0.3, nthread = 1),
    data = xgb.DMatrix(X_clf, label = y_clf),
    nrounds = 50, verbose = 0
  )

  X_reg  <- as.matrix(tr_iris[, 2:4])
  y_reg  <- tr_iris$Sepal.Length
  xgb_reg <- xgb.train(
    params = list(objective = "reg:squarederror",
                  max_depth = 4, eta = 0.3, nthread = 1),
    data = xgb.DMatrix(X_reg, label = y_reg),
    nrounds = 50, verbose = 0
  )

  # ── 3a. Adapter types ──
  check("XGB clf  → 'classification'",     get_ensemble_type(xgb_clf) == "classification")
  check("XGB reg  → 'regression'",         get_ensemble_type(xgb_reg) == "regression")

  # ── 3b. extract_terminal_nodes ──
  nodes_xgb <- extract_terminal_nodes(xgb_clf, tr_iris[, 1:4])
  check("XGB nodes — is data.frame",       is.data.frame(nodes_xgb))
  check("XGB nodes — nrow = n_train",      nrow(nodes_xgb) == nrow(tr_iris))
  check("XGB nodes — ncol > 0",            ncol(nodes_xgb) > 0)
  check("XGB nodes — no NA",               !anyNA(nodes_xgb))

  # ── 3c. Dissimilarity matrix properties ──
  D_xgb_clf <- createDisMatrix(xgb_clf, tr_iris, "Species", parallel = PAR_OFF)
  check("XGB D clf — diagonal = 0",        all(diag(D_xgb_clf) == 0))
  check("XGB D clf — symmetric",           isSymmetric(D_xgb_clf))
  check("XGB D clf — values in [0,1]",     all(D_xgb_clf >= 0) && all(D_xgb_clf <= 1))
  check("XGB D clf — non-trivial",         sum(D_xgb_clf) > 0)

  tr_iris_r <- tr_iris[, c("Sepal.Length", "Sepal.Width", "Petal.Length", "Petal.Width")]
  tr_iris_r$Sepal.Length_y <- tr_iris$Sepal.Length
  D_xgb_reg <- createDisMatrix(xgb_reg,
                                cbind(Sepal.Length = tr_iris$Sepal.Length,
                                      tr_iris[, 2:4]),
                                "Sepal.Length",
                                parallel = PAR_OFF)
  check("XGB D reg — diagonal = 0",        all(diag(D_xgb_reg) == 0))
  check("XGB D reg — symmetric",           isSymmetric(D_xgb_reg))
  check("XGB D reg — values in [0,1]",     all(D_xgb_reg >= 0) && all(D_xgb_reg <= 1))

  # ── 3d. Determinism ──
  D_xgb_clf2 <- createDisMatrix(xgb_clf, tr_iris, "Species", parallel = PAR_OFF)
  check("XGB D clf — deterministic",       identical(D_xgb_clf, D_xgb_clf2))

  # ── 3e. Core logic: intra-class D < inter-class D ──
  sp2 <- tr_iris$Species
  idx_s <- which(outer(sp2, sp2, `==`) & upper.tri(D_xgb_clf))
  idx_d <- which(!outer(sp2, sp2, `==`) & upper.tri(D_xgb_clf))
  check("XGB D clf — intra-class D < inter-class D",
        mean(D_xgb_clf[idx_s]) < mean(D_xgb_clf[idx_d]))

  # ── 3f. Full pipeline ──
  tree_xgb_clf <- e2tree(Species ~ ., tr_iris, D_xgb_clf, xgb_clf, SETTING_CLF)
  check("XGB e2tree clf — builds",         inherits(tree_xgb_clf, "e2tree"))
  acc_xgb <- mean(predict(tree_xgb_clf, te_iris)$fit == te_iris$Species)
  check("XGB e2tree clf — accuracy > 60%", acc_xgb > 0.60)

  # ── 3g. Cross-backend coherence with RF ──
  if (exists("D_rf_clf")) {
    r_xgb_rf <- cor(as.vector(D_rf_clf[upper.tri(D_rf_clf)]),
                    as.vector(D_xgb_clf[upper.tri(D_xgb_clf)]),
                    method = "spearman")
    check("XGB vs RF D clf — Spearman > 0.30", r_xgb_rf > 0.30)
  }

} else {
  SKIP("XGBoost", "package not installed")
}


# =============================================================================
# SECTION 4 — GBM
# =============================================================================
section("4 · GBM")

if (requireNamespace("gbm", quietly = TRUE)) {
  suppressPackageStartupMessages(library(gbm))

  set.seed(42)
  gbm_reg <- gbm(mpg ~ ., data = tr_cars,
                 distribution = "gaussian",
                 n.trees = 100, interaction.depth = 2, shrinkage = 0.1,
                 n.minobsinnode = 1, bag.fraction = 0.8, verbose = FALSE)

  # Binary classification: is_setosa
  tr_bin            <- tr_iris
  tr_bin$is_setosa  <- as.integer(tr_iris$Species == "setosa")
  gbm_clf <- gbm(
    is_setosa ~ Sepal.Length + Sepal.Width + Petal.Length + Petal.Width,
    data = tr_bin, distribution = "bernoulli",
    n.trees = 100, interaction.depth = 2, shrinkage = 0.1,
    n.minobsinnode = 1, bag.fraction = 0.8, verbose = FALSE
  )

  # ── 4a. Types ──
  check("GBM reg  → 'regression'",         get_ensemble_type(gbm_reg) == "regression")
  check("GBM clf  → 'classification'",     get_ensemble_type(gbm_clf) == "classification")

  # ── 4b. extract_terminal_nodes ──
  nodes_gbm <- extract_terminal_nodes(gbm_reg, tr_cars)
  check("GBM nodes — is data.frame",       is.data.frame(nodes_gbm))
  check("GBM nodes — nrow = n_train",      nrow(nodes_gbm) == nrow(tr_cars))
  check("GBM nodes — ncol = n.trees",      ncol(nodes_gbm) == gbm_reg$n.trees)
  check("GBM nodes — positive integers",
        all(unlist(nodes_gbm) > 0) && all(unlist(nodes_gbm) == floor(unlist(nodes_gbm))))

  # ── 4c. Dissimilarity matrix properties ──
  D_gbm_reg <- createDisMatrix(gbm_reg, tr_cars, "mpg", parallel = PAR_OFF)
  check("GBM D reg — diagonal = 0",        all(diag(D_gbm_reg) == 0))
  check("GBM D reg — symmetric",           isSymmetric(D_gbm_reg))
  check("GBM D reg — values in [0,1]",     all(D_gbm_reg >= 0) && all(D_gbm_reg <= 1))

  # For GBM clf: coerce is_setosa to factor BEFORE createDisMatrix
  tr_bin$is_setosa  <- factor(tr_bin$is_setosa)
  D_gbm_clf <- createDisMatrix(gbm_clf, tr_bin, "is_setosa", parallel = PAR_OFF)
  check("GBM D clf — diagonal = 0",        all(diag(D_gbm_clf) == 0))
  check("GBM D clf — symmetric",           isSymmetric(D_gbm_clf))
  check("GBM D clf — values in [0,1]",     all(D_gbm_clf >= 0) && all(D_gbm_clf <= 1))

  # ── 4d. Determinism ──
  D_gbm_reg2 <- createDisMatrix(gbm_reg, tr_cars, "mpg", parallel = PAR_OFF)
  check("GBM D reg — deterministic",       identical(D_gbm_reg, D_gbm_reg2))

  # ── 4e. get_ensemble_predictions ──
  preds_gbm <- get_ensemble_predictions(gbm_reg, tr_cars, "regression")
  check("GBM preds — numeric, no NA",
        is.numeric(preds_gbm) && length(preds_gbm) == nrow(tr_cars) && !anyNA(preds_gbm))

  # ── 4f. Full pipeline (regression) ──
  tree_gbm_reg <- e2tree(mpg ~ ., tr_cars, D_gbm_reg, gbm_reg, SETTING_REG)
  check("GBM e2tree reg — builds",         inherits(tree_gbm_reg, "e2tree"))
  rmse_gbm <- sqrt(mean((as.numeric(predict(tree_gbm_reg, te_cars)$fit) - te_cars$mpg)^2))
  check("GBM e2tree reg — RMSE < sd(y_train)",  rmse_gbm < sd(tr_cars$mpg))

  # ── 4g. Full pipeline (classification) ──
  tree_gbm_clf <- e2tree(
    is_setosa ~ Sepal.Length + Sepal.Width + Petal.Length + Petal.Width,
    tr_bin, D_gbm_clf, gbm_clf, SETTING_CLF
  )
  check("GBM e2tree clf — builds",         inherits(tree_gbm_clf, "e2tree"))

} else {
  SKIP("GBM", "package not installed")
}


# =============================================================================
# SECTION 5 — LightGBM
# =============================================================================
section("5 · LightGBM")

if (requireNamespace("lightgbm", quietly = TRUE)) {
  suppressPackageStartupMessages(library(lightgbm))

  # Use iris for regression (150 obs; avoids 1-tree collapse on mtcars 32 obs)
  set.seed(42)
  X_lgb_clf <- as.matrix(tr_iris[, 1:4])
  y_lgb_clf <- as.integer(tr_iris$Species) - 1L
  lgb_clf <- lgb.train(
    params = list(objective = "multiclass", num_class = 3,
                  num_leaves = 15, min_data_in_leaf = 3, verbose = -1),
    data = lgb.Dataset(X_lgb_clf, label = y_lgb_clf,
                       colnames = colnames(X_lgb_clf)),
    nrounds = 50
  )

  X_lgb_reg <- as.matrix(tr_iris[, 2:4])
  y_lgb_reg <- tr_iris$Sepal.Length
  lgb_reg <- lgb.train(
    params = list(objective = "regression", num_leaves = 8,
                  min_data_in_leaf = 3, verbose = -1),
    data = lgb.Dataset(X_lgb_reg, label = y_lgb_reg,
                       colnames = colnames(X_lgb_reg)),
    nrounds = 50
  )

  # ── 5a. Types ──
  check("LGB clf → 'classification'",      get_ensemble_type(lgb_clf) == "classification")
  check("LGB reg → 'regression'",          get_ensemble_type(lgb_reg) == "regression")

  # ── 5b. extract_terminal_nodes ──
  nodes_lgb <- extract_terminal_nodes(lgb_clf, tr_iris[, 1:4])
  check("LGB nodes — is data.frame",       is.data.frame(nodes_lgb))
  check("LGB nodes — nrow = n_train",      nrow(nodes_lgb) == nrow(tr_iris))
  check("LGB nodes — ncol = num_trees",    ncol(nodes_lgb) == lgb_clf$num_trees())
  check("LGB nodes — no NA",               !anyNA(nodes_lgb))

  nodes_lgb_r <- extract_terminal_nodes(lgb_reg, tr_iris[, 2:4])
  check("LGB reg nodes — nrow = n_train",  nrow(nodes_lgb_r) == nrow(tr_iris))
  check("LGB reg nodes — ncol = num_trees", ncol(nodes_lgb_r) == lgb_reg$num_trees())

  # ── 5c. Dissimilarity matrix properties ──
  D_lgb_clf <- createDisMatrix(lgb_clf, tr_iris, "Species", parallel = PAR_OFF)
  check("LGB D clf — diagonal = 0",        all(diag(D_lgb_clf) == 0))
  check("LGB D clf — symmetric",           isSymmetric(D_lgb_clf))
  check("LGB D clf — values in [0,1]",     all(D_lgb_clf >= 0) && all(D_lgb_clf <= 1))
  check("LGB D clf — non-trivial",         sum(D_lgb_clf) > 0)

  D_lgb_reg <- createDisMatrix(lgb_reg,
                                cbind(Sepal.Length = tr_iris$Sepal.Length,
                                      tr_iris[, 2:4]),
                                "Sepal.Length",
                                parallel = PAR_OFF)
  check("LGB D reg — diagonal = 0",        all(diag(D_lgb_reg) == 0))
  check("LGB D reg — symmetric",           isSymmetric(D_lgb_reg))
  check("LGB D reg — values in [0,1]",     all(D_lgb_reg >= 0) && all(D_lgb_reg <= 1))

  # ── 5d. Determinism ──
  D_lgb_clf2 <- createDisMatrix(lgb_clf, tr_iris, "Species", parallel = PAR_OFF)
  check("LGB D clf — deterministic",       identical(D_lgb_clf, D_lgb_clf2))

  # ── 5e. Core logic: intra-class D < inter-class D ──
  sp3 <- tr_iris$Species
  idx_s3 <- which(outer(sp3, sp3, `==`) & upper.tri(D_lgb_clf))
  idx_d3 <- which(!outer(sp3, sp3, `==`) & upper.tri(D_lgb_clf))
  check("LGB D clf — intra-class D < inter-class D",
        mean(D_lgb_clf[idx_s3]) < mean(D_lgb_clf[idx_d3]))

  # ── 5f. Cross-backend coherence with RF ──
  if (exists("D_rf_clf")) {
    r_lgb_rf <- cor(as.vector(D_rf_clf[upper.tri(D_rf_clf)]),
                    as.vector(D_lgb_clf[upper.tri(D_lgb_clf)]),
                    method = "spearman")
    check("LGB vs RF D clf — Spearman > 0.30", r_lgb_rf > 0.30)
  }

  # ── 5g. Full pipeline (classification) ──
  tree_lgb_clf <- e2tree(Species ~ ., tr_iris, D_lgb_clf, lgb_clf, SETTING_CLF)
  check("LGB e2tree clf — builds",         inherits(tree_lgb_clf, "e2tree"))
  acc_lgb <- mean(predict(tree_lgb_clf, te_iris)$fit == te_iris$Species)
  check("LGB e2tree clf — accuracy > 60%", acc_lgb > 0.60)

  # ── 5h. Full pipeline (regression) ──
  te_iris_r <- cbind(Sepal.Length = te_iris$Sepal.Length, te_iris[, 2:4])
  tree_lgb_reg <- e2tree(
    Sepal.Length ~ Sepal.Width + Petal.Length + Petal.Width,
    cbind(Sepal.Length = tr_iris$Sepal.Length, tr_iris[, 2:4]),
    D_lgb_reg, lgb_reg, SETTING_REG
  )
  check("LGB e2tree reg — builds",         inherits(tree_lgb_reg, "e2tree"))
  pred_lgb_r <- predict(tree_lgb_reg, as.data.frame(te_iris_r))
  rmse_lgb <- sqrt(mean((as.numeric(pred_lgb_r$fit) - te_iris$Sepal.Length)^2))
  check("LGB e2tree reg — RMSE < sd(y_train)",
        rmse_lgb < sd(tr_iris$Sepal.Length))

} else {
  SKIP("LightGBM", "package not installed")
}


# =============================================================================
# SECTION 6 — Cross-backend logic consistency (shared iris classification)
# =============================================================================
section("6 · Cross-backend consistency")

# All available D matrices for iris classification should share the same
# qualitative structure: same-class pairs should cluster together.
backends <- list()
if (exists("D_rf_clf"))  backends[["RF"]]       <- D_rf_clf
if (exists("D_rng_clf")) backends[["ranger"]]   <- D_rng_clf
if (exists("D_xgb_clf")) backends[["XGBoost"]]  <- D_xgb_clf
if (exists("D_lgb_clf")) backends[["LightGBM"]] <- D_lgb_clf

sp_all <- tr_iris$Species
idx_same_all <- which(outer(sp_all, sp_all, `==`) & upper.tri(D_rf_clf))
idx_diff_all <- which(!outer(sp_all, sp_all, `==`) & upper.tri(D_rf_clf))

for (nm in names(backends)) {
  D_b <- backends[[nm]]
  check(sprintf("%s  — intra-class mean D < inter-class mean D", nm),
        mean(D_b[idx_same_all]) < mean(D_b[idx_diff_all]))
}

if (length(backends) >= 2) {
  bnames <- names(backends)
  for (i in seq_len(length(bnames) - 1)) {
    for (j in seq(i + 1, length(bnames))) {
      a <- backends[[bnames[i]]]
      b <- backends[[bnames[j]]]
      r <- cor(as.vector(a[upper.tri(a)]),
               as.vector(b[upper.tri(b)]),
               method = "spearman")
      check(sprintf("%s vs %s — Spearman > 0.30", bnames[i], bnames[j]),
            r > 0.30)
    }
  }
}


# =============================================================================
# SECTION 7 — Error handling and input validation
# =============================================================================
section("7 · Input validation")

check("get_ensemble_type unsupported class → error",
      tryCatch({ get_ensemble_type(list()); FALSE },
               error = function(e) TRUE))

check("createDisMatrix NULL ensemble → error",
      tryCatch({ createDisMatrix(NULL, iris, "Species", PAR_OFF); FALSE },
               error = function(e) TRUE))

check("createDisMatrix non-data.frame data → error",
      tryCatch({
        if (exists("rf_clf")) createDisMatrix(rf_clf, as.matrix(iris[, 1:4]),
                                              "Species", PAR_OFF)
        FALSE
      }, error = function(e) TRUE))

check("createDisMatrix bad label → error",
      tryCatch({
        if (exists("rf_clf")) createDisMatrix(rf_clf, iris, "NOTACOL", PAR_OFF)
        FALSE
      }, error = function(e) TRUE))

check("e2tree bad setting → error",
      tryCatch({
        if (exists("rf_clf") && exists("D_rf_clf"))
          e2tree(Species ~ ., tr_iris, D_rf_clf, rf_clf,
                 list(minsplit = 20))
        FALSE
      }, error = function(e) TRUE))


# =============================================================================
# SUMMARY
# =============================================================================
section("SUMMARY")

total <- .pass_cnt + .fail_cnt
skipped <- length(.results) - total
cat(sprintf(
  "\n  \033[32m%d PASS\033[0m   \033[31m%d FAIL\033[0m   \033[33m%d SKIP\033[0m   (total checked: %d)\n\n",
  .pass_cnt, .fail_cnt, skipped, total
))

if (.fail_cnt > 0) {
  cat("  Failed tests:\n")
  for (r in .results) {
    if (identical(r$status, "FAIL"))
      cat(sprintf("    - %s\n", r$name))
  }
  cat("\n")
  quit(status = 1L)
} else {
  cat("  \033[32mAll tests passed.\033[0m\n\n")
}

# ===========================================================================
# Accessor (extractor) functions for e2tree package classes
# ===========================================================================

# --- nodes() generic and methods ---

#' Extract Tree Node Information
#'
#' Extracts the data frame describing the nodes of an E2Tree model,
#' including split rules, predictions, and node statistics.
#'
#' @param x An e2tree object.
#' @param terminal Logical. If \code{TRUE}, return only terminal (leaf) nodes.
#'   Default is \code{FALSE}.
#' @param ... Additional arguments (ignored).
#'
#' @return A data frame with one row per node.
#'
#' @examples
#' \donttest{
#' data(iris)
#' smp_size <- floor(0.75 * nrow(iris))
#' train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
#' training <- iris[train_ind, ]
#'
#' ensemble <- randomForest::randomForest(Species ~ ., data = training,
#'   importance = TRUE, proximity = TRUE)
#' D <- createDisMatrix(ensemble, data = training, label = "Species",
#'   parallel = list(active = FALSE, no_cores = 1))
#' setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
#' tree <- e2tree(Species ~ ., training, D, ensemble, setting)
#'
#' nodes(tree)
#' nodes(tree, terminal = TRUE)
#' }
#'
#' @export
nodes <- function(x, ...) {
  UseMethod("nodes")
}

#' @rdname nodes
#' @method nodes e2tree
#' @export
nodes.e2tree <- function(x, terminal = FALSE, ...) {
  tree_df <- x$tree
  if (terminal) {
    tree_df[tree_df$terminal == TRUE, , drop = FALSE]
  } else {
    tree_df
  }
}


# --- e2splits() generic and methods ---

#' Extract Split Information from an E2Tree Model
#'
#' Returns the split matrix and categorical split encoding from a fitted
#' E2Tree model.
#'
#' @param x An e2tree object.
#' @param ... Additional arguments (ignored).
#'
#' @return A list with components:
#' \describe{
#'   \item{splits}{The split information matrix.}
#'   \item{csplit}{The categorical split encoding matrix.}
#' }
#'
#' @export
e2splits <- function(x, ...) {
  UseMethod("e2splits")
}

#' @rdname e2splits
#' @method e2splits e2tree
#' @export
e2splits.e2tree <- function(x, ...) {
  list(splits = x$splits, csplit = x$csplit)
}


# --- measures() generic and methods ---

#' Extract Validation Measures
#'
#' Extracts the data frame of validation measures from an eValidation object,
#' including divergence and similarity metrics between the ensemble and
#' E2Tree proximity matrices.
#'
#' @param x An eValidation object.
#' @param ... Additional arguments (ignored).
#'
#' @return A data frame with columns for method name, type, observed value,
#'   and (if permutation tests were performed) null distribution statistics
#'   and p-values.
#'
#' @export
measures <- function(x, ...) {
  UseMethod("measures")
}

#' @rdname measures
#' @method measures eValidation
#' @export
measures.eValidation <- function(x, ...) {
  x$measures
}


# --- proximity() generic and methods ---

#' Extract Proximity Matrices
#'
#' Extracts proximity matrices from an eValidation object. The ensemble
#' proximity matrix is derived from the original ensemble model, while the
#' E2Tree proximity matrix is estimated from the fitted E2Tree.
#'
#' @param x An eValidation object.
#' @param type Character string specifying which proximity matrix to extract.
#'   One of \code{"ensemble"}, \code{"e2tree"}, or \code{"both"} (default).
#' @param ... Additional arguments (ignored).
#'
#' @return A matrix (if \code{type} is \code{"ensemble"} or \code{"e2tree"})
#'   or a list of two matrices (if \code{type} is \code{"both"}).
#'
#' @export
proximity <- function(x, ...) {
  UseMethod("proximity")
}

#' @rdname proximity
#' @method proximity eValidation
#' @export
proximity.eValidation <- function(x, type = c("both", "ensemble", "e2tree"), ...) {
  type <- match.arg(type)
  switch(type,
    ensemble = x$Proximity_matrix_ensemble,
    e2tree   = x$Proximity_matrix_e2tree,
    both     = list(
      ensemble = x$Proximity_matrix_ensemble,
      e2tree   = x$Proximity_matrix_e2tree
    )
  )
}

#' Loss of Interpretability (LoI) Index
#'
#' Computes the LoI index and its decomposition, measuring how well
#' the E2Tree-estimated proximity matrix reconstructs the original
#' ensemble proximity matrix.
#'
#' The statistic is defined as:
#'
#' \deqn{\mathrm{LoI}(O, \hat{O}) = \sum_{i < j}
#'   \frac{(o_{ij} - \hat{o}_{ij})^2}{\max(o_{ij}, \hat{o}_{ij})}}
#'
#' The Normalized LoI divides by the number of pairs \eqn{M = n(n-1)/2}:
#'
#' \deqn{\mathrm{nLoI}(O, \hat{O}) = \frac{1}{M} \mathrm{LoI}(O, \hat{O})}
#'
#' The LoI decomposes into two components:
#' \itemize{
#'   \item \strong{LoI_in}: within-node loss (pairs grouped together by E2Tree)
#'   \item \strong{LoI_out}: between-node loss (pairs separated by E2Tree)
#' }
#' The per-pair averages \code{mean_in} and \code{mean_out} enable direct
#' comparison between the two components despite their different pair counts.
#'
#' @param O Proximity matrix from the ensemble model (n x n), values in the interval 0 to 1
#' @param O_hat Proximity matrix estimated by E2Tree (n x n), values in the interval 0 to 1
#' @param normalize Logical. If TRUE (default), returns nLoI (divided by M).
#'   If FALSE, returns raw LoI.
#'
#' @return An object of class \code{"loi"} containing:
#'   \item{loi}{Raw LoI value (unnormalized)}
#'   \item{nloi}{Normalized LoI (LoI / M)}
#'   \item{loi_in}{Within-node component (total)}
#'   \item{loi_out}{Between-node component (total)}
#'   \item{mean_in}{Per-pair average within-node loss (comparable with mean_out)}
#'   \item{mean_out}{Per-pair average between-node loss (comparable with mean_in)}
#'   \item{n}{Matrix dimension}
#'   \item{m}{Number of unique pairs}
#'   \item{n_within}{Number of within-node pairs}
#'   \item{n_between}{Number of between-node pairs}
#'
#' @details
#' The statistic uses a normalized squared difference, where each cell's
#' contribution is weighted by the maximum of the two proximity values.
#' This gives more weight to discrepancies in high-proximity regions.
#'
#' \strong{Decomposition interpretation (per-pair averages):}
#' \itemize{
#'   \item \code{mean_out}: average ensemble proximity lost by the partition.
#'     Low values (< 0.1) indicate the tree correctly separates low-proximity
#'     pairs. High values (> 0.3) suggest the tree splits apart pairs that
#'     the ensemble considers similar --more terminal nodes may help.
#'   \item \code{mean_in}: average calibration error within nodes. Low values
#'     (< 0.01) indicate excellent within-node reconstruction. Higher values
#'     reflect the inherent fuzzy-to-crisp transition.
#' }
#'
#' @examples
#' \donttest{
#' data(iris)
#' smp_size <- floor(0.75 * nrow(iris))
#' set.seed(42)
#' train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
#' training <- iris[train_ind, ]
#'
#' ensemble <- randomForest::randomForest(Species ~ ., data = training,
#'   importance = TRUE, proximity = TRUE)
#'
#' D <- createDisMatrix(ensemble, data = training, label = "Species",
#'   parallel = list(active = FALSE, no_cores = 1))
#'
#' setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
#' tree <- e2tree(Species ~ ., training, D, ensemble, setting)
#'
#' vs <- eValidation(training, tree, D)
#' prox <- proximity(vs)
#' O <- prox$ensemble
#' O_hat <- prox$e2tree
#'
#' # Compute LoI with decomposition
#' result <- loi(O, O_hat)
#' print(result)
#' summary(result)
#' plot(result)
#'
#' # Permutation test
#' perm <- loi_perm(O, O_hat, n_perm = 999, seed = 42)
#' print(perm)
#' plot(perm)
#' }
#'
#' @importFrom graphics barplot layout text
#' @export
loi <- function(O, O_hat, normalize = TRUE) {

  # -------------------------------------------------------------------------
  # INPUT VALIDATION
  # -------------------------------------------------------------------------

  if (!is.matrix(O) || !is.matrix(O_hat)) {
    stop("Both O and O_hat must be matrices")
  }
  if (nrow(O) != ncol(O) || nrow(O_hat) != ncol(O_hat)) {
    stop("Matrices must be square")
  }
  if (!all(dim(O) == dim(O_hat))) {
    stop("O and O_hat must have the same dimensions")
  }

  n <- nrow(O)
  m <- n * (n - 1) / 2

  # -------------------------------------------------------------------------
  # EXTRACT LOWER TRIANGLE
  # -------------------------------------------------------------------------

  idx <- lower.tri(O, diag = FALSE)
  o_vec <- O[idx]
  ohat_vec <- O_hat[idx]

  # -------------------------------------------------------------------------
  # DECOMPOSITION: within-node vs between-node
  # -------------------------------------------------------------------------

  # Between-node pairs: O_hat == 0 (E2Tree separated them)
  is_between <- ohat_vec == 0
  is_within <- !is_between

  # Within-node component: standard LoI formula
  if (sum(is_within) > 0) {
    o_in <- o_vec[is_within]
    ohat_in <- ohat_vec[is_within]
    denom_in <- pmax(o_in, ohat_in, .Machine$double.eps)
    loi_in <- sum((o_in - ohat_in)^2 / denom_in, na.rm = TRUE)
  } else {
    loi_in <- 0
  }

  # Between-node component: reduces to sum(o_ij) for pairs where O_hat == 0
  # Because (o - 0)^2 / max(o, 0) = o when o > 0, and 0 when o == 0
  loi_out <- sum(o_vec[is_between], na.rm = TRUE)

  # Total
  loi_total <- loi_in + loi_out

  # Normalized
  nloi <- loi_total / m

  # Counts
  n_within <- sum(is_within)
  n_between <- sum(is_between)

  # Per-pair averages (comparable across components)
  mean_in  <- if (n_within > 0)  loi_in  / n_within  else 0
  mean_out <- if (n_between > 0) loi_out / n_between else 0

  # -------------------------------------------------------------------------
  # BUILD RESULT OBJECT
  # -------------------------------------------------------------------------

  result <- list(
    loi       = loi_total,
    nloi      = nloi,
    loi_in    = loi_in,
    loi_out   = loi_out,
    mean_in   = mean_in,
    mean_out  = mean_out,
    n         = n,
    m         = m,
    n_within  = n_within,
    n_between = n_between
  )
  class(result) <- "loi"
  result
}


# ===========================================================================
# PRINT METHOD
# ===========================================================================

#' @method print loi
#' @export
print.loi <- function(x, digits = 4, ...) {

  cat("\n")
  cat("  Loss of Interpretability (LoI)\n")
  cat("  -------------------------------------------\n")
  cat(sprintf("  nLoI (normalized):   %.*f\n", digits, x$nloi))
  cat(sprintf("  LoI (raw):           %.*f\n", digits, x$loi))
  cat(sprintf("  n = %d, pairs = %d\n", x$n, x$m))
  cat("\n")

  invisible(x)
}


# ===========================================================================
# SUMMARY METHOD
# ===========================================================================

#' @method summary loi
#' @export
summary.loi <- function(object, digits = 4, ...) {

  x <- object

  cat("\n")
  cat("##############################################################################\n")
  cat("   Loss of Interpretability (LoI) --Decomposition\n")
  cat("##############################################################################\n\n")

  cat(sprintf("  nLoI (normalized):   %.*f\n", digits, x$nloi))
  cat(sprintf("  LoI (raw):           %.*f\n", digits, x$loi))
  cat(sprintf("  n = %d, pairs = %d (within: %d, between: %d)\n\n",
              x$n, x$m, x$n_within, x$n_between))

  cat("------------------------------------------------------------------------------\n")
  cat("  Component              Total       Pairs     Mean/pair\n")
  cat("------------------------------------------------------------------------------\n")

  cat(sprintf("  Within-node (LoI_in)  %8.*f   %8d     %.*f\n",
              digits, x$loi_in, x$n_within, 6, x$mean_in))
  cat(sprintf("  Between-node (LoI_out)%8.*f   %8d     %.*f\n",
              digits, x$loi_out, x$n_between, 6, x$mean_out))

  cat("------------------------------------------------------------------------------\n\n")

  cat("  Per-pair interpretation (comparable across components):\n\n")

  cat(sprintf("    mean_in  = %.6f  (avg calibration error within nodes)\n", x$mean_in))
  cat(sprintf("    mean_out = %.6f  (avg ensemble proximity lost by separation)\n\n", x$mean_out))

  if (x$mean_out > 0.3) {
    cat("  Diagnostic: HIGH mean_out (> 0.3). The E2Tree is separating pairs\n")
    cat("  with substantial ensemble proximity. Consider more terminal nodes.\n\n")
  } else if (x$mean_out > 0.1) {
    cat("  Diagnostic: MODERATE mean_out. Some ensemble proximity is lost by\n")
    cat("  the partition, but most separated pairs have low ensemble proximity.\n\n")
  } else {
    cat("  Diagnostic: LOW mean_out. The partition correctly separates pairs\n")
    cat("  that have low ensemble proximity --the tree structure is well-placed.\n\n")
  }

  if (x$mean_in > 0.1) {
    cat("  Diagnostic: HIGH mean_in (> 0.1). Within-node proximity values\n")
    cat("  differ substantially from the ensemble. Check within-node calibration.\n\n")
  } else if (x$mean_in > 0.01) {
    cat("  Diagnostic: MODERATE mean_in. Some within-node calibration error,\n")
    cat("  typical of the fuzzy-to-crisp structural transition.\n\n")
  } else {
    cat("  Diagnostic: LOW mean_in. Within-node proximity values closely\n")
    cat("  match the ensemble --excellent calibration.\n\n")
  }

  cat("##############################################################################\n\n")

  invisible(x)
}


# ===========================================================================
# PLOT METHOD
# ===========================================================================

#' @method plot loi
#' @export
plot.loi <- function(x, ...) {

  old_par <- par(no.readonly = TRUE)
  on.exit(par(old_par))

  par(mfrow = c(1, 2), mar = c(5, 4, 4, 2) + 0.1)

  cols <- c("#3498db", "#e74c3c")

  # Panel 1: Per-pair average loss (comparable)
  means <- c(x$mean_in, x$mean_out)
  nms <- c("Within-node\n(mean_in)", "Between-node\n(mean_out)")

  bp <- barplot(means, names.arg = nms, col = cols, border = NA,
                main = sprintf("LoI Per-Pair Average Loss\nnLoI = %.4f", x$nloi),
                ylab = "Mean loss per pair",
                ylim = c(0, max(means) * 1.35))
  text(bp, means, labels = sprintf("%.4f", means),
       pos = 3, cex = 0.9, font = 2)

  # Panel 2: Pair counts
  counts <- c(x$n_within, x$n_between)
  nms2 <- c("Within-node", "Between-node")

  bp2 <- barplot(counts, names.arg = nms2, col = cols, border = NA,
                 main = "Number of Pairs",
                 ylab = "Count",
                 ylim = c(0, max(counts) * 1.25))
  text(bp2, counts, labels = counts,
       pos = 3, cex = 0.9, font = 2)
}


# ===========================================================================
# PERMUTATION TEST
# ===========================================================================

#' Permutation Test for LoI
#'
#' Performs a permutation test using row/column permutation to assess
#' whether the E2Tree reconstruction is significantly better than expected
#' by chance.
#'
#' @param O Proximity matrix from the ensemble model (n x n)
#' @param O_hat Proximity matrix estimated by E2Tree (n x n)
#' @param n_perm Number of permutations (default: 999)
#' @param conf.level Confidence level for intervals (default: 0.95)
#' @param seed Random seed for reproducibility. Default is NULL.
#'
#' @return An object of class \code{"loi_perm"} containing:
#'   \item{observed}{Observed nLoI value and decomposition (loi object)}
#'   \item{statistic}{Observed nLoI value (scalar)}
#'   \item{p.value}{Test p-value (one-sided, less)}
#'   \item{ci}{Permutation-based confidence interval for nLoI}
#'   \item{null_dist}{Null distribution of nLoI values}
#'   \item{null_mean}{Mean of the null distribution}
#'   \item{null_sd}{Standard deviation of the null distribution}
#'   \item{z_stat}{Standardized Z statistic}
#'   \item{n_perm}{Number of permutations}
#'   \item{conf.level}{Confidence level}
#'
#' @details
#' The test uses \strong{simultaneous row/column permutation} of
#' \eqn{\hat{O}}: for each replicate, a random permutation \eqn{\pi}
#' of \eqn{\{1, \ldots, n\}} is drawn and \eqn{\hat{O}^\pi =
#' \hat{O}[\pi, \pi]} is computed. This preserves the block-diagonal
#' structure of \eqn{\hat{O}} while breaking the correspondence with
#' \eqn{O}.
#'
#' The null hypothesis is: the E2Tree labeling is unrelated to the
#' ensemble structure. Under H1 (good reconstruction), the observed
#' nLoI should be significantly \emph{lower} than the null distribution.
#'
#' P-values include the +1 correction of Phipson & Smyth (2010).
#'
#' @examples
#' \donttest{
#' n <- 50
#' O <- matrix(runif(n^2, 0.3, 1), n, n)
#' O <- (O + t(O)) / 2; diag(O) <- 1
#' O_hat <- O + matrix(rnorm(n^2, 0, 0.05), n, n)
#' O_hat <- pmin(pmax((O_hat + t(O_hat)) / 2, 0), 1); diag(O_hat) <- 1
#'
#' result <- loi_perm(O, O_hat, n_perm = 199, seed = 42)
#' print(result)
#' summary(result)
#' plot(result)
#' }
#'
#' @export
loi_perm <- function(O, O_hat, n_perm = 999,
                     conf.level = 0.95,
                     seed = NULL) {

  if (!is.null(seed)) set.seed(seed)

  if (!is.matrix(O) || !is.matrix(O_hat)) {
    stop("Both O and O_hat must be matrices")
  }
  if (!all(dim(O) == dim(O_hat))) {
    stop("O and O_hat must have the same dimensions")
  }
  if (conf.level <= 0 || conf.level >= 1) {
    stop("conf.level must be between 0 and 1")
  }

  n <- nrow(O)

  # -------------------------------------------------------------------------
  # OBSERVED STATISTIC (with full decomposition)
  # -------------------------------------------------------------------------

  obs <- loi(O, O_hat)

  # -------------------------------------------------------------------------
  # NULL DISTRIBUTION via row/column permutation
  # -------------------------------------------------------------------------

  perm_one <- function(dummy) {
    perm_idx <- sample.int(n)
    O_hat_perm <- O_hat[perm_idx, perm_idx]
    loi(O, O_hat_perm)$nloi
  }

  if (requireNamespace("future.apply", quietly = TRUE)) {
    null_dist <- future.apply::future_sapply(seq_len(n_perm), perm_one,
                                              future.seed = TRUE)
  } else {
    null_dist <- vapply(seq_len(n_perm), perm_one, numeric(1))
  }

  # -------------------------------------------------------------------------
  # P-VALUE (one-sided, less: good reconstruction = low nLoI)
  # -------------------------------------------------------------------------

  p_value <- (1 + sum(null_dist <= obs$nloi, na.rm = TRUE)) / (1 + n_perm)

  # -------------------------------------------------------------------------
  # NULL DISTRIBUTION STATISTICS
  # -------------------------------------------------------------------------

  null_mean <- mean(null_dist, na.rm = TRUE)
  null_sd <- sd(null_dist, na.rm = TRUE)
  z_stat <- (obs$nloi - null_mean) / null_sd

  # -------------------------------------------------------------------------
  # CONFIDENCE INTERVAL
  # -------------------------------------------------------------------------

  alpha <- 1 - conf.level
  null_quantiles <- quantile(null_dist, probs = c(alpha / 2, 1 - alpha / 2),
                              na.rm = TRUE)
  p_shift <- null_mean - null_quantiles
  ci_raw <- obs$nloi + p_shift
  ci <- c(ci_raw[2], ci_raw[1])
  ci <- pmax(ci, 0)
  names(ci) <- c(paste0((alpha / 2) * 100, "%"),
                  paste0((1 - alpha / 2) * 100, "%"))

  # -------------------------------------------------------------------------
  # BUILD RESULT
  # -------------------------------------------------------------------------

  result <- list(
    observed   = obs,
    statistic  = obs$nloi,
    p.value    = p_value,
    ci         = ci,
    null_dist  = null_dist,
    null_mean  = null_mean,
    null_sd    = null_sd,
    z_stat     = z_stat,
    n_perm     = n_perm,
    conf.level = conf.level
  )
  class(result) <- "loi_perm"
  result
}


# ===========================================================================
# PRINT METHOD for loi_perm
# ===========================================================================

#' @method print loi_perm
#' @export
print.loi_perm <- function(x, digits = 4, ...) {

  cat("\n")
  cat("==============================================================================\n")
  cat("   Permutation Test for Loss of Interpretability (LoI)\n")
  cat("==============================================================================\n\n")

  cat(sprintf("  Observed nLoI:       %.*f\n", digits, x$statistic))
  cat(sprintf("  Null mean:           %.*f\n", digits, x$null_mean))
  cat(sprintf("  Null SD:             %.*f\n", digits, x$null_sd))
  cat(sprintf("  Z-statistic:         %.*f\n", digits, x$z_stat))

  cat("\n------------------------------------------------------------------------------\n")
  cat("  Hypothesis Test (H1: nLoI < expected by chance)\n")
  cat("------------------------------------------------------------------------------\n")

  if (x$p.value < 0.0001) {
    cat("  p-value:             < 0.0001 ***\n")
  } else {
    stars <- if (x$p.value < 0.001) "***"
             else if (x$p.value < 0.01) "**"
             else if (x$p.value < 0.05) "*"
             else if (x$p.value < 0.1) "."
             else ""
    cat(sprintf("  p-value:             %.*f %s\n", digits, x$p.value, stars))
  }

  cat(sprintf("  %d%% CI:             [%.*f, %.*f]\n",
              round(x$conf.level * 100),
              digits, x$ci[1], digits, x$ci[2]))
  cat(sprintf("  Permutations:        %d (row/column)\n", x$n_perm))

  cat("\n------------------------------------------------------------------------------\n")
  cat("  Decomposition (per-pair averages)\n")
  cat("------------------------------------------------------------------------------\n")

  obs <- x$observed
  cat(sprintf("  mean_in  (within):   %.*f  (n_pairs = %d)\n",
              6, obs$mean_in, obs$n_within))
  cat(sprintf("  mean_out (between):  %.*f  (n_pairs = %d)\n",
              6, obs$mean_out, obs$n_between))

  cat("\n==============================================================================\n")
  cat("  Signif. codes: 0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1\n\n")

  invisible(x)
}


# ===========================================================================
# SUMMARY METHOD for loi_perm
# ===========================================================================

#' @method summary loi_perm
#' @export
summary.loi_perm <- function(object, digits = 4, ...) {

  print(object, digits = digits, ...)
  cat("\n")
  summary(object$observed, digits = digits, ...)

  invisible(object)
}


# ===========================================================================
# PLOT METHOD for loi_perm
# ===========================================================================

#' @method plot loi_perm
#' @export
plot.loi_perm <- function(x, ...) {

  old_par <- par(no.readonly = TRUE)
  on.exit(par(old_par))

  par(mfrow = c(1, 2), mar = c(5, 4, 4, 2) + 0.1)

  # --- Panel 1: Null distribution with observed ---

  xlim_range <- range(c(x$null_dist, x$statistic, x$ci), na.rm = TRUE)

  h <- hist(x$null_dist, plot = FALSE)
  d <- density(x$null_dist, na.rm = TRUE)
  ylim_max <- max(max(h$density), max(d$y)) * 1.15

  hist(x$null_dist,
       main = "Null Distribution of nLoI",
       xlab = "nLoI",
       col = "lightgray",
       border = "white",
       xlim = xlim_range,
       ylim = c(0, ylim_max),
       freq = FALSE)

  lines(d, col = "darkgray", lwd = 2)
  abline(v = x$statistic, col = "red", lwd = 2, lty = 1)
  abline(v = x$null_mean, col = "darkgray", lwd = 2, lty = 2)
  abline(v = x$ci[1], col = "blue", lwd = 1.5, lty = 3)
  abline(v = x$ci[2], col = "blue", lwd = 1.5, lty = 3)

  p_str <- if (x$p.value < 0.0001) "p < 0.0001"
           else sprintf("p = %.4f", x$p.value)

  legend("topright",
         legend = c(sprintf("Observed (%.4f)", x$statistic),
                    sprintf("Null mean (%.4f)", x$null_mean),
                    paste0(round(x$conf.level * 100), "% CI"),
                    p_str),
         col = c("red", "darkgray", "blue", NA),
         lty = c(1, 2, 3, NA),
         lwd = c(2, 2, 1.5, NA),
         bty = "n", cex = 0.85)

  # --- Panel 2: Per-pair decomposition ---

  obs <- x$observed
  means <- c(obs$mean_in, obs$mean_out)
  nms <- c("Within-node\n(mean_in)", "Between-node\n(mean_out)")
  cols <- c("#3498db", "#e74c3c")

  bp <- barplot(means, names.arg = nms, col = cols, border = NA,
                main = sprintf("Per-Pair Average Loss\nnLoI = %.4f", obs$nloi),
                ylab = "Mean loss per pair",
                ylim = c(0, max(means) * 1.35))
  text(bp, means, labels = sprintf("%.4f", means),
       pos = 3, cex = 0.9, font = 2)
}

eImpurity <- function(y, index, S) {
  S_sub <- S[index, , drop = FALSE]
  n <- length(index)

  # Pre-filter: remove splits where one group has < 2 obs
  tab <- colSums(S_sub)
  valid <- which(tab >= 2 & tab <= (n - 2))

  if (length(valid) == 0) {
    imp <- rep(Inf, ncol(S_sub))
    names(imp) <- colnames(S_sub)
    return(imp)
  }

  # Extract the dissimilarity sub-matrix for this node
  y_sub <- y[index, index]

  # Use C++ for fast impurity computation on valid splits only
  S_valid <- matrix(as.integer(as.matrix(S_sub[, valid, drop = FALSE])),
                    nrow = n, ncol = length(valid))
  imp_valid <- compute_impurity_cpp(y_sub, S_valid)

  # Build full result vector
  imp <- rep(Inf, ncol(S_sub))
  names(imp) <- colnames(S_sub)
  imp[valid] <- imp_valid

  return(imp)
}


############
# Kept as fallback for debugging; no longer used in main loop
dissimilarity <- function(y, s) {
  n <- dim(y)[1]
  nR <- sum(s == 0)
  nL <- n - nR
  dR <- y[s == 0, s == 0]
  dL <- y[s == 1, s == 1]
  sR <- sum(dR) / (n * (nR - 1))
  sL <- sum(dL) / (n * (nL - 1))
  return(sR + sL)
}

eStoppingRules <- function(y,index,t, setting, response, ensemble, vart1){
  n <- length(index)
  
  # create type object
  type <- get_ensemble_type(ensemble)
  
  if (n>1){
    impTotal <- meanDis(y[index,index])
    switch(type,
           classification = {
             res <- as.numeric(moda(response[index])[2])
           },
           regression={
             res <- 1-(variance(response[index])/vart1)
           }
    )
  } else {impTotal <- 0
  res <- 1
  }
  
  
  sRule <- isTRUE(impTotal<=setting$impTotal |
                    n<=setting$n |
                    res > 0.95  | ### if the variance is less than 5% of the variance in the root node, stop
                    (t*2)+1 > setting$tMax)
  results <- list(sRule=sRule,impTotal=impTotal,n=n)
  return(results)
  
}

############

meanDis <- function(dis){
  n <- nrow(dis)
  sum(dis)/(n*(n-1))
}

############

moda <- function(x) {
  if ( anyNA(x) ) x = x[!is.na(x)]
  ux <- unique(x)
  tab <- tabulate(match(x, ux))
  mode_val <- ux[which.max(tab)[1]]
  freq     <- max(tab)[1] / sum(tab)
  # `mode_val` may be a factor element, an integer (e.g., 0/1 binary
  # classification fed by gbm bernoulli), a character or a numeric value;
  # coerce uniformly to character so the caller can store it next to a
  # numeric frequency without triggering a length-zero replacement.
  c(as.character(mode_val), freq)
}

utils::globalVariables(c("n","prob", "terminal"))
#' Convert an E2Tree Object to rpart Format
#'
#' Converts an e2tree output into an rpart object.
#'
#' Note: \code{\link{as.rpart.e2tree}} is the preferred coercion method.
#' This function is kept for backward compatibility.
#'
#' @param fit is e2tree object.
#' @param ensemble A trained ensemble model. Supported classes: \code{randomForest}, \code{ranger},
#'   \code{xgb.Booster}, \code{lgb.Booster}, \code{gbm}, \code{catboost.CatBoost}.
#'
#' @return An rpart object. It contains the following components:
#' \tabular{lll}{
#' \code{frame}\tab   \tab  The data frame includes a singular row for each node present in the tree. The row.names within the frame are assigned as unique node numbers, following a binary ordering system indexed by the depth of the nodes. The columns of the frame consist of the following components: (var) this variable denotes the names of the variables employed in the split at each node. In the case of leaf nodes, the level "leaf" is used to indicate their status as terminal nodes; (n) the variable 'n' represents the number of observations that reach a particular node; (wt) 'wt' signifies the sum of case weights associated with the observations reaching a given node; (dev) the deviance of the node, which serves as a measure of the node's impurity or lack of fit; (yval) the fitted value of the response variable at the node; (splits) this two-column matrix presents the labels for the left and right splits associated with each node; (complexity) the complexity parameter indicates the threshold value at which the split is likely to collapse; (ncompete) 'ncompete' denotes the number of competitor splits recorded for a node; (nsurrogate) the variable 'nsurrogate' represents the number of surrogate splits recorded for a node \cr
#' \code{where}\tab   \tab An integer vector that matches the length of observations in the root node. The vector contains the row numbers in the frame that correspond to the leaf nodes where each observation is assigned \cr
#' \code{call}\tab   \tab The matched call \cr
#' \code{terms}\tab   \tab A list of terms and attributes  \cr
#' \code{control}\tab   \tab A list containing the set of stopping rules for the tree building procedure \cr
#' \code{functions}\tab   \tab The summary, print, and text functions are utilized for the specific method required \cr
#' \code{variable.importance}\tab   \tab Variable importance refers to a quantitative measure that assesses the contribution of individual variables within a predictive model towards accurate predictions. It quantifies the influence or impact that each variable has on the model's overall performance. Variable importance provides insights into the relative significance of different variables in explaining the observed outcomes and aids in understanding the underlying relationships and dynamics within the model \cr}
#'
#'
#' @examples
#'
#'\donttest{
#' ## Classification:
#' data(iris)
#'
#' # Create training and validation set:
#' smp_size <- floor(0.75 * nrow(iris))
#' train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
#' training <- iris[train_ind, ]
#' validation <- iris[-train_ind, ]
#' response_training <- training[,5]
#' response_validation <- validation[,5]
#'
#' # Perform training:
#' ## "randomForest" package
#' ensemble <- randomForest::randomForest(Species ~ ., data=training, 
#' importance=TRUE, proximity=TRUE)
#' 
#' ## "ranger" package
#' if (requireNamespace("ranger", quietly = TRUE)) {
#'   ensemble <- ranger::ranger(Species ~ ., data = iris,
#'     num.trees = 1000, importance = 'impurity')
#' }
#'
#' D <- createDisMatrix(ensemble, data=training, label = "Species",
#'                              parallel = list(active=FALSE, no_cores = 1))
#' 
#' setting=list(impTotal=0.1, maxDec=0.01, n=2, level=5)
#' tree <- e2tree(Species ~ ., training, D, ensemble, setting)
#'
#' ## Preferred coercion method:
#' rpart_obj <- as.rpart(tree, ensemble)
#'
#' ## Legacy function (see as.rpart):
#' rpart_obj <- rpart2Tree(tree, ensemble)
#'
#' # Plot using rpart.plot package:
#' rpart.plot::rpart.plot(rpart_obj)
#' 
#'}
#' @export


rpart2Tree <- function(fit, ensemble){
  
  # === Input Validation ===
  
  # Validate 'fit' (must be an 'e2tree' object)
  if (!inherits(fit, "e2tree")) {
    stop("Error: 'fit' must be an 'e2tree' object.")
  }
  
  # Validate 'ensemble' via adapter (supports all ensemble backends).
  # Preserve the legacy error message for NULL / generic-list inputs so that
  # existing user code matching that exact string keeps working.
  if (is.null(ensemble) || is.na(ensemble_backend(ensemble))) {
    stop("Error: 'ensemble' must be a trained 'randomForest' or 'ranger' model.",
         call. = FALSE)
  }
  type <- tryCatch(
    get_ensemble_type(ensemble),
    error = function(e) stop(conditionMessage(e), call. = FALSE)
  )
  if (!tolower(type) %in% c("classification", "regression")) {
    stop("Error: 'type' in ensemble object must be either 'classification' or 'regression'.",
         call. = FALSE)
  }

  # Validate that 'fit$tree' exists and is a data frame
  if (!is.data.frame(fit$tree)) {
    stop("Error: 'fit$tree' must be a data frame.")
  }

  # === Proceed with the function ===
  
  frame <- fit$tree
  
  switch(type,
         classification={
           frame <- frame %>%
             select("node","variable", "n", "pred_val","pred","prob","decImp", starts_with("yval2")) %>%
             rename("var"="variable",
                    "yval"="pred_val")
         },
         regression={
           frame <- frame %>%
             select("node","variable", "n", "pred_val","pred","prob","decImp") %>%
             rename("var"="variable",
                    "yval"="pred")
         })
  
  frame <- frame %>%
    mutate(wt=n,
           ncompete=0,
           nsurrogate=0,
           complexity=1-as.numeric(prob),
           dev=prob) %>%
    as.data.frame()
  
  rownames(frame) <- frame$node
  frame$var[is.na(frame$var)] <- "<leaf>"
  frame$complexity[is.na(frame$complexity)] <- 0.01
  
  
  switch(type,
         classification={
           frame <- frame %>%
             select("var","n","wt","dev","yval","complexity","ncompete","nsurrogate",starts_with("yval2"))
         },
         regression={
           frame <- frame %>%
             select("var","n","wt","dev","yval","complexity","ncompete","nsurrogate")
         })
  
  obs <- fit$tree %>%
    dplyr::filter(terminal) %>%
    select("node","n","obs")
  where <- rep(obs$node,obs$n)
  names(where) <- do.call(c,obs$obs)
  where <- where[order(as.numeric(names(where)))]
  
  variable.importance <- fit$varimp$vimp[[2]]
  names(variable.importance) <- fit$varimp$vimp[[1]]
  
  switch (type,
          classification = {
            obj <- list(frame=frame, where=where, call=fit$call, terms=fit$terms, method="class", control=fit$control, functions=rpartfunctions(),
                        splits=fit$splits, csplit=fit$csplit, variable.importance=variable.importance)
          },
          regression = {
            obj <- list(frame=frame, where=where, call=fit$call, terms=fit$terms, method="anova", control=fit$control, functions=rpartfunctions(),
                        splits=fit$splits, csplit=fit$csplit, variable.importance=variable.importance)
          }
  )
  
  attr(obj, "xlevels") <- attr(fit, "xlevels")
  attr(obj, "ylevels") <- attr(fit, "ylevels")
  #obj$frame <- obj$frame[as.character(fit$N),]
  class(obj) <- "rpart"
  return(obj)
}


rpartfunctions <- function(){
  summary <- function (yval, dev, wt, ylevel, digits)
  {
    nclass <- (ncol(yval) - 2L)/2L
    group <- yval[, 1L]
    counts <- yval[, 1L + (1L:nclass)]
    yprob <- yval[, 1L + nclass + 1L:nclass]
    nodeprob <- yval[, 2L * nclass + 2L]
    if (!is.null(ylevel))
      group <- ylevel[group]
    temp1 <- formatg(counts, format = "%5g")
    temp2 <- formatg(yprob, format = "%5.3f")
    if (nclass > 1) {
      temp1 <- apply(matrix(temp1, ncol = nclass), 1L, paste,
                     collapse = " ")
      temp2 <- apply(matrix(temp2, ncol = nclass), 1L, paste,
                     collapse = " ")
    }
    dev <- dev/(wt[1L] * nodeprob)
    paste0("  predicted class=", format(group, justify = "left"),
           "  expected loss=", formatg(dev, digits), "  P(node) =",
           formatg(nodeprob, digits), "\n", "    class counts: ",
           temp1, "\n", "   probabilities: ", temp2)
  }
  
  print <- function (yval, ylevel, digits, nsmall)
  {
    temp <- if (is.null(ylevel))
      as.character(yval[, 1L])
    else ylevel[yval[, 1L]]
    nclass <- (ncol(yval) - 2L)/2L
    yprob <- if (nclass < 5L)
      format(yval[, 1L + nclass + 1L:nclass], digits = digits,
             nsmall = nsmall)
    else formatg(yval[, 1L + nclass + 1L:nclass], digits = 2L)
    if (!is.matrix(yprob))
      yprob <- matrix(yprob, nrow = 1L)
    temp <- paste0(temp, " (", yprob[, 1L])
    for (i in 2L:ncol(yprob)) temp <- paste(temp, yprob[, i],
                                            sep = " ")
    temp <- paste0(temp, ")")
    temp
  }
  text <- function (yval, dev, wt, ylevel, digits, n, use.n)
  {
    nclass <- (ncol(yval) - 2L)/2L
    group <- yval[, 1L]
    counts <- yval[, 1L + (1L:nclass)]
    if (!is.null(ylevel))
      group <- ylevel[group]
    temp1 <- formatg(counts, digits)
    if (nclass > 1L)
      temp1 <- apply(matrix(temp1, ncol = nclass), 1L, paste,
                     collapse = "/")
    if (use.n)
      paste0(format(group, justify = "left"), "\n", temp1)
    else format(group, justify = "left")
  }
  
  functions <- list(summary=summary, print=print, text=text)
  return(functions)
}

formatg <- function(x, digits = getOption("digits"),
                    format = paste0("%.", digits, "g"))
{
  if (!is.numeric(x)) stop("'x' must be a numeric vector")
  
  temp <- sprintf(format, x)
  if (is.matrix(x)) matrix(temp, nrow = nrow(x)) else temp
}

#' Credit Scoring Dataset
#'
#' A dataset containing socio-economic and banking information for 468 bank
#' clients, used to assess creditworthiness. All variables are categorical.
#'
#' @format A data frame with 468 rows and 12 columns:
#' \describe{
#'   \item{Type_of_client}{Credit evaluation outcome: \code{"Creditworthy"} or
#'     \code{"Non-Creditworthy"}.}
#'   \item{Client_Age}{Age class of the client (e.g., \code{"less than 23 years"},
#'     \code{"from 23 to 35 years"}, \code{"from 35 to 50 years"},
#'     \code{"over 50 years"}).}
#'   \item{Family_Situation}{Marital/family status of the client (e.g.,
#'     \code{"single"}, \code{"married"}, \code{"divorced"}).}
#'   \item{Account_Tenure}{Length of the client's relationship with the bank
#'     (e.g., \code{"1 year or less"}, \code{"from 2 to 5 years"},
#'     \code{"plus 12 years"}).}
#'   \item{Salary_Credited_to_Bank_Account}{Whether the client's salary is
#'     credited to the bank account (e.g., \code{"domicile salary"},
#'     \code{"no domicile salary"}).}
#'   \item{Ammount_of_Savings}{Client's level of savings (e.g.,
#'     \code{"no savings"}, \code{"less than 5 thousand"},
#'     \code{"from 5 to 30 thousand"}, \code{"more than 30 thousand"}).}
#'   \item{Customer_Occupation}{Employment category of the client (e.g.,
#'     \code{"employee"}, \code{"self-employed"}, \code{"retired"}).}
#'   \item{Average_Account_Balance}{Average balance held in the account (e.g.,
#'     \code{"from 2 to 5 thousand"}, \code{"more than 5 thousand"}).}
#'   \item{Average_Account_Turnover}{Average monthly turnover on the account
#'     (e.g., \code{"Less than 10 thousand"}, \code{"from 10 to 50 thousand"},
#'     \code{"more than 50 thousand"}).}
#'   \item{Credit_Card_Transaction_Count_Monthly}{Number of credit card
#'     transactions per month (e.g., \code{"less than 40"}, \code{"from 40 to 100"},
#'     \code{"more than 100"}).}
#'   \item{Authorized_Overdraft_Limit}{Whether the client has an authorized
#'     overdraft facility (\code{"Authorised"} or \code{"forbidden"}).}
#'   \item{Authorized_to_Issue_Bank_Checks}{Whether the client is authorized
#'     to issue bank checks (\code{"Authorised"} or \code{"forbidden"}).}
#' }
"credit"

utils::globalVariables(c("decImp", "n", "pred", "variable", "Nimp", "vimp", "decProb", "pimp",
                         "Pimp", "Nimp", "vimp_resp", "vimp_prob", "Variable", "MeanImpurityDecrease", "MeanAccuracyDecrease")) # to avoid CRAN check errors for tidyverse programming

#' Variable Importance
#'
#' Computes variable importance for an E2Tree model based on mean
#' impurity decrease and (for classification) mean accuracy decrease.
#'
#' @param fit An e2tree object.
#' @param data A data frame containing the variables in the model.
#' @param type Character string: \code{"classification"} or \code{"regression"}.
#'   If \code{NULL} (default), the type is automatically detected from the
#'   e2tree object.
#'
#' @return A list containing:
#' \describe{
#'   \item{vimp}{A data frame with variable importance metrics.}
#'   \item{g_imp}{A ggplot bar chart of Mean Impurity Decrease.}
#'   \item{g_acc}{(Classification only) A ggplot bar chart of Mean Accuracy Decrease.}
#' }
#'
#' @examples
#' \donttest{
#' ## Classification:
#' data(iris)
#'
#' # Create training and validation set:
#' smp_size <- floor(0.75 * nrow(iris))
#' train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
#' training <- iris[train_ind, ]
#'
#' # Perform training:
#' ensemble <- randomForest::randomForest(Species ~ ., data=training,
#' importance=TRUE, proximity=TRUE)
#'
#' D <- createDisMatrix(ensemble, data=training, label = "Species",
#'                              parallel = list(active=FALSE, no_cores = 1))
#'
#' setting=list(impTotal=0.1, maxDec=0.01, n=2, level=5)
#' tree <- e2tree(Species ~ ., training, D, ensemble, setting)
#'
#' vi <- vimp(tree, training)
#' vi$vimp
#' vi$g_imp
#'
#'
#' ## Regression
#' data("mtcars")
#'
#' # Create training and validation set:
#' smp_size <- floor(0.75 * nrow(mtcars))
#' train_ind <- sample(seq_len(nrow(mtcars)), size = smp_size)
#' training <- mtcars[train_ind, ]
#'
#' # Perform training
#' ensemble = randomForest::randomForest(mpg ~ ., data=training, ntree=1000,
#' importance=TRUE, proximity=TRUE)
#'
#' D = createDisMatrix(ensemble, data=training, label = "mpg",
#'                          parallel = list(active=FALSE, no_cores = 1))
#'
#' setting=list(impTotal=0.1, maxDec=(1*10^-6), n=2, level=5)
#' tree <- e2tree(mpg ~ ., training, D, ensemble, setting)
#'
#' vi <- vimp(tree, training)
#' vi$vimp
#' vi$g_imp
#'
#' }
#'
#' @export
#'
vimp <- function(fit, data, type = NULL) {

  # === Input Validation ===

  if (!inherits(fit, "e2tree")) {
    stop("Error: 'fit' must be an 'e2tree' object.")
  }
  if (!is.data.frame(data) || nrow(data) == 0) {
    stop("Error: 'data' must be a non-empty data frame.")
  }
  if (!is.data.frame(fit$tree)) {
    stop("Error: 'fit$tree' must be a data frame.")
  }
  if (is.null(fit$terms)) {
    stop("Error: 'fit$terms' is missing or NULL in the e2tree object.")
  }

  # Auto-detect type from the e2tree object
  if (is.null(type)) {
    type <- if (!is.null(attr(fit, "ylevels"))) "classification" else "regression"
  }
  type <- match.arg(type, c("classification", "regression"))

  # === Prepare data ===

  data <- as.data.frame(data)
  row.names(data) <- NULL
  tree <- fit$tree
  response_var <- names(attr(fit$terms, "dataClasses"))[1]

  if (!response_var %in% colnames(data)) {
    stop("Error: The response variable from 'fit' is not found in 'data'.")
  }

  response <- data[[response_var]]

  # === Compute variable importance ===

  if (type == "classification") {
    res <- .vimp_classification(tree, data, response)
  } else {
    res <- .vimp_regression(tree)
  }

  res
}


# ===========================================================================
# Classification variable importance
# ===========================================================================
.vimp_classification <- function(tree, data, response) {

  row.names(tree) <- tree$node
  tree$prob <- as.numeric(tree$prob)

  # Internal nodes
  t <- row.names(tree)[tree$terminal == FALSE]
  tL <- as.character(tree[t, "node"] * 2)
  tR <- as.character(tree[t, "node"] * 2 + 1)

  # Weighted probability of children (primary split)
  tree[t, "probChildren"] <- (tree[tL, "prob"] * tree[tL, "n"] / tree[t, "n"]) +
    (tree[tR, "prob"] * tree[tR, "n"] / tree[t, "n"])

  # Weighted probability of children (surrogate split)
  names(response) <- row.names(data)
  for (i in t) {
    obs <- row.names(data)[unlist(tree[i, "obs"])]
    split_label <- tree[i, "splitLabelSur"]
    x <- data[obs, , drop = FALSE]
    goes_left <- evaluate_vimp_split(x, split_label)
    indL <- obs[goes_left]
    indR <- obs[!goes_left]
    probL <- as.numeric(moda(response[indL])[2])
    probR <- as.numeric(moda(response[indR])[2])
    tree[i, "probChildrenSur"] <- (probL * length(indL) / tree[i, "n"]) +
      (probR * length(indR) / tree[i, "n"])
  }

  tree[t, "decProb"] <- tree[t, "probChildren"] - tree[t, "probChildrenSur"]

  # Compute weighted importance
  tree_aug <- tree %>%
    mutate(Nimp = decImp * n / n[1],
           Pimp = decProb * n / n[1])

  # Grouped summarize by (variable, pred)
  by_var_pred <- tree_aug %>%
    group_by(variable, pred) %>%
    summarize(vimp = sum(Nimp, na.rm = TRUE),
              pimp = sum(Pimp, na.rm = TRUE), .groups = "drop") %>%
    drop_na(variable)

  vimp_resp <- by_var_pred %>%
    select(variable, pred, vimp) %>%
    pivot_wider(names_from = pred, values_from = vimp)
  names(vimp_resp)[-1] <- paste("ImpDec_", names(vimp_resp)[-1])

  vimp_prob <- by_var_pred %>%
    select(variable, pred, pimp) %>%
    pivot_wider(names_from = pred, values_from = pimp)
  names(vimp_prob)[-1] <- paste("AccDec_", names(vimp_prob)[-1])

  # Overall vimp
  vimp_df <- by_var_pred %>%
    group_by(variable) %>%
    summarize(vimp = sum(vimp, na.rm = TRUE),
              pimp = sum(pimp, na.rm = TRUE), .groups = "drop") %>%
    arrange(desc(vimp)) %>%
    left_join(vimp_resp, by = "variable") %>%
    left_join(vimp_prob, by = "variable")

  names(vimp_df)[1:3] <- c("Variable", "MeanImpurityDecrease", "MeanAccuracyDecrease")

  # --- Plots ---
  # Sort by MeanImpurityDecrease
  vimp_sorted_imp <- vimp_df %>% arrange(desc(MeanImpurityDecrease))
  pImp <- ggplot(vimp_sorted_imp, aes(y = Variable, x = MeanImpurityDecrease)) +
    geom_bar(stat = "identity", fill = "steelblue") +
    scale_y_discrete(limits = rev(vimp_sorted_imp$Variable)) +
    labs(title = "Mean Impurity Decrease",
         x = "Mean Impurity Decrease", y = "Variable") +
    theme_minimal()

  # Sort by MeanAccuracyDecrease
  vimp_sorted_acc <- vimp_df %>% arrange(desc(MeanAccuracyDecrease))
  pAcc <- ggplot(vimp_sorted_acc, aes(y = Variable, x = MeanAccuracyDecrease)) +
    geom_bar(stat = "identity", fill = "steelblue") +
    scale_y_discrete(limits = rev(vimp_sorted_acc$Variable)) +
    labs(title = "Mean Accuracy Decrease",
         x = "Mean Accuracy Decrease", y = "Variable") +
    theme_minimal()

  list(vimp = vimp_df, g_imp = pImp, g_acc = pAcc)
}


# ===========================================================================
# Regression variable importance
# ===========================================================================
.vimp_regression <- function(tree) {

  row.names(tree) <- tree$node

  vimp_df <- tree %>%
    mutate(Nimp = decImp * n / n[1]) %>%
    group_by(variable) %>%
    summarize(MeanImpurityDecrease = sum(Nimp, na.rm = TRUE), .groups = "drop") %>%
    drop_na(variable) %>%
    arrange(desc(MeanImpurityDecrease))

  # Rename for consistency
  names(vimp_df)[1] <- "Variable"

  # --- Plot ---
  pImp <- ggplot(vimp_df, aes(y = Variable, x = MeanImpurityDecrease)) +
    geom_bar(stat = "identity", fill = "steelblue") +
    scale_y_discrete(limits = rev(vimp_df$Variable)) +
    labs(title = "Mean Impurity Decrease",
         x = "Mean Impurity Decrease",
         y = "Variable") +
    theme_minimal()

  list(vimp = vimp_df, g_imp = pImp)
}


## Helper: evaluate surrogate split without eval/parse
evaluate_vimp_split <- function(x, split_label) {
  if (grepl(" <= ?", split_label)) {
    parts <- regmatches(split_label, regexec("^(.+?) <=(.+)$", split_label))[[1]]
    var_name <- trimws(parts[2])
    threshold <- as.numeric(trimws(parts[3]))
    return(x[[var_name]] <= threshold)
  } else if (grepl(" %in% ", split_label)) {
    parts <- regmatches(split_label, regexec("^(.+?) %in% (.+)$", split_label))[[1]]
    var_name <- trimws(parts[2])
    cats <- eval(parse(text = parts[3]))
    return(as.character(x[[var_name]]) %in% cats)
  } else {
    warning(paste("Unknown split format:", split_label))
    return(rep(TRUE, nrow(x)))
  }
}

utils::globalVariables("tree") # to avoid CRAN check errors for tidyverse programming

#' Validate an E2Tree Model via Proximity Matrix Comparison
#'
#' Compares the ensemble proximity matrix with the E2Tree-estimated proximity
#' matrix using multiple divergence and similarity measures. Can perform the
#' Mantel test, permutation tests on divergence/similarity measures (nLoI,
#' Hellinger, wRMSE, RV, SSIM), or both.
#'
#' @param data A data frame containing the variables in the model.
#' @param fit An e2tree object.
#' @param D The dissimilarity matrix obtained with \code{\link{createDisMatrix}}.
#' @param test Character string specifying which tests to perform. One of
#'   \code{"both"} (default), \code{"mantel"} (Mantel test only), or
#'   \code{"measures"} (divergence/similarity measures with permutation tests only).
#' @param graph Logical (default TRUE). If TRUE, heatmaps are displayed.
#' @param n_perm Integer. Number of permutations for the permutation
#'   test on measures. Default is 999. Set to 0 to skip permutation testing.
#'   Ignored when \code{test = "mantel"}.
#' @param conf.level Numeric. Confidence level for intervals. Default is 0.95.
#' @param seed Integer or NULL. Random seed for reproducibility.
#'
#' @return An object of class \code{"eValidation"} containing:
#'   \describe{
#'     \item{Proximity_matrix_ensemble}{Ensemble proximity matrix (reordered)}
#'     \item{Proximity_matrix_e2tree}{E2Tree proximity matrix (reordered)}
#'     \item{mantel_test}{Mantel test result (NULL if \code{test = "measures"})}
#'     \item{loi}{LoI object with decomposition (NULL if \code{test = "mantel"})}
#'     \item{measures}{Data frame with all measures (NULL if \code{test = "mantel"})}
#'     \item{permutation}{Permutation test results for measures (if applicable)}
#'   }
#'
#' @examples
#' \donttest{
#' ## Classification:
#' data(iris)
#' smp_size <- floor(0.75 * nrow(iris))
#' train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
#' training <- iris[train_ind, ]
#'
#' ensemble <- randomForest::randomForest(Species ~ ., data=training,
#'   importance=TRUE, proximity=TRUE)
#'
#' D <- createDisMatrix(ensemble, data=training, label = "Species",
#'   parallel = list(active=FALSE, no_cores = 1))
#'
#' setting <- list(impTotal=0.1, maxDec=0.01, n=2, level=5)
#' tree <- e2tree(Species ~ ., training, D, ensemble, setting)
#'
#' val <- eValidation(training, tree, D, n_perm = 199)
#' print(val)
#' summary(val)
#' plot(val)
#' }
#'
#' @export
eValidation <- function(data, fit, D, test = c("both", "mantel", "measures"),
                        graph = TRUE,
                        n_perm = 999, conf.level = 0.95, seed = NULL) {

  if (!is.null(seed)) set.seed(seed)
  test <- match.arg(test)

  # === Input Validation ===
  if (!is.data.frame(data) || nrow(data) == 0) {
    stop("Error: 'data' must be a non-empty data frame.")
  }
  if (!inherits(fit, "e2tree")) {
    stop("Error: 'fit' must be an 'e2tree' object.")
  }
  if (!is.matrix(D) || nrow(D) != ncol(D)) {
    stop("Error: 'D' must be a square dissimilarity matrix.")
  }
  if (nrow(data) != nrow(D)) {
    stop("Error: The number of rows in 'data' must match the dimensions of 'D'.")
  }

  n <- nrow(data)
  df <- fit$tree
  terminal_nodes <- df$node[df$terminal]

  # === Build E2Tree proximity matrix Ps ===
  Ps <- matrix(0, n, n)
  for (i in terminal_nodes) {
    obs <- unlist(df$obs[df$node == i])
    if (!is.null(df$prob)) {
      Ps[obs, obs] <- df$prob[df$node == i]
    } else {
      Ps[obs, obs] <- df$Wt[df$node == i]
    }
  }
  diag(Ps) <- 1
  rownames(Ps) <- 1:n
  colnames(Ps) <- 1:n

  # === Reorder by hierarchical clustering ===
  D_exp <- D
  clusD <- hclust(as.dist(D_exp))
  ord <- clusD$order

  Ps_ord <- Ps[ord, ord]
  rownames(Ps_ord) <- ord
  colnames(Ps_ord) <- ord

  # === Mantel test ===
  mantel_test <- NULL
  if (test %in% c("both", "mantel")) {
    mantel_test <- ape::mantel.test(
      Ps_ord,
      1 - D_exp[ord, ord],
      graph = graph,
      main = "Mantel test",
      xlab = "z-statistic", ylab = "Density"
    )
  }

  # === Clamp to [0,1] and compute proximity matrices ===
  Ps_ord <- pmin(pmax(Ps_ord, 0), 1)
  prox_e2tree <- sqrt(Ps_ord)
  prox_ens <- sqrt(pmax(1 - D_exp[ord, ord], 0))

  # === Heatmaps ===
  if (graph) {
    e2heatmap(prox_e2tree)
    e2heatmap(prox_ens)
  }

  # === Compute divergence/similarity measures ===
  loi_result <- NULL
  measures_df <- NULL
  perm_result <- NULL

  if (test %in% c("both", "measures")) {
    O <- prox_ens
    O_hat <- prox_e2tree

    loi_result <- loi(O, O_hat)

    hellinger_val <- .e2val_hellinger(O, O_hat)
    wrmse_val     <- .e2val_wrmse(O, O_hat)
    rv_val        <- .e2val_rv(O, O_hat)
    ssim_val      <- .e2val_ssim(O, O_hat)

    measures_df <- data.frame(
      method   = c("nLoI", "Hellinger", "wRMSE", "RV", "SSIM"),
      type     = c("divergence", "divergence", "divergence", "similarity", "similarity"),
      observed = c(loi_result$nloi, hellinger_val, wrmse_val, rv_val, ssim_val),
      stringsAsFactors = FALSE
    )

    # === Permutation test (unified, row/column) ===
    if (n_perm > 0) {
      perm_result <- .e2val_unified_perm(O, O_hat, n_perm, conf.level)
      measures_df$null_mean <- perm_result$null_means
      measures_df$null_sd   <- perm_result$null_sds
      measures_df$z_stat    <- perm_result$z_stats
      measures_df$p_value   <- perm_result$p_values
      measures_df$ci_lower  <- perm_result$ci_lower
      measures_df$ci_upper  <- perm_result$ci_upper
    }
  }

  # === Build result ===
  results <- list(
    Proximity_matrix_ensemble = prox_ens,
    Proximity_matrix_e2tree   = prox_e2tree,
    mantel_test               = mantel_test,
    loi                       = loi_result,
    measures                  = measures_df,
    permutation               = perm_result,
    n                         = n,
    n_perm                    = n_perm,
    conf.level                = conf.level,
    test                      = test
  )
  class(results) <- "eValidation"
  results
}


# ===========================================================================
# INTERNAL: Hellinger distance
# ===========================================================================
.e2val_hellinger <- function(O, O_hat) {
  idx <- lower.tri(O, diag = FALSE)
  m <- sum(idx)
  sq_diff <- (sqrt(O[idx]) - sqrt(O_hat[idx]))^2
  sqrt(sum(sq_diff, na.rm = TRUE) / m)
}

# ===========================================================================
# INTERNAL: Weighted RMSE
# ===========================================================================
.e2val_wrmse <- function(O, O_hat, epsilon = .Machine$double.eps) {
  idx <- lower.tri(O, diag = FALSE)
  o <- O[idx]
  oh <- O_hat[idx]
  w <- pmax(o, oh, epsilon)
  sqrt(sum(w * (o - oh)^2, na.rm = TRUE) / sum(w, na.rm = TRUE))
}

# ===========================================================================
# INTERNAL: RV coefficient
# ===========================================================================
.e2val_rv <- function(O, O_hat) {
  O_c <- O; O_hat_c <- O_hat
  diag(O_c) <- 0; diag(O_hat_c) <- 0
  num <- sum(O_c * O_hat_c)
  denom <- sqrt(sum(O_c * O_c) * sum(O_hat_c * O_hat_c))
  if (denom < .Machine$double.eps) return(0)
  num / denom
}

# ===========================================================================
# INTERNAL: SSIM
# ===========================================================================
.e2val_ssim <- function(O, O_hat, window_size = NULL) {
  n <- nrow(O)
  if (is.null(window_size)) {
    window_size <- min(11L, max(3L, floor(n / 3)))
  }
  window_size <- as.integer(window_size)
  if (window_size %% 2 == 0) window_size <- window_size + 1L

  L <- 1
  C1 <- (0.01 * L)^2
  C2 <- (0.03 * L)^2
  w <- window_size
  k <- w * w

  .box_filter <- function(M, w) {
    n <- nrow(M)
    half_w <- (w - 1L) %/% 2L
    pad <- matrix(0, n + w - 1, n + w - 1)
    pad[(half_w + 1):(half_w + n), (half_w + 1):(half_w + n)] <- M
    cs <- apply(pad, 2, cumsum)
    row_sum <- cs[w:nrow(cs), , drop = FALSE] -
      rbind(rep(0, ncol(cs)), cs[1:(nrow(cs) - w), , drop = FALSE])
    cs2 <- t(apply(row_sum, 1, cumsum))
    out <- cs2[, w:ncol(cs2), drop = FALSE] -
      cbind(rep(0, nrow(cs2)), cs2[, 1:(ncol(cs2) - w), drop = FALSE])
    out[1:n, 1:n, drop = FALSE] / k
  }

  mu_O <- .box_filter(O, w)
  mu_Oh <- .box_filter(O_hat, w)
  mu_O2 <- .box_filter(O * O, w)
  mu_Oh2 <- .box_filter(O_hat * O_hat, w)
  mu_OOh <- .box_filter(O * O_hat, w)

  sigma_O2 <- pmax(mu_O2 - mu_O^2, 0)
  sigma_Oh2 <- pmax(mu_Oh2 - mu_Oh^2, 0)
  sigma_OOh <- mu_OOh - mu_O * mu_Oh

  numerator <- (2 * mu_O * mu_Oh + C1) * (2 * sigma_OOh + C2)
  denominator <- (mu_O^2 + mu_Oh^2 + C1) * (sigma_O2 + sigma_Oh2 + C2)
  ssim_map <- numerator / denominator

  half_w <- (w - 1L) %/% 2L
  valid_rows <- (half_w + 1):(n - half_w)
  valid_cols <- (half_w + 1):(n - half_w)

  if (length(valid_rows) < 1 || length(valid_cols) < 1) {
    o_vec <- O[lower.tri(O)]
    oh_vec <- O_hat[lower.tri(O_hat)]
    mu_o <- mean(o_vec); mu_oh <- mean(oh_vec)
    s_o2 <- var(o_vec); s_oh2 <- var(oh_vec)
    s_cross <- cov(o_vec, oh_vec)
    return(((2*mu_o*mu_oh + C1) * (2*s_cross + C2)) /
             ((mu_o^2 + mu_oh^2 + C1) * (s_o2 + s_oh2 + C2)))
  }

  mean(ssim_map[valid_rows, valid_cols], na.rm = TRUE)
}


# ===========================================================================
# INTERNAL: Unified permutation test for all measures
# ===========================================================================
.e2val_unified_perm <- function(O, O_hat, n_perm, conf.level) {

  n <- nrow(O)
  method_names <- c("nLoI", "Hellinger", "wRMSE", "RV", "SSIM")
  n_methods <- length(method_names)

  # Observed values
  obs_vals <- c(
    loi(O, O_hat)$nloi,
    .e2val_hellinger(O, O_hat),
    .e2val_wrmse(O, O_hat),
    .e2val_rv(O, O_hat),
    .e2val_ssim(O, O_hat)
  )

  is_div <- c(TRUE, TRUE, TRUE, FALSE, FALSE)

  # Permutation loop
  compute_perm <- function(dummy) {
    perm_idx <- sample.int(n)
    O_hat_p <- O_hat[perm_idx, perm_idx]
    c(
      loi(O, O_hat_p)$nloi,
      .e2val_hellinger(O, O_hat_p),
      .e2val_wrmse(O, O_hat_p),
      .e2val_rv(O, O_hat_p),
      .e2val_ssim(O, O_hat_p)
    )
  }

  if (requireNamespace("future.apply", quietly = TRUE)) {
    perm_mat <- future.apply::future_sapply(seq_len(n_perm), compute_perm,
                                             future.seed = TRUE)
  } else {
    perm_mat <- vapply(seq_len(n_perm), compute_perm, numeric(n_methods))
  }
  # perm_mat: n_methods x n_perm

  # Statistics
  null_means <- rowMeans(perm_mat, na.rm = TRUE)
  null_sds <- apply(perm_mat, 1, sd, na.rm = TRUE)
  z_stats <- (obs_vals - null_means) / null_sds

  # P-values
  p_values <- numeric(n_methods)
  for (i in seq_len(n_methods)) {
    if (is_div[i]) {
      p_values[i] <- (1 + sum(perm_mat[i, ] <= obs_vals[i], na.rm = TRUE)) / (1 + n_perm)
    } else {
      p_values[i] <- (1 + sum(perm_mat[i, ] >= obs_vals[i], na.rm = TRUE)) / (1 + n_perm)
    }
  }

  # Confidence intervals
  alpha <- 1 - conf.level
  ci_lower <- ci_upper <- numeric(n_methods)
  for (i in seq_len(n_methods)) {
    null_q <- quantile(perm_mat[i, ], probs = c(alpha/2, 1-alpha/2), na.rm = TRUE)
    p_shift <- null_means[i] - null_q
    ci_raw <- obs_vals[i] + p_shift
    ci <- c(ci_raw[2], ci_raw[1])
    if (is_div[i]) ci <- pmax(ci, 0)
    if (!is_div[i]) {
      lb <- if (method_names[i] == "SSIM") -1 else 0
      ci <- pmin(pmax(ci, lb), 1)
    }
    ci_lower[i] <- ci[1]
    ci_upper[i] <- ci[2]
  }

  list(
    null_means = null_means,
    null_sds   = null_sds,
    z_stats    = z_stats,
    p_values   = p_values,
    ci_lower   = ci_lower,
    ci_upper   = ci_upper,
    null_dists = perm_mat,
    method_names = method_names
  )
}


# ===========================================================================
# PRINT METHOD
# ===========================================================================

#' @method print eValidation
#' @export
print.eValidation <- function(x, digits = 4, ...) {

  cat("\n")
  cat("##############################################################################\n")
  cat("   E2Tree Validation\n")
  cat("##############################################################################\n\n")

  cat(sprintf("  Matrix dimension:    %d x %d\n", x$n, x$n))
  cat(sprintf("  Pairs:               %d\n", x$n * (x$n - 1) / 2))

  # Mantel test
  if (!is.null(x$mantel_test)) {
    cat(sprintf("\n  Mantel test:         z = %.2f, p = %.4f\n",
                x$mantel_test$z.stat, x$mantel_test$p))
  }

  # Measures table
  if (!is.null(x$measures)) {
    tbl <- x$measures

    cat("\n------------------------------------------------------------------------------\n")
    cat("  Measure          Type        Observed")
    if (!is.null(x$permutation)) cat("    Null mean    Z-stat     p-value")
    cat("\n")
    cat("------------------------------------------------------------------------------\n")

    for (i in seq_len(nrow(tbl))) {
      dir <- if (tbl$type[i] == "divergence") "[div]" else "[sim]"

      line <- sprintf("  %-14s %s   %8.*f",
                      tbl$method[i], dir, digits, tbl$observed[i])

      if (!is.null(x$permutation)) {
        stars <- if (tbl$p_value[i] < 0.001) "***"
                 else if (tbl$p_value[i] < 0.01) "**"
                 else if (tbl$p_value[i] < 0.05) "*"
                 else if (tbl$p_value[i] < 0.1) "."
                 else ""

        p_str <- if (tbl$p_value[i] < 0.0001) "< 0.0001"
                 else sprintf("%.*f", digits, tbl$p_value[i])

        line <- paste0(line, sprintf("   %8.*f   %+8.2f   %s %s",
                                      digits, tbl$null_mean[i],
                                      tbl$z_stat[i], p_str, stars))
      }
      cat(line, "\n")
    }
    cat("------------------------------------------------------------------------------\n")

    if (!is.null(x$permutation)) {
      cat(sprintf("  Permutations: %d (row/column), conf.level: %d%%\n",
                  x$n_perm, round(x$conf.level * 100)))
    }

    # LoI decomposition summary
    if (!is.null(x$loi)) {
      lo <- x$loi
      cat(sprintf("\n  LoI Decomposition (per-pair avg):  mean_in = %.6f,  mean_out = %.6f\n",
                  lo$mean_in, lo$mean_out))
    }

    cat("\n")
    cat("  [div] = divergence (lower=better), [sim] = similarity (higher=better)\n")
    cat("  Signif. codes: 0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1\n")
  }

  cat("\n##############################################################################\n\n")

  invisible(x)
}


# ===========================================================================
# SUMMARY METHOD
# ===========================================================================

#' @method summary eValidation
#' @export
summary.eValidation <- function(object, digits = 4, ...) {

  print(object, digits = digits, ...)
  if (!is.null(object$loi)) {
    cat("\n")
    summary(object$loi, digits = digits)
  }

  invisible(object)
}


# ===========================================================================
# PLOT METHOD
# ===========================================================================

#' @method plot eValidation
#' @export
plot.eValidation <- function(x, ...) {

  old_par <- par(no.readonly = TRUE)
  on.exit(par(old_par))

  has_perm <- !is.null(x$permutation)
  has_measures <- !is.null(x$measures)

  # Determine layout based on available components
  n_panels <- 2L  # always show heatmaps
  if (has_measures && has_perm) n_panels <- 4L
  else if (has_measures) n_panels <- 3L

  if (n_panels == 4L) {
    layout(matrix(c(1, 2, 3, 4), 2, 2, byrow = TRUE))
  } else {
    par(mfrow = c(1, n_panels))
  }

  par(mar = c(4, 4, 3, 1))

  # --- Panel 1: Ensemble heatmap ---
  image(x$Proximity_matrix_ensemble,
        main = "Ensemble Proximity (O)",
        col = colorRampPalette(c("white", "black"))(100),
        axes = FALSE, xlab = "", ylab = "")

  # --- Panel 2: E2Tree heatmap ---
  image(x$Proximity_matrix_e2tree,
        main = expression("E2Tree Proximity (" * hat(O) * ")"),
        col = colorRampPalette(c("white", "black"))(100),
        axes = FALSE, xlab = "", ylab = "")

  if (has_measures && has_perm) {
    # --- Panel 3: Null distribution of nLoI ---
    null_nloi <- x$permutation$null_dists[1, ]
    obs_nloi <- x$measures$observed[1]

    xlim_range <- range(c(null_nloi, obs_nloi), na.rm = TRUE)
    h <- hist(null_nloi, plot = FALSE)
    d <- density(null_nloi, na.rm = TRUE)
    ylim_max <- max(max(h$density), max(d$y)) * 1.15

    hist(null_nloi,
         main = "Null Distribution of nLoI",
         xlab = "nLoI", col = "lightgray", border = "white",
         xlim = xlim_range, ylim = c(0, ylim_max), freq = FALSE)
    lines(d, col = "darkgray", lwd = 2)
    abline(v = obs_nloi, col = "red", lwd = 2)
    abline(v = mean(null_nloi), col = "darkgray", lwd = 1.5, lty = 2)

    p_val <- x$measures$p_value[1]
    p_str <- if (p_val < 0.0001) "p < 0.0001" else sprintf("p = %.4f", p_val)
    legend("topright",
           legend = c(sprintf("Observed (%.4f)", obs_nloi), p_str),
           col = c("red", NA), lty = c(1, NA), lwd = c(2, NA),
           bty = "n", cex = 0.85)
  }

  # --- LoI Decomposition (per-pair) ---
  if (has_measures && !is.null(x$loi)) {
    lo <- x$loi
    means <- c(lo$mean_in, lo$mean_out)
    nms <- c("Within\n(mean_in)", "Between\n(mean_out)")
    cols <- c("#3498db", "#e74c3c")

    bp <- barplot(means, names.arg = nms, col = cols, border = NA,
                  main = sprintf("Per-Pair Average Loss\nnLoI = %.4f", lo$nloi),
                  ylab = "Mean loss per pair",
                  ylim = c(0, max(means) * 1.35))
    text(bp, means, labels = sprintf("%.4f", means),
         pos = 3, cex = 0.9, font = 2)
  }
}


# ===========================================================================
# HEATMAP HELPER (internal)
# ===========================================================================

e2heatmap <- function(data_matrix) {
  heatmap(
    data_matrix,
    Rowv = NA,
    Colv = NA,
    scale = "none",
    col = colorRampPalette(c("white", "black"))(100)
  )
}

#X is a data frame of predictors (numerical, ordered or factor classes)
#max_cat: maximum number of categories for exhaustive binary partitioning
#         variables with more categories fall back to ordinal (frequency-based) splits

split <- function(X, max_cat = 10){
  # check column class and generate splits for each column

  xLabels <- names(X)

  S <- lapply(X,function(x){
    type <- class(x)
    if (length(type)>1){
      if ("ordered" %in% type){
        type <- "ordered"  # if ordered is present, use it
      } else {
        type <- type[1]  # otherwise take the first class
      }
    }

    if (type %in% c("character","factor")){
      S <- catSplit(x, max_cat = max_cat)  # split for categorical variables
    } else if (type %in% c("numeric","ordered","integer")){
      S <- ordSplit(x)  # split for numeric or ordered variables
    }
    return(S)
  })
  # merge splits into a single data frame
  lab <- rep(names(X),lengths(S))
  l <- paste(lab,unlist(lapply(S,names)))
  S <- (do.call(cbind, S))
  names(S) <- l
  row.names(S) <- row.names(X)
  res <- list(S=S,lab=lab)
  return(res)
}

# Splitting of categorical variables
# max_cat: if number of categories > max_cat, use frequency-ordered
#          ordinal splits (k-1 splits) instead of exhaustive binary
#          partitions (2^(k-1)-1 splits) to avoid combinatorial explosion
catSplit <- function(x, max_cat = 10){
  # coercing factor to character variable
  x <- as.character(x)
  # take unique values
  values <- sort(unique(x))
  k <- length(values)

  if (k < 2) {
    message("\nYou selected a variable with constant value!!!!\n")
    return()
  }

  if (k == 2) {
    # Binary variables — single split
    Sx <- data.frame(rep(0, length(x)))
    Sx[x == values[1], 1] <- 1
    names(Sx) <- paste("%in% c('", values[1], "')", collapse = "", sep = "")
    return(Sx)
  }

  if (k > max_cat) {
    # HIGH-CARDINALITY FALLBACK: order categories by frequency, then
    # generate k-1 ordinal splits (cumulative from most frequent)
    freq_order <- names(sort(table(x), decreasing = TRUE))
    # Direct factor → numeric rank (1 = most frequent)
    x_rank <- as.numeric(factor(x, levels = freq_order))

    # Generate ordinal splits on the rank — vectorized with outer()
    thresholds <- seq_len(k - 1)
    Sx <- as.data.frame(outer(x_rank, thresholds, "<=") + 0L)

    # Label: list categories in the left group
    lab <- vapply(thresholds, function(thr) {
      cats_left <- freq_order[1:thr]
      paste("%in% c('", paste(cats_left, collapse = "', '"), "')", sep = "")
    }, character(1))
    colnames(Sx) <- lab
    return(Sx)
  }

  # STANDARD: exhaustive binary partitions for k <= max_cat
  # generate all partitions
  p <- partitions::listParts(k, do.set = FALSE)

  # remove not binary partitions
  s <- lapply(p, function(x){
    lab <- as.character(x)
    if (length(lab) == 2){
      x <- eval(parse(text = lab[1]))
    } else {x <- NA}
  })

  s <- s[lengths(s) > 1]

  # generate splits
  S <- lapply(s, function(l){
    cbind(x %in% values[l])
  })

  # labeling splits
  lab <- unlist(lapply(s, function(x){
    paste("%in% c('", paste(values[x], collapse = "', '"), "')", sep = "")
  }))

  Sx <- as.data.frame(do.call(cbind, S))
  colnames(Sx) <- lab
  return(Sx)
}


# Splitting of ordered variables
ordSplit <- function(x){

  # take unique values of X and create split list and labels
  values <- sort(unique(x))
  if (length(values)>1){
    values <- values[-length(values)]
    Sxlabel <- paste("<=",values,sep="")

    # build up the splits — vectorized with outer()
    Sx <- as.data.frame(outer(x, values, "<=") + 0L)
    colnames(Sx) <- Sxlabel

    return(Sx)}
  else{
    message("\nYou selected a variable with constant value!!!!\n")
    return()
  }
}

# ===========================================================================
# S3 methods for the "e2tree" class
# ===========================================================================

#' Print an E2Tree model
#'
#' Displays a compact summary of the fitted E2Tree model including
#' task type, tree size, terminal nodes, and splitting variables.
#'
#' @param x An e2tree object
#' @param ... Additional arguments (ignored)
#'
#' @method print e2tree
#' @export
print.e2tree <- function(x, ...) {

  tree_info <- x$tree
  is_class <- !is.null(attr(x, "ylevels"))
  resp <- all.vars(x$terms)[1]
  preds <- all.vars(x$terms)[-1]

  n_nodes <- nrow(tree_info)
  n_terminal <- sum(tree_info$terminal)
  n_obs <- tree_info$n[1]  # root node contains total observations

  # Max depth
  max_depth <- max(vapply(tree_info$node, function(nd) {
    d <- 0L; while (nd > 1L) { nd <- nd %/% 2L; d <- d + 1L }; d
  }, integer(1)))

  # Splitting variables used
  split_vars <- unique(tree_info$variable[!is.na(tree_info$variable)])

  cat("\n")
  cat("  Explainable Ensemble Tree (E2Tree)\n")
  cat("  -----------------------------------\n")
  cat(sprintf("  Task:            %s\n", if (is_class) "Classification" else "Regression"))
  cat(sprintf("  Response:        %s\n", resp))
  cat(sprintf("  Predictors:      %d (%s)\n", length(preds),
              paste(head(preds, 5), collapse = ", ")))
  if (length(preds) > 5) cat(sprintf("                   ... and %d more\n", length(preds) - 5))
  cat(sprintf("  Observations:    %d\n", n_obs))
  cat(sprintf("  Nodes:           %d (total), %d (terminal)\n", n_nodes, n_terminal))
  cat(sprintf("  Max depth:       %d\n", max_depth))
  cat(sprintf("  Split variables: %s\n", paste(split_vars, collapse = ", ")))

  if (is_class) {
    cat(sprintf("  Classes:         %s\n", paste(attr(x, "ylevels"), collapse = ", ")))
  }

  cat("\n")
  invisible(x)
}


#' Summary of an E2Tree model
#'
#' Displays a comprehensive summary including tree structure, decision rules,
#' terminal node statistics, and variable importance.
#'
#' @param object An e2tree object
#' @param ... Additional arguments (ignored)
#'
#' @method summary e2tree
#' @export
summary.e2tree <- function(object, ...) {

  tree_info <- object$tree
  is_class <- !is.null(attr(object, "ylevels"))
  resp <- all.vars(object$terms)[1]

  n_nodes <- nrow(tree_info)
  n_terminal <- sum(tree_info$terminal)
  n_obs <- tree_info$n[1]  # root node

  max_depth <- max(vapply(tree_info$node, function(nd) {
    d <- 0L; while (nd > 1L) { nd <- nd %/% 2L; d <- d + 1L }; d
  }, integer(1)))

  split_vars <- unique(tree_info$variable[!is.na(tree_info$variable)])

  cat("\n")
  cat(paste(rep("=", 70), collapse = ""), "\n")
  cat("                    E2TREE MODEL SUMMARY\n")
  cat(paste(rep("=", 70), collapse = ""), "\n\n")

  # --- Model info ---
  cat("MODEL INFORMATION\n")
  cat(paste(rep("-", 40), collapse = ""), "\n")
  cat(sprintf("  Task:              %s\n", if (is_class) "Classification" else "Regression"))
  cat(sprintf("  Response:          %s\n", resp))
  cat(sprintf("  Observations:      %d\n", n_obs))
  cat(sprintf("  Total Nodes:       %d\n", n_nodes))
  cat(sprintf("  Terminal Nodes:    %d\n", n_terminal))
  cat(sprintf("  Max Depth:         %d\n", max_depth))
  cat(sprintf("  Split Variables:   %s\n", paste(split_vars, collapse = ", ")))

  if (is_class) {
    cat(sprintf("  Classes:           %s\n", paste(attr(object, "ylevels"), collapse = ", ")))
  }

  # --- Variable importance ---
  if (!is.null(object$varimp) && !is.null(object$varimp$vimp)) {
    cat("\n\nVARIABLE IMPORTANCE\n")
    cat(paste(rep("-", 40), collapse = ""), "\n")
    vimp_df <- object$varimp$vimp
    if (is.data.frame(vimp_df) && ncol(vimp_df) >= 2) {
      vimp_df <- vimp_df[order(vimp_df[[2]], decreasing = TRUE), ]
      max_imp <- max(vimp_df[[2]])
      for (i in seq_len(nrow(vimp_df))) {
        bar_width <- round(vimp_df[[2]][i] / max_imp * 20)
        bar <- paste(rep("#", bar_width), collapse = "")
        cat(sprintf("  %-20s %8.4f  %s\n", vimp_df[[1]][i], vimp_df[[2]][i], bar))
      }
    }
  }

  # --- Terminal node summary ---
  cat("\n\nTERMINAL NODES\n")
  cat(paste(rep("-", 70), collapse = ""), "\n")

  terminal <- tree_info[tree_info$terminal, ]

  if (is_class) {
    cat(sprintf("  %-8s  %-15s  %6s  %6s  %6s\n",
                "Node", "Prediction", "n", "Purity", "Wt"))
    cat(paste(rep("-", 55), collapse = ""), "\n")
    for (i in seq_len(nrow(terminal))) {
      nd <- terminal[i, ]
      purity <- if (!is.null(nd$prob)) sprintf("%.1f%%", nd$prob * 100) else "--"
      wt <- if (!is.null(nd$Wt)) sprintf("%.3f", nd$Wt) else "--"
      cat(sprintf("  %-8d  %-15s  %6d  %6s  %6s\n",
                  nd$node, nd$pred, nd$n, purity, wt))
    }
  } else {
    cat(sprintf("  %-8s  %-12s  %6s  %8s\n",
                "Node", "Prediction", "n", "Wt"))
    cat(paste(rep("-", 45), collapse = ""), "\n")
    for (i in seq_len(nrow(terminal))) {
      nd <- terminal[i, ]
      wt <- if (!is.null(nd$Wt)) sprintf("%.3f", nd$Wt) else "--"
      cat(sprintf("  %-8d  %-12s  %6d  %8s\n",
                  nd$node, format(nd$pred, digits = 4), nd$n, wt))
    }
  }

  # --- Decision rules ---
  cat("\n\nDECISION RULES\n")
  cat(paste(rep("-", 70), collapse = ""), "\n")

  for (i in seq_len(nrow(terminal))) {
    nd <- terminal[i, ]
    cat(sprintf("\nRule %d (Node %d, n=%d):\n", i, nd$node, nd$n))

    if (!is.na(nd$path) && nchar(nd$path) > 0) {
      conditions <- strsplit(nd$path, " & ")[[1]]
      for (j in seq_along(conditions)) {
        cond <- trimws(conditions[j])
        # Clean up negation
        if (startsWith(cond, "!")) {
          cond <- substring(cond, 2)
          if (grepl("<=", cond)) cond <- gsub("<=", ">", cond)
          else if (grepl("%in%", cond)) cond <- gsub("%in%", "NOT IN", cond)
        } else {
          if (grepl("%in%", cond)) cond <- gsub("%in%", "IN", cond)
        }
        cond <- gsub("c\\(", "{", cond)
        cond <- gsub("\\)", "}", cond)
        prefix <- if (j == 1) "  IF  " else "  AND "
        cat(sprintf("  %s %s\n", prefix, cond))
      }
    } else {
      cat("  IF (root -- all observations)\n")
    }
    cat(sprintf("  THEN: %s\n", nd$pred))
  }

  cat("\n")
  cat(paste(rep("=", 70), collapse = ""), "\n\n")

  invisible(object)
}


#' Plot an E2Tree model
#'
#' Displays the tree structure using rpart.plot.
#' This is a convenience wrapper around \code{\link{plot_e2tree}}.
#'
#' @param x An e2tree object
#' @param ensemble The ensemble model (randomForest or ranger).
#'   Required for converting the tree to rpart format. Supported classes:
#'   \code{randomForest}, \code{ranger}, \code{xgb.Booster}, \code{lgb.Booster},
#'   \code{gbm}, \code{catboost.CatBoost}.
#' @param main Plot title. Default is "E2Tree".
#' @param ... Additional arguments passed to \code{rpart.plot::rpart.plot}
#'
#' @method plot e2tree
#' @export
plot.e2tree <- function(x, ensemble = NULL, main = "E2Tree", ...) {

  if (is.null(ensemble)) {
    stop("'ensemble' argument is required for plotting. ",
         "Pass the ensemble model (randomForest, ranger, xgb.Booster, lgb.Booster, gbm, catboost.CatBoost) used to build the E2Tree.")
  }

  if (!requireNamespace("rpart.plot", quietly = TRUE)) {
    stop("Package 'rpart.plot' is required. Install with: install.packages('rpart.plot')")
  }

  rpart_obj <- rpart2Tree(x, ensemble)
  is_class <- !is.null(attr(x, "ylevels"))

  rpart.plot::rpart.plot(
    rpart_obj,
    main = main,
    type = 4,
    extra = if (is_class) 104 else 101,
    under = TRUE,
    faclen = 0,
    cex = 0.9,
    box.palette = if (is_class) "auto" else "Blues",
    shadow.col = "gray",
    nn = TRUE,
    roundint = FALSE,
    ...
  )

  invisible(rpart_obj)
}


#' Predict Responses from an E2Tree Model
#'
#' Predicts classification or regression responses for new data using the
#' fitted E2Tree model.
#'
#' @param object An e2tree object.
#' @param newdata A data frame containing the new observations. If missing,
#'   the fitted values for the training data are returned.
#' @param target Character string specifying the target class for computing
#'   classification scores. Only used for classification trees. Default is
#'   \code{NULL}, which uses the first level.
#' @param ... Additional arguments (ignored).
#'
#' @return For regression: a data frame with columns \code{fit} (predicted
#'   value) and \code{sd} (standard deviation of the response within the
#'   terminal node, computed from the training data).
#'   For classification: a data frame with columns \code{fit} (predicted class),
#'   \code{accuracy} (probability of the predicted class), and \code{score}
#'   (probability of the target class).
#'
#' @examples
#' \donttest{
#' data(iris)
#' smp_size <- floor(0.75 * nrow(iris))
#' train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
#' training <- iris[train_ind, ]
#' validation <- iris[-train_ind, ]
#'
#' ensemble <- randomForest::randomForest(Species ~ ., data = training,
#'   importance = TRUE, proximity = TRUE)
#' D <- createDisMatrix(ensemble, data = training, label = "Species",
#'   parallel = list(active = FALSE, no_cores = 1))
#' setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
#' tree <- e2tree(Species ~ ., training, D, ensemble, setting)
#'
#' ## Predict on new data
#' pred <- predict(tree, newdata = validation)
#' }
#'
#' @method predict e2tree
#' @export
predict.e2tree <- function(object, newdata, target = NULL, ...) {
  is_class <- !is.null(attr(object, "ylevels"))

  if (missing(newdata) || is.null(newdata)) {
    return(object$fitted.values)
  }

  if (is.null(target)) target <- "1"

  result <- ePredTree(fit = object, data = newdata, target = target)

  if (!is_class) {
    # For regression: compute node-level sd from training data
    tree_df <- object$tree
    terminal <- tree_df[tree_df$terminal, , drop = FALSE]
    y <- object$y

    # Compute sd for each terminal node, keyed by node number
    node_sd <- vapply(seq_len(nrow(terminal)), function(i) {
      obs_idx <- unlist(terminal$obs[[i]])
      if (length(obs_idx) < 2) return(0)
      sd(y[obs_idx])
    }, numeric(1))
    names(node_sd) <- as.character(terminal$node)

    # Build lookup: prediction value -> node number (handles duplicate preds)
    pred_to_node <- setNames(as.character(terminal$node), as.character(terminal$pred))

    # Map each prediction to its node's sd via pred value
    # Use match on prediction values to find the correct terminal node
    fit_vals <- as.numeric(result$fit)
    sd_vec <- vapply(fit_vals, function(v) {
      idx <- which(abs(as.numeric(terminal$pred) - v) < 1e-10)
      if (length(idx) == 0) return(NA_real_)
      node_sd[idx[1]]
    }, numeric(1))

    return(data.frame(
      fit = fit_vals,
      sd  = as.numeric(sd_vec),
      row.names = NULL
    ))
  }

  result
}


#' Extract Fitted Values from an E2Tree Model
#'
#' Returns the fitted values (predictions) for the training data used to
#' build the E2Tree model.
#'
#' @param object An e2tree object.
#' @param ... Additional arguments (ignored).
#'
#' @return A vector of fitted values.
#'
#' @examples
#' \donttest{
#' data(iris)
#' smp_size <- floor(0.75 * nrow(iris))
#' train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
#' training <- iris[train_ind, ]
#'
#' ensemble <- randomForest::randomForest(Species ~ ., data = training,
#'   importance = TRUE, proximity = TRUE)
#' D <- createDisMatrix(ensemble, data = training, label = "Species",
#'   parallel = list(active = FALSE, no_cores = 1))
#' setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
#' tree <- e2tree(Species ~ ., training, D, ensemble, setting)
#'
#' fitted(tree)
#' }
#'
#' @method fitted e2tree
#' @export
fitted.e2tree <- function(object, ...) {
  object$fitted.values
}


#' Extract Residuals from an E2Tree Model
#'
#' Returns the residuals (observed minus fitted) for regression E2Tree models.
#' Not available for classification models.
#'
#' @param object An e2tree object.
#' @param ... Additional arguments (ignored).
#'
#' @return A numeric vector of residuals.
#'
#' @examples
#' \donttest{
#' data("mtcars")
#' smp_size <- floor(0.75 * nrow(mtcars))
#' train_ind <- sample(seq_len(nrow(mtcars)), size = smp_size)
#' training <- mtcars[train_ind, ]
#'
#' ensemble <- randomForest::randomForest(mpg ~ ., data = training, ntree = 500,
#'   importance = TRUE, proximity = TRUE)
#' D <- createDisMatrix(ensemble, data = training, label = "mpg",
#'   parallel = list(active = FALSE, no_cores = 1))
#' setting <- list(impTotal = 0.1, maxDec = 1e-6, n = 2, level = 5)
#' tree <- e2tree(mpg ~ ., training, D, ensemble, setting)
#'
#' residuals(tree)
#' }
#'
#' @method residuals e2tree
#' @export
residuals.e2tree <- function(object, ...) {
  if (!is.null(attr(object, "ylevels"))) {
    stop("Residuals are not available for classification E2Tree models.")
  }
  as.numeric(object$y) - as.numeric(object$fitted.values)
}

# ===========================================================================
# Coercion methods for the "e2tree" class
# ===========================================================================

#' Convert an E2Tree Object to rpart Format
#'
#' Coerces an \code{e2tree} object into an \code{rpart} object, which can
#' then be used with standard rpart methods for printing, plotting
#' (e.g., via \code{rpart.plot}), and prediction.
#'
#' @param x An e2tree object.
#' @param ensemble The ensemble model used to build the E2Tree. Supported classes:
#'   \code{randomForest}, \code{ranger}, \code{xgb.Booster}, \code{lgb.Booster},
#'   \code{gbm}, \code{catboost.CatBoost}.
#' @param ... Additional arguments (ignored).
#'
#' @return An \code{rpart} object.
#'
#' @seealso \code{\link{as.party.e2tree}} for conversion to partykit format.
#'
#' @examples
#' \donttest{
#' data(iris)
#' smp_size <- floor(0.75 * nrow(iris))
#' train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
#' training <- iris[train_ind, ]
#'
#' ensemble <- randomForest::randomForest(Species ~ ., data = training,
#'   importance = TRUE, proximity = TRUE)
#' D <- createDisMatrix(ensemble, data = training, label = "Species",
#'   parallel = list(active = FALSE, no_cores = 1))
#' setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
#' tree <- e2tree(Species ~ ., training, D, ensemble, setting)
#'
#' rpart_obj <- as.rpart(tree, ensemble)
#' }
#'
#' @export
as.rpart <- function(x, ...) {
  UseMethod("as.rpart")
}

#' @rdname as.rpart
#' @method as.rpart e2tree
#' @export
as.rpart.e2tree <- function(x, ensemble, ...) {
  expected <- attr(x, "ensemble_backend")
  actual   <- ensemble_backend(ensemble)
  if (!is.null(expected) && !is.na(expected) && !is.na(actual) &&
      expected != actual) {
    warning(sprintf(
      paste0("as.rpart(): the e2tree was built from a '%s' ensemble but ",
             "you passed a '%s'. The split labels will still print, but ",
             "any backend-specific information will be incorrect."),
      expected, actual
    ), call. = FALSE)
  }
  rpart2Tree(fit = x, ensemble = ensemble)
}


#' Convert an E2Tree Object to partykit Format
#'
#' Coerces an \code{e2tree} object into a \code{party} object from the
#' \pkg{partykit} package. This enables the use of partykit's infrastructure
#' for printing, plotting, and predicting with the E2Tree model.
#'
#' @param x An e2tree object.
#' @param ... Additional arguments (ignored).
#'
#' @return A \code{party} object (from \pkg{partykit}).
#'
#' @seealso \code{\link{as.rpart.e2tree}} for conversion to rpart format.
#'
#' @examples
#' \donttest{
#' data(iris)
#' smp_size <- floor(0.75 * nrow(iris))
#' train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
#' training <- iris[train_ind, ]
#'
#' ensemble <- randomForest::randomForest(Species ~ ., data = training,
#'   importance = TRUE, proximity = TRUE)
#' D <- createDisMatrix(ensemble, data = training, label = "Species",
#'   parallel = list(active = FALSE, no_cores = 1))
#' setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
#' tree <- e2tree(Species ~ ., training, D, ensemble, setting)
#'
#' if (requireNamespace("partykit", quietly = TRUE)) {
#'   party_obj <- partykit::as.party(tree)
#'   plot(party_obj)
#' }
#' }
#'
#' @keywords internal
as.party.e2tree <- function(x, ...) {
  check_package("partykit")

  tree_df <- x$tree
  data <- x$data
  is_class <- !is.null(attr(x, "ylevels"))

  # Build partynode recursively from the e2tree structure
  node_map <- base::split(tree_df, seq_len(nrow(tree_df)))
  names(node_map) <- tree_df$node

  # Counter for sequential partykit node IDs
  id_counter <- 0L

  build_partynode <- function(node_id) {
    nd <- node_map[[as.character(node_id)]]
    id_counter <<- id_counter + 1L
    current_id <- id_counter

    # Terminal node
    if (nd$terminal) {
      return(partykit::partynode(id = current_id))
    }

    # Internal node: determine split
    split_label <- nd$splitLabel
    var_name <- nd$variable
    var_idx <- which(names(data) == var_name)

    if (length(var_idx) == 0) {
      return(partykit::partynode(id = current_id))
    }

    # Create the split
    if (grepl("<=", split_label)) {
      threshold <- as.numeric(sub(".*<=\\s*", "", split_label))
      sp <- partykit::partysplit(
        varid = as.integer(var_idx),
        breaks = threshold,
        right = FALSE
      )
    } else if (grepl("%in%", split_label)) {
      var_levels <- levels(factor(data[[var_name]]))
      cats_str <- sub(".*%in%\\s*", "", split_label)
      left_cats <- eval(parse(text = cats_str))
      index_vec <- ifelse(var_levels %in% left_cats, 1L, 2L)
      sp <- partykit::partysplit(
        varid = as.integer(var_idx),
        index = index_vec
      )
    } else {
      return(partykit::partynode(id = current_id))
    }

    # Recurse on children (left then right)
    left_id <- node_id * 2L
    right_id <- node_id * 2L + 1L

    kids <- list(
      build_partynode(left_id),
      build_partynode(right_id)
    )

    partykit::partynode(id = current_id, split = sp, kids = kids)
  }

  # Build the tree starting from root (node 1)
  pnode <- build_partynode(1L)

  # Build the fitted data frame with terminal node IDs
  # We need a mapping from e2tree node numbers to partykit sequential IDs
  # Traverse the tree again to build the mapping
  id_counter2 <- 0L
  node_id_map <- integer(0)

  map_node_ids <- function(node_id) {
    id_counter2 <<- id_counter2 + 1L
    nd <- node_map[[as.character(node_id)]]
    node_id_map[as.character(node_id)] <<- id_counter2
    if (!nd$terminal) {
      map_node_ids(node_id * 2L)
      map_node_ids(node_id * 2L + 1L)
    }
  }
  map_node_ids(1L)

  # Assign each observation to its terminal node's partykit ID
  terminal_nodes <- tree_df[tree_df$terminal, ]
  n_obs <- nrow(data)
  node_ids <- rep(NA_integer_, n_obs)
  for (i in seq_len(nrow(terminal_nodes))) {
    obs_idx <- unlist(terminal_nodes$obs[[i]])
    pk_id <- node_id_map[as.character(terminal_nodes$node[i])]
    node_ids[obs_idx] <- pk_id
  }

  if (is_class) {
    fitted_df <- data.frame(
      `(fitted)` = node_ids,
      `(response)` = factor(x$y, levels = attr(x, "ylevels")),
      check.names = FALSE
    )
  } else {
    fitted_df <- data.frame(
      `(fitted)` = node_ids,
      `(response)` = as.numeric(x$y),
      check.names = FALSE
    )
  }

  party_obj <- partykit::party(
    node = pnode,
    data = data,
    fitted = fitted_df,
    terms = x$terms
  )

  # Set constparty class for proper terminal node rendering
  class(party_obj) <- c("constparty", class(party_obj))

  party_obj
}

idImpurity2 <- function(y,index,S){
  y <- y[index,index]
  lab <- row.names(y)
  S <- S[index,]
  #Matrix::diag(dis) <- -1

  y[(lower.tri(y, diag=TRUE))] <- -1
  D <- y %>% as.matrix() %>%
    as.data.frame() %>%
    mutate(obs_row=colnames(.)) %>%
    pivot_longer(cols = 1:(ncol(.)-1), names_to = "obs_col") %>%
    dplyr::filter(.data$value >-1)

  ind <- seq_len(ncol(S))

  n <- length(index)
  tab <- colSums(S)
  ind <- ind[(tab>1 & tab < (n-1))]
  imp=rep(Inf,ncol(S))

  for (i in ind){
    #print(i)
    s2 <- S[,i]
    s <- rep(-1,nrow(D))
    s[D$obs_row %in% lab[s2==1] & D$obs_col %in% lab[s2==1]] <- 1
    s[D$obs_row %in% lab[s2==0] & D$obs_col %in% lab[s2==0]] <- 0
    #n1 <- sum(s2)
    #n0 <- n-n1
    #items <- c(n0*(n0-1),n1*(n1-1))/2
    # Tasso ben classificati in tutti i nodi terminali dell'i-esimo albero
    R <- cbind(D,s=s) %>%
      dplyr::filter(.data$s>-1) %>%
      group_by(s) %>%
      summarize(m=mean(.data$value))
    imp[i] <- sum(R$m*c(n-sum(s2),sum(s2)))/n

  }

  names(imp)=colnames(S)
  return(imp)
}

utils::globalVariables(c("resp", "W", "data_XGB", "i"))

#' Create a Dissimilarity Matrix from an Ensemble Model
#'
#' The function createDisMatrix creates a dissimilarity matrix among
#' observations from an ensemble tree. This optimized version is designed
#' for large datasets (50K-500K observations) with improved memory management
#' and chunking capabilities.
#'
#' @param ensemble is an ensemble tree object
#' @param data is a data frame containing the variables in the model. It is the
#'   data frame used for ensemble learning.
#' @param label is a character. It indicates the response label.
#' @param parallel A list with two elements: \code{active} (logical) and
#'   \code{no_cores} (integer). If \code{active = TRUE}, the function performs
#'   parallel computation using the number of cores specified in
#'   \code{no_cores}. If \code{no_cores} is NULL or equal to 0, it defaults to
#'   using all available cores minus one. If \code{active = FALSE}, the function
#'   runs on a single core. Default: \code{list(active = FALSE, no_cores = 1)}.
#' @param verbose Logical. If TRUE, the function prints progress messages and
#'   other information during execution. If FALSE (the default), messages are
#'   suppressed.
#' @param chunk_size Integer. Number of rows to process in each chunk. If NULL,
#'   automatically determined based on available memory and dataset size.
#'   Default: NULL (auto).
#' @param memory_limit Numeric. Maximum memory to use in GB. Default: NULL (no limit).
#' @param use_disk Logical. If TRUE and dataset is very large, intermediate results
#'   are saved to disk. Default: FALSE.
#' @param temp_dir Character. Directory for temporary files if use_disk = TRUE.
#'   Default: tempdir().
#' @param batch_aggregate Integer. Number of tree results to aggregate at once
#'   before adding to main matrix (reduces memory peaks). Default: 10.
#'
#' @return A dissimilarity matrix. This is a dissimilarity matrix measuring the
#'   discordance between two observations concerning a given random forest
#'   model.
#'
#' @details This optimized version implements several strategies for handling
#'   large datasets:
#'
#' \itemize{
#'   \item **Memory-efficient aggregation**: Results from parallel trees are
#'     aggregated in batches to avoid memory peaks
#'   \item **Chunking**: For very large matrices, computation can be split into
#'     manageable chunks
#'   \item **Sparse matrix optimization**: Maintains sparsity throughout computation
#'   \item **Automatic garbage collection**: Explicit memory cleanup at critical points
#'   \item **Disk-based computation**: Optional saving of intermediate results for
#'     datasets exceeding memory capacity
#' }
#'
#' Supported ensemble types for *classification* or *regression* tasks:
#' \itemize{
#'   \item `randomForest`
#'   \item `ranger`
#'   \item `xgb.Booster` (xgboost)
#'   \item `lgb.Booster` (lightgbm)
#'   \item `gbm` (gbm)
#'   \item `catboost.CatBoost` (catboost)
#' }
#'
#' @section Interpretation note (RF vs boosting):
#' For *bagging* ensembles (\code{randomForest}, \code{ranger}) the trees are
#' grown independently on bootstrap samples; co-occurrence in the same leaf
#' captures local similarity in the predictor space. For *boosting* ensembles
#' (\code{xgb.Booster}, \code{lgb.Booster}, \code{gbm}, \code{catboost})
#' each tree is fit to the residual of the previous ones, so leaf
#' co-occurrence reflects similarity in the *error-correction trajectory*
#' rather than in the final prediction space. The resulting dissimilarity
#' matrices therefore have systematically different scales (typically
#' \eqn{\bar D \in [0.85, 0.95]} for bagging vs. \eqn{[0.35, 0.70]} for
#' boosting). The surrogate tree built on top of \code{D} should be
#' interpreted accordingly.
#'
#' The returned matrix carries an \code{ensemble_backend} attribute identifying
#' the backend used, which downstream functions check to detect mismatched
#' \code{(D, ensemble)} pairs.
#'
#' @examples
#' \donttest{
## Classification
#' data("iris")
#'
#' # Create training and validation set:
#' smp_size <- floor(0.75 * nrow(iris))
#' train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
#' training <- iris[train_ind, ]
#' validation <- iris[-train_ind, ]
#' response_training <- training[,5]
#' response_validation <- validation[,5]
#'
#' # Perform training:
#' ## "randomForest" package
#' ensemble <- randomForest::randomForest(Species ~ ., data=training,
#' importance=TRUE, proximity=TRUE)
#'
#' ## "ranger" package
#' if (requireNamespace("ranger", quietly = TRUE)) {
#'   ensemble <- ranger::ranger(Species ~ ., data = iris,
#'     num.trees = 1000, importance = 'impurity')
#' }
#'
#' # Compute dissimilarity matrix with optimizations
#' D <- createDisMatrix(
#'   ensemble,
#'   data = training,
#'   label = "Species",
#'   parallel = list(active = FALSE, no_cores = 1),
#'   chunk_size = 10000,  # Process 10K rows at a time
#'   batch_aggregate = 20, # Aggregate 20 trees at once
#'   verbose = TRUE
#' )
#' }
#'
#' @export

createDisMatrix <- function(
  ensemble,
  data,
  label,
  parallel = list(active = FALSE, no_cores = 1),
  verbose = FALSE,
  chunk_size = NULL,
  memory_limit = NULL,
  use_disk = FALSE,
  temp_dir = tempdir(),
  batch_aggregate = 10
) {
  # === Input Validation ===
  if (is.null(ensemble)) {
    stop(
      paste0("Error: 'ensemble' cannot be NULL. ",
           "Please provide a trained randomForest, ranger, xgb.Booster, ",
           "lgb.Booster, gbm, or catboost.CatBoost model.")
    )
  }

  # Validate ensemble via adapter dispatch (throws informative error for unsupported classes)
  tryCatch(
    get_ensemble_type(ensemble),
    error = function(e) stop(conditionMessage(e), call. = FALSE)
  )

  if (!is.data.frame(data)) {
    stop("Error: 'data' must be a valid data frame.")
  }

  if (!is.null(label)) {
    if (!is.character(label) || length(label) != 1 || !(label %in% colnames(data))) {
      stop("Error: 'label' must be a valid column name in 'data'.")
    }
  }

  row.names(data) <- NULL
  n_obs <- nrow(data)

  # === Memory Management Setup ===
  if (verbose) {
    cat(sprintf("\n=== OPTIMIZED DISSIMILARITY MATRIX COMPUTATION ===\n"))
    cat(sprintf("Dataset size: %d observations\n", n_obs))
    cat(sprintf(
      "Matrix size: %d x %d = %.2f M elements\n",
      n_obs,
      n_obs,
      (n_obs * n_obs) / 1e6
    ))

    est_memory_gb <- (n_obs * n_obs * 8) / (1024^3)
    cat(sprintf("Estimated dense matrix memory: %.2f GB\n", est_memory_gb))

    if (.Platform$OS.type == "unix") {
      tryCatch(
        {
          mem_info <- system(
            "free -g 2>/dev/null || vm_stat 2>/dev/null | head -1",
            intern = TRUE
          )
          if (length(mem_info) > 0) {
            cat("System memory info available\n")
          }
        },
        error = function(e) {},
        warning = function(w) {}
      )
    }
  }

  if (is.null(chunk_size)) {
    chunk_size <- determine_chunk_size(n_obs, memory_limit, verbose)
  }

  use_chunking <- (chunk_size < n_obs) && (chunk_size > 0)

  if (verbose && use_chunking) {
    cat(sprintf("Using chunked computation with chunk size: %d\n", chunk_size))
  }

  # === Determine Ensemble Type ===
  type <- get_ensemble_type(ensemble)
  if (!(type %in% c("classification", "regression"))) {
    stop("'ensemble' must be a classification or regression model.", call. = FALSE)
  }

  # `label` is mandatory for regression (used to compute the dissimilarity
  # scale).  Validate up front so the error surfaces before any leaf
  # extraction work is wasted.
  if (type == "regression" && is.null(label)) {
    stop("Error: 'label' is required for regression models (needed to compute dissimilarity scale).",
         call. = FALSE)
  }

  # === Extract Terminal Nodes ===
  if (verbose) {
    cat("\nExtracting terminal nodes...\n")
  }

  backend <- ensemble_backend(ensemble)
  # CatBoost adapters strip the response column from the predictor pool via
  # the `e2tree_label` attribute; set it transparently when the user passes
  # `label` so the workflow matches the other backends.
  if (!is.null(label) &&
      backend %in% c("catboost.CatBoost", "catboost.Model") &&
      is.null(attr(ensemble, "e2tree_label"))) {
    attr(ensemble, "e2tree_label") <- label
  }
  obs <- extract_terminal_nodes(ensemble, data)
  validate_terminal_nodes(obs, data, backend = backend)

  class(data) <- "data.frame"
  # `label` is mandatory for regression (used to compute the dissimilarity
  # scale) but optional for classification (where it only annotates `resp`
  # for downstream code). Guard against NULL label before any data[[label]]
  # access; classification with NULL label proceeds with a dummy resp.
  if (!is.null(label) && !inherits(data[[label]], "factor")) {
    data[[label]] <- factor(data[[label]])
  }

  obs <- cbind(row.names(obs), obs)
  names(obs) <- c("OBS", paste("Tree", seq(1, (ncol(obs) - 1L)), sep = ""))
  row.names(obs) <- NULL

  if (type == "regression") {
    if (is.null(label)) {
      stop("Error: 'label' is required for regression models (needed to compute dissimilarity scale).")
    }
    obs$resp <- as.numeric(as.character(data[obs$OBS, label]))
  } else {
    # classification: resp is used only as metadata; use dummy when label is NULL
    obs$resp <- if (!is.null(label)) data[as.numeric(obs$OBS), label] else
      factor(rep(1L, nrow(obs)))
  }

  ntree <- ncol(obs) - 2L

  # Ensure tree columns are numeric (ranger may return double, randomForest integer)
  # C++ reads them as NumericVector and converts to int internally
  tree_col_idx <- 2:(ntree + 1)  # R 1-based columns Tree1..TreeN
  for (ci in tree_col_idx) {
    if (!is.numeric(obs[[ci]])) {
      obs[[ci]] <- as.numeric(obs[[ci]])
    }
  }
  # Ensure resp is plain numeric
  obs$resp <- as.numeric(obs$resp)

  if (verbose) {
    cat(sprintf("Number of trees: %d\n", ntree))
    cat(sprintf("Batch aggregation size: %d trees\n", batch_aggregate))
  }

  # === Determine Parallelism ===
  if (isTRUE(parallel$active)) {
    if (is.null(parallel$no_cores) || parallel$no_cores < 1L) {
      no_cores <- max(1L, parallel::detectCores() - 1L)
    } else {
      no_cores <- parallel$no_cores
    }
  } else {
    no_cores <- 1L
  }

  if (verbose) {
    cat(sprintf("\nParallel mode: %d cores (C++ OpenMP)\n", no_cores))
  }

  # === Main Computation — single C++ call for all trees ===
  if (verbose) {
    cat("\n=== COMPUTING CO-OCCURRENCE MATRIX (C++) ===\n")
  }

  maxvar <- if (type == "regression") diff(range(obs$resp))^2 / 9L else NA_real_

  a <- compute_all_cooccurrences_cpp(type, obs, as.integer(ntree), as.integer(no_cores), as.double(maxvar))

  if (verbose) {
    cat("Co-occurrence matrix computed.\n")
    cat("\n=== COMPUTING FINAL DISSIMILARITY MATRIX (C++) ===\n")
  }

  # === Dissimilarity: 1 - a[i,j] / max(a[i,i], a[j,j]) ===
  if (use_chunking) {
    if (verbose) {
      cat(sprintf("Using chunked R path (chunk size: %d)\n", chunk_size))
    }
    dis <- compute_dissimilarity_chunked(a, obs, chunk_size, verbose = verbose)
  } else {
    dis <- compute_dissimilarity_from_cooc_cpp(a)
    row.names(dis) <- colnames(dis) <- obs$OBS
  }

  rm(a)

  if (isTRUE(use_disk)) {
    if (!dir.exists(temp_dir)) dir.create(temp_dir, recursive = TRUE)
    out_path <- file.path(temp_dir,
                          sprintf("e2tree_dismatrix_%d.rds", as.integer(Sys.time())))
    saveRDS(dis, file = out_path)
    if (verbose) cat(sprintf("Dissimilarity matrix written to %s\n", out_path))
    attr(dis, "disk_path") <- out_path
  }

  gc(verbose = FALSE, full = TRUE)

  # Tag with backend identity so downstream functions (as.rpart, e2tree, ...)
  # can detect mismatched (D, ensemble) pairs.
  attr(dis, "ensemble_backend") <- backend

  if (verbose) {
    cat("\n=== COMPUTATION COMPLETED ===\n")
    cat(sprintf(
      "Final matrix sparsity: %.2f%%\n",
      100 * (1 - sum(dis != 0) / length(dis))
    ))
  }

  return(dis)
}

determine_chunk_size <- function(n_obs, memory_limit = NULL, verbose = FALSE) {
  bytes_per_element <- 8
  bytes_per_row <- n_obs * bytes_per_element

  if (is.null(memory_limit)) {
    if (.Platform$OS.type == "unix") {
      mem_info <- try(
        system("free -b 2>/dev/null", intern = TRUE),
        silent = TRUE
      )
      if (!inherits(mem_info, "try-error") && length(mem_info) > 1) {
        mem_parts <- strsplit(mem_info[2], "\\s+")[[1]]
        if (length(mem_parts) >= 7) {
          available_bytes <- as.numeric(mem_parts[7])
          if (!is.na(available_bytes) && available_bytes > 0) {
            memory_limit <- (available_bytes / (1024^3)) * 0.5
          }
        }
      }
    }

    if (is.null(memory_limit)) {
      memory_limit <- 8
    }
  }

  memory_limit_bytes <- memory_limit * (1024^3)
  target_bytes_per_chunk <- memory_limit_bytes * 0.25
  chunk_size <- floor(target_bytes_per_chunk / bytes_per_row)
  chunk_size <- max(1000, min(chunk_size, n_obs))

  if (verbose) {
    cat(sprintf(
      "Determined chunk size: %d (based on %.2f GB memory limit)\n",
      chunk_size,
      memory_limit
    ))
  }

  return(chunk_size)
}

compute_dissimilarity_chunked <- function(a, obs, chunk_size, verbose = FALSE) {
  n_obs <- nrow(obs)
  n_chunks <- ceiling(n_obs / chunk_size)

  if (verbose) {
    cat(sprintf("Computing dissimilarity in %d chunks...\n", n_chunks))
    pb <- txtProgressBar(min = 0, max = n_chunks, style = 3)
  }

  aa <- Matrix::diag(a)
  dis <- matrix(0, nrow = n_obs, ncol = n_obs)

  for (i in seq_len(n_chunks)) {
    start_row <- (i - 1) * chunk_size + 1
    end_row <- min(i * chunk_size, n_obs)
    rows_idx <- start_row:end_row

    ## Keep sparse as long as possible; only convert the chunk rows
    a_chunk <- a[rows_idx, , drop = FALSE]  # stays sparse (dgCMatrix)
    aa_rows <- aa[rows_idx]
    aa_max  <- outer(aa_rows, aa, pmax)     # dense but only chunk_size x n
    ## Convert sparse chunk to dense only for the small chunk
    dis[rows_idx, ] <- 1 - (as.matrix(a_chunk) / aa_max)

    if (verbose) {
      setTxtProgressBar(pb, i)
    }

    rm(a_chunk, aa_max)
    if (i %% 5 == 0) gc(verbose = FALSE)
  }

  if (verbose) {
    close(pb)
  }

  row.names(dis) <- colnames(dis) <- obs$OBS

  return(dis)
}

maxValue <- function(x, y) {
  pmax(x, y)
}

## Variance — delegates to shared utility
variance <- e2_variance

library(testthat)
library(e2tree)  # Ensure the package is loaded
library(randomForest)

test_that("createDisMatrix works correctly for classification task", {
  set.seed(42)
  
  # Prepare data
  data(iris)
  train_idx <- sample(seq_len(nrow(iris)), size = 0.75 * nrow(iris))
  training <- iris[train_idx, ]
  
  # Train Random Forest
  ensemble <- randomForest(Species ~ ., data = training, importance = TRUE, proximity = TRUE)
  
  # Compute dissimilarity matrix
  D <- createDisMatrix(ensemble, data = training, label = "Species", parallel = list(active=FALSE, no_cores = 1))
  
  # Tests
  expect_type(D, "double")  # Should return a numeric matrix
  expect_true(is.matrix(D)) # Must be a matrix
  expect_equal(dim(D), c(nrow(training), nrow(training)))  # Should match input size
  expect_equal(as.numeric(diag(D)), rep(0, nrow(training))) # Diagonal should be 0 (self-similarity)
})

test_that("createDisMatrix works correctly for regression task", {
  set.seed(42)
  
  # Prepare data
  data(mtcars)
  train_idx <- sample(seq_len(nrow(mtcars)), size = 0.75 * nrow(mtcars))
  training <- mtcars[train_idx, ]
  
  # Train Random Forest
  ensemble <- randomForest(mpg ~ ., data = training, ntree = 100, importance = TRUE, proximity = TRUE)
  
  # Compute dissimilarity matrix
  D <- createDisMatrix(ensemble, data = training, label = "mpg", parallel = list(active=FALSE, no_cores = 1))
  
  # Tests
  expect_type(D, "double")
  expect_true(is.matrix(D))
  expect_equal(dim(D), c(nrow(training), nrow(training)))
  expect_equal(as.numeric(diag(D)), rep(0, nrow(training)))
})

test_that("createDisMatrix handles incorrect input types", {
  set.seed(42)
  
  data(iris)
  train_idx <- sample(seq_len(nrow(iris)), size = 0.75 * nrow(iris))
  training <- iris[train_idx, ]
  
  ensemble <- randomForest(Species ~ ., data = training, importance = TRUE, proximity = TRUE)
  
  expect_error(createDisMatrix(NULL, data = training, label = "Species"), 
               "Error: 'ensemble' cannot be NULL")
  expect_error(createDisMatrix(ensemble, data = NULL, label = "Species"), 
               "Error: 'data' must be a valid data frame")
  expect_error(createDisMatrix(ensemble, data = training, label = "InvalidLabel"), 
               "Error: 'label' must be a valid column name in 'data'")
})

test_that("createDisMatrix works with parallelization", {

  set.seed(42)
  
  data(iris)
  train_idx <- sample(seq_len(nrow(iris)), size = 0.75 * nrow(iris))
  training <- iris[train_idx, ]
  
  ensemble <- randomForest(Species ~ ., data = training, importance = TRUE, proximity = TRUE)

  # Compute dissimilarity matrix in parallel
  D <- createDisMatrix(ensemble, data = training, label = "Species", parallel = list(active=TRUE, no_cores = 1))
  
  expect_true(is.matrix(D))
  expect_equal(dim(D), c(nrow(training), nrow(training)))
})

################################################################################
# Equivalent tests for ranger
################################################################################

library(testthat)

test_that("createDisMatrix works correctly for classification task (ranger)", {
  skip_if_not_installed("ranger")
  set.seed(42)

  # Prepare data
  data(iris)
  train_idx <- sample(seq_len(nrow(iris)), size = 0.75 * nrow(iris))
  training  <- iris[train_idx, ]

  # Train Ranger for classification
  # write.forest = TRUE is crucial for predict(type = "terminalNodes")
  ensemble <- ranger::ranger(Species ~ .,
                             data = training,
                             num.trees = 100,
                             write.forest = TRUE
                             )

  # Compute dissimilarity matrix 
  D <- createDisMatrix(ensemble,
                       data = training,
                       label = "Species",
                       parallel = list(active = FALSE, no_cores = 1)
                       )

  # Tests
  expect_type(D, "double") # Should return a numeric matrix
  expect_true(is.matrix(D)) # Must be a matrix
  expect_equal(dim(D), c(nrow(training), nrow(training))) 
  expect_equal(as.numeric(diag(D)), rep(0, nrow(training)))
})

test_that("createDisMatrix works correctly for regression task (ranger)", {
  skip_if_not_installed("ranger")
  set.seed(42)

  # Prepare data
  data(mtcars)
  train_idx <- sample(seq_len(nrow(mtcars)), size = 0.75 * nrow(mtcars))
  training <- mtcars[train_idx, ]

  # Train Ranger for regression
  ensemble <- ranger::ranger(mpg ~ .,
                             data = training, 
                             num.trees = 100, 
                             write.forest = TRUE
                             )

  # Compute dissimilarity matrix
  D <- createDisMatrix(ensemble, 
                       data = training, 
                       label = "mpg", 
                       parallel = list(active = FALSE, no_cores = 1)
                       )

  # Tests
  expect_type(D, "double")
  expect_true(is.matrix(D))
  expect_equal(dim(D), c(nrow(training), nrow(training)))
  expect_equal(as.numeric(diag(D)), rep(0, nrow(training)))
})

test_that("createDisMatrix handles incorrect input types (ranger)", {
  skip_if_not_installed("ranger")
  set.seed(42)

  data(iris)
  train_idx <- sample(seq_len(nrow(iris)), size = 0.75 * nrow(iris))
  training <- iris[train_idx, ]

  # Train Ranger
  ensemble <- ranger::ranger(Species ~ .,
                             data = training, 
                             num.trees = 100, 
                             write.forest = TRUE
                             )

  expect_error(
    createDisMatrix(NULL, data = training, label = "Species"),
    "Error: 'ensemble' cannot be NULL"
  ) 
  expect_error(
    createDisMatrix(ensemble, data = NULL, label = "Species"),
    "Error: 'data' must be a valid data frame"
  )
  expect_error(
    createDisMatrix(ensemble, data = training, label = "InvalidLabel"),
    "Error: 'label' must be a valid column name in 'data'"
  )
})

test_that("createDisMatrix works with parallelization (ranger)", {
  skip_if_not_installed("ranger")
  set.seed(42)

  data(iris)
  train_idx <- sample(seq_len(nrow(iris)), size = 0.75 * nrow(iris))
  training  <- iris[train_idx, ]

  # Train Ranger
  ensemble <- ranger::ranger(Species ~ .,
                             data = training,
                             num.trees = 100,
                             write.forest = TRUE
                             )

  # Compute dissimilarity matrix in parallel 
  # (assuming createDisMatrix handles ranger and parallel argument)
  # Note: ranger itself has built-in parallelization with 'num.threads'.
  # predict sets num.threads to 1
  # createDisMatrix is designed to to have a layer of parallelization, 
  # For this test, we assume createDisMatrix's parallel argument is handled.
  D <- createDisMatrix(ensemble,
                       data = training,
                       label = "Species",
                       parallel = list(active = TRUE, no_cores = 1)
                       )

  expect_true(is.matrix(D))
  expect_equal(dim(D), c(nrow(training), nrow(training)))
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

compute_impurity_cpp <- function(y, S) {
    .Call('_e2tree_compute_impurity_cpp', PACKAGE = 'e2tree', y, S)
}

compute_cooccurrences_sparse_cpp <- function(type, obs, tree_index, n_total, maxvar = NA_real_) {
    .Call('_e2tree_compute_cooccurrences_sparse_cpp', PACKAGE = 'e2tree', type, obs, tree_index, n_total, maxvar)
}

compute_cooccurrences_cpp <- function(type, obs, w, tree_index, maxvar = NA_real_) {
    .Call('_e2tree_compute_cooccurrences_cpp', PACKAGE = 'e2tree', type, obs, w, tree_index, maxvar)
}

compute_all_cooccurrences_cpp <- function(type, obs, ntree, n_cores, maxvar = NA_real_) {
    .Call('_e2tree_compute_all_cooccurrences_cpp', PACKAGE = 'e2tree', type, obs, ntree, n_cores, maxvar)
}

compute_dissimilarity_from_cooc_cpp <- function(a) {
    .Call('_e2tree_compute_dissimilarity_from_cooc_cpp', PACKAGE = 'e2tree', a)
}

#' Interactive E2Tree Plot with visNetwork
#'
#' @description
#' Displays an E2Tree as an interactive network plot using visNetwork.
#' Features: drag nodes anywhere, zoom, pan, click for details.
#' Starts with hierarchical layout, then you can freely move nodes.
#'
#' @param fit An e2tree object
#' @param data The training data used to build the tree
#' @param ensemble The ensemble model (randomForest or ranger)
#' @param width Width of the widget (default: "100%")
#' @param height Height of the widget (default: "100%")
#' @param direction Layout direction: "UD" (top-down), "DU" (bottom-up), 
#'                  "LR" (left-right), "RL" (right-left)
#' @param node_spacing Spacing between nodes at same level (default: 200)
#' @param level_separation Spacing between levels (default: 200)
#' @param colors Named vector of colors for classes, or NULL for auto
#' @param show_percent Show percentage in nodes (default: TRUE)
#' @param show_prob Show class probabilities in nodes (default: TRUE)
#' @param show_n Show observation count in nodes (default: TRUE)
#' @param font_size Font size for node labels (default: 14)
#' @param edge_font_size Font size for edge labels (default: 12)
#' @param split_label_style How to display split information:
#'   \itemize{
#'     \item "rpart" - Variable name in node, threshold on edges (like rpart.plot)
#'     \item "full" - Full split rule on edges (variable + condition)
#'     \item "threshold" - Only threshold values on edges (< 47, >= 47)
#'     \item "yesno" - Simple yes/no on edges
#'     \item "none" - No labels on edges (hover for details)
#'     \item "innode" - Full split rule displayed IN the node (above stats)
#'   }
#' @param max_label_length Maximum characters for edge labels before truncating (default: 50)
#' @param details_on When to show node details:
#'   \itemize{
#'     \item "hover" - Show on mouse hover (default, but may cover other nodes)
#'     \item "click" - Show only on click (avoids covering highlighted nodes)
#'     \item "none" - No tooltips (use for cleaner visualization)
#'   }
#' @param navigation_buttons Show navigation buttons (default: FALSE)
#' @param free_drag If TRUE, nodes can be dragged in ALL directions (horizontal, 
#'                  vertical, diagonal). If FALSE (default), nodes can only be 
#'                  moved horizontally within their level.
#'
#' @return A visNetwork htmlwidget object
#'
#' @examples
#' \donttest{
#' data(iris)
#' set.seed(42)
#' smp_size <- floor(0.75 * nrow(iris))
#' train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
#' training <- iris[train_ind, ]
#'
#' ensemble <- randomForest::randomForest(Species ~ ., data = training,
#'   importance = TRUE, proximity = TRUE)
#' D <- createDisMatrix(ensemble, data = training, label = "Species",
#'   parallel = list(active = FALSE, no_cores = 1))
#' setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
#' tree <- e2tree(Species ~ ., training, D, ensemble, setting)
#'
#' # Basic usage
#' plot_e2tree_vis(tree, training, ensemble)
#' }
#'
#' @export
plot_e2tree_vis <- function(fit, data, ensemble,
                            width = "100%",
                            height = "100%",
                            direction = "UD",
                            node_spacing = 200,
                            level_separation = 200,
                            colors = NULL,
                            show_percent = TRUE,
                            show_prob = TRUE,
                            show_n = TRUE,
                            font_size = 14,
                            edge_font_size = 12,
                            split_label_style = "rpart",
                            max_label_length = 50,
                            details_on = "hover",
                            navigation_buttons = FALSE,
                            free_drag = FALSE) {
  
  # Check required packages
  if (!requireNamespace("visNetwork", quietly = TRUE)) {
    stop("Package 'visNetwork' is required. Install with: install.packages('visNetwork')")
  }
  
  if (!requireNamespace("jsonlite", quietly = TRUE)) {
    stop("Package 'jsonlite' is required. Install with: install.packages('jsonlite')")
  }
  
  if (!inherits(fit, "e2tree")) {
    stop("'fit' must be an e2tree object")
  }
  
  # Validate split_label_style
  valid_styles <- c("rpart", "full", "threshold", "yesno", "none", "innode")
  if (!split_label_style %in% valid_styles) {
    stop("split_label_style must be one of: ", paste(valid_styles, collapse = ", "))
  }
  
  # Validate details_on
  valid_details <- c("hover", "click", "none")
  if (!details_on %in% valid_details) {
    stop("details_on must be one of: ", paste(valid_details, collapse = ", "))
  }
  
  # Get tree info
  tree_info <- fit$tree
  
  # Determine task type
  is_classification <- !is.null(attr(fit, "ylevels"))
  response_var <- all.vars(fit$terms)[1]
  
  # Get class levels
  if (is_classification) {
    class_levels <- attr(fit, "ylevels")
    n_classes <- length(class_levels)
    
    # Default colors (rpart.plot style)
    if (is.null(colors)) {
      if (n_classes == 2) {
        colors <- c("#5B9F5B", "#6BAED6")
        names(colors) <- class_levels
      } else if (n_classes <= 8) {
        palette <- c("#5B9F5B", "#6BAED6", "#FDB863", "#B2ABD2", 
                     "#E66101", "#5E3C99", "#1B9E77", "#D95F02")
        colors <- palette[1:n_classes]
        names(colors) <- class_levels
      } else {
        colors <- rainbow(n_classes)
        names(colors) <- class_levels
      }
    }
  } else {
    colors <- NULL
  }
  
  # =========================================================================
  # Helper functions
  # =========================================================================
  
  calc_depth <- function(node_id) {
    depth <- 0
    n <- node_id
    while (n > 1) {
      n <- floor(n / 2)
      depth <- depth + 1
    }
    depth
  }
  
  # Get variable name from split
  get_split_variable <- function(split_label) {
    if (is.na(split_label) || split_label == "") return("")
    
    if (grepl("<=", split_label)) {
      parts <- strsplit(split_label, "<=")[[1]]
      return(trimws(parts[1]))
    } else if (grepl("%in%", split_label)) {
      parts <- strsplit(split_label, "%in%")[[1]]
      return(trimws(parts[1]))
    }
    return("")
  }
  
  # Get threshold/values from split
  get_split_condition <- function(split_label, is_left) {
    if (is.na(split_label) || split_label == "") return("")
    
    if (grepl("<=", split_label)) {
      parts <- strsplit(split_label, "<=")[[1]]
      threshold <- trimws(parts[2])
      if (is_left) {
        return(paste0("< ", threshold))
      } else {
        return(paste0("\u2265 ", threshold))
      }
    } else if (grepl("%in%", split_label)) {
      parts <- strsplit(split_label, "%in%")[[1]]
      values <- trimws(parts[2])
      values <- gsub("c\\(|\\)", "", values)
      values <- gsub("\"", "", values)
      if (is_left) {
        return(paste0("= ", values))
      } else {
        return(paste0("\u2260 ", values))
      }
    }
    return("")
  }
  
  # Get full condition for tooltip
  get_full_condition <- function(split_label, is_left) {
    if (is.na(split_label) || split_label == "") return("")
    
    var_name <- get_split_variable(split_label)
    
    if (grepl("<=", split_label)) {
      parts <- strsplit(split_label, "<=")[[1]]
      threshold <- trimws(parts[2])
      if (is_left) {
        return(paste0(var_name, " < ", threshold))
      } else {
        return(paste0(var_name, " \u2265 ", threshold))
      }
    } else if (grepl("%in%", split_label)) {
      parts <- strsplit(split_label, "%in%")[[1]]
      values <- trimws(parts[2])
      values <- gsub("c\\(", "{", values)
      values <- gsub("\\)", "}", values)
      values <- gsub("\"", "", values)
      if (is_left) {
        return(paste0(var_name, " IN ", values))
      } else {
        return(paste0(var_name, " NOT IN ", values))
      }
    }
    return("")
  }
  
  # Format edge label based on style
  format_edge_label <- function(split_label, is_left, style, max_len) {
    if (is.na(split_label) || split_label == "") return("")
    
    if (style == "none") {
      return("")
    }
    
    if (style == "yesno") {
      return(if (is_left) "yes" else "no")
    }
    
    # For "innode" style, use simple yes/no on edges since split is in node
    if (style == "innode") {
      return(if (is_left) "yes" else "no")
    }
    
    var_name <- get_split_variable(split_label)
    condition <- get_split_condition(split_label, is_left)
    
    if (style == "threshold" || style == "rpart") {
      # Just the condition (< 47, >= 47, or category values)
      label <- condition
    } else if (style == "full") {
      # Full: variable + condition
      label <- paste0(var_name, " ", condition)
    } else {
      label <- condition
    }
    
    # Truncate if too long
    if (nchar(label) > max_len) {
      label <- paste0(substr(label, 1, max_len - 3), "...")
    }
    
    return(label)
  }
  
  # Format path for tooltip (HTML)
  format_path_html <- function(path) {
    if (is.na(path) || path == "") return("<b>Root node</b>")
    
    conditions <- strsplit(path, " & ")[[1]]
    
    formatted <- sapply(conditions, function(cond) {
      cond <- trimws(cond)
      if (startsWith(cond, "!")) {
        cond <- substring(cond, 2)
        if (grepl("<=", cond)) cond <- gsub("<=", " > ", cond)
        else if (grepl("%in%", cond)) cond <- gsub("%in%", " NOT IN ", cond)
      } else {
        if (grepl("<=", cond)) cond <- gsub("<=", " \u2264 ", cond)
        if (grepl("%in%", cond)) cond <- gsub("%in%", " IN ", cond)
      }
      cond <- gsub("c\\(", "{", cond)
      cond <- gsub("\\)", "}", cond)
      cond <- gsub("\"", "", cond)
      cond
    })
    
    paste0("<b>Path:</b><br>", paste(formatted, collapse = "<br>AND "))
  }
  
  # Format split rule for tooltip
  format_split_html <- function(split_label) {
    if (is.na(split_label) || split_label == "") return("")
    
    split_display <- split_label
    split_display <- gsub("<=", " \u2264 ", split_display)
    split_display <- gsub("%in%", " IN ", split_display)
    split_display <- gsub("c\\(", "{", split_display)
    split_display <- gsub("\\)", "}", split_display)
    split_display <- gsub("\"", "", split_display)
    
    return(split_display)
  }
  
  # Format split rule for display IN the node (compact version)
  format_split_for_node <- function(split_label, max_chars = 25) {
    if (is.na(split_label) || split_label == "") return("")
    
    if (grepl("<=", split_label)) {
      parts <- strsplit(split_label, "<=")[[1]]
      var_name <- trimws(parts[1])
      threshold <- trimws(parts[2])
      result <- paste0(var_name, " \u2264 ", threshold)
    } else if (grepl("%in%", split_label)) {
      parts <- strsplit(split_label, "%in%")[[1]]
      var_name <- trimws(parts[1])
      values <- trimws(parts[2])
      values <- gsub("c\\(|\\)", "", values)
      values <- gsub("\"", "", values)
      # Truncate values if too long
      if (nchar(values) > max_chars - nchar(var_name) - 4) {
        values <- paste0(substr(values, 1, max_chars - nchar(var_name) - 7), "...")
      }
      result <- paste0(var_name, " = ", values)
    } else {
      result <- split_label
    }
    
    # Final truncation if still too long
    if (nchar(result) > max_chars + 10) {
      result <- paste0(substr(result, 1, max_chars + 7), "...")
    }
    
    return(result)
  }
  
  # =========================================================================
  # Build nodes dataframe
  # =========================================================================
  
  total_n <- tree_info$n[tree_info$node == 1]
  max_depth <- max(sapply(tree_info$node, calc_depth))
  
  # Function to calculate node position for free_drag mode
  calc_node_position <- function(node_id, node_spacing, level_separation, direction) {
    depth <- calc_depth(node_id)
    
    if (node_id == 1) {
      x_pos <- 0
      y_pos <- 0
    } else {
      # Trace path from root: 0 = left, 1 = right
      path <- c()
      n <- node_id
      while (n > 1) {
        path <- c(n %% 2, path)
        n <- floor(n / 2)
      }
      
      # Calculate x position following the path
      # Base width depends on tree depth for good spacing
      base_width <- node_spacing * (2 ^ (max_depth - 1))
      x_pos <- 0
      width <- base_width
      for (dir in path) {
        if (dir == 0) {
          x_pos <- x_pos - width
        } else {
          x_pos <- x_pos + width
        }
        width <- width / 2
      }
      
      y_pos <- depth * level_separation
    }
    
    # Adjust based on direction
    if (direction == "DU") {
      y_pos <- -y_pos
    } else if (direction == "LR") {
      tmp <- x_pos
      x_pos <- y_pos
      y_pos <- tmp
    } else if (direction == "RL") {
      tmp <- x_pos
      x_pos <- -y_pos
      y_pos <- tmp
    }
    
    list(x = x_pos, y = y_pos)
  }
  
  nodes_list <- lapply(seq_len(nrow(tree_info)), function(i) {
    row <- tree_info[i, ]
    node_id <- row$node
    depth <- calc_depth(node_id)
    
    # Calculate position for free_drag mode
    pos <- calc_node_position(node_id, node_spacing, level_separation, direction)
    
    # Calculate percentage
    pct <- round(row$n / total_n * 100)
    
    # Build label
    if (is_classification) {
      main_label <- as.character(row$pred)
      
      # For rpart style: add split variable name in internal nodes
      if (split_label_style == "rpart" && !row$terminal && !is.na(row$splitLabel)) {
        split_var <- get_split_variable(row$splitLabel)
        main_label <- paste0(main_label, "\n", split_var)
      }
      
      # For innode style: add full split rule in internal nodes
      if (split_label_style == "innode" && !row$terminal && !is.na(row$splitLabel)) {
        split_rule <- format_split_for_node(row$splitLabel, max_chars = 30)
        main_label <- paste0(main_label, "\n", split_rule)
      }
      
      # Secondary info - each on separate line
      info_lines <- c()
      if (show_prob) {
        prob_str <- sprintf("%.2f  %.2f", 1 - row$prob, row$prob)
        info_lines <- c(info_lines, prob_str)
      }
      if (show_percent) {
        info_lines <- c(info_lines, paste0(pct, "%"))
      }
      if (show_n) {
        info_lines <- c(info_lines, paste0("n=", row$n))
      }
      
      if (length(info_lines) > 0) {
        label <- paste0(main_label, "\n", paste(info_lines, collapse = "\n"))
      } else {
        label <- main_label
      }
      
      bg_color <- colors[as.character(row$pred)]
      if (is.null(bg_color) || is.na(bg_color)) bg_color <- "#cccccc"
      
    } else {
      # Regression
      main_label <- sprintf("%.2f", row$pred)
      
      if (split_label_style == "rpart" && !row$terminal && !is.na(row$splitLabel)) {
        split_var <- get_split_variable(row$splitLabel)
        main_label <- paste0(main_label, "\n", split_var)
      }
      
      # For innode style: add full split rule in internal nodes
      if (split_label_style == "innode" && !row$terminal && !is.na(row$splitLabel)) {
        split_rule <- format_split_for_node(row$splitLabel, max_chars = 30)
        main_label <- paste0(main_label, "\n", split_rule)
      }
      
      # Secondary info - each on separate line
      info_lines <- c()
      if (show_percent) info_lines <- c(info_lines, paste0(pct, "%"))
      if (show_n) info_lines <- c(info_lines, paste0("n=", row$n))
      
      if (length(info_lines) > 0) {
        label <- paste0(main_label, "\n", paste(info_lines, collapse = "\n"))
      } else {
        label <- main_label
      }
      
      bg_color <- "#6BAED6"
    }
    
    # Shape
    shape <- if (row$terminal) "box" else "ellipse"
    border_color <- if (row$terminal) "#333333" else "#666666"
    border_width <- if (row$terminal) 2 else 1
    
    # Build tooltip
    tooltip <- paste0(
      "<div style='padding: 10px; font-family: Arial, sans-serif; max-width: 500px;'>",
      "<h4 style='margin: 0 0 10px 0;'>Node ", node_id, 
      if (row$terminal) " (Terminal)" else " (Internal)", "</h4>",
      "<table style='border-collapse: collapse;'>",
      "<tr><td style='padding: 3px 10px 3px 0;'><b>Prediction:</b></td><td>", row$pred, "</td></tr>",
      "<tr><td style='padding: 3px 10px 3px 0;'><b>Observations:</b></td><td>", row$n, " (", pct, "%)</td></tr>"
    )
    
    if (is_classification) {
      tooltip <- paste0(tooltip,
                        "<tr><td style='padding: 3px 10px 3px 0;'><b>Accuracy:</b></td><td>", 
                        round(row$prob * 100, 1), "%</td></tr>"
      )
    }
    
    tooltip <- paste0(tooltip,
                      "<tr><td style='padding: 3px 10px 3px 0;'><b>Depth:</b></td><td>", depth, "</td></tr>",
                      "</table>",
                      "<hr style='margin: 10px 0;'>",
                      format_path_html(row$path)
    )
    
    # Add full split rule in tooltip
    if (!row$terminal && !is.na(row$splitLabel)) {
      tooltip <- paste0(tooltip,
                        "<hr style='margin: 10px 0;'>",
                        "<b>Split Rule:</b><br>",
                        "<span style='font-family: monospace; background: #f0f0f0; padding: 2px 5px; word-wrap: break-word;'>",
                        format_split_html(row$splitLabel),
                        "</span>"
      )
    }
    
    tooltip <- paste0(tooltip, "</div>")
    
    list(
      id = node_id,
      label = label,
      level = depth,
      x = pos$x,
      y = pos$y,
      shape = shape,
      color.background = bg_color,
      color.border = border_color,
      borderWidth = border_width,
      title = tooltip,
      is_terminal = row$terminal,
      split_label = if (!is.na(row$splitLabel)) row$splitLabel else ""
    )
  })
  
  nodes_df <- data.frame(
    id = sapply(nodes_list, function(x) x$id),
    label = sapply(nodes_list, function(x) x$label),
    level = sapply(nodes_list, function(x) x$level),
    shape = sapply(nodes_list, function(x) x$shape),
    color.background = sapply(nodes_list, function(x) x$color.background),
    color.border = sapply(nodes_list, function(x) x$color.border),
    color.highlight.background = "#FFD700",
    color.highlight.border = "#FFA500",
    borderWidth = sapply(nodes_list, function(x) x$borderWidth),
    font.size = font_size,
    font.color = "#000000",
    shadow = TRUE,
    stringsAsFactors = FALSE
  )
  
  # Add x, y coordinates only for free_drag mode
  if (free_drag) {
    nodes_df$x <- sapply(nodes_list, function(x) x$x)
    nodes_df$y <- sapply(nodes_list, function(x) x$y)
  }
  
  # Add tooltips only if details_on is not "none"
  if (details_on != "none") {
    nodes_df$title <- sapply(nodes_list, function(x) x$title)
  }
  
  # =========================================================================
  # Build edges dataframe
  # =========================================================================
  
  edges_list <- list()
  
  for (i in seq_len(nrow(tree_info))) {
    row <- tree_info[i, ]
    node_id <- row$node
    
    left_child <- node_id * 2
    right_child <- node_id * 2 + 1
    
    left_exists <- left_child %in% tree_info$node
    right_exists <- right_child %in% tree_info$node
    
    if (left_exists || right_exists) {
      split_label <- row$splitLabel
      
      if (left_exists) {
        edge_label <- format_edge_label(split_label, is_left = TRUE, 
                                        style = split_label_style, max_len = max_label_length)
        
        # Edge tooltip with FULL condition (never truncated)
        full_condition <- get_full_condition(split_label, is_left = TRUE)
        edge_tooltip <- paste0(
          "<div style='padding: 8px; font-family: Arial; max-width: 400px; word-wrap: break-word;'>",
          "<b>Condition:</b><br>",
          "<span style='font-family: monospace; background: #e8f4e8; padding: 3px 6px;'>",
          full_condition,
          "</span>",
          "</div>"
        )
        
        edges_list <- c(edges_list, list(data.frame(
          from = node_id,
          to = left_child,
          label = edge_label,
          title = edge_tooltip,
          arrows = "to",
          color = "#666666",
          width = 1.5,
          font.size = edge_font_size,
          font.color = "#333333",
          font.align = "middle",
          stringsAsFactors = FALSE
        )))
      }
      
      if (right_exists) {
        edge_label <- format_edge_label(split_label, is_left = FALSE, 
                                        style = split_label_style, max_len = max_label_length)
        
        # Edge tooltip with FULL condition (never truncated)
        full_condition <- get_full_condition(split_label, is_left = FALSE)
        edge_tooltip <- paste0(
          "<div style='padding: 8px; font-family: Arial; max-width: 400px; word-wrap: break-word;'>",
          "<b>Condition:</b><br>",
          "<span style='font-family: monospace; background: #f4e8e8; padding: 3px 6px;'>",
          full_condition,
          "</span>",
          "</div>"
        )
        
        edges_list <- c(edges_list, list(data.frame(
          from = node_id,
          to = right_child,
          label = edge_label,
          title = edge_tooltip,
          arrows = "to",
          color = "#666666",
          width = 1.5,
          font.size = edge_font_size,
          font.color = "#333333",
          font.align = "middle",
          stringsAsFactors = FALSE
        )))
      }
    }
  }
  
  if (length(edges_list) > 0) {
    edges_df <- do.call(rbind, edges_list)
  } else {
    edges_df <- data.frame(from = integer(), to = integer(), stringsAsFactors = FALSE)
  }
  
  # =========================================================================
  # Create visNetwork
  # =========================================================================
  
  vis <- visNetwork::visNetwork(
    nodes = nodes_df,
    edges = edges_df,
    width = width,
    height = height
  )
  
  # Layout: hierarchical (fixed vertical) or free (using pre-calculated x,y)
  if (!free_drag) {
    # Standard hierarchical layout - nodes can only move horizontally
    vis <- visNetwork::visHierarchicalLayout(
      vis,
      direction = direction,
      levelSeparation = level_separation,
      nodeSpacing = node_spacing,
      sortMethod = "directed",
      shakeTowards = "roots"
    )
  }
  # If free_drag = TRUE, nodes use x,y coordinates and can be dragged anywhere
  
  # Physics disabled - allows dragging without snapping back
  vis <- visNetwork::visPhysics(vis, enabled = FALSE)
  
  # Interaction - depends on details_on setting
  if (details_on == "hover") {
    vis <- visNetwork::visInteraction(
      vis,
      dragNodes = TRUE,
      dragView = TRUE,
      zoomView = TRUE,
      hover = TRUE,
      hoverConnectedEdges = TRUE,
      tooltipDelay = 100,
      tooltipStay = 500,
      navigationButtons = navigation_buttons
    )
  } else {
    # For "click" or "none" - disable hover tooltips
    vis <- visNetwork::visInteraction(
      vis,
      dragNodes = TRUE,
      dragView = TRUE,
      zoomView = TRUE,
      hover = FALSE,
      hoverConnectedEdges = TRUE,
      tooltipDelay = 0,
      navigationButtons = navigation_buttons
    )
  }
  
  # Options - remove node selection dropdown to save space
  vis <- visNetwork::visOptions(
    vis,
    highlightNearest = list(enabled = TRUE, degree = 1, hover = TRUE),
    nodesIdSelection = FALSE
  )
  
  # Add click event for details_on = "click"
  if (details_on == "click") {
    # Store node info in a data attribute for JavaScript access
    node_info_json <- sapply(nodes_list, function(x) {
      jsonlite::toJSON(list(
        id = x$id,
        title = x$title
      ), auto_unbox = TRUE)
    })
    
    vis <- visNetwork::visEvents(
      vis,
      click = sprintf("function(params) {
        if (params.nodes.length > 0) {
          var nodeId = params.nodes[0];
          var nodeInfo = %s;
          var info = nodeInfo[nodeId];
          if (info && info.title) {
            // Create a popup div
            var popup = document.getElementById('e2tree-popup');
            if (!popup) {
              popup = document.createElement('div');
              popup.id = 'e2tree-popup';
              popup.style.cssText = 'position:fixed; top:10px; right:10px; max-width:400px; max-height:80vh; overflow-y:auto; background:white; border:2px solid #333; border-radius:8px; box-shadow:0 4px 20px rgba(0,0,0,0.3); z-index:9999; padding:0;';
              document.body.appendChild(popup);
            }
            // Add close button and content
            popup.innerHTML = '<div style=\"display:flex; justify-content:space-between; align-items:center; background:#f0f0f0; padding:8px 12px; border-bottom:1px solid #ddd;\"><span style=\"font-weight:bold;\">Node Details</span><button onclick=\"this.parentElement.parentElement.style.display=\\'none\\'\" style=\"border:none; background:#e74c3c; color:white; border-radius:4px; padding:4px 10px; cursor:pointer;\">Close</button></div><div style=\"padding:10px;\">' + info.title + '</div>';
            popup.style.display = 'block';
          }
        }
      }", jsonlite::toJSON(setNames(lapply(nodes_list, function(x) list(id = x$id, title = x$title)), sapply(nodes_list, function(x) x$id))))
    )
  }
  
  # Export
  vis <- visNetwork::visExport(vis, type = "png", name = "e2tree_plot")
  
  return(vis)
}


#' Save E2Tree visNetwork Plot to HTML
#'
#' @param vis A visNetwork object from plot_e2tree_vis()
#' @param file Output file path (should end with .html)
#' @param selfcontained Include all dependencies in single file
#'
#' @export
save_e2tree_html <- function(vis, file = "e2tree_plot.html", selfcontained = TRUE) {
  if (!requireNamespace("htmlwidgets", quietly = TRUE)) {
    stop("Package 'htmlwidgets' is required.")
  }
  
  htmlwidgets::saveWidget(vis, file = file, selfcontained = selfcontained)
  message("Plot saved to: ", normalizePath(file))
}

#' Interactive E2Tree Plot for R Graphics Device
#'
#' @description
#' Displays an E2Tree as an interactive plot in the R graphics device.
#' Click on nodes to see detailed information in the console.
#' Right-click or press ESC to exit interactive mode.
#'
#' @param fit An e2tree object
#' @param data The training data used to build the tree
#' @param ensemble The ensemble model (randomForest or ranger)
#' @param main Plot title (default: "E2Tree - Click on nodes (ESC to exit)")
#' @param ... Additional arguments passed to rpart.plot
#'
#' @return Invisibly returns the rpart object
#'
#' @details
#' This function converts the e2tree object to an rpart object and displays it
#' using rpart.plot. You can then click on any node to see:
#' - Node ID and type (terminal/internal)
#' - Number of observations
#' - Prediction and probability/purity
#' - Decision path to reach the node
#' - Class distribution (for classification)
#' - Split rule (for internal nodes)
#' - Observations in the node (for terminal nodes)
#'
#' @examples
#' \donttest{
#' # After creating an e2tree object (requires interactive session)
#' if (interactive()) {
#'   plot_e2tree_click(tree, training, ensemble)
#' }
#' }
#'
#' @export
plot_e2tree_click <- function(fit, data, ensemble, 
                               main = "E2Tree - Click on nodes (ESC to exit)",
                               ...) {
  

  # Check required packages
  if (!requireNamespace("rpart.plot", quietly = TRUE)) {
    stop("Package 'rpart.plot' is required. Install with: install.packages('rpart.plot')")
  }
  
  # Validate inputs
  if (!inherits(fit, "e2tree")) {
    stop("'fit' must be an e2tree object")
  }
  
  # Convert e2tree to rpart
  rpart_obj <- rpart2Tree(fit, ensemble)
  
  # Get tree info
  tree_info <- fit$tree
  
  # Determine task type
  is_classification <- !is.null(attr(fit, "ylevels"))
  response_var <- all.vars(fit$terms)[1]
  
  # Get class levels and colors
  if (is_classification) {
    class_levels <- attr(fit, "ylevels")
    n_classes <- length(class_levels)
    if (n_classes <= 3) {
      class_colors <- c("#e74c3c", "#3498db", "#27ae60")[1:n_classes]
    } else {
      class_colors <- rainbow(n_classes)
    }
  }
  
  # =========================================================================
  # Helper function to format path
  # =========================================================================
  format_path <- function(path) {
    if (is.na(path) || path == "") return("Root node")
    
    # Split by " & "
    conditions <- strsplit(path, " & ")[[1]]
    
    formatted <- sapply(conditions, function(cond) {
      cond <- trimws(cond)
      # Handle negation
      if (startsWith(cond, "!")) {
        cond <- substring(cond, 2)
        if (grepl("<=", cond)) {
          cond <- gsub("<=", ">", cond)
        } else if (grepl("%in%", cond)) {
          cond <- gsub("%in%", "NOT IN", cond)
        }
      } else {
        if (grepl("%in%", cond)) {
          cond <- gsub("%in%", "IN", cond)
        }
      }
      # Format c() values
      if (grepl("c\\(", cond)) {
        cond <- gsub("c\\(", "{", cond)
        cond <- gsub("\\)", "}", cond)
      }
      cond
    })
    
    paste(formatted, collapse = "\n  AND ")
  }
  
  # =========================================================================
  # Get node positions from rpart.plot
  # =========================================================================
  get_node_positions <- function(rpart_obj) {
    # Get frame info
    frame <- rpart_obj$frame
    nodes <- as.integer(row.names(frame))
    n_nodes <- length(nodes)
    
    # Calculate depth for each node
    get_depth <- function(node) {
      d <- 0
      n <- node
      while (n > 1) {
        n <- floor(n / 2)
        d <- d + 1
      }
      d
    }
    
    depths <- sapply(nodes, get_depth)
    max_depth <- max(depths)
    
    # Calculate x positions based on binary tree structure
    # Nodes at each level are positioned based on their path
    x_pos <- numeric(n_nodes)
    y_pos <- numeric(n_nodes)
    
    for (i in seq_along(nodes)) {
      node <- nodes[i]
      depth <- depths[i]
      
      # Y position based on depth (inverted, root at top)
      y_pos[i] <- 1 - depth / (max_depth + 1)
      
      # X position based on binary path
      if (node == 1) {
        x_pos[i] <- 0.5
      } else {
        # Trace path from root
        path <- integer(0)
        n <- node
        while (n > 1) {
          path <- c(n %% 2, path)  # 0 = left, 1 = right
          n <- floor(n / 2)
        }
        
        # Calculate x position
        x <- 0.5
        width <- 0.25
        for (dir in path) {
          if (dir == 0) {
            x <- x - width
          } else {
            x <- x + width
          }
          width <- width / 2
        }
        x_pos[i] <- x
      }
    }
    
    data.frame(
      node = nodes,
      x = x_pos,
      y = y_pos,
      depth = depths,
      is_terminal = frame$var == "<leaf>",
      n = frame$n
    )
  }
  
  # =========================================================================
  # Print node information
  # =========================================================================
  print_node_info <- function(node_id) {
    
    # Find node in tree_info
    node_idx <- which(tree_info$node == node_id)
    
    if (length(node_idx) == 0) {
      cat("\n  Node not found in tree structure\n")
      return(invisible(NULL))
    }
    
    node_row <- tree_info[node_idx, ]
    
    # Header
    cat("\n")
    cat(paste(rep("=", 60), collapse = ""), "\n")
    if (node_row$terminal) {
      cat(sprintf("  TERMINAL NODE #%d\n", node_id))
    } else {
      cat(sprintf("  INTERNAL NODE #%d\n", node_id))
    }
    cat(paste(rep("=", 60), collapse = ""), "\n\n")
    
    # Basic info
    cat(sprintf("  Observations:    %d\n", node_row$n))
    cat(sprintf("  Prediction:      %s\n", node_row$pred))
    
    if (is_classification) {
      cat(sprintf("  Purity:          %.1f%%\n", node_row$prob * 100))
    }
    
    # Path
    cat("\n  PATH TO NODE:\n")
    cat("  ", format_path(node_row$path), "\n")
    
    # Split rule for internal nodes
    if (!node_row$terminal && !is.na(node_row$splitLabel)) {
      cat("\n  SPLIT RULE:\n")
      split_label <- node_row$splitLabel
      if (grepl("<=", split_label)) {
        parts <- strsplit(split_label, "<=")[[1]]
        cat(sprintf("    If %s <= %s -> LEFT child (node %d)\n", 
                    trimws(parts[1]), trimws(parts[2]), node_id * 2))
        cat(sprintf("    If %s >  %s -> RIGHT child (node %d)\n", 
                    trimws(parts[1]), trimws(parts[2]), node_id * 2 + 1))
      } else if (grepl("%in%", split_label)) {
        cat(sprintf("    %s\n", split_label))
        cat(sprintf("    TRUE  -> LEFT child (node %d)\n", node_id * 2))
        cat(sprintf("    FALSE -> RIGHT child (node %d)\n", node_id * 2 + 1))
      }
    }
    
    # Class distribution for classification
    if (is_classification && !is.null(node_row$obs) && length(node_row$obs[[1]]) > 0) {
      obs_indices <- node_row$obs[[1]]
      if (length(obs_indices) > 0 && max(obs_indices) <= nrow(data)) {
        node_data <- data[obs_indices, ]
        class_table <- table(factor(node_data[[response_var]], levels = class_levels))
        
        cat("\n  CLASS DISTRIBUTION:\n")
        for (cls in names(class_table)) {
          pct <- round(class_table[cls] / sum(class_table) * 100, 1)
          bar_width <- round(pct / 5)
          bar <- paste(rep("#", bar_width), collapse = "")
          cat(sprintf("    %-15s %4d (%5.1f%%) %s\n", cls, class_table[cls], pct, bar))
        }
      }
    }
    
    # Show observations for terminal nodes
    if (node_row$terminal && !is.null(node_row$obs) && length(node_row$obs[[1]]) > 0) {
      obs_indices <- node_row$obs[[1]]
      cat(sprintf("\n  OBSERVATION IDs: (%d total)\n", length(obs_indices)))
      
      if (length(obs_indices) <= 20) {
        cat("    ", paste(obs_indices, collapse = ", "), "\n")
      } else {
        cat("    ", paste(head(obs_indices, 10), collapse = ", "), 
            ", ..., ", 
            paste(tail(obs_indices, 5), collapse = ", "), "\n")
      }
    }
    
    cat("\n")
    cat(paste(rep("-", 60), collapse = ""), "\n")
    
    invisible(NULL)
  }
  
  # =========================================================================
  # Main plotting and interaction loop
  # =========================================================================
  
  # Set up the plot
  # Use a clean device
  if (dev.cur() > 1) {
    # Keep current device
  } else {
    dev.new()
  }
  
  # Plot the tree with rpart.plot
  rpart.plot::rpart.plot(
    rpart_obj,
    main = main,
    type = 4,
    extra = if(is_classification) 104 else 101,
    under = TRUE,
    faclen = 0,
    cex = 0.9,
    box.palette = if(is_classification) "auto" else "Blues",
    shadow.col = "gray",
    nn = TRUE,  # Show node numbers
    roundint = FALSE,
    ...
  )

  # Get node positions
  node_pos <- get_node_positions(rpart_obj)
  
  # Get plot coordinates
  usr <- par("usr")
  
  # Scale node positions to plot coordinates
  node_pos$x_plot <- usr[1] + node_pos$x * (usr[2] - usr[1])
  node_pos$y_plot <- usr[3] + node_pos$y * (usr[4] - usr[3])
  
  # Instructions
  cat("\n")
  cat(paste(rep("=", 60), collapse = ""), "\n")
  cat("  E2TREE INTERACTIVE PLOT\n")
  cat(paste(rep("=", 60), collapse = ""), "\n")
  cat("\n")
  cat("  Click on any node to see detailed information.\n")
  cat("  Press ESC or right-click to exit interactive mode.\n")
  cat("\n")
  cat(paste(rep("-", 60), collapse = ""), "\n")
  
  # Interactive loop
  repeat {
    # Get click location
    click <- tryCatch({
      locator(1)
    }, error = function(e) {
      NULL
    })
    
    # Exit if no click (ESC pressed or window closed)
    if (is.null(click) || length(click$x) == 0) {
      cat("\n  Exiting interactive mode.\n\n")
      break
    }
    
    # Find closest node
    distances <- sqrt((node_pos$x_plot - click$x)^2 + (node_pos$y_plot - click$y)^2)
    closest_idx <- which.min(distances)
    closest_node <- node_pos$node[closest_idx]
    
    # Check if click is reasonably close to a node
    # (within 10% of plot width)
    threshold <- (usr[2] - usr[1]) * 0.1
    if (distances[closest_idx] > threshold) {
      cat("\n  Click closer to a node.\n")
      next
    }
    
    # Print node information
    print_node_info(closest_node)
    
    # Highlight the selected node briefly
    points(node_pos$x_plot[closest_idx], node_pos$y_plot[closest_idx], 
           pch = 1, cex = 4, col = "red", lwd = 3)
    
    # Redraw after a moment to remove highlight
    Sys.sleep(0.3)
    rpart.plot::rpart.plot(
      rpart_obj,
      main = main,
      type = 4,
      extra = if(is_classification) 104 else 101,
      under = TRUE,
      faclen = 0,
      cex = 0.9,
      box.palette = if(is_classification) "auto" else "Blues",
      shadow.col = "gray",
      nn = TRUE,
      roundint = FALSE,
      ...
    )
  }

  invisible(rpart_obj)
}


#' Quick E2Tree Plot (Non-Interactive)
#'
#' @description
#' Displays an E2Tree as a static plot using rpart.plot.
#' For interactive exploration, use plot_e2tree_click().
#'
#' @param fit An e2tree object
#' @param ensemble The ensemble model (randomForest or ranger)
#' @param main Plot title
#' @param ... Additional arguments passed to rpart.plot
#'
#' @return Invisibly returns the rpart object
#'
#' @export
plot_e2tree <- function(fit, ensemble, main = "E2Tree", ...) {
  
  if (!requireNamespace("rpart.plot", quietly = TRUE)) {
    stop("Package 'rpart.plot' is required. Install with: install.packages('rpart.plot')")
  }
  
  if (!inherits(fit, "e2tree")) {
    stop("'fit' must be an e2tree object")
  }
  
  rpart_obj <- rpart2Tree(fit, ensemble)
  is_classification <- !is.null(attr(fit, "ylevels"))
  
  rpart.plot::rpart.plot(
    rpart_obj,
    main = main,
    type = 4,
    extra = if(is_classification) 104 else 101,
    under = TRUE,
    faclen = 0,
    cex = 0.9,
    box.palette = if(is_classification) "auto" else "Blues",
    shadow.col = "gray",
    nn = TRUE,
    roundint = FALSE,
    ...
  )

  invisible(rpart_obj)
}


#' Print E2Tree Summary
#'
#' @description
#' Prints a comprehensive summary of an E2Tree model including
#' all decision rules, variable importance, and node statistics.
#'
#' @param fit An e2tree object
#' @param data The training data
#'
#' @export
print_e2tree_summary <- function(fit, data) {
  
  if (!inherits(fit, "e2tree")) {
    stop("'fit' must be an e2tree object")
  }
  
  tree_info <- fit$tree
  is_classification <- !is.null(attr(fit, "ylevels"))
  response_var <- all.vars(fit$terms)[1]
  
  if (is_classification) {
    class_levels <- attr(fit, "ylevels")
  }
  
  # Header
  cat("\n")
  cat(paste(rep("=", 70), collapse = ""), "\n")
  cat("                    E2TREE MODEL SUMMARY\n")
  cat(paste(rep("=", 70), collapse = ""), "\n\n")
  
  # Model info
  cat("MODEL INFORMATION\n")
  cat(paste(rep("-", 40), collapse = ""), "\n")
  cat(sprintf("  Task:              %s\n", if(is_classification) "Classification" else "Regression"))
  cat(sprintf("  Response:          %s\n", response_var))
  cat(sprintf("  Observations:      %d\n", nrow(data)))
  cat(sprintf("  Total Nodes:       %d\n", nrow(tree_info)))
  cat(sprintf("  Terminal Nodes:    %d\n", sum(tree_info$terminal)))
  cat(sprintf("  Max Depth:         %d\n", max(sapply(tree_info$node, function(n) {
    d <- 0; while(n > 1) { n <- floor(n/2); d <- d + 1 }; d
  }))))
  
  # Variable importance
  if (!is.null(fit$varimp) && !is.null(fit$varimp$vimp)) {
    cat("\n\nVARIABLE IMPORTANCE\n")
    cat(paste(rep("-", 40), collapse = ""), "\n")
    vimp <- fit$varimp$vimp
    if (is.data.frame(vimp) && ncol(vimp) >= 2) {
      vimp <- vimp[order(vimp[[2]], decreasing = TRUE), ]
      max_imp <- max(vimp[[2]])
      for (i in 1:nrow(vimp)) {
        bar_width <- round(vimp[[2]][i] / max_imp * 20)
        bar <- paste(rep("#", bar_width), collapse = "")
        cat(sprintf("  %-20s %8.4f  %s\n", vimp[[1]][i], vimp[[2]][i], bar))
      }
    }
  }
  
  # Decision rules
  cat("\n\nDECISION RULES\n")
  cat(paste(rep("-", 40), collapse = ""), "\n")
  
  terminal_nodes <- tree_info[tree_info$terminal, ]
  
  for (i in 1:nrow(terminal_nodes)) {
    node <- terminal_nodes[i, ]
    
    cat(sprintf("\nRule %d (Node %d):\n", i, node$node))
    
    # Format path
    if (!is.na(node$path) && node$path != "") {
      conditions <- strsplit(node$path, " & ")[[1]]
      cat("  IF:\n")
      for (j in seq_along(conditions)) {
        cond <- trimws(conditions[j])
        # Handle negation
        if (startsWith(cond, "!")) {
          cond <- substring(cond, 2)
          if (grepl("<=", cond)) cond <- gsub("<=", ">", cond)
          else if (grepl("%in%", cond)) cond <- gsub("%in%", "NOT IN", cond)
        } else {
          if (grepl("%in%", cond)) cond <- gsub("%in%", "IN", cond)
        }
        if (grepl("c\\(", cond)) {
          cond <- gsub("c\\(", "{", cond)
          cond <- gsub("\\)", "}", cond)
        }
        if (j == 1) {
          cat(sprintf("       %s\n", cond))
        } else {
          cat(sprintf("   AND %s\n", cond))
        }
      }
    } else {
      cat("  IF: (root - all observations)\n")
    }
    
    cat(sprintf("  THEN: %s", node$pred))
    if (is_classification) {
      cat(sprintf(" (purity: %.1f%%, n=%d)", node$prob * 100, node$n))
    } else {
      cat(sprintf(" (n=%d)", node$n))
    }
    cat("\n")
  }
  
  cat("\n")
  cat(paste(rep("=", 70), collapse = ""), "\n\n")
  
  invisible(fit)
}

---
title: "Introduction to e2tree: Explainable Ensemble Trees"
author: "Massimo Aria, Agostino Gnasso"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to e2tree: Explainable Ensemble Trees}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 5
)
```

## Overview

The **e2tree** package implements the Explainable Ensemble Tree (E2Tree)
methodology, which constructs a single interpretable decision tree that
approximates the predictions of an ensemble model (such as a random forest).
The key idea is to use the dissimilarity structure induced by the ensemble
to guide the construction of the tree, thereby preserving the predictive
information encoded in the ensemble while providing a transparent,
human-readable model.

The methodology is described in:

- Aria, M., Gnasso, A., Iorio, C., & Pandolfo, G. (2024). Explainable
  ensemble trees. *Computational Statistics*, 39(1), 3--19.
- Aria, M., Gnasso, A., Iorio, C., & Fokkema, M. (2026). Extending
  Explainable Ensemble Trees to Regression Contexts. *Applied Stochastic
  Models in Business and Industry*, 42(1), e70064.

## Installation

```{r eval = FALSE}
# Install from CRAN (when available):
install.packages("e2tree")

# Or install the development version from GitHub:
# devtools::install_github("massimoaria/e2tree")
```

## Workflow

A typical e2tree analysis follows these steps:

1. **Train an ensemble model** (random forest via `randomForest` or `ranger`)
2. **Compute the dissimilarity matrix** from the ensemble using `createDisMatrix()`
3. **Build the E2Tree** using `e2tree()`
4. **Inspect and visualize** the tree (print, summary, plot)
5. **Predict** on new data with `predict()`
6. **Validate** the E2Tree structure with `eValidation()` and `loi()`

## Classification Example

We use the classic `iris` dataset to demonstrate a classification workflow.

### Step 1: Prepare data and train ensemble

```{r classification-setup}
library(e2tree)

data(iris)
set.seed(42)
smp_size <- floor(0.75 * nrow(iris))
train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
training <- iris[train_ind, ]
validation <- iris[-train_ind, ]
```

```{r classification-ensemble}
if (!require("randomForest")) install.packages("randomForest", 
                                               repos="https://cran.r-project.org")

library(randomForest)
  
ensemble <- randomForest(Species ~ ., data = training,
                         importance = TRUE, proximity = TRUE)
ensemble
```

### Step 2: Compute dissimilarity matrix

The ensemble induces a co-occurrence structure that captures how frequently
pairs of observations fall into the same terminal nodes across the ensemble's
trees. This structure is encoded in the co-occurrence matrix
$\mathbf{O} = [o_{ij}]$, where each entry $o_{ij} \in [0, 1]$ reflects the
weighted frequency with which observations $i$ and $j$ are grouped together.
The corresponding dissimilarity matrix is $\mathbf{D} = \mathbf{1} - \mathbf{O}$.

```{r classification-dismatrix}
D <- createDisMatrix(ensemble, data = training, label = "Species",
                     parallel = list(active = FALSE, no_cores = 1))
dim(D)
```

### Step 3: Build the E2Tree

The `e2tree()` function constructs a single tree that best represents the
ensemble's dissimilarity structure. The `setting` argument controls the
stopping rules:

- `impTotal`: minimum impurity to attempt a split
- `maxDec`: minimum decrease in impurity required
- `n`: minimum number of observations in a node
- `level`: maximum tree depth

```{r classification-tree}
setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
tree <- e2tree(Species ~ ., training, D, ensemble, setting)
```

### Step 4: Inspect the tree

```{r classification-print}
print(tree)
```

```{r classification-summary}
summary(tree)
```

The `nodes()` accessor extracts the tree structure as a data frame:

```{r classification-nodes}
# Terminal nodes only
nodes(tree, terminal = TRUE)
```

### Step 5: Visualization

The `plot()` method renders the tree using `rpart.plot`:

```{r classification-plot, fig.width = 8, fig.height = 6}
plot(tree, ensemble = ensemble, main = "E2Tree - Iris Classification")
```

You can also convert the tree to `rpart` or `partykit` formats for
alternative visualizations:

```{r classification-rpart, fig.width = 8, fig.height = 6}
# Convert to rpart
rpart_obj <- as.rpart(tree, ensemble)

# Convert to partykit (if installed)
if (requireNamespace("partykit", quietly = TRUE)) {
  party_obj <- partykit::as.party(tree)
  plot(party_obj)
}
```

### Step 6: Prediction

Use the standard `predict()` method on new data:

```{r classification-predict}
pred <- predict(tree, newdata = validation)
head(pred)
```

Fitted values for the training data:

```{r classification-fitted}
head(fitted(tree))
```

### Step 7: Variable importance

```{r classification-vimp}
vi <- vimp(tree, data = training)
vi$vimp
vi$g_imp
```

## Regression Example

The same workflow applies to regression problems. Here we use the `mtcars` dataset.

```{r regression-setup}
data(mtcars)
set.seed(123)
smp_size <- floor(0.75 * nrow(mtcars))
train_ind <- sample(seq_len(nrow(mtcars)), size = smp_size)
training_reg <- mtcars[train_ind, ]
validation_reg <- mtcars[-train_ind, ]
```

```{r regression-ensemble}
ensemble_reg <- randomForest(mpg ~ ., data = training_reg, ntree = 500,
                             importance = TRUE, proximity = TRUE)
```

```{r regression-dismatrix}
D_reg <- createDisMatrix(ensemble_reg, data = training_reg, label = "mpg",
                         parallel = list(active = FALSE, no_cores = 1))
```

```{r regression-tree}
setting_reg <- list(impTotal = 0.1, maxDec = 1e-6, n = 2, level = 5)
tree_reg <- e2tree(mpg ~ ., training_reg, D_reg, ensemble_reg, setting_reg)
print(tree_reg)
```

```{r regression-plot, fig.width = 8, fig.height = 6}
plot(tree_reg, ensemble = ensemble_reg, main = "E2Tree - MPG Regression")
```

For regression models, predictions include the standard deviation within
each terminal node, providing a measure of prediction uncertainty:

```{r regression-predict}
pred_reg <- predict(tree_reg, newdata = validation_reg)
head(pred_reg)
```

Residuals are also available:

```{r regression-residuals}
res <- residuals(tree_reg)
summary(res)
```


## Validation of the E2Tree Structure

A fundamental question arises when using E2Tree: how well does the single
interpretable tree capture the original ensemble's structure? The E2Tree
reconstructs the ensemble's co-occurrence matrix $\mathbf{O}$ through
its terminal node structure, producing an approximation
$\hat{\mathbf{O}}_T$. Assessing the quality of this reconstruction is
essential to ensure that the explanatory model faithfully represents
the ensemble's decision logic.

The `eValidation()` function provides two complementary approaches to
validation, selectable through the `test` argument:

- **`test = "mantel"`**: The Mantel test, a classical non-parametric
  test of *association* between similarity matrices.
- **`test = "measures"`**: A family of divergence and similarity measures
  that assess *agreement* between the matrices, with permutation-based
  inference.
- **`test = "both"`** (default): Performs both.

### Association vs. Agreement

The distinction between these two approaches mirrors the classical
distinction between *correlation* and *concordance* in method comparison
studies (Bland & Altman, 1986; Lin, 1989). Two proximity matrices can be
perfectly correlated ($r = 1$) and yet systematically disagree: if
$\hat{\mathbf{O}}_T = c \cdot \mathbf{O}$ for any constant $c \neq 1$,
the Mantel test would declare perfect association, yet the actual
proximity values would not be faithfully reproduced.

This distinction matters in the E2Tree context. The ensemble matrix
$\mathbf{O}$ contains continuous values in $(0, 1)$, reflecting varying
degrees of co-occurrence across the ensemble's trees. The E2Tree matrix
$\hat{\mathbf{O}}_T$, by contrast, has a crisp block-diagonal structure:
observations within the same terminal node share a common co-occurrence
value, while observations in different nodes have zero co-occurrence.
Even when the block structure perfectly matches the grouping pattern of
$\mathbf{O}$, the element-wise values will differ because $\mathbf{O}$
is fuzzy and $\hat{\mathbf{O}}_T$ is crisp. A scale-sensitive divergence
measure can quantify this discrepancy; a correlation-based measure cannot.

### The Mantel Test

The Mantel test evaluates whether the E2Tree proximity matrix is
*associated* with the ensemble proximity matrix. It computes a
correlation-like statistic between the two matrices and assesses
significance through permutation.

```{r val-mantel}
val_mantel <- eValidation(training, tree, D, test = "mantel", graph = FALSE)
print(val_mantel)
```

A significant Mantel test confirms that E2Tree preserves the overall
*pattern* of the ensemble's proximity structure. However, it does not
tell us how closely the actual proximity values are reproduced, nor
does it indicate *where* the reconstruction fails.

### Divergence and Similarity Measures

To assess *agreement* rather than mere association, we use a family of
five measures, each capturing a distinct aspect of reconstruction quality:

| Measure | Type | Range | What it captures |
|---------|------|-------|------------------|
| **nLoI** (normalized Loss of Interpretability) | divergence | [0, 1] | Scale-sensitive per-pair discrepancy with diagnostic decomposition |
| **Hellinger** distance | divergence | [0, 1] | Variance-stabilized element-wise difference (a true metric) |
| **wRMSE** (weighted Root Mean Squared Error) | divergence | [0, 1] | Weighted average discrepancy emphasizing high-proximity regions |
| **RV** coefficient | similarity | [0, 1] | Global structural similarity (matrix-level correlation) |
| **SSIM** (Structural Similarity Index) | similarity | [-1, 1] | Local spatial pattern similarity (luminance, contrast, structure) |

For *divergence* measures, lower values indicate better reconstruction.
For *similarity* measures, higher values indicate better reconstruction.

Statistical significance is assessed through a unified permutation testing
framework based on simultaneous row/column permutation of $\hat{\mathbf{O}}_T$.
This preserves the block-diagonal structure of the E2Tree matrix while
breaking the correspondence between observation labels and node assignments,
providing the appropriate null hypothesis: under $H_0$, E2Tree groups
observations randomly with respect to the ensemble structure.

```{r val-measures}
val_meas <- eValidation(training, tree, D, test = "measures",
                        graph = FALSE, n_perm = 499)
print(val_meas)
```

All divergence measures are significantly lower than their null
expectations (negative Z-statistics), while all similarity measures are
significantly higher (positive Z-statistics), confirming that the E2Tree
faithfully reconstructs the ensemble's proximity structure.

```{r val-measures-table}
measures(val_meas)
```

### Visualizing the Reconstruction

The `plot()` method for `eValidation` objects provides a visual summary
including heatmaps of both proximity matrices, the null distribution of
nLoI (when permutation tests are performed), and the LoI decomposition:

```{r val-plot, fig.width = 8, fig.height = 7}
val_full <- eValidation(training, tree, D, test = "both",
                        graph = FALSE, n_perm = 499)
plot(val_full)
```

### The Normalized Loss of Interpretability (nLoI)

Among the divergence measures, the nLoI deserves special attention because
of its unique *decomposability* into within-node and between-node components.

The nLoI is defined as:
$$\text{nLoI}(\mathbf{O}, \hat{\mathbf{O}}_T) = \frac{1}{M} \sum_{i<j} \frac{(o_{ij} - \hat{o}_{ij})^2}{\max(o_{ij}, \hat{o}_{ij})}$$
where $M = n(n-1)/2$ is the number of unique pairs.

This statistic is connected to the Cressie-Read power divergence family
at $\lambda = -2$ (Neyman-type statistic), with a robustified denominator
that avoids singularities when $\hat{o}_{ij} = 0$ (which is common in
E2Tree due to the block-diagonal structure of $\hat{\mathbf{O}}_T$).

The `loi()` function computes the nLoI and its decomposition directly
from two proximity matrices:

```{r loi-classification}
prox <- proximity(val_full)
result <- loi(prox$ensemble, prox$e2tree)
summary(result)
```

#### LoI Decomposition: Diagnosing Reconstruction Quality

The LoI decomposes into two interpretable components:

- **Within-node component** ($\text{LoI}_{in}$): Measures reconstruction
  fidelity for pairs that E2Tree groups together. It captures the
  calibration error --- whether the single tree assigns co-occurrence
  values that match the ensemble's.

- **Between-node component** ($\text{LoI}_{out}$): Measures the cost of
  the partition. It accumulates the ensemble co-occurrence values
  ($o_{ij}$) for pairs that E2Tree separates into different terminal
  nodes (where $\hat{o}_{ij} = 0$), representing irrecoverable loss.

Since between-node pairs vastly outnumber within-node pairs (especially
with many terminal nodes), the raw totals are not directly comparable.
The per-pair averages (`mean_in` and `mean_out`) provide the meaningful
diagnostic:

- A **high `mean_out`** (e.g., > 0.3) indicates that E2Tree is separating
  pairs with substantial ensemble proximity. The practitioner should
  consider increasing the tree's complexity (more terminal nodes) or
  relaxing pruning constraints.

- A **high `mean_in`** (e.g., > 0.1) indicates poor within-node calibration:
  the partition boundaries are well-placed, but the within-node proximity
  values deviate from the ensemble's. This typically reflects the inherent
  fuzzy-to-crisp structural transition.

- **Low values of both** confirm that the E2Tree faithfully captures the
  ensemble structure --- the partition is well-placed and the calibration
  is accurate.

This diagnostic capability is fundamentally impossible with the Mantel test
or any other scalar association measure, which can only report that the
overall reconstruction is "good" or "poor" without indicating the source
of the discrepancy.

```{r loi-plot}
plot(result)
```

#### Permutation Test for LoI

The `loi_perm()` function provides a standalone permutation test for the
nLoI, useful when a quick assessment of reconstruction significance is
needed without computing all five measures:

```{r loi-perm}
perm <- loi_perm(prox$ensemble, prox$e2tree, n_perm = 499, seed = 42)
print(perm)
plot(perm)
```

### Validation of the Regression Example

The same validation framework applies to regression E2Trees:

```{r val-regression}
val_reg <- eValidation(training_reg, tree_reg, D_reg, test = "measures",
                       graph = FALSE, n_perm = 499)
print(val_reg)
```

Regression E2Trees typically show higher nLoI and Hellinger values than
classification, consistent with the greater complexity of regression
proximity structures and the typically smaller sample sizes. Nonetheless,
all measures should achieve significant p-values, confirming that the
reconstruction captures genuine structure.


## Comparison with Related Packages

The **e2tree** package complements existing R infrastructure for tree-based
models:

- **partykit**: Provides generic classes for representing trees. E2Tree
  objects can be converted to `party` objects via `as.party()`, enabling
  the use of partykit's rich visualization and prediction infrastructure.
- **stablelearner**: Explores the stability of ensemble trees via
  `stabletree()`. While stablelearner focuses on variable and cutpoint
  stability, e2tree constructs a single tree that captures the ensemble's
  dissimilarity-based structure --- a complementary perspective.
- **rpart**: The standard recursive partitioning package. E2Tree objects can
  be converted to `rpart` objects via `as.rpart()` for visualization with
  `rpart.plot`.

## Session Info

```{r session-info}
sessionInfo()
```

---
title: "Using e2tree with XGBoost, GBM, LightGBM, and CatBoost"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Using e2tree with XGBoost, GBM, LightGBM, and CatBoost}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  eval     = FALSE   # examples require optional packages; set eval=TRUE locally
)
```

## Overview

Starting from version 1.1.0, `e2tree` supports the following tree ensemble
backends in addition to `randomForest` and `ranger`:

| Package | Model class | Task |
|---------|-------------|------|
| **xgboost** | `xgb.Booster` | classification, regression |
| **gbm** | `gbm` | classification, regression |
| **lightgbm** | `lgb.Booster` | classification, regression |
| **catboost** | `catboost.CatBoost` / `catboost.Model` | classification, regression |

The workflow is identical regardless of the backend: train a model, build the
dissimilarity matrix with `createDisMatrix()`, then call `e2tree()`.

---

## XGBoost

### Classification (iris)

```{r xgb-clf}
library(e2tree)
if (!require("xgboost")) install.packages("xgboost", 
                                               repos="https://cran.r-project.org")
library(xgboost)

data(iris)
set.seed(42)
n  <- floor(0.75 * nrow(iris))
tr <- iris[sample(nrow(iris), n), ]
va <- iris[setdiff(seq_len(nrow(iris)), as.integer(rownames(tr))), ]

# XGBoost requires a numeric matrix and 0-indexed integer labels
X_tr <- as.matrix(tr[, 1:4])
y_tr <- as.integer(tr$Species) - 1L
dm_tr <- xgb.DMatrix(data = X_tr, label = y_tr)

ensemble <- xgb.train(
  params  = list(objective  = "multi:softmax",
                 num_class  = 3,
                 max_depth  = 4,
                 eta        = 0.3),
  data    = dm_tr,
  nrounds = 100,
  verbose = 0
)

# Attach the response back to the data.frame so the formula in e2tree()
# can find it; createDisMatrix() will use it to annotate the dissimilarity
# matrix (in classification, `label` is optional but recommended).
tr_xgb         <- tr[, 1:4]
tr_xgb$Species <- tr$Species

D <- createDisMatrix(ensemble, data = tr_xgb, label = "Species",
                     parallel = list(active = FALSE, no_cores = 1))

setting  <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
tree_xgb <- e2tree(Species ~ ., data = tr_xgb, D = D,
                   ensemble = ensemble, setting = setting)
print(tree_xgb)
```

### Regression (mtcars)

For regression backends, `createDisMatrix()` needs the response column to
compute the dissimilarity scale. Pass the full data frame (predictors plus
response) and the name of the response column via the `label` argument.

```{r xgb-reg}
library(xgboost)

data(mtcars)
set.seed(42)
n  <- floor(0.75 * nrow(mtcars))
tr <- mtcars[sample(nrow(mtcars), n), ]

X_tr  <- as.matrix(tr[, -1])
y_tr  <- tr$mpg
dm_tr <- xgb.DMatrix(data = X_tr, label = y_tr)

ensemble <- xgb.train(
  params  = list(objective = "reg:squarederror", max_depth = 4, eta = 0.3),
  data    = dm_tr,
  nrounds = 100,
  verbose = 0
)

# `data = tr` carries the response column too; the XGBoost adapter
# automatically trims the matrix to the features used at training time.
D    <- createDisMatrix(ensemble, data = tr, label = "mpg",
                        parallel = list(active = FALSE, no_cores = 1))
tree <- e2tree(mpg ~ ., data = tr, D = D, ensemble = ensemble,
               setting = list(impTotal = 0.1, maxDec = 1e-6, n = 2, level = 5))
print(tree)
```

---

## GBM

### Classification (iris, binary)

`gbm` expects a 0/1 numeric response for the `bernoulli` distribution, while
`e2tree` expects a factor response for classification. We therefore train
`gbm` on the integer column and pass a factor copy of the same column to
`e2tree`.

```{r gbm-clf}
if (!require("gbm")) install.packages("gbm", 
                                               repos="https://cran.r-project.org")
library(gbm)

data(iris)
set.seed(42)
df <- iris
df$is_setosa     <- as.integer(df$Species == "setosa")
df$is_setosa_fct <- factor(df$is_setosa, levels = c(0L, 1L))
n  <- floor(0.75 * nrow(df))
tr <- df[sample(nrow(df), n), ]

ensemble <- gbm(is_setosa ~ Sepal.Length + Sepal.Width + Petal.Length + Petal.Width,
                data              = tr,
                distribution      = "bernoulli",
                n.trees           = 200,
                interaction.depth = 4,
                verbose           = FALSE)

D    <- createDisMatrix(ensemble,
                        data     = tr[, c("Sepal.Length","Sepal.Width",
                                          "Petal.Length","Petal.Width",
                                          "is_setosa_fct")],
                        label    = "is_setosa_fct",
                        parallel = list(active = FALSE, no_cores = 1))
tree <- e2tree(is_setosa_fct ~ Sepal.Length + Sepal.Width +
                               Petal.Length + Petal.Width,
               data = tr, D = D, ensemble = ensemble,
               setting = list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5))
print(tree)
```

### Regression (mtcars)

`gbm` requires `nTrain * bag.fraction > 2 * n.minobsinnode + 1`, which fails
on small training sets such as 24-row `mtcars` with the default settings.
Lower `n.minobsinnode` and raise `bag.fraction` to keep the example
self-contained.

```{r gbm-reg}
library(gbm)

data(mtcars)
set.seed(42)
n  <- floor(0.75 * nrow(mtcars))
tr <- mtcars[sample(nrow(mtcars), n), ]

ensemble <- gbm(mpg ~ ., data = tr,
                distribution      = "gaussian",
                n.trees           = 200,
                interaction.depth = 4,
                n.minobsinnode    = 2,
                bag.fraction      = 0.8,
                verbose           = FALSE)

D    <- createDisMatrix(ensemble, data = tr, label = "mpg",
                        parallel = list(active = FALSE, no_cores = 1))
tree <- e2tree(mpg ~ ., data = tr, D = D, ensemble = ensemble,
               setting = list(impTotal = 0.1, maxDec = 1e-6, n = 2, level = 5))
print(tree)
```

---

## LightGBM

### Classification (iris)

```{r lgb-clf}
if (!require("lightgbm")) install.packages("lightgbm", 
                                               repos="https://cran.r-project.org")
library(lightgbm)

data(iris)
set.seed(42)
n  <- floor(0.75 * nrow(iris))
tr <- iris[sample(nrow(iris), n), ]

X_tr <- as.matrix(tr[, 1:4])
y_tr <- as.integer(tr$Species) - 1L
ds   <- lgb.Dataset(X_tr, label = y_tr)

ensemble <- lgb.train(
  params  = list(objective  = "multiclass",
                 num_class  = 3,
                 num_leaves = 15,
                 verbose    = -1),
  data    = ds,
  nrounds = 100
)

tr_lgb         <- tr[, 1:4]
tr_lgb$Species <- tr$Species

D <- createDisMatrix(ensemble, data = tr_lgb, label = "Species",
                     parallel = list(active = FALSE, no_cores = 1))

tree <- e2tree(Species ~ ., data = tr_lgb, D = D, ensemble = ensemble,
               setting = list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5))
print(tree)
```

### Regression (mtcars)

```{r lgb-reg}
library(lightgbm)

data(mtcars)
set.seed(42)
n  <- floor(0.75 * nrow(mtcars))
tr <- mtcars[sample(nrow(mtcars), n), ]

X_tr <- as.matrix(tr[, -1])
y_tr <- tr$mpg
ds   <- lgb.Dataset(X_tr, label = y_tr)

ensemble <- lgb.train(
  params  = list(objective        = "regression",
                 num_leaves       = 8,
                 min_data_in_leaf = 2,
                 learning_rate    = 0.1,
                 verbose          = -1),
  data    = ds,
  nrounds = 200
)

# Pass the response column to createDisMatrix() via `label`. The
# LightGBM adapter selects the columns it needs through the booster's
# stored feature names, so any extra columns in `data` are ignored.
D    <- createDisMatrix(ensemble, data = tr, label = "mpg",
                        parallel = list(active = FALSE, no_cores = 1))
tree <- e2tree(mpg ~ ., data = tr, D = D, ensemble = ensemble,
               setting = list(impTotal = 0.1, maxDec = 1e-6, n = 2, level = 5))
print(tree)
```


---

## Adding a new backend

To support a further model class `MyEnsemble`, implement three S3 methods and
register them in `NAMESPACE`:

```r
# In R/adapters.R (or a separate R/adapter_mymodel.R)

get_ensemble_type.MyEnsemble <- function(ensemble) {
  # return "classification" or "regression"
}

extract_terminal_nodes.MyEnsemble <- function(ensemble, data) {
  # return data.frame of (n_obs × n_trees) terminal node IDs
}

get_ensemble_predictions.MyEnsemble <- function(ensemble, data, type) {
  # return numeric vector of length n_obs
}
```

No changes to `createDisMatrix()`, `e2tree()`, or any other core function are
required.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{nodes}
\alias{nodes}
\alias{nodes.e2tree}
\title{Extract Tree Node Information}
\usage{
nodes(x, ...)

\method{nodes}{e2tree}(x, terminal = FALSE, ...)
}
\arguments{
\item{x}{An e2tree object.}

\item{...}{Additional arguments (ignored).}

\item{terminal}{Logical. If \code{TRUE}, return only terminal (leaf) nodes.
Default is \code{FALSE}.}
}
\value{
A data frame with one row per node.
}
\description{
Extracts the data frame describing the nodes of an E2Tree model,
including split rules, predictions, and node statistics.
}
\examples{
\donttest{
data(iris)
smp_size <- floor(0.75 * nrow(iris))
train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
training <- iris[train_ind, ]

ensemble <- randomForest::randomForest(Species ~ ., data = training,
  importance = TRUE, proximity = TRUE)
D <- createDisMatrix(ensemble, data = training, label = "Species",
  parallel = list(active = FALSE, no_cores = 1))
setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
tree <- e2tree(Species ~ ., training, D, ensemble, setting)

nodes(tree)
nodes(tree, terminal = TRUE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods_e2tree.R
\name{fitted.e2tree}
\alias{fitted.e2tree}
\title{Extract Fitted Values from an E2Tree Model}
\usage{
\method{fitted}{e2tree}(object, ...)
}
\arguments{
\item{object}{An e2tree object.}

\item{...}{Additional arguments (ignored).}
}
\value{
A vector of fitted values.
}
\description{
Returns the fitted values (predictions) for the training data used to
build the E2Tree model.
}
\examples{
\donttest{
data(iris)
smp_size <- floor(0.75 * nrow(iris))
train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
training <- iris[train_ind, ]

ensemble <- randomForest::randomForest(Species ~ ., data = training,
  importance = TRUE, proximity = TRUE)
D <- createDisMatrix(ensemble, data = training, label = "Species",
  parallel = list(active = FALSE, no_cores = 1))
setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
tree <- e2tree(Species ~ ., training, D, ensemble, setting)

fitted(tree)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{proximity}
\alias{proximity}
\alias{proximity.eValidation}
\title{Extract Proximity Matrices}
\usage{
proximity(x, ...)

\method{proximity}{eValidation}(x, type = c("both", "ensemble", "e2tree"), ...)
}
\arguments{
\item{x}{An eValidation object.}

\item{...}{Additional arguments (ignored).}

\item{type}{Character string specifying which proximity matrix to extract.
One of \code{"ensemble"}, \code{"e2tree"}, or \code{"both"} (default).}
}
\value{
A matrix (if \code{type} is \code{"ensemble"} or \code{"e2tree"})
or a list of two matrices (if \code{type} is \code{"both"}).
}
\description{
Extracts proximity matrices from an eValidation object. The ensemble
proximity matrix is derived from the original ensemble model, while the
E2Tree proximity matrix is estimated from the fitted E2Tree.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods_e2tree.R
\name{summary.e2tree}
\alias{summary.e2tree}
\title{Summary of an E2Tree model}
\usage{
\method{summary}{e2tree}(object, ...)
}
\arguments{
\item{object}{An e2tree object}

\item{...}{Additional arguments (ignored)}
}
\description{
Displays a comprehensive summary including tree structure, decision rules,
terminal node statistics, and variable importance.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa_utils.R
\name{validate_terminal_nodes}
\alias{validate_terminal_nodes}
\title{Validate the output of \code{extract_terminal_nodes()}}
\usage{
validate_terminal_nodes(nodes, data, backend = NA_character_)
}
\description{
Boosting backends store their tree structures in opaque containers; a tiny
API change can silently produce a malformed leaf matrix (e.g. all zeros),
yielding a degenerate dissimilarity matrix without raising any error.
This function asserts the shape and type contract so problems surface
immediately at extraction time rather than much later, after the C++
co-occurrence call has already produced nonsense.
}
\details{
Contract: \code{nodes} must be a \code{data.frame} with \code{nrow(data)}
rows and at least one column; every column must be coercible to integer;
at least one column must contain more than one distinct value.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods_e2tree.R
\name{residuals.e2tree}
\alias{residuals.e2tree}
\title{Extract Residuals from an E2Tree Model}
\usage{
\method{residuals}{e2tree}(object, ...)
}
\arguments{
\item{object}{An e2tree object.}

\item{...}{Additional arguments (ignored).}
}
\value{
A numeric vector of residuals.
}
\description{
Returns the residuals (observed minus fitted) for regression E2Tree models.
Not available for classification models.
}
\examples{
\donttest{
data("mtcars")
smp_size <- floor(0.75 * nrow(mtcars))
train_ind <- sample(seq_len(nrow(mtcars)), size = smp_size)
training <- mtcars[train_ind, ]

ensemble <- randomForest::randomForest(mpg ~ ., data = training, ntree = 500,
  importance = TRUE, proximity = TRUE)
D <- createDisMatrix(ensemble, data = training, label = "mpg",
  parallel = list(active = FALSE, no_cores = 1))
setting <- list(impTotal = 0.1, maxDec = 1e-6, n = 2, level = 5)
tree <- e2tree(mpg ~ ., training, D, ensemble, setting)

residuals(tree)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loi.R
\name{loi}
\alias{loi}
\title{Loss of Interpretability (LoI) Index}
\usage{
loi(O, O_hat, normalize = TRUE)
}
\arguments{
\item{O}{Proximity matrix from the ensemble model (n x n), values in the interval 0 to 1}

\item{O_hat}{Proximity matrix estimated by E2Tree (n x n), values in the interval 0 to 1}

\item{normalize}{Logical. If TRUE (default), returns nLoI (divided by M).
If FALSE, returns raw LoI.}
}
\value{
An object of class \code{"loi"} containing:
\item{loi}{Raw LoI value (unnormalized)}
\item{nloi}{Normalized LoI (LoI / M)}
\item{loi_in}{Within-node component (total)}
\item{loi_out}{Between-node component (total)}
\item{mean_in}{Per-pair average within-node loss (comparable with mean_out)}
\item{mean_out}{Per-pair average between-node loss (comparable with mean_in)}
\item{n}{Matrix dimension}
\item{m}{Number of unique pairs}
\item{n_within}{Number of within-node pairs}
\item{n_between}{Number of between-node pairs}
}
\description{
Computes the LoI index and its decomposition, measuring how well
the E2Tree-estimated proximity matrix reconstructs the original
ensemble proximity matrix.
}
\details{
The statistic is defined as:

\deqn{\mathrm{LoI}(O, \hat{O}) = \sum_{i < j}
  \frac{(o_{ij} - \hat{o}_{ij})^2}{\max(o_{ij}, \hat{o}_{ij})}}

The Normalized LoI divides by the number of pairs \eqn{M = n(n-1)/2}:

\deqn{\mathrm{nLoI}(O, \hat{O}) = \frac{1}{M} \mathrm{LoI}(O, \hat{O})}

The LoI decomposes into two components:
\itemize{
\item \strong{LoI_in}: within-node loss (pairs grouped together by E2Tree)
\item \strong{LoI_out}: between-node loss (pairs separated by E2Tree)
}
The per-pair averages \code{mean_in} and \code{mean_out} enable direct
comparison between the two components despite their different pair counts.

The statistic uses a normalized squared difference, where each cell's
contribution is weighted by the maximum of the two proximity values.
This gives more weight to discrepancies in high-proximity regions.

\strong{Decomposition interpretation (per-pair averages):}
\itemize{
\item \code{mean_out}: average ensemble proximity lost by the partition.
Low values (< 0.1) indicate the tree correctly separates low-proximity
pairs. High values (> 0.3) suggest the tree splits apart pairs that
the ensemble considers similar --more terminal nodes may help.
\item \code{mean_in}: average calibration error within nodes. Low values
(< 0.01) indicate excellent within-node reconstruction. Higher values
reflect the inherent fuzzy-to-crisp transition.
}
}
\examples{
\donttest{
data(iris)
smp_size <- floor(0.75 * nrow(iris))
set.seed(42)
train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
training <- iris[train_ind, ]

ensemble <- randomForest::randomForest(Species ~ ., data = training,
  importance = TRUE, proximity = TRUE)

D <- createDisMatrix(ensemble, data = training, label = "Species",
  parallel = list(active = FALSE, no_cores = 1))

setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
tree <- e2tree(Species ~ ., training, D, ensemble, setting)

vs <- eValidation(training, tree, D)
prox <- proximity(vs)
O <- prox$ensemble
O_hat <- prox$e2tree

# Compute LoI with decomposition
result <- loi(O, O_hat)
print(result)
summary(result)
plot(result)

# Permutation test
perm <- loi_perm(O, O_hat, n_perm = 999, seed = 42)
print(perm)
plot(perm)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createDisMatrix.R
\name{createDisMatrix}
\alias{createDisMatrix}
\title{Create a Dissimilarity Matrix from an Ensemble Model}
\usage{
createDisMatrix(
  ensemble,
  data,
  label,
  parallel = list(active = FALSE, no_cores = 1),
  verbose = FALSE,
  chunk_size = NULL,
  memory_limit = NULL,
  use_disk = FALSE,
  temp_dir = tempdir(),
  batch_aggregate = 10
)
}
\arguments{
\item{ensemble}{is an ensemble tree object}

\item{data}{is a data frame containing the variables in the model. It is the
data frame used for ensemble learning.}

\item{label}{is a character. It indicates the response label.}

\item{parallel}{A list with two elements: \code{active} (logical) and
\code{no_cores} (integer). If \code{active = TRUE}, the function performs
parallel computation using the number of cores specified in
\code{no_cores}. If \code{no_cores} is NULL or equal to 0, it defaults to
using all available cores minus one. If \code{active = FALSE}, the function
runs on a single core. Default: \code{list(active = FALSE, no_cores = 1)}.}

\item{verbose}{Logical. If TRUE, the function prints progress messages and
other information during execution. If FALSE (the default), messages are
suppressed.}

\item{chunk_size}{Integer. Number of rows to process in each chunk. If NULL,
automatically determined based on available memory and dataset size.
Default: NULL (auto).}

\item{memory_limit}{Numeric. Maximum memory to use in GB. Default: NULL (no limit).}

\item{use_disk}{Logical. If TRUE and dataset is very large, intermediate results
are saved to disk. Default: FALSE.}

\item{temp_dir}{Character. Directory for temporary files if use_disk = TRUE.
Default: tempdir().}

\item{batch_aggregate}{Integer. Number of tree results to aggregate at once
before adding to main matrix (reduces memory peaks). Default: 10.}
}
\value{
A dissimilarity matrix. This is a dissimilarity matrix measuring the
discordance between two observations concerning a given random forest
model.
}
\description{
The function createDisMatrix creates a dissimilarity matrix among
observations from an ensemble tree. This optimized version is designed
for large datasets (50K-500K observations) with improved memory management
and chunking capabilities.
}
\details{
This optimized version implements several strategies for handling
large datasets:

\itemize{
\item \strong{Memory-efficient aggregation}: Results from parallel trees are
aggregated in batches to avoid memory peaks
\item \strong{Chunking}: For very large matrices, computation can be split into
manageable chunks
\item \strong{Sparse matrix optimization}: Maintains sparsity throughout computation
\item \strong{Automatic garbage collection}: Explicit memory cleanup at critical points
\item \strong{Disk-based computation}: Optional saving of intermediate results for
datasets exceeding memory capacity
}

Supported ensemble types for \emph{classification} or \emph{regression} tasks:
\itemize{
\item \code{randomForest}
\item \code{ranger}
\item \code{xgb.Booster} (xgboost)
\item \code{lgb.Booster} (lightgbm)
\item \code{gbm} (gbm)
\item \code{catboost.CatBoost} (catboost)
}
}
\section{Interpretation note (RF vs boosting)}{

For \emph{bagging} ensembles (\code{randomForest}, \code{ranger}) the trees are
grown independently on bootstrap samples; co-occurrence in the same leaf
captures local similarity in the predictor space. For \emph{boosting} ensembles
(\code{xgb.Booster}, \code{lgb.Booster}, \code{gbm}, \code{catboost})
each tree is fit to the residual of the previous ones, so leaf
co-occurrence reflects similarity in the \emph{error-correction trajectory}
rather than in the final prediction space. The resulting dissimilarity
matrices therefore have systematically different scales (typically
\eqn{\bar D \in [0.85, 0.95]} for bagging vs. \eqn{[0.35, 0.70]} for
boosting). The surrogate tree built on top of \code{D} should be
interpreted accordingly.

The returned matrix carries an \code{ensemble_backend} attribute identifying
the backend used, which downstream functions check to detect mismatched
\code{(D, ensemble)} pairs.
}

\examples{
\donttest{
data("iris")

# Create training and validation set:
smp_size <- floor(0.75 * nrow(iris))
train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
training <- iris[train_ind, ]
validation <- iris[-train_ind, ]
response_training <- training[,5]
response_validation <- validation[,5]

# Perform training:
## "randomForest" package
ensemble <- randomForest::randomForest(Species ~ ., data=training,
importance=TRUE, proximity=TRUE)

## "ranger" package
if (requireNamespace("ranger", quietly = TRUE)) {
  ensemble <- ranger::ranger(Species ~ ., data = iris,
    num.trees = 1000, importance = 'impurity')
}

# Compute dissimilarity matrix with optimizations
D <- createDisMatrix(
  ensemble,
  data = training,
  label = "Species",
  parallel = list(active = FALSE, no_cores = 1),
  chunk_size = 10000,  # Process 10K rows at a time
  batch_aggregate = 20, # Aggregate 20 trees at once
  verbose = TRUE
)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ePredTree.R
\name{ePredTree}
\alias{ePredTree}
\title{Predict Responses Using an Explainable Ensemble Tree}
\usage{
ePredTree(fit, data, target = "1")
}
\arguments{
\item{fit}{An e2tree object.}

\item{data}{A data frame with new observations.}

\item{target}{Target class for classification scoring.}
}
\value{
A data frame with predictions.
}
\description{
Predicts classification and regression tree responses.
}
\details{
\strong{Deprecated:} Use \code{\link{predict.e2tree}} instead.
}
\examples{
\donttest{
## Classification:
data(iris)

# Create training and validation set:
smp_size <- floor(0.75 * nrow(iris))
train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
training <- iris[train_ind, ]
validation <- iris[-train_ind, ]
response_training <- training[,5]
response_validation <- validation[,5]

# Perform training:
ensemble <- randomForest::randomForest(Species ~ ., data=training,
importance=TRUE, proximity=TRUE)

D <- createDisMatrix(ensemble, data=training, label = "Species",
                             parallel = list(active=FALSE, no_cores = 1))

setting=list(impTotal=0.1, maxDec=0.01, n=2, level=5)
tree <- e2tree(Species ~ ., training, D, ensemble, setting)

## Preferred method:
predict(tree, newdata = validation, target = "1")

## Legacy function (deprecated):
ePredTree(tree, validation, target = "1")


## Regression
data("mtcars")

# Create training and validation set:
smp_size <- floor(0.75 * nrow(mtcars))
train_ind <- sample(seq_len(nrow(mtcars)), size = smp_size)
training <- mtcars[train_ind, ]
validation <- mtcars[-train_ind, ]
response_training <- training[,1]
response_validation <- validation[,1]

# Perform training
ensemble = randomForest::randomForest(mpg ~ ., data=training, ntree=1000,
importance=TRUE, proximity=TRUE)

D = createDisMatrix(ensemble, data=training, label = "mpg",
                              parallel = list(active=FALSE, no_cores = 1))

setting=list(impTotal=0.1, maxDec=(1*10^-6), n=2, level=5)
tree <- e2tree(mpg ~ ., training, D, ensemble, setting)

## Preferred method:
predict(tree, newdata = validation)

## Legacy function (deprecated):
ePredTree(tree, validation)

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods_e2tree.R
\name{print.e2tree}
\alias{print.e2tree}
\title{Print an E2Tree model}
\usage{
\method{print}{e2tree}(x, ...)
}
\arguments{
\item{x}{An e2tree object}

\item{...}{Additional arguments (ignored)}
}
\description{
Displays a compact summary of the fitted E2Tree model including
task type, tree size, terminal nodes, and splitting variables.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/roc.R
\name{roc}
\alias{roc}
\title{ROC Curve}
\usage{
roc(response, scores, target = "1")
}
\arguments{
\item{response}{is the response variable vector}

\item{scores}{is the probability vector of the prediction}

\item{target}{is the target response class}
}
\value{
an object.
}
\description{
Computes and plots the Receiver Operating Characteristic (ROC) curve for a binary classification model,
along with the Area Under the Curve (AUC). The ROC curve is a graphical representation of a classifier’s
performance across all classification thresholds.
}
\examples{
\donttest{
## Classification:
data(iris)

# Create training and validation set:
smp_size <- floor(0.75 * nrow(iris))
train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
training <- iris[train_ind, ]
validation <- iris[-train_ind, ]
response_training <- training[,5]
response_validation <- validation[,5]

# Perform training:
ensemble <- randomForest::randomForest(Species ~ ., data=training, 
importance=TRUE, proximity=TRUE)

D <- createDisMatrix(ensemble, data=training, label = "Species", 
                            parallel = list(active=FALSE, no_cores = 1))

setting=list(impTotal=0.1, maxDec=0.01, n=2, level=5)
tree <- e2tree(Species ~ ., training, D, ensemble, setting)

pr <- ePredTree(tree, validation, target="setosa")

roc(response_training, scores = pr$score, target = "setosa")
 
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_e2tree_vis.R
\name{plot_e2tree_vis}
\alias{plot_e2tree_vis}
\title{Interactive E2Tree Plot with visNetwork}
\usage{
plot_e2tree_vis(
  fit,
  data,
  ensemble,
  width = "100\%",
  height = "100\%",
  direction = "UD",
  node_spacing = 200,
  level_separation = 200,
  colors = NULL,
  show_percent = TRUE,
  show_prob = TRUE,
  show_n = TRUE,
  font_size = 14,
  edge_font_size = 12,
  split_label_style = "rpart",
  max_label_length = 50,
  details_on = "hover",
  navigation_buttons = FALSE,
  free_drag = FALSE
)
}
\arguments{
\item{fit}{An e2tree object}

\item{data}{The training data used to build the tree}

\item{ensemble}{The ensemble model (randomForest or ranger)}

\item{width}{Width of the widget (default: "100\%")}

\item{height}{Height of the widget (default: "100\%")}

\item{direction}{Layout direction: "UD" (top-down), "DU" (bottom-up),
"LR" (left-right), "RL" (right-left)}

\item{node_spacing}{Spacing between nodes at same level (default: 200)}

\item{level_separation}{Spacing between levels (default: 200)}

\item{colors}{Named vector of colors for classes, or NULL for auto}

\item{show_percent}{Show percentage in nodes (default: TRUE)}

\item{show_prob}{Show class probabilities in nodes (default: TRUE)}

\item{show_n}{Show observation count in nodes (default: TRUE)}

\item{font_size}{Font size for node labels (default: 14)}

\item{edge_font_size}{Font size for edge labels (default: 12)}

\item{split_label_style}{How to display split information:
\itemize{
\item "rpart" - Variable name in node, threshold on edges (like rpart.plot)
\item "full" - Full split rule on edges (variable + condition)
\item "threshold" - Only threshold values on edges (< 47, >= 47)
\item "yesno" - Simple yes/no on edges
\item "none" - No labels on edges (hover for details)
\item "innode" - Full split rule displayed IN the node (above stats)
}}

\item{max_label_length}{Maximum characters for edge labels before truncating (default: 50)}

\item{details_on}{When to show node details:
\itemize{
\item "hover" - Show on mouse hover (default, but may cover other nodes)
\item "click" - Show only on click (avoids covering highlighted nodes)
\item "none" - No tooltips (use for cleaner visualization)
}}

\item{navigation_buttons}{Show navigation buttons (default: FALSE)}

\item{free_drag}{If TRUE, nodes can be dragged in ALL directions (horizontal,
vertical, diagonal). If FALSE (default), nodes can only be
moved horizontally within their level.}
}
\value{
A visNetwork htmlwidget object
}
\description{
Displays an E2Tree as an interactive network plot using visNetwork.
Features: drag nodes anywhere, zoom, pan, click for details.
Starts with hierarchical layout, then you can freely move nodes.
}
\examples{
\donttest{
data(iris)
set.seed(42)
smp_size <- floor(0.75 * nrow(iris))
train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
training <- iris[train_ind, ]

ensemble <- randomForest::randomForest(Species ~ ., data = training,
  importance = TRUE, proximity = TRUE)
D <- createDisMatrix(ensemble, data = training, label = "Species",
  parallel = list(active = FALSE, no_cores = 1))
setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
tree <- e2tree(Species ~ ., training, D, ensemble, setting)

# Basic usage
plot_e2tree_vis(tree, training, ensemble)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coercion.R
\name{as.rpart}
\alias{as.rpart}
\alias{as.rpart.e2tree}
\title{Convert an E2Tree Object to rpart Format}
\usage{
as.rpart(x, ...)

\method{as.rpart}{e2tree}(x, ensemble, ...)
}
\arguments{
\item{x}{An e2tree object.}

\item{...}{Additional arguments (ignored).}

\item{ensemble}{The ensemble model used to build the E2Tree. Supported classes:
\code{randomForest}, \code{ranger}, \code{xgb.Booster}, \code{lgb.Booster},
\code{gbm}, \code{catboost.CatBoost}.}
}
\value{
An \code{rpart} object.
}
\description{
Coerces an \code{e2tree} object into an \code{rpart} object, which can
then be used with standard rpart methods for printing, plotting
(e.g., via \code{rpart.plot}), and prediction.
}
\examples{
\donttest{
data(iris)
smp_size <- floor(0.75 * nrow(iris))
train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
training <- iris[train_ind, ]

ensemble <- randomForest::randomForest(Species ~ ., data = training,
  importance = TRUE, proximity = TRUE)
D <- createDisMatrix(ensemble, data = training, label = "Species",
  parallel = list(active = FALSE, no_cores = 1))
setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
tree <- e2tree(Species ~ ., training, D, ensemble, setting)

rpart_obj <- as.rpart(tree, ensemble)
}

}
\seealso{
\code{\link{as.party.e2tree}} for conversion to partykit format.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods_e2tree.R
\name{predict.e2tree}
\alias{predict.e2tree}
\title{Predict Responses from an E2Tree Model}
\usage{
\method{predict}{e2tree}(object, newdata, target = NULL, ...)
}
\arguments{
\item{object}{An e2tree object.}

\item{newdata}{A data frame containing the new observations. If missing,
the fitted values for the training data are returned.}

\item{target}{Character string specifying the target class for computing
classification scores. Only used for classification trees. Default is
\code{NULL}, which uses the first level.}

\item{...}{Additional arguments (ignored).}
}
\value{
For regression: a data frame with columns \code{fit} (predicted
value) and \code{sd} (standard deviation of the response within the
terminal node, computed from the training data).
For classification: a data frame with columns \code{fit} (predicted class),
\code{accuracy} (probability of the predicted class), and \code{score}
(probability of the target class).
}
\description{
Predicts classification or regression responses for new data using the
fitted E2Tree model.
}
\examples{
\donttest{
data(iris)
smp_size <- floor(0.75 * nrow(iris))
train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
training <- iris[train_ind, ]
validation <- iris[-train_ind, ]

ensemble <- randomForest::randomForest(Species ~ ., data = training,
  importance = TRUE, proximity = TRUE)
D <- createDisMatrix(ensemble, data = training, label = "Species",
  parallel = list(active = FALSE, no_cores = 1))
setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
tree <- e2tree(Species ~ ., training, D, ensemble, setting)

## Predict on new data
pred <- predict(tree, newdata = validation)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adapters.R
\name{extract_terminal_nodes}
\alias{extract_terminal_nodes}
\title{Extract Terminal Node Assignments from an Ensemble Model}
\usage{
extract_terminal_nodes(ensemble, data)
}
\arguments{
\item{ensemble}{A trained ensemble model.}

\item{data}{A \code{data.frame} of observations (may include the response
column; it is ignored internally).}
}
\value{
A \code{data.frame} with \code{n_obs} rows and \code{n_trees} columns
of integer terminal-node identifiers.
}
\description{
Returns a \code{data.frame} with \code{n_obs} rows and \code{n_trees}
columns where each cell is the terminal-node index assigned to that
observation by that tree.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_e2tree_click.R
\name{plot_e2tree}
\alias{plot_e2tree}
\title{Quick E2Tree Plot (Non-Interactive)}
\usage{
plot_e2tree(fit, ensemble, main = "E2Tree", ...)
}
\arguments{
\item{fit}{An e2tree object}

\item{ensemble}{The ensemble model (randomForest or ranger)}

\item{main}{Plot title}

\item{...}{Additional arguments passed to rpart.plot}
}
\value{
Invisibly returns the rpart object
}
\description{
Displays an E2Tree as a static plot using rpart.plot.
For interactive exploration, use plot_e2tree_click().
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eValidation.R
\name{eValidation}
\alias{eValidation}
\title{Validate an E2Tree Model via Proximity Matrix Comparison}
\usage{
eValidation(
  data,
  fit,
  D,
  test = c("both", "mantel", "measures"),
  graph = TRUE,
  n_perm = 999,
  conf.level = 0.95,
  seed = NULL
)
}
\arguments{
\item{data}{A data frame containing the variables in the model.}

\item{fit}{An e2tree object.}

\item{D}{The dissimilarity matrix obtained with \code{\link{createDisMatrix}}.}

\item{test}{Character string specifying which tests to perform. One of
\code{"both"} (default), \code{"mantel"} (Mantel test only), or
\code{"measures"} (divergence/similarity measures with permutation tests only).}

\item{graph}{Logical (default TRUE). If TRUE, heatmaps are displayed.}

\item{n_perm}{Integer. Number of permutations for the permutation
test on measures. Default is 999. Set to 0 to skip permutation testing.
Ignored when \code{test = "mantel"}.}

\item{conf.level}{Numeric. Confidence level for intervals. Default is 0.95.}

\item{seed}{Integer or NULL. Random seed for reproducibility.}
}
\value{
An object of class \code{"eValidation"} containing:
\describe{
\item{Proximity_matrix_ensemble}{Ensemble proximity matrix (reordered)}
\item{Proximity_matrix_e2tree}{E2Tree proximity matrix (reordered)}
\item{mantel_test}{Mantel test result (NULL if \code{test = "measures"})}
\item{loi}{LoI object with decomposition (NULL if \code{test = "mantel"})}
\item{measures}{Data frame with all measures (NULL if \code{test = "mantel"})}
\item{permutation}{Permutation test results for measures (if applicable)}
}
}
\description{
Compares the ensemble proximity matrix with the E2Tree-estimated proximity
matrix using multiple divergence and similarity measures. Can perform the
Mantel test, permutation tests on divergence/similarity measures (nLoI,
Hellinger, wRMSE, RV, SSIM), or both.
}
\examples{
\donttest{
## Classification:
data(iris)
smp_size <- floor(0.75 * nrow(iris))
train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
training <- iris[train_ind, ]

ensemble <- randomForest::randomForest(Species ~ ., data=training,
  importance=TRUE, proximity=TRUE)

D <- createDisMatrix(ensemble, data=training, label = "Species",
  parallel = list(active=FALSE, no_cores = 1))

setting <- list(impTotal=0.1, maxDec=0.01, n=2, level=5)
tree <- e2tree(Species ~ ., training, D, ensemble, setting)

val <- eValidation(training, tree, D, n_perm = 199)
print(val)
summary(val)
plot(val)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa_utils.R
\name{ensemble_backend}
\alias{ensemble_backend}
\title{Identify the canonical class of a supported ensemble model}
\usage{
ensemble_backend(ensemble)
}
\description{
Returns one of \code{"randomForest"}, \code{"ranger"}, \code{"xgb.Booster"},
\code{"lgb.Booster"}, \code{"gbm"}, \code{"catboost.CatBoost"} or
\code{"catboost.Model"} (the same class used by the S3 adapter dispatch),
or \code{NA_character_} when no supported class is matched.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adapters.R
\name{get_ensemble_type}
\alias{get_ensemble_type}
\title{Determine Task Type from a Trained Ensemble Model}
\usage{
get_ensemble_type(ensemble)
}
\arguments{
\item{ensemble}{A trained ensemble model. Supported classes:
\code{randomForest}, \code{ranger}, \code{xgb.Booster},
\code{lgb.Booster}, \code{gbm}, \code{catboost.CatBoost}.}
}
\value{
Character scalar: \code{"classification"} or \code{"regression"}.
}
\description{
Returns \code{"classification"} or \code{"regression"} depending on the
objective used to train the ensemble.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coercion.R
\name{as.party.e2tree}
\alias{as.party.e2tree}
\title{Convert an E2Tree Object to partykit Format}
\usage{
as.party.e2tree(x, ...)
}
\arguments{
\item{x}{An e2tree object.}

\item{...}{Additional arguments (ignored).}
}
\value{
A \code{party} object (from \pkg{partykit}).
}
\description{
Coerces an \code{e2tree} object into a \code{party} object from the
\pkg{partykit} package. This enables the use of partykit's infrastructure
for printing, plotting, and predicting with the E2Tree model.
}
\examples{
\donttest{
data(iris)
smp_size <- floor(0.75 * nrow(iris))
train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
training <- iris[train_ind, ]

ensemble <- randomForest::randomForest(Species ~ ., data = training,
  importance = TRUE, proximity = TRUE)
D <- createDisMatrix(ensemble, data = training, label = "Species",
  parallel = list(active = FALSE, no_cores = 1))
setting <- list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
tree <- e2tree(Species ~ ., training, D, ensemble, setting)

if (requireNamespace("partykit", quietly = TRUE)) {
  party_obj <- partykit::as.party(tree)
  plot(party_obj)
}
}

}
\seealso{
\code{\link{as.rpart.e2tree}} for conversion to rpart format.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_e2tree_click.R
\name{plot_e2tree_click}
\alias{plot_e2tree_click}
\title{Interactive E2Tree Plot for R Graphics Device}
\usage{
plot_e2tree_click(
  fit,
  data,
  ensemble,
  main = "E2Tree - Click on nodes (ESC to exit)",
  ...
)
}
\arguments{
\item{fit}{An e2tree object}

\item{data}{The training data used to build the tree}

\item{ensemble}{The ensemble model (randomForest or ranger)}

\item{main}{Plot title (default: "E2Tree - Click on nodes (ESC to exit)")}

\item{...}{Additional arguments passed to rpart.plot}
}
\value{
Invisibly returns the rpart object
}
\description{
Displays an E2Tree as an interactive plot in the R graphics device.
Click on nodes to see detailed information in the console.
Right-click or press ESC to exit interactive mode.
}
\details{
This function converts the e2tree object to an rpart object and displays it
using rpart.plot. You can then click on any node to see:
\itemize{
\item Node ID and type (terminal/internal)
\item Number of observations
\item Prediction and probability/purity
\item Decision path to reach the node
\item Class distribution (for classification)
\item Split rule (for internal nodes)
\item Observations in the node (for terminal nodes)
}
}
\examples{
\donttest{
# After creating an e2tree object (requires interactive session)
if (interactive()) {
  plot_e2tree_click(tree, training, ensemble)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/e2tree.R
\name{e2tree}
\alias{e2tree}
\title{Explainable Ensemble Tree}
\usage{
e2tree(
  formula,
  data,
  D,
  ensemble,
  setting = list(impTotal = 0.1, maxDec = 0.01, n = 2, level = 5)
)
}
\arguments{
\item{formula}{is a formula describing the model to be fitted, with a response but no interaction terms.}

\item{data}{a data frame containing the variables in the model. It is a data frame in which to interpret the variables named in the formula.}

\item{D}{is the dissimilarity matrix. This is a dissimilarity matrix measuring the discordance between two observations concerning a given classifier of a random forest model. The dissimilarity matrix is obtained with the \link{createDisMatrix} function.}

\item{ensemble}{is an ensemble tree object (for the moment ensemble works only with random forest objects)}

\item{setting}{is a list containing the set of stopping rules for the tree building procedure.
\tabular{lll}{
\code{impTotal}\tab   \tab The threshold for the impurity in the node\cr
\code{maxDec}\tab   \tab The threshold for the maximum impurity decrease of the node\cr
\code{n}\tab   \tab The minimum number of the observations in the node \cr
\code{level}\tab   \tab The maximum depth of the tree (levels) \cr}
Default is \code{setting=list(impTotal=0.1, maxDec=0.01, n=2, level=5)}.}
}
\value{
A e2tree object, which is a list with the following components:
\tabular{lll}{
\code{tree}\tab   \tab  A data frame representing the main structure of the tree aimed at explaining and graphically representing the relationships and interactions between the variables used to perform an ensemble method. \cr
\code{call}\tab   \tab The matched call\cr
\code{terms}\tab   \tab A list of terms and attributes \cr
\code{control}\tab   \tab A list containing the set of stopping rules for the tree building procedure  \cr
\code{varimp}\tab   \tab A list containing a table and a plot for the variable importance. Variable importance refers to a quantitative measure that assesses the contribution of individual variables within a predictive model towards accurate predictions. It quantifies the influence or impact that each variable has on the model's overall performance. Variable importance provides insights into the relative significance of different variables in explaining the observed outcomes and aids in understanding the underlying relationships and dynamics within the model \cr}
}
\description{
It creates an explainable tree for Random Forest. Explainable Ensemble Trees (E2Tree) aimed to generate a “new tree” that can explain and represent the relational structure between the response variable and the predictors. This lead to providing a tree structure similar to those obtained for a decision tree exploiting the advantages of a dendrogram-like output.
}
\examples{
\donttest{
## Classification:
data(iris)

# Create training and validation set:
smp_size <- floor(0.75 * nrow(iris))
train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
training <- iris[train_ind, ]
validation <- iris[-train_ind, ]
response_training <- training[,5]
response_validation <- validation[,5]

# Perform training:
## "randomForest" package
ensemble <- randomForest::randomForest(Species ~ ., data=training,
importance=TRUE, proximity=TRUE)

## "ranger" package
if (requireNamespace("ranger", quietly = TRUE)) {
  ensemble <- ranger::ranger(Species ~ ., data = iris,
    num.trees = 1000, importance = 'impurity')
}

D <- createDisMatrix(ensemble, data=training, label = "Species",
                              parallel = list(active=FALSE, no_cores = 1))

setting=list(impTotal=0.1, maxDec=0.01, n=2, level=5)
tree <- e2tree(Species ~ ., training, D, ensemble, setting)



## Regression
data("mtcars")

# Create training and validation set:
smp_size <- floor(0.75 * nrow(mtcars))
train_ind <- sample(seq_len(nrow(mtcars)), size = smp_size)
training <- mtcars[train_ind, ]
validation <- mtcars[-train_ind, ]
response_training <- training[,1]
response_validation <- validation[,1]

# Perform training
## "randomForest" package
ensemble = randomForest::randomForest(mpg ~ ., data=training, ntree=1000,
importance=TRUE, proximity=TRUE)

## "ranger" package
if (requireNamespace("ranger", quietly = TRUE)) {
  ensemble <- ranger::ranger(formula = mpg ~ ., data = training,
    num.trees = 1000, importance = "permutation")
}

D = createDisMatrix(ensemble, data=training, label = "mpg",
                               parallel = list(active=FALSE, no_cores = 1))

setting=list(impTotal=0.1, maxDec=(1*10^-6), n=2, level=5)
tree <- e2tree(mpg ~ ., training, D, ensemble, setting)

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa_utils.R
\name{check_package}
\alias{check_package}
\title{Check Availability of Suggested Packages}
\usage{
check_package(pkg)
}
\description{
Check Availability of Suggested Packages
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa_utils.R
\name{e2_variance}
\alias{e2_variance}
\title{Population Variance}
\usage{
e2_variance(x)
}
\description{
Population Variance
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vimp.R
\name{vimp}
\alias{vimp}
\title{Variable Importance}
\usage{
vimp(fit, data, type = NULL)
}
\arguments{
\item{fit}{An e2tree object.}

\item{data}{A data frame containing the variables in the model.}

\item{type}{Character string: \code{"classification"} or \code{"regression"}.
If \code{NULL} (default), the type is automatically detected from the
e2tree object.}
}
\value{
A list containing:
\describe{
\item{vimp}{A data frame with variable importance metrics.}
\item{g_imp}{A ggplot bar chart of Mean Impurity Decrease.}
\item{g_acc}{(Classification only) A ggplot bar chart of Mean Accuracy Decrease.}
}
}
\description{
Computes variable importance for an E2Tree model based on mean
impurity decrease and (for classification) mean accuracy decrease.
}
\examples{
\donttest{
## Classification:
data(iris)

# Create training and validation set:
smp_size <- floor(0.75 * nrow(iris))
train_ind <- sample(seq_len(nrow(iris)), size = smp_size)
training <- iris[train_ind, ]

# Perform training:
ensemble <- randomForest::randomForest(Species ~ ., data=training,
importance=TRUE, proximity=TRUE)

D <- createDisMatrix(ensemble, data=training, label = "Species",
                             parallel = list(active=FALSE, no_cores = 1))

setting=list(impTotal=0.1, maxDec=0.01, n=2, level=5)
tree <- e2tree(Species ~ ., training, D, ensemble, setting)

vi <- vimp(tree, training)
vi$vimp
vi$g_imp


## Regression
data("mtcars")

# Create training and validation set:
smp_size <- floor(0.75 * nrow(mtcars))
train_ind <- sample(seq_len(nrow(mtcars)), size = smp_size)
training <- mtcars[train_ind, ]

# Perform training
ensemble = randomForest::randomForest(mpg ~ ., data=training, ntree=1000,
importance=TRUE, proximity=TRUE)

D = createDisMatrix(ensemble, data=training, label = "mpg",
                         parallel = list(active=FALSE, no_cores = 1))

setting=list(impTotal=0.1, maxDec=(1*10^-6), n=2, level=5)
tree <- e2tree(mpg ~ ., training, D, ensemble, setting)

vi <- vimp(tree, training)
vi$vimp
vi$g_imp

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adapters.R
\name{get_ensemble_predictions}
\alias{get_ensemble_predictions}
\title{Get Training Predictions from an Ensemble Model}
\usage{
get_ensemble_predictions(ensemble, data, type)
}
\arguments{
\item{ensemble}{A trained ensemble model.}

\item{data}{The training \code{data.frame} that was used to fit the model.}

\item{type}{Character: \code{"classification"} or \code{"regression"}.}
}
\value{
Numeric vector of length \code{nrow(data)}.
}
\description{
Returns a numeric vector of length \code{n_obs} with the ensemble's
prediction for every training observation. For models that store
out-of-bag (OOB) predictions (\code{randomForest}, \code{ranger}) the
stored OOB vector is returned; for other models in-sample predictions
are computed from the training data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_e2tree_vis.R
\name{save_e2tree_html}
\alias{save_e2tree_html}
\title{Save E2Tree visNetwork Plot to HTML}
\usage{
save_e2tree_html(vis, file = "e2tree_plot.html", selfcontained = TRUE)
}
\arguments{
\item{vis}{A visNetwork object from plot_e2tree_vis()}

\item{file}{Output file path (should end with .html)}

\item{selfcontained}{Include all dependencies in single file}
}
\description{
Save E2Tree visNetwork Plot to HTML
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{measures}
\alias{measures}
\alias{measures.eValidation}
\title{Extract Validation Measures}
\usage{
measures(x, ...)

\method{measures}{eValidation}(x, ...)
}
\arguments{
\item{x}{An eValidation object.}

\item{...}{Additional arguments (ignored).}
}
\value{
A data frame with columns for method name, type, observed value,
and (if permutation tests were performed) null distribution statistics
and p-values.
}
\description{
Extracts the data frame of validation measures from an eValidation object,
including divergence and similarity metrics between the ensemble and
E2Tree proximity matrices.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods_e2tree.R
\name{plot.e2tree}
\alias{plot.e2tree}
\title{Plot an E2Tree model}
\usage{
\method{plot}{e2tree}(x, ensemble = NULL, main = "E2Tree", ...)
}
\arguments{
\item{x}{An e2tree object}

\item{ensemble}{The ensemble model (randomForest or ranger).
Required for converting the tree to rpart format. Supported classes:
\code{randomForest}, \code{ranger}, \code{xgb.Booster}, \code{lgb.Booster},
\code{gbm}, \code{catboost.CatBoost}.}

\item{main}{Plot title. Default is "E2Tree".}

\item{...}{Additional arguments passed to \code{rpart.plot::rpart.plot}}
}
\description{
Displays the tree structure using rpart.plot.
This is a convenience wrapper around \code{\link{plot_e2tree}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loi.R
\name{loi_perm}
\alias{loi_perm}
\title{Permutation Test for LoI}
\usage{
loi_perm(O, O_hat, n_perm = 999, conf.level = 0.95, seed = NULL)
}
\arguments{
\item{O}{Proximity matrix from the ensemble model (n x n)}

\item{O_hat}{Proximity matrix estimated by E2Tree (n x n)}

\item{n_perm}{Number of permutations (default: 999)}

\item{conf.level}{Confidence level for intervals (default: 0.95)}

\item{seed}{Random seed for reproducibility. Default is NULL.}
}
\value{
An object of class \code{"loi_perm"} containing:
\item{observed}{Observed nLoI value and decomposition (loi object)}
\item{statistic}{Observed nLoI value (scalar)}
\item{p.value}{Test p-value (one-sided, less)}
\item{ci}{Permutation-based confidence interval for nLoI}
\item{null_dist}{Null distribution of nLoI values}
\item{null_mean}{Mean of the null distribution}
\item{null_sd}{Standard deviation of the null distribution}
\item{z_stat}{Standardized Z statistic}
\item{n_perm}{Number of permutations}
\item{conf.level}{Confidence level}
}
\description{
Performs a permutation test using row/column permutation to assess
whether the E2Tree reconstruction is significantly better than expected
by chance.
}
\details{
The test uses \strong{simultaneous row/column permutation} of
\eqn{\hat{O}}: for each replicate, a random permutation \eqn{\pi}
of \eqn{\{1, \ldots, n\}} is drawn and \eqn{\hat{O}^\pi =
\hat{O}[\pi, \pi]} is computed. This preserves the block-diagonal
structure of \eqn{\hat{O}} while breaking the correspondence with
\eqn{O}.

The null hypothesis is: the E2Tree labeling is unrelated to the
ensemble structure. Under H1 (good reconstruction), the observed
nLoI should be significantly \emph{lower} than the null distribution.

P-values include the +1 correction of Phipson & Smyth (2010).
}
\examples{
\donttest{
n <- 50
O <- matrix(runif(n^2, 0.3, 1), n, n)
O <- (O + t(O)) / 2; diag(O) <- 1
O_hat <- O + matrix(rnorm(n^2, 0, 0.05), n, n)
O_hat <- pmin(pmax((O_hat + t(O_hat)) / 2, 0), 1); diag(O_hat) <- 1

result <- loi_perm(O, O_hat, n_perm = 199, seed = 42)
print(result)
summary(result)
plot(result)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{e2splits}
\alias{e2splits}
\alias{e2splits.e2tree}
\title{Extract Split Information from an E2Tree Model}
\usage{
e2splits(x, ...)

\method{e2splits}{e2tree}(x, ...)
}
\arguments{
\item{x}{An e2tree object.}

\item{...}{Additional arguments (ignored).}
}
\value{
A list with components:
\describe{
\item{splits}{The split information matrix.}
\item{csplit}{The categorical split encoding matrix.}
}
}
\description{
Returns the split matrix and categorical split encoding from a fitted
E2Tree model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_e2tree_click.R
\name{print_e2tree_summary}
\alias{print_e2tree_summary}
\title{Print E2Tree Summary}
\usage{
print_e2tree_summary(fit, data)
}
\arguments{
\item{fit}{An e2tree object}

\item{data}{The training data}
}
\description{
Prints a comprehensive summary of an E2Tree model including
all decision rules, variable importance, and node statistics.
}
